#===============================================================
#
#   filerep.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#print "*FILEREP*\n";
use strict;
use integer;

BEGIN
{
my $cols = 80;
if (-t STDOUT)
{
use Term::ReadKey qw(GetTerminalSize);
($cols, undef) = GetTerminalSize();
}
print '=' x ($cols - 5) . "\n";
}

use File::Glob ':glob';      # Override glob built-in




sub filerep_file($);
sub find_files_recursively(@);
sub recurse_dir($);
sub find_files(@);
sub warning($;@);
sub error($;@);
sub error_exit($;@);
sub fatal($;@);
sub usage();
sub env_say($@);




my $PREFIX = 'FILEREP';
my $PREFIX_ = ' ' x ((length $PREFIX) + 1);

$| = 1;
my $RC = 0;
my @MSG_LINES;

my $IS_WIN32 = ($^O eq 'MSWin32');

my @REL_FILE_SPECS;
my @RECURSE_WILD_SPECS;
my $IS_TEST_MODE = 0;
my $MUST_RECURSE = 0;

my @REP_REFS;


my $RENAMES_PRESENT = 0;
my $RENAME_COUNT = 0;
my $DELETES_PRESENT = 0;
my $DELETE_COUNT = 0;






=optdoc
use glo::genopt;

my @genopts = (
[ '<1>', 'command_or_commandsfile',  'ssm', "", [ "Rename/Delete Command or filespec of file containing Rename/Delete Commands",
"Rename/Delete Command: <delim><find><delim><replace><delim>[g|r][i]",
"- g = glob rename (default)",
"- r = Perl regular expression rename",
"- i = ignore case of search string",
"Empty <replace> deletes the file",
"commandsfile:  Empty lines and lines beginning with '#' are ignored" ] ],
[ '<*>', 'files',	 'sso',	"", "Files to process" ],
[ 't', 'test',	 'bso',	 0, "test only - do not make modifications" ],
[ 'r', 'recurse',	 'bso',	 0, "recurse directory looking for -f= files" ],
[ 'f', 'wild-files', 'sao',	"", "wild-card files to process" ],
);
GENOPT_set_flag_prefix( '-');
GENOPT_set_optdefs( 'filerep', \@genopts,
'Rename/Delete files',
"Directory/Files starting with '.' are skipped");
GENOPT_parse();

=cut




{

if (@ARGV == 1 && $ARGV[0] =~ /^--?(h|help)$/)
{
usage();
exit 0;
} else
{
if (@ARGV < 2 || substr($ARGV[0], 0, 1) eq '-')
{
env_say( 1, "Not enough arguments");
usage();
exit 9;
}
}





my @finds;
{
my $find = shift @ARGV;
if (-f $find)
{
env_say( 1, "FileRep File: '$find' found");
open( IN, $find) or fatal( "Cannot open '$find'", $!);
my @lines = <IN>;
close IN;
foreach my $line (@lines)
{
$line =~ s/\s+$//;
if ($line ne '' && substr($line, 0, 1) ne '#')
{
push @finds, $line;
}
}
} else
{
my $delim = substr( $find, 0, 1);
my $replace;
my $replace_options;
my $command = substr( $find, 1);
($find, $replace, $replace_options) = split( quotemeta $delim, $command);
while (@ARGV && !defined $replace_options)
{
$command .= ' ' . shift( @ARGV);

($find, $replace, $replace_options) = split( quotemeta $delim, $command);
}

error_exit( "No find string specified")
if ($find eq '');
error_exit( "No replace string specified ($find)")
if (!defined $replace_options);
push @finds, $delim . $find . $delim . $replace . $delim . $replace_options;
}
}




my @rel_file_defs;
{
foreach my $arg (@ARGV)
{
if (substr($arg,0,1) eq '-')
{
if ($arg eq '-t' || $arg eq '-t+')
{
$IS_TEST_MODE = 1;
} elsif ($arg eq '-t-')
{
$IS_TEST_MODE = 0;
} elsif ($arg eq '-r' || $arg eq '-r+')
{
$MUST_RECURSE = 1;
} elsif ($arg eq '-r-')
{
$MUST_RECURSE = 0;
} elsif (substr( $arg, 0, 3) eq '-r=')
{
$MUST_RECURSE = 1;
$arg = substr ($arg, 3);
push @rel_file_defs, split( ',', $arg);
push @RECURSE_WILD_SPECS, split( ',', $arg);
} elsif (substr( $arg, 0, 3) eq '-f=')
{
$arg = substr ($arg, 3);
push @rel_file_defs, split( ',', $arg);
push @RECURSE_WILD_SPECS, split( ',', $arg);
} else
{
env_say( 1, "Invalid option '$arg'. Allowed '-t', '-r', '-r=<files>', '-f=<files>'");
usage();
exit 9;
}
} else
{
push @rel_file_defs, $arg;
}
}
if ($MUST_RECURSE && @RECURSE_WILD_SPECS == 0)
{
env_say( 1, "No recurse (wild-)files specified. Assume *.*");
push @RECURSE_WILD_SPECS, '*.*';
}
}




foreach my $find (@finds)
{
env_say( 1, "Command: $find");

my $delim = substr( $find, 0, 1);
my $command = substr( $find, 1);
my ($find, $replace, $replace_options) = split( quotemeta $delim, $command);
error_exit( "- No find string specified")
if ($find eq '');
error_exit( "- No replace string specified ($find)")
if (!defined $replace_options);
if ($replace_options ne '')
{
$replace_options = lc $replace_options;
} else
{
$replace_options = 'g';		# glob rename
}

my @replace_options = split( '', $replace_options);
my $replace_mode;			# g = glob, r = RegExp
my $must_ignore_case = 0;		# i
my $is_eval_mode = 0;
foreach my $option (@replace_options)
{
if (index( 'gr', $option) >= 0)
{
if (defined $replace_mode)
{
error_exit( "- Replace mode ('$option') already defined to '$replace_mode'");
} else
{
$replace_mode = $option;
}
} elsif ($option eq 'i')
{
$must_ignore_case = 1;
} else
{
error_exit( "- Invalid rep option '$option'. Allowed: [g|r] [i]");
}
}
$replace_mode = 'g'
if (!defined $replace_mode);




my $search;
if ($replace_mode eq 'g')				# glob rename
{
$find =~ tr!\\!/!;
if ($find !~ m!/! && $replace !~ m!/!)
{
$find = "*/$find";
$replace = "*/$replace";
}
$search = quotemeta( $find);
$search =~ s/\\\*/(.*)/g;
$search =~ s/\\\?/(.)/g;

if ($must_ignore_case)
{
$search = qr/^$search$/i;
} else
{
$search = qr/^$search$/;
}

if ($replace =~ /(\*|\?)/)
{
my $i = 1;
$replace =~ s/(\*|\?)/'$' . $i++/ge;	# concat to $1, $2, $3, etc
$replace = '"' . $replace . '"';
$is_eval_mode = 1;
}
} elsif ($replace_mode eq 'r')			# regexp rename
{
$search = $find;
if ($must_ignore_case)
{
$search = qr/$search/i;
} else
{
$search = qr/$search/;
}

if ($replace =~ /\$\d/)
{
$replace = '"' . $replace . '"';
$is_eval_mode = 1;
}
} else
{
error_exit( "- Invalid option '$replace_mode'. Allowed:[g|r]");
}
if ($replace ne '')
{
$RENAMES_PRESENT++;
} else
{
$DELETES_PRESENT++;
}




push @REP_REFS, [$search, $replace, $is_eval_mode];
env_say( 2, "- find=$find ($search), replace=$replace, options=$replace_options");
}




if ($MUST_RECURSE)
{
@REL_FILE_SPECS = find_files_recursively( @rel_file_defs);
} else
{
@REL_FILE_SPECS = find_files( @rel_file_defs);
}





{
my %seen;
@REL_FILE_SPECS = grep { ! $seen{ $_ }++ } @REL_FILE_SPECS;
}
}




{
if (@REL_FILE_SPECS)
{
env_say( 1, "Found " . scalar @REL_FILE_SPECS . " file(s).");
foreach my $file (@REL_FILE_SPECS)
{
filerep_file( $file);
}
} else
{
warning( "No files found");
}
}




{
env_say( 0, '');
if ($IS_TEST_MODE)
{
env_say( 1, "TestMode: $RENAME_COUNT files selected for rename")
if ($RENAMES_PRESENT);
env_say( 1, "TestMode: $DELETE_COUNT files selected for delete")
if ($DELETES_PRESENT);
} else
{
env_say( 1, "$RENAME_COUNT files renamed")
if ($RENAMES_PRESENT);
env_say( 1, "$DELETE_COUNT files deleted")
if ($DELETES_PRESENT);
}

if (@MSG_LINES)
{
env_say( 1, "The following errors were encountered:");
map { env_say( 0, "  $_"); } @MSG_LINES;
$RC = 2;
}
}

exit $RC;




sub END
{
if (@MSG_LINES)
{
env_say( 1, "The following warnings/errors were encountered:");
map { env_say( 2, "$_"); } @MSG_LINES;
}

my $rc = $?;
print "\a"
if ($rc > 0 || @MSG_LINES > 0);

env_say( 1, ($rc) ? "Failed ($?)" : "Done");
}




sub filerep_file($)
{
my ($old_file) = @_;


foreach my $ref (@REP_REFS)
{



my ($search, $replace, $is_eval_mode) = @{$ref};

if ($replace ne '')
{



my $new_file = $old_file;
my $changed;

if ($is_eval_mode)
{
$changed = $new_file =~ s/$search/$replace/gee;
} else
{
$changed = $new_file =~ s/$search/$replace/g;
}
if ($changed)
{
print "  $old_file => $new_file\n";
if (! $IS_TEST_MODE)
{
print "    rename( $old_file, $new_file)\n";
if (!rename( $old_file, $new_file))
{
error( 'Rename failed', $!);

}
}
$RENAME_COUNT++;
last;
}
} else
{



if ($old_file =~ /$search/)
{
print "  Delete $old_file\n";
if (! $IS_TEST_MODE)
{
print "    unlink( $old_file)\n";
if (!unlink( $old_file))
{
error( 'Delete failed:', $!);

}
}
$DELETE_COUNT++;
last;
}
}
}
}




sub find_files_recursively(@)
{
my @wild_dirs = @_;
my @rel_file_specs;


foreach my $wild_dir_spec (@wild_dirs)
{
$wild_dir_spec =~ s!//!/!g;
$wild_dir_spec =~ s!/$!!g;
}
env_say( 1, "Recursing (@RECURSE_WILD_SPECS) files in (@wild_dirs) ...");
foreach my $wild_dir_spec (@wild_dirs)
{
my @dir_specs;
if ($wild_dir_spec =~ /[\*\?]/)
{
push @dir_specs, grep -d $_, glob( $wild_dir_spec);
} else
{
if (-d $wild_dir_spec)
{
push @dir_specs, $wild_dir_spec;
} elsif (-f _)
{
warning( "FileSpec must be directory: '$wild_dir_spec'");
} else
{
warning( "No such directory '$wild_dir_spec'");
}
}
foreach my $dir_spec (@dir_specs)
{
push @rel_file_specs, recurse_dir( $dir_spec);
}
}


return @rel_file_specs;
}





sub recurse_dir($)
{
my ($dir,
) = @_;
my @rel_file_specs;





foreach my $wild_file (@RECURSE_WILD_SPECS)
{
my $file_spec = "$dir/$wild_file";
if ($wild_file =~ /[\*\?]/)
{
push @rel_file_specs, grep ! -d $_, glob( $file_spec);
} else
{
if (-f $file_spec)
{
push @rel_file_specs, $file_spec;
}
}
}




opendir( DIR, "$dir") || fatal( "Cannot opendir '$dir'", $?);
my @entries = readdir( DIR);
closedir( DIR);
foreach my $entry (@entries)
{
next if (substr( $entry, 0, 1) eq '.');	    # Skip directories/files starting with '.'
my $this_dir = "$dir/$entry";
push @rel_file_specs, recurse_dir( $this_dir)
if (-d $this_dir);
}

return @rel_file_specs;
}




sub find_files(@)
{
my @file_specs = @_;
my @rel_file_specs;

env_say( 1, "Files: (@file_specs)");

foreach my $file_spec (@file_specs)
{
if ($file_spec =~ /[\*\?]/)
{
push @rel_file_specs, glob( $file_spec);
} else
{
if (-f $file_spec)
{
push @rel_file_specs, $file_spec
} elsif (-d _)
{
warning( "FileSpec must be file: '$file_spec'");
} else
{
warning( "No such file '$file_spec'");
}
}
}

return @rel_file_specs;
}




sub warning($;@)
{
my ($first_line,
@rest_lines,
) = @_;

env_say( 1, "*WARNING* $first_line");
if (@rest_lines)
{
map { env_say( 2, "$_") } @rest_lines;
}
push @MSG_LINES, "WRN: $first_line";
}




sub error($;@)
{
my ($first_line,
@rest_lines,
) = @_;

env_say( 1, "*ERROR* $first_line");
if (@rest_lines)
{
map { env_say( 2, "$_") } @rest_lines;
}
push @MSG_LINES, "ERR: $first_line";
$RC = 2
if ($RC < 2);
}




sub error_exit($;@)
{
my ($first_line,
@rest_lines,
) = @_;

env_say( 1, "*ERROR* $first_line");
if (@rest_lines)
{
map { env_say( 2, "$_") } @rest_lines;
}
push @MSG_LINES, "ERR: $first_line";

exit 2;
}




sub fatal($;@)
{
my ($first_line,
@rest_lines,
) = @_;

env_say( 1, "*FATAL* $first_line");
if (@rest_lines)
{
map { env_say( 2, "$_") } @rest_lines;
}
push @MSG_LINES, "FTL: $first_line";

exit 9;
}




sub usage()
{
print "usage is: filerep <command> | <commands-file> <files/dirs...> [-t] [-r] [-r=<files>] [-f=<files>]\n";
print "  command = <delim><find><delim><replace><delim>[g|r][i]\n";
print "    g = glob rename\n";
print "    r = perl regular expresion rename\n";
print "    i = ignore case of search string\n";
print "    Empty <replace> deletes the file\n";
print "    -t == test-only\n";
print "    -r == recurse directory\n";
print "    -r=<files> == recurse directory using comma-separated wild files\n";
print "    -f=<files> == comma-separated wild files\n";
print "  commands-file: Empty lines and lines beginning with '#' are ignored\n";
print "  wild files: Directories and Files starting with '.' are skipped\n";
}




sub env_say($@)
{
my ($indent,    # 0 = no prefix, 1 = $prefix, 2 = $prefix_ - Negative = prefix_level - 1
@lines_and_or_refs
) = @_;

my @lines = map { (ref $_) ? @{$_} : $_ } @lines_and_or_refs;
if ($indent == 0)
{
map { CORE::say( $_) } @lines;
} else
{
if (abs( $indent) == 1)
{
my $first_line = shift @lines;
CORE::say( "$PREFIX: $first_line");
}
map { CORE::say( "$PREFIX_ $_") } @lines;
}
}


