#===============================================================
#
#   bld.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package bld;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
BLD_init_build
BLD_init_audit
BLD_get_commands
);

use strict;
use integer;

use glo::env;
use glo::spit;
use gbsenv;
use sys;
use target;
use audit;
use flincs;
use gbsbg;




sub BLD_init_build();
sub BLD_init_audit($);
sub BLD_get_commands($$@);

sub get_incs($$);
sub get_flags($$$);
sub format_flags($$);




my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();
my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_SYSGEN_PATH = "$GBS_ROOT_PATH/sysgen";
my $GBS_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $GBS_COMP_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/comp";
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my $GBS_AUDIT;
my $TARGET_PLUGIN_NAME;
my $AUDIT_PLUGIN_NAME;
my $SCA_PLUGIN_NAME;

my $VIA_FORMAT;
my $VIA_FILE_TYPE;

my %BANNERS_CACHE;


my %FLAGS_CACHE;


my %INCS_CACHE;





my $SUBS_TOKEN_1 = (ENV_is_win32()) ? '%1' : '\\$1';	    # %1 or $1

my @FLAGS_INCS_ORDER = qw( sysflags flags sysincs incs);




sub BLD_init_build()
{


$GBS_AUDIT = undef;
TARGET_set_envs( $GBS_TARGET);	# Also sets: GBS_BLD_*, GBS_MODE, GBS_OPT, etc
SYS_set_envs();			# Checks and sets GBS_APP_*
my ($target_plugin_name, $plugin_rel) = TARGET_get_plugin_data( $GBS_TARGET);
ENV_say( 1, "$GBS_TARGET [$target_plugin_name ($plugin_rel)]");

($VIA_FORMAT, $VIA_FILE_TYPE) = TARGET_get_section_items( $GBS_TARGET, SETUP => 'VIA_FORMAT');

return '-';	# exec_data
}




sub BLD_init_audit($)
{
($GBS_AUDIT) = @_;	# can have multiple audit-runs in one execution of gbsaudit command




if (!defined $TARGET_PLUGIN_NAME)
{
($TARGET_PLUGIN_NAME, my $plugin_rel) = TARGET_get_plugin_data( $GBS_TARGET);
ENV_say( 1, "$GBS_TARGET [$TARGET_PLUGIN_NAME ($plugin_rel)]");
}

GBSENV_setenv( GBS_AUDIT => $GBS_AUDIT);
$SCA_PLUGIN_NAME = AUDIT_set_envs( $GBS_AUDIT);
($AUDIT_PLUGIN_NAME, my $plugin_rel) = AUDIT_get_plugin_data( $GBS_AUDIT);
my @src_types = AUDIT_get_src_types( $GBS_AUDIT);

ENV_say( 1, "$GBS_AUDIT [$AUDIT_PLUGIN_NAME ($plugin_rel)] (@src_types)");

($VIA_FORMAT, $VIA_FILE_TYPE) = AUDIT_get_section_items( $GBS_AUDIT, SETUP => 'VIA_FORMAT');

%BANNERS_CACHE = ();
%FLAGS_CACHE = ();
%INCS_CACHE = ();

return $TARGET_PLUGIN_NAME;
}







sub BLD_get_commands($$@)
{
my ($component,
$in_file,
@command_line_flags,
) = @_;
my @lines;


my ($file_name, $src_type) = ENV_split_spec_nt( $in_file);


my $src_dir = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/comp/$component/src";


my $out_file_path;
my $out_file_base;
my $gen_type;
my $out_file_type;
my $src_abs_path;	    # bool
my $gen_type_verb;
my @out_types;
my @opt_out_types;
my @out_files;
my $out_keep_file;
if (defined $GBS_AUDIT)
{
$out_file_path	= "../aud/$GBS_AUDIT/$GBS_TARGET";
$out_file_base	= "$out_file_path/$in_file";
$gen_type	= '';
$out_file_type	= AUDIT_get_primary_out_type( $GBS_AUDIT, $src_type);
$src_abs_path	= 0;
$gen_type_verb	= 'Auditing';
@out_types	= (AUDIT_get_src_items( $GBS_AUDIT, $src_type, 'OUT_TYPES'),
qw( .garg .gerr .gmet .gout .html ));
@opt_out_types	= (qw( .log.html ));
@out_files	= ();

$out_keep_file	= "$out_file_base.gout";
} else
{
$out_file_path	= "../bld/$GBS_TARGET";
$out_file_base	= "$out_file_path/$file_name";
$gen_type	= TARGET_get_src_items( $GBS_TARGET, $src_type, 'TYPE')->[0];
$out_file_type	= TARGET_get_primary_bld_type( $GBS_TARGET, $src_type);
$src_abs_path	= TARGET_get_src_items( $GBS_TARGET, $src_type, 'SRC_ABS_PATH');
$gen_type_verb	= TARGET_get_type_verb( $gen_type);	# E.g.: Compiling, Linking
@out_types	= TARGET_get_src_items( $GBS_TARGET, $src_type, 'OUT_TYPES');
@opt_out_types	= TARGET_get_src_items( $GBS_TARGET, $src_type, 'OPT_OUT_TYPES');
@out_files	= TARGET_get_src_items( $GBS_TARGET, $src_type, 'OUT_FILES');

if ($gen_type eq 'glk')
{
$out_keep_file = "$out_file_base.map";		    # .map file
} elsif ($gen_type eq 'glt')
{
$out_keep_file = "$out_file_base$out_file_type";	    # .log file
} else
{

}
}




my @flinc_file_specs;
my $flincs_file_type = ($VIA_FILE_TYPE) ? $VIA_FILE_TYPE : '.gbs';
foreach my $flinc (@FLAGS_INCS_ORDER)   # = qw( sysflags flags sysincs incs)
{
push @flinc_file_specs, "$out_file_path/$in_file.$flinc$flincs_file_type";
}

unlink @flinc_file_specs;




my @out_del_files  = map { "$out_file_base$_" } (@out_types, @opt_out_types);
push @out_del_files, map { "$out_file_path/$_" } @out_files;
unlink @out_del_files;




@out_del_files = grep( $_ ne $out_keep_file, @out_del_files)
if (defined $out_keep_file);




my @flincs;

push @flincs, get_flags( $component, $src_type, "$out_file_base.map");
push @flincs, get_incs(  $component, $src_type);

my @flinc_env_refs;
for (my $i = 0; $i < @FLAGS_INCS_ORDER; $i++) # = qw( sysflags flags sysincs incs)
{
my $flinc_type = $flincs[$i*2];
my $flincs_ref = $flincs[$i*2+1];

my $envvar_value;
if ($flinc_type eq 'FILE')
{
foreach my $item (@{$flincs_ref})
{
$item = ENV_join_quoted_space( $item);
}
my $flinc_file_spec = $flinc_file_specs[$i];
SPIT_file_nl( $flinc_file_spec, $flincs_ref);
$envvar_value = sprintf( $VIA_FORMAT, $flinc_file_spec);
} else	#$flincs_type eq 'VARIABLE'
{
$envvar_value = ENV_join_quoted_space( @{$flincs_ref});
}
my $flinc_uc = uc $FLAGS_INCS_ORDER[$i];
push @flinc_env_refs, [ "GBS_$flinc_uc"  => $envvar_value ];
}




my $in_file_spec = ($src_abs_path) ? "$src_dir/$in_file" : $in_file;
my $out_file_spec = "$out_file_base$out_file_type";
my @pos_args = ($file_name, $in_file_spec, $out_file_spec, $out_file_path, $out_file_type);
my @command_data_refs;
if (defined $GBS_AUDIT)
{
@command_data_refs = AUDIT_get_src_command_data_refs_exp(  $GBS_AUDIT,  $src_type, \@pos_args, \@command_line_flags);
} else
{
@command_data_refs = TARGET_get_src_command_data_refs_exp( $GBS_TARGET, $src_type, \@pos_args, \@command_line_flags);
}




{



push @lines, GBSBG_echo( "** $gen_type_verb ${GBS_SUBSYS}:$component:$in_file to $GBS_TARGET:$file_name$out_file_type...",
"** $BANNERS_CACHE{$src_type}");




foreach my $ref (@flinc_env_refs)
{
push @lines, GBSBG_set( $ref->[0], $ref->[1]);	# $envvar_name $envvar_value
}




if (defined $GBS_AUDIT)
{
my $audit_flags = ENV_getenv( 'GBS_FLAGS_'. uc $GBS_AUDIT);
push @lines, GBSBG_set( 'GBS_AUDIT_FLAGS', $audit_flags);
}




if (defined $GBS_AUDIT)
{
if ($SCA_PLUGIN_NAME eq '')
{



push @lines, GBSBG_exec( $in_file, undef, \@out_del_files, @command_data_refs); # $must_print_commands, out_files
} else
{



my $unix_style_search = (TARGET_get_src_items( $GBS_TARGET, $src_type, 'INC_SEARCH_STYLE') eq 'UNIX') ? 1 : 0;
push @lines, GBSBG_sca( $in_file, $unix_style_search, undef, @command_data_refs);  # out_files
}
} else
{
my ($glkb_type, undef) = TARGET_get_src_items( $GBS_TARGET, $src_type, 'GLKB');
if (defined $glkb_type)
{
my $glkb_args_ref = [ $glkb_type, $in_file, $out_file_spec, $VIA_FORMAT ];
push @lines, GBSBG_glkb( $in_file, undef, $glkb_args_ref, \@out_del_files, @command_data_refs);	    # $must_print_commands
} else
{
push @lines, GBSBG_exec( $in_file, undef, \@out_del_files, @command_data_refs);			    # $must_print_commands
}
}

}

return @lines;
}




sub get_incs($$)
{
my ($component,
$src_type,
) = @_;
my ( $sysincs_type, $sysincs_ref, $incs_type, $incs_ref);

my $incs_src_ref = $INCS_CACHE{$component}->{$src_type};
if (defined $incs_src_ref)
{
( $sysincs_type, $sysincs_ref, $incs_type, $incs_ref ) = @{$incs_src_ref};
} else
{
my ( @sysincs, @incs );




my ($inc_format, $sysinc_format, $inc_abs_path);
if (defined $GBS_AUDIT)
{
my ($inc_format_ref, $sysinc_format_ref) =
AUDIT_get_src_items( $GBS_AUDIT, $src_type, qw( INC_FORMAT SYSINC_FORMAT));
($incs_type, $inc_format) = ENV_undef2empty( @{$inc_format_ref});
($sysincs_type, $sysinc_format) = ENV_undef2empty( @{$sysinc_format_ref});
$inc_abs_path = 0;
} else
{
(my $inc_format_ref, my $sysinc_format_ref, $inc_abs_path) =
TARGET_get_src_items( $GBS_TARGET, $src_type, qw( INC_FORMAT SYSINC_FORMAT INC_ABS_PATH));
($incs_type, $inc_format) = ENV_undef2empty( @{$inc_format_ref});
($sysincs_type, $sysinc_format) = ENV_undef2empty( @{$sysinc_format_ref});
}
my $src_dir = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/comp/$component/src";




if ($sysinc_format)
{
@sysincs = FLINCS_get_bld_incs( $src_type);
@sysincs = ($inc_abs_path) ?
ENV_abs_path( $src_dir, [ @sysincs ]) :
ENV_shortest_path( $src_dir, [ @sysincs ]);
@sysincs = map { $_ = sprintf( $sysinc_format, $_) } @sysincs;  # format the include
}
if ($inc_format)
{
@incs = FLINCS_get_incs( $src_type, $component);
@incs = ($inc_abs_path) ?
ENV_abs_path( $src_dir, [ @incs ]) :
ENV_shortest_path( $src_dir, [ @incs ]);
@incs = map { $_ = sprintf( $inc_format, $_) } @incs;	    # format the include
}


$sysincs_ref = [ @sysincs ];
$incs_ref    = [ @incs ];
$INCS_CACHE{$component}->{$src_type} = [ $sysincs_type, $sysincs_ref, $incs_type, $incs_ref ];
}

return ( $sysincs_type, $sysincs_ref, $incs_type, $incs_ref );
}




sub get_flags($$$)
{
my ($component,
$src_type,
$map_file,
) = @_;
my ( $sysflags_type, $sysflags_ref, $flags_type, $flags_ref, $map_option_index );

my $flags_src_ref = $FLAGS_CACHE{$component}->{$src_type};
if (defined $flags_src_ref)
{
( $sysflags_type, $sysflags_ref, $flags_type, $flags_ref, $map_option_index ) = @{$flags_src_ref};
} else
{
my ( @sysflags, @flags );




my ($flag_format, $sysflag_format);
if (defined $GBS_AUDIT)
{
my ($flag_format_ref, $sysflag_format_ref) =
AUDIT_get_src_items( $GBS_AUDIT, $src_type, qw( FLAG_FORMAT SYSFLAG_FORMAT));
($flags_type, $flag_format) = ENV_undef2empty( @{$flag_format_ref});
($sysflags_type, $sysflag_format) = ENV_undef2empty( @{$sysflag_format_ref});
} else
{
my ($flag_format_ref, $sysflag_format_ref) =
TARGET_get_src_items( $GBS_TARGET, $src_type, qw( FLAG_FORMAT SYSFLAG_FORMAT));
($flags_type, $flag_format) = ENV_undef2empty( @{$flag_format_ref});
($sysflags_type, $sysflag_format) = ENV_undef2empty( @{$sysflag_format_ref});
}




if ($sysflag_format)
{
@sysflags = FLINCS_get_bld_flags( $src_type);




@sysflags = grep( /^"?-D/, @sysflags)
if (defined $GBS_AUDIT);
}




{



push @flags, FLINCS_get_flags( $src_type, $component);




{
my $src_type_name = substr( $src_type, 1);
my $key = 'GBS_FLAGS_'. uc $src_type_name;
my $gbs_flags = ENV_getenv( $key);
push @flags, $gbs_flags
if ($gbs_flags);
}






$map_option_index = -1;
my %wanted_options;
if (defined $GBS_AUDIT)
{
$wanted_options{MODE} = 1;
} else
{
map { $wanted_options{$_} = 1 } TARGET_get_options_order();
}
my @banner;
foreach my $opt_ref (TARGET_get_default_options( $GBS_TARGET, $src_type))
{
my ($option, $default_choice) = @{$opt_ref};    #  MODE OPT DEBUGGER MAP
if ($wanted_options{$option})
{
my $choice = ENV_getenv( "GBS_$option");
$choice = $default_choice
if ($choice eq '');
my @values = TARGET_get_src_subitems( $GBS_TARGET, $src_type, $option, $choice);
if ($option eq 'MAP')
{
foreach my $value (@values)
{
push @flags, $value;
$map_option_index = $#flags
if ($value =~ /$SUBS_TOKEN_1/);	    # $SUBS_TOKEN_1 = (ENV_is_win32()) ? '%1' : '\\$1';
}
} else
{
push @flags, @values;
}

my $banner = "$option=$choice";
$banner .= " ($default_choice)"
if ($choice ne $default_choice);
push @banner, $banner;
}
}
$BANNERS_CACHE{$src_type} = join( ', ', @banner)
if (!exists $BANNERS_CACHE{$src_type});




@flags = grep( /^"?-D/, @flags)
if (defined $GBS_AUDIT);
}




if ($sysflag_format ne '-D%s')
{
format_flags( \@sysflags, $sysflag_format);
}
$sysflags_ref = [ @sysflags ];

if ($flag_format ne '-D%s')
{
format_flags( \@flags, $flag_format);
}
$flags_ref = [ @flags ];

$FLAGS_CACHE{$component}->{$src_type} = [ $sysflags_type, $sysflags_ref, $flags_type, $flags_ref, $map_option_index ];
}




if ($map_option_index >= 0)
{



my @flags = @{$flags_ref};					# we must copy the flags for the MAP


$flags[$map_option_index] =~ s/$SUBS_TOKEN_1/$map_file/g;	# $SUBS_TOKEN_1 = (ENV_is_win32()) ? '%1' : '\\$1';
$flags_ref = \@flags;
}

return ( $sysflags_type, $sysflags_ref, $flags_type, $flags_ref);
}





sub format_flags($$)
{
my ($flags_ref,	# In/OUT
$flag_format,
) = @_;


foreach my $flag (@{$flags_ref})
{
if (substr( $flag, 0, 2) eq '-D')
{
$flag = sprintf( $flag_format, substr( $flag, 2));
} elsif (substr( $flag, 0, 3) eq '"-D')
{
$flag = '"' . sprintf( $flag_format, substr( $flag, 3));
} else
{

}
}

}

return 1;

