#===============================================================
#
#   fcreate.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package fcreate;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
FCREATE_get_any
FCREATE_any
FCREATE_gen_type
FCREATE_audit_target
FCREATE_export
FCREATE_full_gbs
FCREATE_plugin_file
FCREATE_plugin_ref_file
FCREATE_scope
FCREATE_owners
FCREATE_system
FCREATE_steps
FCREATE_switch
FCREATE_sys
FCREATE_subsys_script
FCREATE_gbsall
);

use strict;
use integer;

use glo::env;
use glo::slurp;
use glo::spit;
use glo::file;
use gbsenv;
use gbsglo;
use fspec;




sub FCREATE_get_any($);
sub FCREATE_any($$$);
sub FCREATE_gen_type($$$$$);
sub FCREATE_audit_target($$$$$$);
sub FCREATE_export($$$$);
sub FCREATE_full_gbs($$);
sub FCREATE_plugin_file($$$);
sub FCREATE_plugin_ref_file($$$$);
sub FCREATE_scope($$$$$);
sub FCREATE_owners($$);
sub FCREATE_system($$);
sub FCREATE_steps($$$);
sub FCREATE_switch($$$);
sub FCREATE_sys($$$);
sub FCREATE_subsys_script($$$);
sub FCREATE_gbsall($$);

sub create_gbs_file($$$@);
sub create_non_template_gbs_file($$%);
sub create_empty_file($);




my $COMMAND_FILE_EXT = ENV_commandfile_ext();

my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();
my $GBS_SYSGEN_PATH = ENV_getenv_perl_path( 'GBS_SYSGEN_PATH');
my $GBS_SCRIPTS_TEMPLATE_PATH = "$GBS_SCRIPTS_PATH/template";
my $GBS_SYSGEN_TEMPLATES_PATH = "$GBS_SYSGEN_PATH/templates";
my $GBS_TEMPLATES_PATH = ENV_getenv_perl_path( 'GBS_TEMPLATES_PATH');




sub FCREATE_get_any($)
{
my ($file_types_ref,    # [ $primary_file_type, $sec_file_type ] secondary = from TARGET.GBS
) = @_;
my @anyfile_specs;

my @any_types = @{$file_types_ref};





my @template_paths;
push @template_paths, $GBS_TEMPLATES_PATH
if ($GBS_TEMPLATES_PATH ne '');
push @template_paths, ($GBS_SYSGEN_TEMPLATES_PATH, $GBS_SCRIPTS_TEMPLATE_PATH);


my %files_found;

foreach my $path (@template_paths)
{
my @all_files = SLURP_dir_files( $path, 0);
my @any_files;
foreach my $type (@any_types)
{
my $q_type = quotemeta $type;
push @any_files, grep /^any.*$q_type$/, @all_files;
last if @any_files;	    # no need to search secondary files if at least one primary found
}





foreach my $file (@any_files)
{
if (!exists $files_found{$file})
{
push @anyfile_specs, "$path/$file";
$files_found{$file} = 1;
}
}
}


return @anyfile_specs;
}




sub FCREATE_any($$$)
{
my ($file_spec,
$anyfile_spec_or_comment_ref,	# $anyfile_spec or [ $begin_comment, $end_comment ]
$subsys_component_ref,		# [ $subsys, $component ]
) = @_;



my ($file_name, $file_type) = ENV_split_spec_nt( $file_spec);
my $file = "$file_name$file_type";
my ($subsys, $component) = @{$subsys_component_ref};
my $system = ENV_getenv( 'GBS_SYSTEM_NAME');

if (ref $anyfile_spec_or_comment_ref)
{



my ($begin_com, $end_com) = @{$anyfile_spec_or_comment_ref};

my $begin_com_l = length( $begin_com);
my $end_com_l = length( $end_com);
my $max_line_length = 72;
my $border_length = $max_line_length - $begin_com_l - $end_com_l;
my $left_com = $begin_com . ' ' x (4 - $begin_com_l);
my $left_com_l = length( $left_com);
my $file_text = "File: $file";
my $file_text_l = length( $file_text);
my @lines;
push @lines, $begin_com . '=' x $border_length . $end_com;
push @lines, $left_com  . $file_text . ' ' x ($max_line_length - $left_com_l - $file_text_l - $end_com_l) . $end_com;
push @lines, $begin_com . '=' x $border_length . $end_com;
push @lines, '';
push @lines, $begin_com . ' ===EOF=== ' . $end_com;
map { $_ =~ s/\s*$/\n/ } @lines;	# remove trailing white and add nl

SPIT_file( $file_spec, \@lines);
} elsif ($anyfile_spec_or_comment_ref ne '')
{



my $anyfile_spec = $anyfile_spec_or_comment_ref;

my $include_guard = uc $file;
$include_guard =~ s/\./_/g;

FILE_copy_template( $anyfile_spec, $file_spec,
SYSTEM_NAME => $system,
SUBSYS => $subsys,
COMPONENT => $component,
FILE => $file,
FILE_NAME => $file_name,
FILE_TYPE => $file_type,
UC_FILE => uc $file,
UC_FILE_NAME => uc $file_name,
UC_FILE_TYPE => uc $file_type,
INCLUDE_GUARD => $include_guard,
);
} else
{
SPIT_file( $file_spec);
}

return $file_spec;
}








sub FCREATE_gen_type($$$$$)
{
my ($file,		    # incs_<type>.gbs, incs_<type>.usr, etc. see above
$root,
$subsys,
$component,
$target) = @_;
my $file_spec;

my $title;
if ($subsys)
{
$title = $subsys.'::';
if ($component)
{
$title .= $component;
} else
{
$title .= 'GEN';
}
} else
{
$title = 'SYSGEN';
}
if ($target)
{
$title .= ":$target";
}

$file_spec = FSPEC_gen_type( $file, $root, $subsys, $component, $target);
create_gbs_file( $file_spec, undef, undef,
TITLE=>$title);

return $file_spec;
}





sub FCREATE_audit_target($$$$$$)
{
my ($file,		    # sca_<target>.gbs, sca_<target>.usr, etc. see above
$root,
$subsys,
$component,
$audit,
$target) = @_;
my $file_spec;

my $title;
if ($subsys)
{
$title = $subsys.'::';
if ($component)
{
$title .= $component;
} else
{
$title .= 'AUDIT';
}
} else
{
$title = 'SYSAUDIT';
}
if ($audit)
{
$title .= ":$audit";
}
if ($target)
{
$title .= "[$target]";
}

$file_spec = FSPEC_audit_target( $file, $root, $subsys, $component, $audit, $target);
create_gbs_file( $file_spec, undef, undef,
TITLE=>$title);

return $file_spec;
}




sub FCREATE_full_gbs($$)
{
my ($root,
$subsys,
) = @_;
my $file_spec;

$file_spec = FSPEC_full_gbs( $root, $subsys);
create_empty_file( $file_spec);

return $file_spec;
}




sub FCREATE_export($$$$)
{
my ($file,	    	#export.gbs or export.usr
$root,
$subsys,
$component,	# can be '' in case of gen/export.gbs or undef for non-gbs subsys
) = @_;
my $file_spec;

my $title = $subsys;
$title .= '::' . $component if ($component);
$file_spec = FSPEC_export( $file, $root, $subsys, $component);
create_gbs_file( $file_spec, undef, undef,
TITLE=>$title);

return $file_spec;
}




sub FCREATE_plugin_file($$$)
{
my ($file,		# e.g. target.gbs, audit.usr, tool.gbs, etc
$root,
$item_name,
) = @_;
my $file_spec;

(my $dir, $file) = FSPEC_plugin_file( $file, $root, $item_name);
$file_spec = "$dir/$file";

create_gbs_file( $file_spec, undef, (),
TITLE=>$item_name);

return $file_spec;
}




sub FCREATE_plugin_ref_file($$$$)
{
my ($file,		# e.g. target.gbs, audit.usr, tool.gbs, etc
$root,
$item_name,
$plugin_name,
) = @_;
my $file_spec;

(my $dir, $file) = FSPEC_plugin_file( $file, $root, $item_name);
$file_spec = "$dir/$file";

my $file_name = ENV_split_spec_n( $file);
my $filename_gbs = "$file_name.gbs";

my $template_file = "${file_name}_ref.gbs";
create_gbs_file( $file_spec, $template_file, undef,
PLUGIN => $plugin_name,
INCLUDE=> $filename_gbs,
TITLE  => $item_name);

return $file_spec;
}




sub FCREATE_scope($$$$$)
{
my ($file,		# scope.gbs or scope.usr
$root,
$subsys,
$component,
$lines_ref) = @_;
my $file_spec;

$file_spec = FSPEC_scope( $file, $root, $subsys, $component);
create_gbs_file( $file_spec, undef, $lines_ref,
TITLE=>"$subsys\:\:$component");

return $file_spec;
}




sub FCREATE_owners($$)
{
my ($file,		    #  with or without .gbs or .usr
$root,
) = @_;
my $file_spec;

$file_spec = FSPEC_owners( $file, $root);
create_empty_file( $file_spec);

return $file_spec;
}




sub FCREATE_system($$)
{
my ($file,		    #  with or without .gbs or .usr
$root,
) = @_;
my $file_spec;

$file_spec = FSPEC_system( $file, $root);
create_empty_file( $file_spec);

return $file_spec;
}




sub FCREATE_steps($$$)
{
my ($file,		# steps.gbs or steps.usr
$root,
$system_name) = @_;
my $file_spec;

$file_spec = FSPEC_steps( $file, $root);

my @lines;
foreach my $subsys (GBSGLO_subsystems())
{
push @lines,
"STEP $subsys",
"     EXEC       = SELECT",
'';
}
create_gbs_file( $file_spec, undef, \@lines,
SYSTEM_NAME=>$system_name);

return $file_spec;
}




sub FCREATE_switch($$$)
{
my ($file,			# switch.gbs or switch.gbs.bat or switch.gbs.sh or

$root,
$system_name) = @_;
my $file_spec;

my $site = ENV_getenv( 'GBS_SITE');
$site = 'EIN' if ($site eq '');

$file_spec = FSPEC_switch( $file, $root);
create_gbs_file( $file_spec, undef, undef,
SYSTEM_NAME=>$system_name, SITE=>$site);

return $file_spec;
}




sub FCREATE_sys($$$)
{
my ($file,		# sys.gbs or sys.usr
$root,
$system_name) = @_;
my $file_spec;

$file_spec = FSPEC_sys( $file, $root);
create_gbs_file( $file_spec, undef, undef,
SYSTEM_NAME=>$system_name);

return $file_spec;
}




sub FCREATE_subsys_script($$$)
{
my ($file_name,	# gbssubgen, gbssubmake, etc with or without .sh or .bat
$root,
$subsys,
) = @_;
my $file_spec;

$file_spec = FSPEC_subsys_script( $file_name, $root, $subsys);
create_gbs_file( $file_spec, undef, undef,
SUBSYS => $subsys);

return $file_spec;
}




sub FCREATE_gbsall($$)
{
my ($file_type,	    # full_name+type or .sh or .bat or ''
$in_install,	    # In GBS_SCRIPTS_ROOT else in GBS_BASE_PATH
) = @_;
my $file_spec;

$file_spec = FSPEC_gbsall( $file_type, $in_install);
my $location = ($in_install) ? 'Install' : 'User';
create_gbs_file( $file_spec, undef, undef,
LOCATION => $location);

return $file_spec;
}




sub create_gbs_file($$$@)
{
my ($file_spec,	    # path/name.gbs or .usr
$template_file,	    # may be undef
$lines_ref,	    # may be undef
@translate) = @_;

my $file = ENV_split_spec_f( $file_spec);
if (!-e $file_spec || -z $file_spec)
{
my $filename_gbs ;
if (defined $template_file)
{
$filename_gbs = $template_file;
} else
{
$filename_gbs = $file;
$filename_gbs =~ s/usr$/gbs/;
}
my $template_spec = "$GBS_SCRIPTS_TEMPLATE_PATH/$filename_gbs";
if (-f $template_spec)
{
ENV_say( 1, "Creating $file file from template $filename_gbs");
FILE_copy_template( $template_spec, $file_spec, @translate,
FILE=>$file, _LINES_=>$lines_ref);
} else
{
create_non_template_gbs_file( $file_spec, $lines_ref, @translate);
}
chmod 0755, $file_spec
if (substr( $file_spec, -3, 3) eq '.sh');	# set file to 'x'
} else
{
ENV_say( 1, "$file file already exists - not created");
}
}




sub create_non_template_gbs_file($$%)
{
my ($file_spec,
$lines_ref,
%translate) = @_;

my ($dir, $file) = ENV_split_spec_df( $file_spec);
if (!-d $dir)
{
ENV_say( 1, "Creating $dir...");
ENV_mkdir( $dir);
}
my $nr_lines = (defined $lines_ref) ? @{$lines_ref} : 0;

if ($nr_lines > 0)
{
ENV_say( 1, "Creating $file file ($nr_lines)");
} else
{
ENV_say( 1, "Creating empty $file file");
}
my @lines;
push @lines, "#========================================================";
push @lines, "#   [$translate{TITLE}] $file";
push @lines, "#========================================================";
push @lines, @{$lines_ref}
if ($nr_lines > 0);
push @lines, '';
push @lines, "###EOF###";

SPIT_file_nl( $file_spec, @lines);
}




sub create_empty_file($)
{
my ($file_spec) = @_;

my ($dir, $file) = ENV_split_spec_df( $file_spec);
if (!-f $file_spec)
{
ENV_say( 1, "Creating empty $file file");
SPIT_file( $file_spec);
chmod 0755, $file_spec
if (substr( $file_spec, -3, 3) eq '.sh');	# set file to 'x'
} else
{
ENV_say( 1, "$file file already exists - not created");
}
}

return 1;

