#=================================================
#
#   fix.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package fix;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
FIX_main
FIX_structure
);

use strict;
use integer;

use glo::env;
use glo::ask;
use glo::file;
use glo::scm;
use gbsenv;
use gbsscm;
use files;




sub FIX_main();
sub FIX_structure();

sub check_illegal($);
sub check_missing($);
sub check_scm_control($);
sub add_scm($$);
sub fix_execute_bit();




my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();




sub FIX_main()
{
ENV_say( 1, "Check GBS Directory Structure");

my @line_refs = FILES_all( $GBS_ROOT_PATH);








check_illegal( \@line_refs);
check_missing( \@line_refs);
ENV_say( 1, "Check Directory Structure: Done");

GBSSCM_preset( 0);

ENV_say( 0, '');
if (SCM_get_current_id())
{
check_scm_control( \@line_refs);
} else
{
ENV_say( 1, "No_SCM: Check SCM control: Not executed");
}

if (ENV_is_lunix())
{
fix_execute_bit();
}
}




sub FIX_structure()
{
my $is_ok;

ENV_say( 1, "Check GBS Directory Structure");

my @line_refs = FILES_all( $GBS_ROOT_PATH);








my $illegal_ok = check_illegal( \@line_refs);
my $missing_ok = check_missing( \@line_refs);
$is_ok = ($illegal_ok && $missing_ok) ? 1 : 0;

ENV_say( 1, "Check: " . (($is_ok) ? 'Done' : 'Failed'));
return $is_ok;
}




sub check_illegal($)
{
my ($line_refs_ref) = @_;





my $is_ok = 1;




ENV_say( 1, "Checking for illegal directories/files...");
my $nr_illegal_files = 0;
foreach my $ref (@{$line_refs_ref})
{
my ($type, $avail, undef, $spec) = @{$ref};
if ($avail eq 'I')
{
ENV_say( 0, " - Illegal: $type $spec");
$nr_illegal_files++;
} elsif ($avail eq 'i')
{
ENV_say( 0, " - TypeErr: $type $spec");
$nr_illegal_files++;
}
}




if ($nr_illegal_files > 0)
{
$is_ok = 0;
ENV_error( "$nr_illegal_files illegal files/directories");
ENV_say( 1, "GBS can try to delete these files/direcories",
"This may not allways succeed",
"GBS is not able to remove them from SCM control if applicable");
if (ASK_YN( 'Try to delete anyway? (selective)', 'N') eq 'Y')
{
$is_ok = 1;
my $ans = '';
foreach my $ref (@{$line_refs_ref})
{
my ($type, $avail, undef, $spec) = @{$ref};
my $full_spec = "$GBS_ROOT_PATH/$spec";
if ($avail eq 'I' && -e $full_spec)	# may have been deleted in recursion
{
$ans = ASK_YNAE( " - Delete: $spec?", 'Y')
if ($ans ne 'A');
last if ($ans eq 'E');
if ($ans eq 'A' || $ans eq 'Y')
{
my ($nr_dirs_deleted, $nr_files_deleted, $total_scanned) = (0,0,0);
if ($type eq 'D')
{

($nr_dirs_deleted, $nr_files_deleted, $total_scanned) = FILE_del_tree( $full_spec, undef, 0);
if ($nr_dirs_deleted + $nr_files_deleted == $total_scanned)
{
$nr_dirs_deleted++
if (rmdir( $full_spec));
}
$total_scanned++;
} else
{

$total_scanned = 1;
$nr_files_deleted++
if (unlink( $full_spec));
}
my $total_deleted = $nr_dirs_deleted + $nr_files_deleted;
ENV_say( 2, "   $total_deleted/$total_scanned deleted ($spec)");
$is_ok = 0
if ($total_deleted != $total_scanned);
} else
{
$is_ok = 0;
}
}
}
}
}

return $is_ok;
}




sub check_missing($)
{
my ($line_refs_ref) = @_;







my $dirs_ok = 1;
my $files_ok = 1;




ENV_say( 1, "Collecting missing Directories/Files...");
my @missing_dir_refs;
my @missing_file_refs;
foreach my $ref (@{$line_refs_ref})
{
my ($type, $avail, undef, $spec) = @{$ref};
if ($avail eq 'm')
{
if ($type eq 'D')
{
push @missing_dir_refs, $ref;
} else
{
push @missing_file_refs, $ref;
}
}
}




if (@missing_dir_refs)
{
$dirs_ok = 0;
ENV_say( 1, "Missing Directories:");
map {  ENV_say( 0, " - $_->[3]") } @missing_dir_refs;
my $count = @missing_dir_refs;
ENV_say( 1, "$count Directories missing");
if (ASK_YN( 'Create missing directories?', 'Y') eq 'Y')
{
ENV_say( 1, "Creating directories...");
foreach my $ref (@missing_dir_refs)
{
my ($type, $avail, $scm, $spec) = @{$ref};
my $full_spec = "$GBS_ROOT_PATH/$ref->[3]";

ENV_mkdir( $full_spec);
$ref->[1] = $avail = 'M';
add_scm( $type, $spec)
if ($scm);
}
$dirs_ok = 1;
}
} else
{
ENV_say( 1, "All required Directories are present.");
}




if (@missing_file_refs)
{
$files_ok = 0;
ENV_say( 1, "Missing Files:");
map { ENV_say( 0, " - $_->[3]") } @missing_file_refs;
my $count = @missing_file_refs;
ENV_say( 1, "$count Files missing",
"Please create files with the gbsedit command");
} else
{
ENV_say( 1, "All required Files are present.");
}

return ($dirs_ok && $files_ok) ? 1 : 0;
}




sub check_scm_control($)
{
my ($line_refs_ref) = @_;






ENV_say( 1, "Checking directories/files under SCM control may take a VERY long time",
"You may want to run a cleanup first...");
if (ASK_YN( 'Check directories/files under SCM control?', 'N') eq 'Y')
{
ENV_chdir( $GBS_ROOT_PATH);
GBSSCM_connect();

ENV_say( 1, "Caching file_states...");
SCM_cache_states( $GBS_ROOT_PATH);

ENV_say( 1, "Checking Directories/Files under SCM control...");
my @scm_not_allowed_refs;	# [ $type, $spec ]
my @not_scm_dir_refs;		# [ $state, $spec ]
my @not_scm_file_refs;		# [ $state, $spec ]
my %dirs_not_scm;
my $previous_display = '';
ENV_say( 1, "...");
foreach my $ref (@{$line_refs_ref})
{
my ($type, $avail, $scm, $spec) = @{$ref};







if ($spec =~ m|(.+?)/.+|)
{
my $main_dir = $1;

if ($main_dir eq 'dev')
{
if ($spec =~ m|(dev/.+?)/.+|)
{
my $subsys_dir = $1;

ENV_say( 1, "$subsys_dir...")
if ($subsys_dir ne $previous_display);
$previous_display = $subsys_dir;
} else
{
ENV_say( 1, "$main_dir...")
if ($main_dir ne $previous_display);
$previous_display = $main_dir;
}
} else
{
ENV_say( 1, "$main_dir...")
if ($main_dir ne $previous_display);
$previous_display = $main_dir;
}
}
if (($avail eq 'M' || $avail eq 'm' || $avail eq 'O') && $scm >= 0)
{
my ($parent) = $spec =~ m|(.*)/.+$|;



my $state;
if ($parent && exists( $dirs_not_scm{$parent}))
{

$state = -1;
} else
{
my $full_spec = "$GBS_ROOT_PATH/$spec";
$state = SCM_get_states( $full_spec);    #  -2 = nofile, -1 = notscm, 0 = Ci  1 = co
}

if ($scm == 0)
{



if ($state >= 0)	# ci or co
{

push @scm_not_allowed_refs, [ $type, $spec ];
} else		# not_scm
{
$dirs_not_scm{$spec} = 1
if ($type eq 'D');
}
} else # $scm = 1 or -1:  scm mandory if exists
{
if ($state >= 0)	# ci or co
{

} else		# not_exist = -2 not_scm = -1
{
if ($type eq 'D')
{
push @not_scm_dir_refs, [ $state, $spec ];
$dirs_not_scm{$spec} = 1;
} else
{
push @not_scm_file_refs, [ $state, $spec ];
}
}
}
}
}




if (@not_scm_dir_refs)
{
my $count = @not_scm_dir_refs;
ENV_say( 1, "$count Directories not under SCM control:");
foreach my $ref (@not_scm_dir_refs)
{
my ($state, $spec) = @{$ref};
my $txt = ($state == -2) ? 'Missing' : 'Not SCM';
ENV_say( 0, " - $txt: $spec");
}
if (ASK_YN( 'Add missing Directories? (selective)', 'Y') eq 'Y')
{
my $ans = '';
foreach my $ref (@not_scm_dir_refs)
{
my ($state, $spec) = @{$ref};
$ans = ASK_YNAE( " - Add: $spec?", 'Y')
if ($ans ne 'A');
last if ($ans eq 'E');
if ($ans eq 'A' || $ans eq 'Y')
{
add_scm( 'D', $spec);
}
}
}
} else
{
ENV_say( 1, "All Directories under SCM control");
}




if (@not_scm_file_refs)
{
my $count = @not_scm_file_refs;
ENV_say( 1, "$count Files not under SCM control:");
my @non_scm_files;
foreach my $ref (@not_scm_file_refs)
{
my ($state, $spec) = @{$ref};
my $txt = ($state == -2) ? 'Missing' : 'Not SCM';
ENV_say( 0, " - $txt: $spec");
push @non_scm_files, $spec
if ($state == -1);
}
if (@non_scm_files)
{
if (ASK_YN( 'Add Non SCM Files? (selective)', 'Y') eq 'Y')
{
my $ans = '';
foreach my $spec (@non_scm_files)
{
$ans = ASK_YNAE( " - Add: $spec?", 'Y')
if ($ans ne 'A');
last if ($ans eq 'E');
if ($ans eq 'A' || $ans eq 'Y')
{
add_scm( undef, $spec);
}
}
}
}
my $nr_missing_files = $count - @non_scm_files;
if ($nr_missing_files > 0)
{
ENV_say( 1, "$nr_missing_files Files missing",
"Please create files with the gbsedit command");
}
} else
{
ENV_say( 1, "All Files under SCM control");
}




if (@scm_not_allowed_refs)
{
my $count = @scm_not_allowed_refs;
ENV_say( 1, "$count Directories/Files illegaly under SCM control:");
foreach my $ref (@scm_not_allowed_refs)
{
my ($type, $spec) = @{$ref};
ENV_say( 0, " - $type: $spec");
}
} else
{
ENV_say( 1, "No Directories/Files illegaly under SCM control");
}

ENV_say( 1, "Done");
}
}




sub add_scm($$)
{
my ($type,		# 'F', 'D' or undef (existing, add only!)
$spec,
) = @_;

my $full_spec = "$GBS_ROOT_PATH/$spec";

$type = (-d $full_spec) ? 'D' : 'F'
if (!defined $type);




my $parent_dir = ENV_parent_path( $full_spec);
SCM_add_dirs( $parent_dir, \&FILES_get_ignore_spec, 0, 1);




if ($type eq 'D')
{
SCM_add_dirs( $full_spec, \&FILES_get_ignore_spec, 1, 1);	# $prevalidated, $verbose
} else
{
SCM_add_txt_files( $full_spec, 1, 1);				# $prevalidated, $verbose
}
}





sub fix_execute_bit()
{
if (ASK_YN( 'Fix execute-bit of .sh files? (Unix/Linux only)', 'N') eq 'Y')
{
ENV_chdir( $GBS_ROOT_PATH);
my $command = '\find . -name "*.sh" -exec \chmod +x {} \;';
ENV_say( 1, "$command");
ENV_system( $command, 0, undef);
ENV_say( 1, "Done");
}
}

return 1;


