#===============================================================
#
#   fspec.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package fspec;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
FSPEC_gen_type
FSPEC_audit_target
FSPEC_export
FSPEC_plugin_file
FSPEC_scope
FSPEC_owners
FSPEC_system
FSPEC_steps
FSPEC_switch
FSPEC_sys
FSPEC_subsys_script
FSPEC_subsys_script_names
FSPEC_full_gbs
FSPEC_gbsdb
FSPEC_gbsall
);

use strict;
use integer;

use glo::env;




sub FSPEC_gen_type($$$$$);
sub FSPEC_audit_target($$$$$$);
sub FSPEC_export($$$$);
sub FSPEC_plugin_file($$$);
sub FSPEC_scope($$$$);
sub FSPEC_owners($$);
sub FSPEC_system($$);
sub FSPEC_steps($$);
sub FSPEC_switch($$);
sub FSPEC_sys($$);
sub FSPEC_subsys_script($$$);
sub FSPEC_subsys_script_names();
sub FSPEC_full_gbs($$);
sub FSPEC_gbsdb();
sub FSPEC_gbsall($$);

sub set_file_spec($$);




my $COMMAND_FILE_EXT = ENV_commandfile_ext();

my %NON_GBS_SCRIPTS =
(
gbssubaudit		=> 1,
gbssubcleanup_aud	=> 1,
gbssubcleanup_bld	=> 1,
gbssubgen		=> 1,
gbssubmake		=> 1,
);








sub FSPEC_gen_type($$$$$)
{
my ($file,		    # incs_<type>.gbs, incs_<type>.usr, etc. see above
$root,
$subsys,
$component,
$target) = @_;


my $dir = $root;
if ($subsys)
{
$dir .= "/dev/$subsys";
if ($component)
{
$dir .= "/comp/$component/opt";
} else
{
$dir .= '/gen';
}
} else
{
$dir .= '/sysgen';
}
if ($target)
{
$dir .= "/$target";
}

return (wantarray) ? ($dir, $file) : "$dir/$file";
}





sub FSPEC_audit_target($$$$$$)
{
my ($file,		    # sca_<target>, etc. see above
$root,
$subsys,
$component,
$audit,
$target) = @_;


my $dir = $root;
if ($subsys)
{
$dir .= "/dev/$subsys";
if ($component)
{
$dir .= "/comp/$component/opt";
} else
{
$dir .= '/audit';
}
} else
{
$dir .= '/sysaudit';
}
if ($audit)
{
$dir .= "/$audit";
}

return (wantarray) ? ($dir, $file) : "$dir/$file";
}





sub FSPEC_export($$$$)
{
my ($file,		    # with or without .gbs or .usr
$root,
$subsys,
$component,
) = @_;


my $dir;
if (!defined $component)
{
$dir = "$root/dev/$subsys";
} elsif ($component eq '')
{
$dir = "$root/dev/$subsys/gen";
} else
{
$dir = "$root/dev/$subsys/comp/$component";
}

return set_file_spec( $dir, $file);
}







sub FSPEC_plugin_file($$$)
{
my ($file,			# e.g. target.gbs, audit.usr, target
$root,
$plugin_item,		# e.g. $GBS_TARGET
) = @_;


my $dir;
{
my $file_name = ENV_split_spec_n( $file);
if ($file_name  eq 'target')
{
$dir = "$root/sysgen/$plugin_item";
} elsif ($file_name eq 'audit')
{
$dir = "$root/sysaudit/$plugin_item";
} elsif ($file_name eq 'tool')
{
$dir = "$root/systools/$plugin_item";
} else
{
ENV_fatal( "Invalid plugin_file '$file'");
}
}

return set_file_spec( $dir, $file);
}





sub FSPEC_scope($$$$)
{
my ($file,		    # with or without .gbs or .usr
$root,
$subsys,
$component,
) = @_;


return set_file_spec( "$root/dev/$subsys/comp/$component", $file);
}




sub FSPEC_owners($$)
{
my ($file,		    #  with or without .gbs or .usr
$root,
) = @_;


return set_file_spec( "$root/sys", $file);
}




sub FSPEC_system($$)
{
my ($file,		    #  with or without .gbs or .usr
$root,
) = @_;


return set_file_spec( $root, $file);
}





sub FSPEC_steps($$)
{
my ($file,		    # 'steps', with or without .gbs or .usr
$root,
) = @_;


return set_file_spec( "$root/sys", $file);
}






sub FSPEC_switch($$)
{
my ($file,			# switch or


$root) = @_;


my $dir = $root;
$file .= (-e "$dir/$file.usr") ? '.usr' : '.gbs'
if ($file eq 'switch');
$file .= $COMMAND_FILE_EXT
if ($file !~ /(\.sh$)|(\.bat$)/);

return (wantarray) ? ($dir, $file) : "$dir/$file";
}





sub FSPEC_sys($$)
{
my ($file,		    # 'sysgen', with or without .gbs or .usr
$root,
) = @_;


return set_file_spec( "$root/sys", $file);
}






sub FSPEC_subsys_script($$$)
{
my ($file,	    	# gbssubgen, gbssubmake, etc with or without .sh or .bat
$root,
$subsys,
) = @_;


my ($name, $type) = ENV_split_spec_nt( $file);
ENV_fatal( "No such non-gbs script name '$name'")
if (!exists $NON_GBS_SCRIPTS{$name});

my $dir = "$root/dev/$subsys";
$file .= $COMMAND_FILE_EXT
if ($type eq '');

return (wantarray) ? ($dir, $file) : "$dir/$file";
}




sub FSPEC_subsys_script_names()
{
return sort( keys( %NON_GBS_SCRIPTS));
}





sub FSPEC_full_gbs($$)
{
my ($root,
$subsys,
) = @_;


my $dir = "$root/dev/$subsys";
my $file = 'full.gbs';

return (wantarray) ? ($dir, $file) : "$dir/$file";
}




sub FSPEC_gbsdb()
{
my $gbs_base_dir = ENV_getenv_perl_path( 'GBS_BASE_PATH');
my $file = 'gbsdb.dat';

return (wantarray) ? ($gbs_base_dir, $file) : "$gbs_base_dir/$file";
}




sub FSPEC_gbsall($$)
{
my ($file_type,	    # full_name+type or .sh or .bat or ''
$in_install,	    # In GBS_SCRIPTS_ROOT else in GBS_BASE_PATH
) = @_;

my $path = ($in_install) ? ENV_getenv_perl_path( 'GBS_SCRIPTS_ROOT') : ENV_getenv_perl_path( 'GBS_BASE_PATH');
my $file;

my ($name, $type) = ENV_split_spec_nt( $file_type);
ENV_fatal( "Filename in '$file_type' must be 'gbsall'")
if ($name ne '' && $name ne 'gbsall');
if ($type eq '')
{
$file = "gbsall$COMMAND_FILE_EXT";
} else
{
ENV_fatal( "Filetype in '$file_type' must be .bat or .sh")
if ($type ne '.bat' && $type ne '.sh');
$file = "gbsall$type";
}

return (wantarray) ? ($path, $file) : "$path/$file";
}




sub set_file_spec($$)
{
my ($dir,
$file,	    # type == .gbs or .usr or ''
) = @_;

if (index( $file, '.') == -1)
{
$file .= (-e "$dir/$file.usr") ? '.usr' : '.gbs';
}

return (wantarray) ? ($dir, $file) : "$dir/$file";
}

return 1;

