#==============================================================
#
#   gbs.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbs;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBS_init
GBS_validate
GBS_set_title
);

use strict;
use integer;

use glo::env;
use glo::shell;
use glo::build;
use gbsenv;
use gbscmd;




sub GBS_init($);
sub GBS_validate();
sub GBS_set_title(@);




my $GBS_BASE_PATH = ENV_getenv_perl_path( 'GBS_BASE_PATH');
ENV_fatal( "GBS_BASE_PATH not defined")
if ($GBS_BASE_PATH eq '');





sub GBS_init($)
{
my ($wanted_rel,
) = @_;
my @lines;

$wanted_rel = ''
if ($wanted_rel eq '.');
if ($wanted_rel eq '')
{
ENV_say( 1, "Initializing GBS...");
} else
{
ENV_say( 1, "Initializing GBS ($wanted_rel)...");
}




foreach my $env (qw( GBS_SCRIPTS_PATH GBS_SCRIPTS_REL GBS_SCRIPTS_PATH))
{
ENV_fatal( "'$env' not defined")
if (ENV_getenv( $env) eq '');
}




my $gbs_scripts;
if ($wanted_rel ne '')
{



my $gbs_scripts_root = ENV_getenv_perl_path( 'GBS_SCRIPTS_ROOT');
my $new_gbs_scripts = "$gbs_scripts_root/$wanted_rel";
ENV_error_exit( "No such GBS_SCRIPTS_REL ($wanted_rel) in $gbs_scripts_root")
if (!-d $new_gbs_scripts);
$gbs_scripts = $new_gbs_scripts;
ENV_setenv( GBS_SCRIPTS_REL => $wanted_rel);
ENV_set_application( undef, $gbs_scripts, undef, undef, undef);   # only update scripts
if ($new_gbs_scripts eq $gbs_scripts)
{
ENV_say( 1, "GBS_SCRIPTS_REL '$wanted_rel' was already current")
unless ($wanted_rel eq 'beta');
} else
{
ENV_say( 1, "GBS_SCRIPTS_REL set to '$wanted_rel'");
}
} else
{
$gbs_scripts = GBSENV_perl_gbs_scripts();
}
GBSENV_setenv( GBS_SCRIPTS_PATH => ENV_delink_spec( $gbs_scripts));

my $gbs_version = BUILD_get_version();
my $gbs_build = BUILD_version_2_build( $gbs_version, BUILD_get_version_date());
GBSENV_setenv( GBS_VERSION => $gbs_version);
GBSENV_setenv( GBS_BUILD => $gbs_build);






my $gbs_home = ENV_getenv( 'GBS_HOME');
if ($gbs_home eq '')
{
$gbs_home = ENV_get_application_home_path();
GBSENV_setenv( GBS_HOME => $gbs_home);

}




GBSENV_setenv( GBS_PLATFORM => $^O);

GBSENV_setenv( GBS_ROOT_PATH => '');
GBSENV_setenv( GBS_SUBSYS => '');
GBSENV_setenv( GBS_COMPONENT => '');
GBSENV_setenv( GBS_TARGET => '');

GBSENV_setenv( GBS_PID => '$$')
if (ENV_is_lunix());








{
my $site_gbsall_file = ENV_commandfile_spec( "$gbs_scripts/../gbsall");
if (-f $site_gbsall_file)
{
push @lines, SHELL_echo( "Executing Install gbsall..");
push @lines, SHELL_source( $site_gbsall_file);
}

my $base_gbsall_file = ENV_commandfile_spec( "$GBS_BASE_PATH/gbsall");
if (-f $base_gbsall_file)
{
push @lines, SHELL_echo( "Executing Base gbsall..");
push @lines, SHELL_source( $base_gbsall_file);
}
}








my $gbs_exec_mode = ENV_getenv( 'GBS_EXEC_MODE');
if ($gbs_exec_mode eq '')
{
$gbs_exec_mode = GBSENV_setenv( GBS_EXEC_MODE => 'INTERACTIVE');	# initial
}






eval( "use Tkx;");
if ($@)
{

ENV_say( 1, "Perl Package 'Tkx' not found");
GBSENV_setenv( GBS_GUI => 'None');
} else
{
GBSENV_setenv( GBS_GUI => 'Tkx');
}




push @lines, GBSENV_changed_setenv_commands( 0);


if ($gbs_exec_mode eq 'INTERACTIVE')
{



if (ENV_is_win32())
{

push @lines, "doskey /insert /macrofile=%GBS_SCRIPTS_PATH%\\gbsdef.doskey";
} else
{

push @lines, SHELL_source( "$gbs_scripts/gbsdef.sh");
}
}

return @lines;
}





sub GBS_validate()
{



my $gbs_log_dir = ENV_getenv_perl_path( 'GBS_LOG_PATH');
ENV_error_exit( "'GBS_LOG_PATH' not set")
if ($gbs_log_dir eq '');
ENV_error_exit( "GBS_LOG_PATH ($gbs_log_dir) does not exist!")
if (! -d $gbs_log_dir);
ENV_error_exit( "GBS_LOG_PATH ($gbs_log_dir) is not writable!")
if (! -w $gbs_log_dir);




my $gbs_site = ENV_getenv_perl_path( 'GBS_SITE');
ENV_error_exit( "'GBS_SITE' not set")
if ($gbs_site eq '');
}





sub GBS_set_title(@)
{
my @text = @_;
my @lines;

if (!GBSENV_mode_is_background())	    # INTERACTIVE or FOREGROUND
{
my $text = (@text) ? join ( ' - ', @text) : '-';
push @lines, GBSCMD_set_xtitle( "GBS: $text");
}

return @lines;
}

return 1;


