#===============================================================
#
#   gbsbg.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbsbg;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSBG_banner
GBSBG_echo
GBSBG_set
GBSBG_exec_short
GBSBG_exec
GBSBG_glkb
GBSBG_sca
GBSBG_batch_init
GBSBG_batch_run
GBSBG_batch_queue
GBSBG_batch_wait
GBSBG_batch_finish
);

use strict;
use integer;

use glo::env;
use glo::spit;
use gbsenv;
use exec;
use fail;




sub GBSBG_banner(@);
sub GBSBG_echo(@);
sub GBSBG_set($$);
sub GBSBG_exec_short($$$$);
sub GBSBG_exec($$$@);
sub GBSBG_glkb($$$$@);
sub GBSBG_sca($$$@);
sub GBSBG_batch_init($);
sub GBSBG_batch_run($$$$);
sub GBSBG_batch_queue($$$$);
sub GBSBG_batch_wait();
sub GBSBG_batch_finish();

sub get_command_data_lines($);
sub set_batch_type($);
sub create_gbsbg_lines($@);








my @COMMAND_FILE_TYPES = (ENV_is_win32()) ? qw(.bat) : qw(.sh);

my %BATCH_PROGS = (

exec    => 'gbsbgexec.pl',
sca	    => 'gbsbgsca.pl',
scasum  => 'gbsbgexec.pl',	#   intend to change to gbsbgscasum
);




my $BATCH_TYPE;	    # exec, sca, scasum

my @QUEUED_FILES;




sub GBSBG_banner(@)
{
my @lines = @_;

return create_gbsbg_lines( BANNER => @lines);
}




sub GBSBG_echo(@)
{
my (@lines) = @_;

return create_gbsbg_lines( ECHO => @lines);
}




sub GBSBG_set($$)
{
my ($name,
$value,
) = @_;


return create_gbsbg_lines( SET => $name, $value);
}




sub GBSBG_exec_short($$$$)
{
my ($file,			# for failure logging purposes
$must_print_commands,	# undef = ok
$out_files_ref,		# undef = ok
$command_ref,
) = @_;

$must_print_commands = '-'
if (!defined $must_print_commands);
my @out_files = (defined $out_files_ref) ? @{$out_files_ref} : ();
my @command_data_lines = get_command_data_lines( [ [ 1, undef, ENV_command_line( $command_ref), '0', '-'] ]);




set_batch_type( 'exec');

return create_gbsbg_lines( EXEC => $file, $must_print_commands, "@out_files", @command_data_lines );
}




sub GBSBG_exec($$$@)
{
my ($file,			# for failure logging purposes
$must_print_commands,	# undef = ok
$out_files_ref,		# undef = ok
@command_data_refs,     # [ $command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition ]

) = @_;

$must_print_commands = '-'
if (!defined $must_print_commands);
my @out_files = (defined $out_files_ref) ? @{$out_files_ref} : ();
my @command_data_lines = get_command_data_lines( \@command_data_refs);


set_batch_type( 'exec');

return create_gbsbg_lines( EXEC => $file, $must_print_commands, "@out_files", @command_data_lines );
}




sub GBSBG_glkb($$$$@)
{
my ($file,			# for failure logging purposes
$must_print_commands,	# undef = OK
$glkb_args_ref,		# [ $gen_type, $glkb_type, $glkb_file, $out_file_spec ]
$out_files_ref,		# undef = OK
@command_data_refs,     # [ $command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition ]

) = @_;

$must_print_commands = '-'
if (!defined $must_print_commands);
my @glkb_args = @{$glkb_args_ref};
my @out_files = (defined $out_files_ref) ? @{$out_files_ref} : ();
my @command_data_lines = get_command_data_lines( \@command_data_refs);


set_batch_type( 'exec');

return create_gbsbg_lines( GLKB => $file, $must_print_commands, "@glkb_args", "@out_files", @command_data_lines);
}




sub GBSBG_sca($$$@)
{
my ($file,			# for failure logging purposes
$unix_style_search,
$out_files_ref,		# undef = ok
@command_data_refs,     # [ $command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition ]

) = @_;

my @out_files = ($out_files_ref) ? @{$out_files_ref} : ();
my @command_data_lines = get_command_data_lines( \@command_data_refs);


set_batch_type( 'sca');

return create_gbsbg_lines( SCA => $file, $unix_style_search, "@out_files", @command_data_lines );
}




sub GBSBG_batch_init($)
{
my ($max_parallel_jobs) = @_;









$BATCH_TYPE = '';

EXEC_batch_init( $max_parallel_jobs);
}




sub GBSBG_batch_run($$$$)
{
my ($batch_name,	# gbsgen_post, gbsaudit_post
$component,	# '-' == none
$exec_data,	# '-' == undef
$gbsbg_lines_ref,
) = @_;
my $rc;

ENV_fatal( "BATCH_type was not initialised", @{$gbsbg_lines_ref})
if ($BATCH_TYPE eq '');


map { $_ = "$_\n" } @{$gbsbg_lines_ref};
my $gbsbg_file = SPIT_tmp_file( "gbsbg_batch_$batch_name", $gbsbg_lines_ref);

my $batch_prog = $BATCH_PROGS{$BATCH_TYPE};	    # e.g.: 'gbsbg.pl'
$rc = EXEC_batch_run( ENV_command_ref( $batch_prog, [ $batch_name, $component, $exec_data, $gbsbg_file ]));
$BATCH_TYPE = '';

unlink $gbsbg_file;

return $rc;
}




sub GBSBG_batch_queue($$$$)
{
my ($batch_name,	# gbsgen gbs_audit
$component,	# '-' == none
$exec_data,	# '-' == undef
$gbsbg_lines_ref,
) = @_;
my $queued;		    # 1 = queued, 0 == not queued because of failed steps

ENV_fatal( "BATCH_type was not initialised", @{$gbsbg_lines_ref})
if ($BATCH_TYPE eq '');

map { $_ = "$_\n" } @{$gbsbg_lines_ref};
my $gbsbg_file = SPIT_tmp_file( "gbsbg_batch_$batch_name", $gbsbg_lines_ref);
push @QUEUED_FILES, $gbsbg_file;

my $batch_prog = $BATCH_PROGS{$BATCH_TYPE};	    # e.g.: 'gbsbg.pl'
$queued = EXEC_batch_queue( ENV_command_ref( $batch_prog, [ $batch_name, $component, $exec_data, $gbsbg_file ]));
$BATCH_TYPE = '';

return $queued;
}




sub GBSBG_batch_wait()
{
my $rc;

$rc = EXEC_batch_wait();

unlink @QUEUED_FILES;
@QUEUED_FILES = ();

return $rc;
}




sub GBSBG_batch_finish()
{
if (GBSENV_get_exec_level() == 0)	# may have be called by gbsmake->gbsgen!
{



my @lines = FAIL_read();
if (@lines)
{
ENV_say( 1, "Failed:");
map { ENV_say( 0, "###  $_") } @lines;
unlink FAIL_filespec()
if (!GBSENV_mode_is_background());
}
}


set_batch_type( undef);
}




sub get_command_data_lines($)
{
my ($command_data_refs_ref) = @_;
my @command_data_lines;

foreach my $ref (@{$command_data_refs_ref})
{
my ($command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition) = @{$ref};
if ($command_type != 2)	# Special command (e.g.: echo)
{
my ($command, $args) = split( ' ', $command_line, 2);   #
my $command_ref = ENV_command_ref( $command, undef);    # TBS Move to gbssfile???
$command_line = ENV_command_line( $command_ref);	    #
$command_line = "$command_line $args"		    #
if (defined $args);				    #
}
push @command_data_lines, $command_type;
push @command_data_lines, $command_line;
push @command_data_lines, $ok_values;
push @command_data_lines, $exec_condition;
}

return @command_data_lines;
}




sub set_batch_type($)
{
my ($type) = @_;

ENV_fatal( "GBSBG_batch_init was not called")
if (!defined $BATCH_TYPE);

if ($BATCH_TYPE eq '')
{
$BATCH_TYPE = $type;
} elsif ($BATCH_TYPE eq $type)
{

} else
{
ENV_fatal( "BATCH_TYPE mismatch ($BATCH_TYPE <- $type)");

}
}




sub create_gbsbg_lines($@)
{
my ($name,
@lines
) = @_;

if (@lines)
{
return ( $name, scalar @lines, @lines);
} else
{
return ();
}
}

return 1;


