#===============================================================
#
#   gbsbldaudit.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSBLDAUDIT*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::genopt;
use gbsenv;
use gbsglo;
use target;
use validate;




sub audit_subsys($);
sub audit_component($$);
sub audit_file($$$);




my $GBS_ROOT_PATH;

my $CUR_TARGET;






$| = 1;
my $RC = 0;
GBSENV_init();




my @genopts = (
[ '<*>', 'subsystems',  'sao', "", "Specific SubSystems" ],
[ 'm' , 'max_missing',  'iso', 50, "Cancel after this many missing files" ],
[ 'targets', 'targets', 'sao', '', "(wild-)Targets to Build-Audit. '' or '*' == All, '.' == Current" ],
);
GENOPT_set_optdefs( 'gbsbldaudit', \@genopts,
'Check if all files have generated',
undef);
GENOPT_parse();
my @opt_subsystems = GENOPT_get( 'subsystems');
my $MAX_MISSING = GENOPT_get( 'max_missing');
my @targets = GENOPT_get( 'targets');




$GBS_ROOT_PATH = VALIDATE_root();
@targets = ('*') if (!@targets);
@targets = VALIDATE_targets( @targets);
ENV_say( 1, "TARGETS = @targets");

my @SRC_TYPES;
my %BLD_TYPES;
my %BLD_FILES;

my @MISSING;


foreach my $target (@targets)
{
ENV_say( 1, "TARGET = $target");
$CUR_TARGET = $target;

my $gbs_subsys = ENV_getenv( 'GBS_SUBSYS');
my @subsystems;
if (@opt_subsystems)
{
foreach my $subsys (@opt_subsystems)
{
$subsys = $gbs_subsys if ($subsys eq '.');
ENV_error_exit( "No current SubSystem") if ($subsys eq '');
my $subsys_root = "$GBS_ROOT_PATH/dev/$subsys";
ENV_error_exit( "No such SubSystem: '$subsys") if (!-d $subsys_root);
ENV_error_exit( "SubSystem '$subsys' not Full GBS") if (!GBSGLO_subsystem_is_full_gbs( $subsys));
if (-d "$subsys_root/gen/$target")
{
push @subsystems, $subsys;
} else
{
ENV_warn( "SubSystem '$subsys' does not generate for Target '$target'");
}
}
} else
{
foreach my $subsys (GBSGLO_subsystems())
{
if (-f GBSGLO_subsystem_is_full_gbs( $subsys))
{
push @subsystems, $subsys;
} else
{
ENV_say( 1, "SubSystem '$subsys' not Full GBS - Skipped")
}
}
}




@SRC_TYPES = TARGET_get_src_types( $target);
%BLD_TYPES = ();
foreach my $src_type (@SRC_TYPES)
{
$BLD_TYPES{$src_type} = [ TARGET_get_src_items( $target, $src_type, 'OUT_TYPES') ];
$BLD_FILES{$src_type} = [ TARGET_get_src_items( $target, $src_type, 'OUT_FILES') ];
}




ENV_say( 1, "Begin $target");
my $target_checked = 0;
my $missing = 0;
subsystems: foreach my $subsys (@subsystems)
{
my $gen_root = "$GBS_ROOT_PATH/dev/$subsys/gen";
if (-d "$gen_root/$target")
{
ENV_say( 1, "SubSystem: $subsys...");
$missing += audit_subsys( $subsys);
$target_checked = 1;
} else
{

}
if ($missing > $MAX_MISSING)
{
ENV_error( "$target: More than $MAX_MISSING missing files. Audit cancelled!");
last subsystems;
}
}
ENV_say( 1, "$target: Missing: $missing file(s)");
if ($target_checked)
{
push @MISSING, [$target, $missing ];
} else
{
push @MISSING, [$target, undef ];
}
}




ENV_say( 1, "Verdict:");
foreach my $ref (@MISSING)
{
my ($target, $missing) = @{$ref};
ENV_print( 2, "Target: '$target': ");
if (defined $missing)
{
if ($missing > 0)
{
my $at_least = ($missing < $MAX_MISSING) ? '' : 'At least ';
ENV_say( 0, "Failed. $at_least$missing file(s) missing");
$RC = 2;
} else
{
ENV_say( 0, 'OK');
}
} else
{
ENV_say( 0, 'Not Checked');
$RC = 1;
}
}

ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}




sub audit_subsys($)
{
my ($subsys) = @_;
my $missing = 0;

my $comp_root = "$GBS_ROOT_PATH/dev/$subsys/comp";
foreach my $component (GBSGLO_components( $subsys))
{
my $src_dir = "$comp_root/$component/src";
if (-d $src_dir)
{
my $target_dir = "$comp_root/$component/bld/$CUR_TARGET";
if (-d $target_dir)
{
$missing += audit_component( $subsys, $component);
} else
{
ENV_say( 1, "Component '$component' does not generate for target '$CUR_TARGET' - Skipped");
}
} else
{
ENV_warn( "Component $component is not a GBS component!");
}
}

return $missing;
}




sub audit_component($$)
{
my ($subsys,
$component,
) = @_;
my $missing = 0;

my $src_dir = "$GBS_ROOT_PATH/dev/$subsys/comp/$component/src";
ENV_chdir( $src_dir);
foreach my $src_file (glob '*.*')
{
$missing += audit_file( $subsys, $component, $src_file);
}

return $missing;
}




sub audit_file($$$)
{
my ($subsys,
$component,
$src_file,
) = @_;
my $missing = 0;

my ($name, $src_type) = ENV_split_spec_nt( $src_file);
my $src_spec = "${subsys}::${component}:$src_file";
if (GBSGLO_exclude_direntries( $src_file))
{
ENV_say( 1, "File '$src_spec' excluded");
} elsif (!grep $src_type eq $_, @SRC_TYPES)
{
ENV_warn( "Invalid src-type in file '$src_spec'");
} else
{

my $bld_target_root = "$GBS_ROOT_PATH/dev/$subsys/comp/$component/bld/$CUR_TARGET";
my $bld_prefix = "$bld_target_root/$name";




my @out_types = @{$BLD_TYPES{$src_type}};
my $primary_out_type = $out_types[0];
if (-f "$bld_prefix$primary_out_type")
{
foreach my $out_type (@out_types[1..$#out_types])
{

if (!-f "$bld_prefix$out_type")
{
ENV_say( 1, "$CUR_TARGET: '$src_spec'  => Secondary Bld ($name$out_type) not found");
$missing++;
}
}
} else
{
ENV_say( 1, "$CUR_TARGET: '$src_spec'  => Primary Bld ($primary_out_type) not found");
$missing++;
}




foreach my $out_file (@{$BLD_FILES{$src_type}})
{

if (!-f "$bld_target_root/$out_file")
{
ENV_say( 1, "$CUR_TARGET: '$src_spec'  => Bld File ($out_file) not found");
$missing++;
}
}
}
return $missing;
}


