#===============================================================
#
#   gbscmd.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbscmd;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSCMD_command
GBSCMD_base_command
GBSCMD_batch
GBSCMD_submit
GBSCMD_subwin
GBSCMD_make
GBSCMD_view_file
GBSCMD_browse_file
GBSCMD_edit_file
GBSCMD_set_xtitle
GBSCMD_bg_notifier
);

use strict;
use integer;

use glo::env;
use glo::shell;
use gbsenv;




sub GBSCMD_command($@);
sub GBSCMD_base_command($@);
sub GBSCMD_batch($);
sub GBSCMD_submit($$$);
sub GBSCMD_subwin($$);
sub GBSCMD_make();
sub GBSCMD_view_file($);
sub GBSCMD_browse_file($);
sub GBSCMD_edit_file($$);
sub GBSCMD_set_xtitle($);
sub GBSCMD_bg_notifier();

sub get_env_command_ref($$@);
sub get_target_env($);
sub substitute_args($$$$);




my $IS_WIN32 = ENV_is_win32();
my $IS_LUNIX = !$IS_WIN32;

my %ENV_COMMAND_REFS;	    # Do not inilialize!


my $GBS_XTITLE; 	    # Do not initialize!

my %TARGET_ENVS;





my $DEFAULT_WIN32_EDITOR  = ENV_getenv( 'WINDIR') . '\notepad.exe %1%';
my $DEFAULT_LINUX_EDITOR  = (ENV_is_linux()) ? 'gedit $1' : 'vi $1';
my $DEFAULT_LUNIX_BROWSER =  $ENV{GBS_SCRIPTS_PATH} . '/glo/run_mozilla.sh firefox $1';
my %GEN_DEFS = (

GBS_VIEWER	=> [ undef,    $DEFAULT_WIN32_EDITOR, 'less $1' ],
GBS_BROWSER	=> [ undef,    'start %1%',           $DEFAULT_LUNIX_BROWSER ],
GBS_EDITOR	=> [ 'EDITOR', $DEFAULT_WIN32_EDITOR, $DEFAULT_LINUX_EDITOR ],
);
my %COMMANDS_REQ_SHELL = (
less    => 1,
more    => 1,
);

my $OS_GBS_PERL_CMD = ENV_getenv( 'GBS_PERL_CMD');
my $OS_GBS_WPERL_CMD = ENV_getenv( 'GBS_WPERL_CMD');
my $OS_GBS_SCRIPTS = ENV_getenv( 'GBS_SCRIPTS_PATH');
my $SCRIPT_FILETYPE = ENV_commandfile_ext();
my %GBS_COMMAND_REFS = (


gbs		=> [ 's', 1],
gbsaudit	=> [ 's', 1 ],
gbsbldaudit	=> [ 'p', 0 ],
gbsedit	=> [ 'p', 0 ],
gbsexit	=> [ 's', 1 ],
gbsexport	=> [ 'p', 0 ],
gbsgen	=> [ 'p', 0 ],
gbshelp	=> [ 'p', 0 ],
gbslocate	=> [ 's', 1 ],
gbsmaint	=> [ 's', 1 ],
gbsmake	=> [ 'p', 0 ],
gbsmakemake	=> [ 'p', 0 ],
gbsman	=> [ 'd', 0 , "more $OS_GBS_SCRIPTS/gbsman.txt" ],
gbsnew	=> [ 'p', 1 ],			# special case!
gbsscm	=> [ 'p', 0 ],
gbssetup	=> [ 's', 1 ],
gbsshow	=> [ 'p', 0 ],
gbssilo	=> [ 'p', 0 ],
gbsstart	=> [ 's', 1 ],
gbsstats	=> [ 'p', 0 ],
gbsswc	=> [ 's', 1 ],
gbsswr	=> [ 's', 1 ],
gbssws	=> [ 's', 1 ],
gbsswt	=> [ 's', 1 ],
gbssysaudit	=> [ 'p', 0 ],
gbssysgen	=> [ 'p', 0 ],
gbssysmake	=> [ 'p', 0 ],
gbswhich	=> [ 'p', 0 ],
gbsxref	=> [ 'p', 0 ],
);

my $OS_GBS_BASE_PATH = ENV_getenv( 'GBS_BASE_PATH');
my %GBS_BASE_COMMANDS = (

define	=> 1,
gbsinit	=> 1,
gbssetup	=> 1,
gbsstart	=> 1,
);




sub GBSCMD_command($@)
{
my ($name,
@args,
) = @_;
my $command;

my $ref = $GBS_COMMAND_REFS{$name};
ENV_fatal( "No such GBS command '$name'")
if (! defined $ref);

my ($type, $must_be_sourced) = @{$ref};
if ($type eq 'p')		# perl
{
$command = "$OS_GBS_PERL_CMD -w $OS_GBS_SCRIPTS/$name.pl";
} elsif ($type eq 's')	# script
{
$command = "$OS_GBS_SCRIPTS/$name$SCRIPT_FILETYPE";
$command = "source $command"
if ($IS_LUNIX && $must_be_sourced);
} elsif ($type eq 'd')	# defined
{
$command = $ref->[2];
} else # ($type eq 'g')	# script + needs GUI
{
if (GBSENV_gui_available())
{
$command = "$OS_GBS_SCRIPTS/$name$SCRIPT_FILETYPE";
$command = "source $command"
if ($IS_LUNIX && $must_be_sourced);
} else
{
$command = SHELL_concat( SHELL_echo( "*** Cannot startup $name ***",
"- Perl Tkx Package is not available"));
}
}
$command =~ tr!\\!/!;
$command = "$command @args"
if (@args);

return $command;
}




sub GBSCMD_base_command($@)
{
my ($name,
@args,
) = @_;
my $command;

ENV_fatal( "no such GBS command '$name'")
if (! exists $GBS_BASE_COMMANDS{$name});

$command = "$OS_GBS_BASE_PATH/$name$SCRIPT_FILETYPE";
$command = "source $command"
if ($IS_LUNIX);

$command =~ tr!\\!/!;
$command = "$command @args"
if (@args);

return $command;
}








sub GBSCMD_batch($)
{
my ($command_ref,		    # ADT
) = @_;
my $full_command;

$full_command = ENV_command_line( $command_ref);


my $gbs_batch = get_target_env( 'GBS_BATCH');
if ($gbs_batch ne '')
{
$full_command = substitute_args( $gbs_batch, '1', $full_command, 1);
}

return $full_command;
}








sub GBSCMD_submit($$$)
{
my ($command,
$command_name,
$os_log_file) = @_;
my $submit_cmd;

my $gbs_submit = get_target_env( 'GBS_SUBMIT');
if ($gbs_submit eq '')
{
$submit_cmd = $command;
} else
{
$submit_cmd = substitute_args( $gbs_submit, '1', $command, 1);
$submit_cmd = substitute_args( $submit_cmd, '2', $command_name, 0);
$submit_cmd = substitute_args( $submit_cmd, '3', $os_log_file, 0);
}



return $submit_cmd;
}








sub GBSCMD_subwin($$)
{
my ($command,
$command_name) = @_;
my $subwin_cmd;

my $gbs_subwin = get_target_env( 'GBS_SUBWIN');
if ($gbs_subwin eq '')
{
$subwin_cmd = $command;
} else
{
$subwin_cmd = substitute_args( $gbs_subwin, '1', $command, 1);
$subwin_cmd = substitute_args( $subwin_cmd, '2', $command_name, 0);
}



return $subwin_cmd;
}





sub GBSCMD_make()
{
my $gbs_make = get_target_env( 'GBS_MAKE');
if ($gbs_make eq '')
{
$gbs_make = ($IS_WIN32) ? "$ENV{GBS_SCRIPTS_PATH}\\win32utils\\make.exe" : 'make';
}
return $gbs_make;
}






sub GBSCMD_view_file($)
{
my ($file_spec,
) = @_;


return get_env_command_ref( GBS_VIEWER => $file_spec);
}





sub GBSCMD_browse_file($)
{
my ($file_spec,
) = @_;


return get_env_command_ref( GBS_BROWSER => $file_spec);
}





sub GBSCMD_edit_file($$)
{
my ($file_spec,
$editor_opt,
) = @_;


my @opts = (defined $editor_opt && $editor_opt ne '') ? ($editor_opt) : ();

return get_env_command_ref( GBS_EDITOR => $file_spec, @opts);
}




sub get_env_command_ref($$@)
{
my ($env_name,  # GBS_VIEWER, GBS_BROWSER, GBS_EDITOR
$file_spec,
@options,
) = @_;
my $command_ref;




my $env_command_ref = $ENV_COMMAND_REFS{$env_name};
my ($os_command, $args_ref);
if (defined $env_command_ref)
{
($os_command, $args_ref) = @{$env_command_ref};
} else
{
my $command_line = ENV_getenv( $env_name);

if ($command_line eq '')
{



my ($sec_env_name, $win_default, $unix_default) = @{$GEN_DEFS{$env_name}};

$command_line = ENV_getenv( $sec_env_name)
if (defined $sec_env_name);
$command_line = ($IS_WIN32) ? $win_default : $unix_default
if ($command_line eq '');
}

($os_command, $args_ref) = ENV_command_split( $command_line);
$os_command =~ tr!/!\\!
if ($IS_WIN32);
}




$file_spec =~ s!/!\\!g
if ($IS_WIN32);
my @args = substitute_args( $args_ref, '1', $file_spec, 1);
push @args, @options;

print "os_command=$os_command\n";
$command_ref = ENV_command_ref( $os_command, \@args);
return $command_ref;
}





sub GBSCMD_set_xtitle($)
{
my ($title) = @_;
my $command;

$title =~ s!/!\\!g
if ($IS_WIN32);

$GBS_XTITLE = ENV_getenv( 'GBS_XTITLE') if (!defined $GBS_XTITLE);
if ($GBS_XTITLE eq '')
{
$command = SHELL_xtitle( $title);
} else
{
$command = substitute_args( $GBS_XTITLE, '1', $title, 1);
}


return $command;
}





sub GBSCMD_bg_notifier()
{
my $gbs_bg_notifier = get_target_env( 'GBS_BG_NOTIFIER');

return $gbs_bg_notifier;
}




sub get_target_env($)
{
my ($base_env) = @_;
my $env_value;

my $env_name;
my $gbs_target = uc ENV_getenv( 'GBS_TARGET');
my $gbs_audit = uc ENV_getenv( 'GBS_AUDIT');

my $key = join( '_', $base_env, $gbs_target, $gbs_audit);
my $ref = $TARGET_ENVS{$key};
if (defined $ref)
{
($env_name, $env_value) = @{$ref};

} else
{
my @envs;
if ($gbs_audit ne '')
{
push @envs, "${base_env}_${gbs_target}_${gbs_audit}";
push @envs, "${base_env}_${gbs_audit}";
} else
{
push @envs, "${base_env}_${gbs_target}";
}
push @envs, "${base_env}";

$env_name = '';
foreach my $env (@envs)
{

$env_value = ENV_getenv( $env);
if ($env_value ne '')
{
$env_name = $env;
last;
}
}
if ($env_name ne '')
{
ENV_say( 1, "$env_name EnvVar defined");
$TARGET_ENVS{$key} = [ $env_name, $env_value ];
}
}


return $env_value;
}




sub substitute_args($$$$)
{
my ($strings_ref,	    # $string == OK
$name,
$value,
$is_mandatory,
) = @_;
my @strings;    # or $string

@strings = (ref $strings_ref) ? @{$strings_ref} : ($strings_ref);

my $token_name = ($IS_WIN32) ? "\%$name\%" : "\$$name";
my $token = quotemeta $token_name;
my $count = 0;

foreach my $string (@strings)
{
$count += $string =~ s/$token/$value/g;
}

if ($count == 0 && $is_mandatory)
{
ENV_error_exit( "Argument '$token_name' required in string '@strings'");
}

return (wantarray) ? @strings : $strings[0];
}

return 1;


