#===============================================================
#
#   gbsedit.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSEDIT*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::ask;
use glo::genopt;
use glo::format;
use glo::scm;
use gbsenv;
use gbsglo;
use gbsscm;
use target;
use validate;
use files;
use fcreate;
use fspec;
use run;




sub analyse_file();
sub check_file_location();
sub create_file_if_necessary();
sub do_the_edit();
sub pre_select_audit();
sub pre_bld_type_gbs_file();
sub pre_type_gbs_file();
sub pre_select_sca();
sub pre_export_gbs_file();
sub pre_select_tool();
sub pre_gbssub_file();
sub pre_gbsall_file();
sub pre_unknown_gbs_file();
sub pre_any_file();
sub validate_sub_comp_subsys($);
sub validate_sub_path($);
sub validate_system();
sub validate_subsys();
sub validate_component();
sub validate_target();
sub spec_any();
sub create_any();
sub get_gbs_gen_type($);
sub is_src_type($);
sub is_inc_type($);




my $CWD = ENV_cwd();
my $GBS_ROOT_PATH = ENV_getenv_perl_path( 'GBS_ROOT_PATH');
my $GBS_SYSTEM_NAME = ENV_getenv( 'GBS_SYSTEM_NAME');
my $GBS_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $GBS_COMPONENT = ENV_getenv( 'GBS_COMPONENT');
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my $GBS_AUDIT = ENV_getenv( 'GBS_AUDIT');
my $GBS_TOOL = ENV_getenv( 'GBS_TOOL');

my $CWD_IS_IN_CURRENT_GBS_PATH = 0;
my $CWD_IS_IN_A_GBS_PATH = 0;
my @SUBDIR_COMP_SUBSYS;

my $SUBSYS;
my $FULL_GBS;
my $COMPONENT;
my $SUBDIR;
my $TARGET;
my $AUDIT;
my $TOOL;
my $SPECIFIED_SUBSYS = 0;
my $SPECIFIED_COMPONENT = 0;
my $FILE_SPEC;			# FULL file_spec

my $SUB_PATH;
my $FILE_PATH;
my $FILE_NAME;
my $FILE_TYPE;
my $FILE_NAME_AND_TYPE;

my $FILE_TYPE_GBS = '';	    # .gbs .usr
my $FILE_IS_GBS_FILE = 0;

my $GENERIC_FILE_NAME;
my $GENERIC_FILE_NAME_VAR_PART;

my $IN_INSTALL;		# bool: used for gbsall

my $FILE_CREATE_FUNC;	# defaults to ENV_touch_file




my %BARE_COMMAND_FILES = (
'switch.gbs'    => 1,
'switch.usr'    => 1,
'gbsall'	    => 1,







);

my %SUB_COMMAND_FILES;
foreach my $name (FSPEC_subsys_script_names())    # gbssubaudit, gbssubcleanup_aud, gbssubcleanup_bld, gbssubgen, gbssubmake
{
$BARE_COMMAND_FILES{$name} = 1;
$SUB_COMMAND_FILES{"$name.bat"} = 1;
$SUB_COMMAND_FILES{"$name.sh"} = 1;
}
my @SUB_COMMAND_FILES = sort( keys( %SUB_COMMAND_FILES));




my %SWITCH_NAMES =
(
'switch.gbs.bat'	=> '.gbs',
'switch.gbs.sh'	=> '.gbs',
'switch.usr.bat'	=> '.usr',
'switch.usr.sh'	=> '.usr',
);

my @SUBDIR_NAMES = qw( src inc loc bld dat sav);

my @GENERIC_FILE_NAMES_REFS = (


[ qr/^audit$/,	    'audit',	    ],
[ qr/^bld_flags_(.+)$/, 'bld_flags',    ],
[ qr/^bld_incs_(.+)$/,  'bld_incs',     ],
[ qr/^flags_(.+)$/,	    'flags',	    ],
[ qr/^incs_(.+)$/,	    'incs',	    ],
[ qr/^export$/,	    'export',	    ],
[ qr/^gbsall$/,	    'gbsall',	    ],	    # batch-file
[ qr/^gbssub(.*)$/,	    'gbssub',	    ],	    # batch-file
[ qr/^owners$/,	    'owners',	    ],
[ qr/^sca_(.+)$/,	    'sca',	    ],
[ qr/^scope$/,	    'scope',	    ],
[ qr/^steps$/,	    'steps',	    ],
[ qr/^switch\.gbs$/,    'switch',	    ],	    # batch-file
[ qr/^switch\.usr$/,    'switch',	    ],	    # batch-file
[ qr/^sys$/,	    'sys',	    ],
[ qr/^system$/,	    'system',	    ],
[ qr/^target$/,	    'target',	    ],
[ qr/^tool$/,	    'tool',	    ],
);

my %FILE_DEFS = (




audit   => [
0, 1, 0, 0,	    \&pre_select_audit,
sub { FSPEC_plugin_file(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $AUDIT) },
sub { FCREATE_plugin_file( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $AUDIT) } ],
bld_flags   => [
0, 1, 0, 0,	    \&pre_bld_type_gbs_file,
sub { FSPEC_gen_type(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, '', '', $TARGET) },
sub { FCREATE_gen_type( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, '', '', $TARGET) } ],
bld_incs   => [
0, 1, 0, 0,	    \&pre_bld_type_gbs_file,
sub { FSPEC_gen_type(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, '', '', $TARGET) },
sub { FCREATE_gen_type( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, '', '', $TARGET) } ],
flags   => [
0, 1, 0, 0,	    \&pre_type_gbs_file,
sub { FSPEC_gen_type(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT, $TARGET) },
sub { FCREATE_gen_type( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT, $TARGET) } ],
incs    => [
0, 1, 0, 0,	    \&pre_type_gbs_file,
sub { FSPEC_gen_type(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT, $TARGET) },
sub { FCREATE_gen_type( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT, $TARGET) } ],
export  => [
3, 2, 1, 0,	    \&pre_export_gbs_file,
sub { FSPEC_export(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT) },
sub { FCREATE_export( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT) } ],
gbsall  => [
0, 0, 0, 0,	    \&pre_gbsall_file,
sub { FSPEC_gbsall(   $FILE_NAME_AND_TYPE, $IN_INSTALL) },
sub { FCREATE_gbsall( $FILE_NAME_AND_TYPE, $IN_INSTALL) } ],
gbssub  => [
1, 2, 0, 0,	    \&pre_gbssub_file,
sub { FSPEC_subsys_script(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS) },
sub { FCREATE_subsys_script( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS) } ],
owners  => [
0, 1, 0, 1,	    undef,
sub { FSPEC_owners(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) },
sub { FCREATE_owners( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) } ],
sca    => [
0, 1, 0, 0,	    \&pre_select_sca,
sub { FSPEC_audit_target(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, '', '', $AUDIT, $TARGET) },
sub { FCREATE_audit_target( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, '', '', $AUDIT, $TARGET) } ],
scope   => [
3, 3, 0, 0,	    undef,
sub { FSPEC_scope(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT) },
sub { FCREATE_scope( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT, undef) } ],
steps   => [
0, 1, 0, 0,	    undef,
sub { FSPEC_steps(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) },
sub { FCREATE_steps( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $GBS_SYSTEM_NAME)} ],
switch  => [
0, 1, 0, 0,	    undef,
sub { FSPEC_switch(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) },
sub { FCREATE_switch( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $GBS_SYSTEM_NAME) } ],
sys	    => [
0, 1, 0, 0,	    undef,
sub { FSPEC_sys(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) },
sub { FCREATE_sys( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $GBS_SYSTEM_NAME) } ],
system  => [
0, 1, 0, 1,	    undef,
sub { FSPEC_system(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) },
sub { FCREATE_system( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH) } ],
target  => [
0, 1, 0, 0,	    \&VALIDATE_target,
sub { FSPEC_plugin_file(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $TARGET) },
sub { FCREATE_plugin_file( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $TARGET) } ],
tool    => [
0, 1, 0, 0,	    \&pre_select_tool,
sub { FSPEC_plugin_file(   $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $TOOL) },
sub { FCREATE_plugin_file( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $TOOL) } ],
unkgbs  => [
0, 0, 0, 0,	    \&pre_unknown_gbs_file,
sub { spec_any()   },
sub { create_any() } ],
any  => [
0, 0, 1, 0,	    \&pre_any_file,
sub { spec_any() },
sub { create_any() } ],
);

my $FILE_DEF_REF;






$| = 1;
my $RC = 0;
GBSENV_init();




my @genopts = (
[ '<1>',  'file',	     'ssm', "", '[[[subsys:]component:]subdir:[<subpath>/]]file'],
[ '<2>',  'editor_opt',  'sso', "", 'Editor option. e.g. Line-nr to jump to'],
[ 'nco',  'no_checkout', 'bso', 0 , '.usr files will always be forced to no_checkout'],
);
GENOPT_set_optdefs( 'gbsedit', \@genopts,
[ 'Create/edit GBS-files and user-files. GBS-files end with .gbs or .usr',
'The switch.gbs, gbssub* and gbsall batch files are also considered GBS_files',
'Source, Include and glkbt-files are considered user-files',
'Relevant templates will be used with file-creation' ],
undef);
GENOPT_parse();
my $FILE = GENOPT_get( 'file');
my $EDITOR_OPT = GENOPT_get( 'editor_opt');
my $NO_CHECKOUT = GENOPT_get( 'no_checkout');

my $MUST_CHECKOUT = !$NO_CHECKOUT;




{
ENV_say( 1, "No current GBS system.")
if ($GBS_ROOT_PATH eq '');

if ($GBS_ROOT_PATH ne '' && ENV_is_in_path( $CWD, $GBS_ROOT_PATH))
{

$CWD_IS_IN_CURRENT_GBS_PATH = 1;
$CWD_IS_IN_A_GBS_PATH = 1;
} else
{
if (FILES_find_root_path( $CWD) ne '')
{

$CWD_IS_IN_A_GBS_PATH = 1;
} else
{

}
}
}


$FILE = ENV_perl_paths( $FILE);
if (-e $FILE)
{



$FILE_SPEC = ENV_abs_path( $CWD, $FILE);
if ($GBS_ROOT_PATH eq '' || !ENV_is_in_path( $FILE_SPEC, $GBS_ROOT_PATH))
{
ENV_say( 1, "File '$FILE' is not in located in current GBS System ($GBS_SYSTEM_NAME).");
if ($MUST_CHECKOUT)
{
$MUST_CHECKOUT = 0;
ENV_say( 2, "CheckOut disabled.");
}
}
do_the_edit();
} else
{



if ($FILE =~ /:/)
{
ENV_error_exit( "The [[[subsys:]component:]subdir:]file notation is only possible in a GBS System")
if ($GBS_ROOT_PATH eq '');

(my $file, @SUBDIR_COMP_SUBSYS) = reverse( split( ':', $FILE));
ENV_error_exit( "Syntax is [[[<subsys>:][<component>:]][<subdir>:][<subpath>/]]<file>")
if (@SUBDIR_COMP_SUBSYS > 3);

($SUB_PATH, $FILE_NAME, $FILE_TYPE) = ENV_split_spec_dnt( $file);
$FILE_PATH = '';
} else
{
$SUB_PATH = '';
($FILE_PATH, $FILE_NAME, $FILE_TYPE) = ENV_split_spec_dnt( $FILE);
}




analyse_file();




if ($CWD_IS_IN_CURRENT_GBS_PATH)
{
VALIDATE_gbs();
} elsif ($CWD_IS_IN_A_GBS_PATH)
{
ENV_say( 1, "CWD is not in current GBS system.");
} else
{
ENV_say( 1, "CWD is not in a GBS system.");
}




check_file_location();
create_file_if_necessary();
do_the_edit();
}




ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}





sub analyse_file()
{
$FILE_NAME_AND_TYPE = "$FILE_NAME$FILE_TYPE";




if ($BARE_COMMAND_FILES{$FILE_NAME_AND_TYPE})
{
$FILE_TYPE = ENV_commandfile_ext();
$FILE_NAME = $FILE_NAME_AND_TYPE;
$FILE_NAME_AND_TYPE = "$FILE_NAME$FILE_TYPE";
ENV_say( 1, "Assumed $FILE_NAME_AND_TYPE");
}

ENV_exit()
if ($FILE_TYPE eq '' && ASK_YNQ( "No file-type specified ($FILE_NAME). Continue?", 'N', 1) eq 'N');




if ($FILE_TYPE eq '.gbs' || $FILE_TYPE eq '.usr')
{
$FILE_TYPE_GBS = $FILE_TYPE;
} else
{
$FILE_TYPE_GBS = $SWITCH_NAMES{$FILE_NAME_AND_TYPE};
}
if (defined $FILE_TYPE_GBS)
{
ENV_say( 1, "GBS File");
$FILE_IS_GBS_FILE = 1;
} else
{
$FILE_TYPE_GBS = '';
if ($BARE_COMMAND_FILES{$FILE_NAME})
{
ENV_say( 1, "GBS Batch File");
$FILE_IS_GBS_FILE = 1;
} else
{
ENV_say( 1, "User File");
$FILE_IS_GBS_FILE = 0;
}
}




foreach my $ref (@GENERIC_FILE_NAMES_REFS)
{
if ($FILE_NAME =~ $ref->[0])    # $re
{
$GENERIC_FILE_NAME = $ref->[1];
$GENERIC_FILE_NAME_VAR_PART = $1;
last;
}
}
if (!defined $GENERIC_FILE_NAME)
{
if ($FILE_TYPE_GBS ne '')
{
$GENERIC_FILE_NAME = 'unkgbs';
} else
{
$GENERIC_FILE_NAME = 'any';
}
}

$FILE_DEF_REF = $FILE_DEFS{$GENERIC_FILE_NAME};



}




sub check_file_location()
{



my ($subsys_comp, $minimum_level, $sub_path_allowed, $special_file, $pre_func, $fspec_func, $create_func) = @{$FILE_DEF_REF};



validate_sub_comp_subsys( $subsys_comp);

if ($minimum_level == 1)		    # system
{
validate_system();
} elsif ($minimum_level == 2)	    # subsys
{
validate_subsys();
} elsif ($minimum_level == 3)	    # component
{
validate_component();
}

ENV_error_exit( "SubPath '$SUB_PATH' not allowed for this kind of GBS file")
if ($SUB_PATH ne '' && !$sub_path_allowed);

if ($FILE_TYPE_GBS eq '.gbs')
{
VALIDATE_administrator();
} elsif ($FILE_TYPE_GBS eq '.usr')
{
$MUST_CHECKOUT = 0;
}




if ($special_file)
{
ENV_warn( "**",
"** Special GBS file ($FILE_NAME_AND_TYPE) You are on dangerous ground!",
"***");
ENV_exit()
if (ASK_YNQ('Edit anyway?', 'N', 1) eq 'N');
}




if (defined $pre_func)
{
&$pre_func();
}




$FILE_SPEC = &$fspec_func();		# get the file_spec




$FILE_CREATE_FUNC = $create_func;
}




sub create_file_if_necessary()
{
if (!-f $FILE_SPEC || -z $FILE_SPEC)
{
ENV_say( 1, "File '$FILE_SPEC' does not exist");
ENV_exit()
if (ASK_YNQ( 'Create?', 'Y', 1) eq 'N');
ENV_say( 1, "Creating '$FILE_SPEC'...");




if (defined $FILE_CREATE_FUNC)
{
&$FILE_CREATE_FUNC();
} else
{
ENV_touch_file( $FILE_SPEC);
}
}
}




sub do_the_edit()
{
if ($MUST_CHECKOUT)
{
GBSSCM_preset( 0);
GBSSCM_connect();
SCM_assert_co_file( $FILE_SPEC);
}

ENV_say( 1, "Editing '$FILE_SPEC'...");
RUN_editor( $FILE_SPEC, $EDITOR_OPT);
}




sub pre_select_audit()
{



my @all_audits = ENV_getenv( 'GBS_ALL_AUDITS');
ENV_error_exit( 'No Audits defined')
if (!@all_audits);

my $default_audit = $GBS_AUDIT;
$default_audit = $all_audits[0]
if ($default_audit eq '' && @all_audits == 1);
my $audit = ASK_value_from_menu( 'Select Audit', \$default_audit, @all_audits);
VALIDATE_audit_no_subsys( $audit);
$ENV{GBS_AUDIT} = $audit;
$AUDIT = $audit;
}




sub pre_bld_type_gbs_file()
{



validate_target();
my $src = ".$GENERIC_FILE_NAME_VAR_PART";
if (!is_src_type( $src))
{
ENV_warn( "Unknown src type '$src' in this context ($TARGET)");
ENV_exit()
if (ASK_YNQ( 'Continue anyway?', 'Y', 1) eq 'N');
}
}




sub pre_type_gbs_file()
{



validate_target();
my $src = ".$GENERIC_FILE_NAME_VAR_PART";
if (!is_src_type( $src))
{
ENV_warn( "Unknown src type '$src' in this context ($TARGET)");
ENV_exit()
if (ASK_YNQ( 'Continue anyway?', 'Y', 1) eq 'N');
}

my $default = -1;

my @dir_refs;
push @dir_refs,
[ "$GBS_ROOT_PATH/sysgen/$TARGET",
[ '', '', $TARGET ],
"sysgen/$TARGET" ]
if ($TARGET);
push @dir_refs,
[ "$GBS_ROOT_PATH/sysgen",
[ '', '', '' ],
"sysgen" ];
if ($FULL_GBS && $SUBSYS)
{
push @dir_refs,
[ "$GBS_ROOT_PATH/dev/$SUBSYS/gen/$TARGET",
[ $SUBSYS, '', $TARGET ],
"gen/$TARGET" ]
if ($TARGET);
push @dir_refs,
[ "$GBS_ROOT_PATH/dev/$SUBSYS/gen",
[ $SUBSYS, '', '' ],
"gen" ];
if ($COMPONENT)
{
push @dir_refs,
[ "$GBS_ROOT_PATH/dev/$SUBSYS/comp/$COMPONENT/opt/$TARGET",
[ $SUBSYS, $COMPONENT, $TARGET ],
"opt/$TARGET" ]
if ($TARGET);
push @dir_refs,
[ "$GBS_ROOT_PATH/dev/$SUBSYS/comp/$COMPONENT/opt",
[ $SUBSYS, $COMPONENT, '' ],
"opt" ];
}
}

foreach my $i (0..$#dir_refs)
{
if ($CWD eq $dir_refs[$i]->[0])
{
$default = $i;
last;
}
}

{
my @item_refs;
foreach my $ref (@dir_refs)
{
my ($dir, $locs_ref, $item) = @{$ref};
my $e = (-f "$dir/$FILE_NAME_AND_TYPE") ? '*' : '';
push @item_refs, [ "$item$e", @{$locs_ref} ];
}
ENV_say( 0, scalar @item_refs . " Items");		# DBG?
ENV_say( 1, "Currencies: $SUBSYS : $COMPONENT : $TARGET");
my $index = ASK_index_from_menu( "Select Level for '$FILE_NAME_AND_TYPE'", $default, @item_refs);
(undef, $SUBSYS, $COMPONENT, $TARGET) = @{$item_refs[$index]};
}

ENV_say( 1, "$SUBSYS, $COMPONENT, $TARGET");
if ($MUST_CHECKOUT)
{




my $file_spec = FSPEC_gen_type( $FILE_NAME_AND_TYPE, $GBS_ROOT_PATH, $SUBSYS, $COMPONENT, $TARGET);
GBSSCM_preset( 0);
GBSSCM_connect();

my $dir = ENV_split_spec_d( $file_spec);
my $dir_state = SCM_get_states( $dir);

SCM_add_dirs( $dir, \&FILES_get_ignore_spec, 0, 0)	    # $pre_validated, $verbose
if ($dir_state < 0);	# not_scm, not_exist
}
}




sub pre_gbsall_file()
{



if ($MUST_CHECKOUT)
{
$MUST_CHECKOUT = 0;
ENV_say( 1, "CheckOut disabled.");
}




my @locations = ("In User dir (GBS_BASE_PATH)",  "In Install dir (GBS_SCRIPTS_ROOT)");
$IN_INSTALL = ASK_index_from_menu( "Select Level for gbsall'", 0, @locations);
}





sub pre_gbssub_file()
{



ENV_error_exit( "No current SubSys")
if ($SUBSYS eq '');
ENV_error_exit( "This file is only allowed for NON Full GBS SubSystems")
if ($FULL_GBS);

if (!exists( $SUB_COMMAND_FILES{$FILE_NAME_AND_TYPE}))
{
ENV_warn( "Unknown GBS command file ($FILE_NAME_AND_TYPE)");
ENV_say( 0, FORMAT_hanging( 0, 9, 4, "Known gbs files are: @SUB_COMMAND_FILES"));
ENV_exit()
if (ASK_YNQ( 'Edit anyway?', 'N', 1) eq 'N');
$GENERIC_FILE_NAME = 'any';
}
}




sub pre_select_sca()
{



my @all_targets = ENV_getenv( 'GBS_ALL_TARGETS');
ENV_error_exit( 'No Targets defined')
if (!@all_targets);

my $target = $GENERIC_FILE_NAME_VAR_PART;
if (!grep( $_ eq $target, @all_targets))
{
ENV_error_exit( "Target '$target' does not exist (@all_targets)");
}

pre_select_audit();
}




sub pre_export_gbs_file()
{








if ($FULL_GBS)
{
if ($SPECIFIED_COMPONENT)
{
validate_sub_path( undef);
} else
{
if ($SUB_PATH ne '')
{
validate_sub_path( 'gen');
$COMPONENT = '';
} else
{
if ($COMPONENT ne '')
{
my $se = (-e "$GBS_ROOT_PATH/dev/$SUBSYS/gen/$FILE_NAME_AND_TYPE") ? '*' : '';
my $ce = (-e "$GBS_ROOT_PATH/dev/$SUBSYS/comp/$COMPONENT/$FILE_NAME_AND_TYPE") ? '*' : '';
my @list = ( "SubSys$ce ($SUBSYS)",
"Component$ce ($COMPONENT)");
my $cur_entry = @list - 1;  # last entry

my $index = ASK_index_from_menu( 'Select Level of export.gbs', $cur_entry, @list);
$COMPONENT = ''
if ($index == 0);
}
}
}
} else
{
$COMPONENT = undef;
}
}




sub pre_select_tool()
{



my @all_tools = ENV_getenv( 'GBS_TOOLS');
ENV_error_exit( 'No Tools defined')
if (!@all_tools);

my $default_index = (@all_tools == 1) ? 0 : -1;
my $tool = ASK_value_from_menu( 'Select Tool', $default_index, @all_tools);
$ENV{GBS_TOOL} = $tool;
$TOOL = $tool;
}




sub pre_unknown_gbs_file()
{



if ($FILE_TYPE eq '.gbs')
{
ENV_error_exit( "Unknown GBS file ($FILE_NAME_AND_TYPE)");
} else
{
ENV_warn( "Unknown GBS file ($FILE_NAME_AND_TYPE)");
ENV_exit()
if (ASK_YNQ( 'Edit anyway?', 'N', 1) eq 'N');
}
}




sub pre_any_file()
{






my $path;
if ($SUBDIR ne '')
{
validate_component();
$path = "$GBS_ROOT_PATH/dev/$SUBSYS/comp/$COMPONENT/$SUBDIR";
} else
{
$path = $CWD;
ENV_say( 1, "Currently in: $path");
if ($CWD_IS_IN_CURRENT_GBS_PATH)
{
my $this_dir = ENV_parent_dir( $path, -1);
if (grep( $_ eq $this_dir, @SUBDIR_NAMES))
{
$SUBDIR = $this_dir;
}
}
}

if ($SUBDIR eq 'src')
{
if (!is_src_type( $FILE_TYPE))
{
ENV_warn( "Unknown src type '$FILE_TYPE' in this context ($TARGET)");
ENV_exit()
if (ASK_YNQ( 'Continue anyway?', 'Y', 1) eq 'N');
}
ENV_error_exit( "Cannot specify SubPath ($SUB_PATH) in SubDir $SUBDIR")
if ($SUB_PATH ne '');
} elsif ($SUBDIR eq 'loc'|| $SUBDIR eq 'inc')
{
if (!is_inc_type( $FILE_TYPE))
{
ENV_warn( "Unknown $SUBDIR type '$FILE_TYPE' in this context ($TARGET)");
ENV_exit()
if (ASK_YNQ( 'Continue anyway?', 'Y', 1) eq 'N');
}
ENV_error_exit( "Cannot specify SubPath ($SUB_PATH) in SubDir $SUBDIR")
if ($SUB_PATH ne '');
} else
{
$path .= "/$SUB_PATH"
if ($SUB_PATH ne '');
ENV_error_exit( "Path '$path' does not exist")
if (!-d $path);
}

$FILE_PATH = $path;
}




sub validate_sub_comp_subsys($)
{
my ($subsys_comp) = @_;	    # 0 = none, 1 = subsys, 2 = component, 3 = subsys+comp

if (@SUBDIR_COMP_SUBSYS > 0)
{
if ($subsys_comp == 0)		# any
{
($SUBDIR, $COMPONENT, $SUBSYS) = @SUBDIR_COMP_SUBSYS;
} elsif ($subsys_comp == 1)	# subsys only
{
if (@SUBDIR_COMP_SUBSYS == 1)
{
$SUBSYS = $SUBDIR_COMP_SUBSYS[0];
} else
{
ENV_error_exit( "Can only specify SubSys: for GBS file $FILE_NAME_AND_TYPE");
}
} else # $subsys_comp == 2 or $subsys_comp == 3
{
if (@SUBDIR_COMP_SUBSYS == 1)
{
$SUBSYS = $SUBDIR_COMP_SUBSYS[0];
} elsif (@SUBDIR_COMP_SUBSYS == 2)
{
($COMPONENT, $SUBSYS) = @SUBDIR_COMP_SUBSYS;
} else
{
ENV_error_exit( "can only specify SubSys: and/or Component: for GBS file $FILE_NAME_AND_TYPE");
}
}
}




if ($SUBSYS)
{
$SPECIFIED_SUBSYS = 1;
$SUBSYS = $GBS_SUBSYS
if ($SUBSYS eq '.');
ENV_error_exit( "No such SubSystem '$SUBSYS'")
if (! -e "$GBS_ROOT_PATH/dev/$SUBSYS");
} else
{
$SUBSYS = $GBS_SUBSYS;
}
if ($SUBSYS)
{
$FULL_GBS = GBSGLO_subsystem_is_full_gbs( $SUBSYS);
}




if ($COMPONENT)
{
$SPECIFIED_COMPONENT = 1;
$COMPONENT = $GBS_COMPONENT
if ($COMPONENT eq '.');
ENV_error_exit( "No such Component '$COMPONENT' in SubSystem '$SUBSYS'")
if (! -e "$GBS_ROOT_PATH/dev/$SUBSYS/comp/$COMPONENT");
} else
{
$COMPONENT = $GBS_COMPONENT;
}



if ($SUBDIR)
{
ENV_error_exit( "SubDir '$SUBDIR' must be one of (@SUBDIR_NAMES)")
if (!grep( $SUBDIR eq $_, @SUBDIR_NAMES));
} else
{
$SUBDIR = '';
}




$TARGET = $GBS_TARGET;
$AUDIT = $GBS_AUDIT;
$TOOL = $GBS_TOOL;




if ($SUBDIR ne '')
{
ENV_say( 1, "$SUBSYS:$COMPONENT:$SUBDIR");
} elsif ($COMPONENT ne '')
{
ENV_say( 1, "$SUBSYS:$COMPONENT");
} elsif ($SUBSYS ne '')
{
ENV_say( 1, "$SUBSYS");
}
}




sub validate_sub_path($)
{
my ($alloweds_ref) = @_;

if ($SUB_PATH ne '')
{
if (defined $alloweds_ref)
{
$alloweds_ref = [ $alloweds_ref ] if (!ref $alloweds_ref);
ENV_error_exit( "'$FILE': SubPath must be one of (@{$alloweds_ref})")
if (!grep( $_ eq $SUB_PATH, @{$alloweds_ref}));
} else
{
ENV_error_exit( "'$FILE': SubPath cannot be used in this context");
}
}
}




sub validate_system()
{
ENV_error_exit( "No current GBS System")
if ($GBS_ROOT_PATH eq '');
}




sub validate_subsys()
{
validate_system();
ENV_error_exit( "No current SubSys")
if ($SUBSYS eq '');
}




sub validate_component()
{
validate_subsys();
ENV_error_exit( "No current Component")
if ($COMPONENT eq '');
}




sub validate_target()
{
ENV_error_exit( "No current Target")
if ($TARGET eq '');
}




sub spec_any()
{
my $file_spec;

if ($FILE_PATH eq '')
{
$file_spec = "$CWD/$FILE_NAME_AND_TYPE";
} else
{
$file_spec = "$FILE_PATH/$FILE_NAME_AND_TYPE";
}

return $file_spec;
}




sub create_any()
{

my @any_file_types = ($FILE_TYPE);
if ($SUBDIR eq 'src')
{
my $src_file_type = '.' . get_gbs_gen_type( $FILE_TYPE);
push @any_file_types, $src_file_type
unless ($src_file_type eq $FILE_TYPE);
} elsif ($SUBDIR eq 'inc' || $SUBDIR eq 'loc')
{



foreach my $src_type (TARGET_get_src_types( $TARGET))
{
my @inc_types = TARGET_get_src_items( $TARGET, $src_type, 'INC_TYPES');
if (grep $_ eq $FILE_TYPE, @inc_types)
{
push @any_file_types, $src_type;
my $src_file_type = '.' . get_gbs_gen_type( $src_type);
push @any_file_types, $src_file_type;
last;
}
}
}
my (@anyfile_specs) = FCREATE_get_any( \@any_file_types);

my $anyfile_spec_or_comment_ref;
if (@anyfile_specs)
{



my $anyfile_spec;
if (@anyfile_specs == 1)
{
$anyfile_spec = $anyfile_specs[0];
} else
{
ENV_say( 1, "Multiple Templates found:");

my @menu_items;
foreach my $spec (@anyfile_specs)
{
my $file = ENV_split_spec_f( $spec);
$file =~ s/any_//;
push @menu_items, [ $file, $spec ];
}
$anyfile_spec = ASK_value_from_menu( "  Select Template:", 0, @menu_items);
}
ENV_say( 1, "- Using $anyfile_spec");
$anyfile_spec_or_comment_ref = $anyfile_spec;
} else
{



ENV_say( 1, "No Template File found");
my $begin_com = ASK_text( '- Enter Begin Comments character(s)', '#', [ 0 ,3 ]);
if ($begin_com ne '')
{
my $end_com   = ASK_text( '- Enter   End Comments character(s)', '',  [ 0, 3 ]);
$anyfile_spec_or_comment_ref = [ $begin_com, $end_com ];
} else
{
$anyfile_spec_or_comment_ref = '';
ENV_say( 1, "Creating empty file...");
}
}


return FCREATE_any( $FILE_SPEC, $anyfile_spec_or_comment_ref, [ $SUBSYS, $COMPONENT ]);
}




sub get_gbs_gen_type($)
{
my ($src) = @_;
my $gbs_src_type;

if (is_src_type( $src))
{
$gbs_src_type = TARGET_get_src_items( $TARGET, $src, 'TYPE')->[0];	# $type e.g.: gen, asm, c, glb, ...
} else
{
$gbs_src_type = '';
}

return $gbs_src_type;
}




sub is_src_type($)
{
my ($type) = @_;
my $is_src_type = 0;

VALIDATE_target();	    # Error-exit if current component does not generate for the current Target

if (-f FSPEC_plugin_file( 'target', $GBS_ROOT_PATH, $TARGET) &&
TARGET_is_src_type( $TARGET, $type))
{
$is_src_type = 1;
}

return $is_src_type;
}




sub is_inc_type($)
{
my ($type) = @_;
my $is_inc_type = 0;

VALIDATE_target();	    # Error-exit if current component does not generate for the current Target

if (-f FSPEC_plugin_file( 'target', $GBS_ROOT_PATH, $TARGET))
{
my @inc_types = TARGET_get_inc_types( $TARGET);
if (grep $_ eq $type, @inc_types)
{
$is_inc_type = 1;
}
}
return $is_inc_type;
}

