#===============================================================
#
#   gbshtml.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbshtml;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
GBSHTML_doc_start
GBSHTML_doc_top
GBSHTML_section_head
GBSHTML_doc_bottom
GBSHTML_doc_end
GBSHTML_img
GBSHTML_insert_toc
GBSHTML_insert_nav_2
GBSHTML_insert_nav_3
);

use strict;
use integer;

use glo::env;
use glo::html;
use glo::file;
use glo::build;
use gbsres;
use gbsdb;
use gbsenv;




sub GBSHTML_doc_start($$$);
sub GBSHTML_doc_top($@);
sub GBSHTML_section_head($$);
sub GBSHTML_doc_bottom($$);
sub GBSHTML_doc_end();
sub GBSHTML_img($);
sub GBSHTML_insert_toc($);
sub GBSHTML_insert_nav_2($$;$$);
sub GBSHTML_insert_nav_3($$$;$$$$);

sub insert_nav($);
sub init();




my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();
my $GBS_SCRIPTS_DOC_PATH = "$GBS_SCRIPTS_PATH/doc";
my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();

my $INITIALISED = 0;




my $SILO_HTML_DIR = "$GBS_ROOT_PATH/silo/.gbs/cache/html";




my $IMAGES_DIR = "images";	# set to $SILO_HTML_DIR after init, so that this file can also be used


my %FILES = (
'gbs.css'	    => 1,	    # Keep for backward compatibility
'gbs2.css'	    => 1,
'gbsprint.css'  => 1,
'favicon.ico'   => 1,
);

my %SCRIPTS = (
'sorttable.js'  => 1,
);

my %IMGS =  (

up_small	 => [ 'index_up_3.gif',		'level_3',    'Up Chapter Button' ],	# 14x10
down_small	 => [ 'index_down_3.gif',	'level_3',    'Down Chapter Button' ],
blank_small  => [ 'blank.gif',		'level_3',    'Blank' ],
up_medium	 => [ 'index_up_2.gif',		'level_2',    'Up Section Button' ],	# 16x18
down_medium  => [ 'index_down_2.gif',	'level_2',    'Down Section Button' ],
blank_medium => [ 'blank.gif',		'level_2',    'Blank' ],
up_big	 => [ 'index_up_1.gif',		'level_1',    'Top Document Button' ],	# 24x24
down_big	 => [ 'index_down_1.gif',	'level_1',    'Bottom Document Button' ],
blank_big	 => [ 'blank.gif',		'level_1',    'Blank' ],
gbslogo	 => [ 'gbslogo.gif',		'gbs_logo',   'GBS Logo' ],		# 233x145
gbslogo_small=> [ 'gbslogo_small.gif',	'gbs_logo_s', 'GBS Logo' ],		# 100x62
rmclogo_wide => [ 'Banner_Wide_240x31.png', 'rmclogo_w',  'RMC Logo' ],		# 240x31
bullet	 => [ 'greysmallbullet.gif',	'bullet_s',   'Bullet', ]			# 5x5
);





sub GBSHTML_doc_start($$$)
{
my ($file_spec,	    # name of the html-file to be created (is not created yet)
$title,
$want_scripts,
) = @_;
my @lines;

init()
if (!$INITIALISED);

my $css_file_spec = $FILES{ 'gbs2.css'};
my $cssprint_file_spec = $FILES{ 'gbsprint.css'};
my $sortable_spec = ($want_scripts) ? $SCRIPTS{ 'sorttable.js'} : '';
my $favicon_file_spec = $FILES{ 'favicon.ico'};
push @lines,
HTML_doc_s( $file_spec, "GBS - $title",
css => $css_file_spec,
cssprint => $cssprint_file_spec,
script => $sortable_spec,
favicon => $favicon_file_spec,
),
HTML_body_s( bgcolor => "#ffffff", text => "#000000");

return @lines;
}




sub GBSHTML_doc_top($@)
{
my ($title_or_ref,	    # [ $title, $t_right ]
@item_refs	    # [ $label, $value, $right ]...
) = @_;
my @lines;

my ($title, $t_right);
($title, $t_right) = (ref $title_or_ref) ?  @{$title_or_ref} : ($title_or_ref, '');
if (@item_refs == 0)
{
if ($t_right eq '')
{
$t_right = GBSHTML_insert_nav_2( '', 'Bottom');
} else
{
@item_refs = ( [ '','','' ] );
}
}




push @lines,
HTML_table_s( id => 'Top', class => 'header', width => '100%');




push @lines,
HTML_trtd( {rowspan => 2}, HTML_link( 'http://www.genericbuildsupport.com', GBSHTML_img( 'gbslogo_small'))),
HTML_td( {class => 'gbs-head', width => '100%'}, GBSRES_title_html()),
HTML_td( {align => 'right'}, HTML_link( 'http://www.randymarques.com', HTML_bold( GBSRES_company_html()))),
HTML_trtd( {class => 'license', colspan => 2}, scalar BUILD_get_lic());




if ($t_right eq '')
{
push @lines,
HTML_trtd( {class => 'heading', colspan => 3}, $title);
} else
{
push @lines,
HTML_trtd( {class => 'heading', colspan => 2}, $title),
HTML_td( {align => 'right'}, $t_right);
}





if (@item_refs)
{
$item_refs[$#item_refs]->[2] .= GBSHTML_insert_nav_2( '', 'Bottom');
foreach my $item_ref (@item_refs)
{
my ( $label, $value, $right ) = @{$item_ref};

if ($value eq '' && $right eq '')
{
push @lines, HTML_trtd( {class => 'subheading', colspan => 3}, $label);
} elsif ($value eq '')
{
push @lines, HTML_trtd( {class => 'subheading', colspan => 2}, $label);
push @lines,   HTML_td( {class => 'subheading'}, $right);
} elsif ($right eq '')
{
push @lines, HTML_trtd( {class => 'subheading'}, $label);
push @lines,   HTML_td( {class => 'subheading', colspan => 2}, ": $value");
} else
{
push @lines, HTML_trtd( {class => 'subheading'}, $label);
push @lines,   HTML_td( {class => 'subheading'}, ": $value");
push @lines,   HTML_td( {align => 'right'}, $right);
}
}
}




push @lines,
HTML_table_e();

return @lines;
}




sub GBSHTML_section_head($$)
{
my ($name,
$sect_nr,   	# 1 == first (no prev), < 0 == last (no next)
) = @_;
my @lines;

my $this = abs $sect_nr;
my $prev = $this - 1;
my $next = ($sect_nr < 0) ? 0 : $this + 1;

my $sect_prev = ($prev == 0) ? undef : "SECT_$prev";
my $sect_next = ($next == 0) ? undef : "SECT_$next";

push @lines, HTML_h( 2, {id => "SECT_$this"} ,$name . GBSHTML_insert_nav_2( 'Top', 'Bottom', $sect_prev, $sect_next));

return @lines;
}




sub GBSHTML_doc_bottom($$)
{
my ($title,
$last_sect,	    # undef = none.
) = @_;
my @lines;

my $sect_last = (defined $last_sect) ? "SECT_$last_sect" : undef;

push @lines, HTML_div( {id => 'Bottom', class => 'doc-footer'}, $title . GBSHTML_insert_nav_2( 'Top', '', $sect_last, ''));

return @lines;
}




sub GBSHTML_doc_end()
{
return ( HTML_body_e(), HTML_doc_e());
}




sub GBSHTML_img($)
{
my ($name) = @_;

my ($url, $class, $alt) = @{$IMGS{$name}};
$url = "$IMAGES_DIR/$url";

my $html = HTML_img( $url, class => $class, alt => $alt);

return $html;
}




sub GBSHTML_insert_toc($)
{
my ($in_lines_ref) = @_;
my @out_lines;

my @toc_refs;


my @heading_refs;

my %heading_indexes;






{

my $i = 0;
my @heads = ('', 0, 0);



foreach my $line (@{$in_lines_ref})
{
if ($line ne '')
{
if ($line =~ /^<h(\d)(.*?)>(.*)<\/h\d>/)
{
my $h = $1;
my $h_arg = (defined $2) ? $2 : '';
my $title = $3;

$heads[$h]++;
map { $_ = 0 } @heads[$h+1..$#heads]
if ($h + 1 <= $#heads);

my $id;
if ($h_arg =~ /id="(.*)"/)
{
$id = $1;
} else
{
$id = "TOC" . join( '_', @heads);
$line = "<h$h id=\"$id\">$title</h$h>";	    # we are changing an input-line here!
}
if ($h == 1)
{
push @toc_refs, [ $id, $title ];
}
$heading_refs[$heads[1]]->[$heads[2]] = $id;
$heading_indexes{$id} = [ @heads ];

}
}
$i++;
}
}




foreach my $line (@{$in_lines_ref})
{
if ($line =~ /^\.TOC/)
{



push @out_lines, '<div class="indent">';
foreach my $ref (@toc_refs)
{
my ( $id, $title) = @{$ref};
push @out_lines, HTML_llink( $id, $title) . HTML_br();
}
push @out_lines, '</div>';
push @out_lines, '<br>';
} elsif ($line =~ /^<h(\d) id="(.*?)">/)
{








my $h = $1;
my $id = $2;
my @head = @{$heading_indexes{$id}};

my $sect = $head[1];
my $chap = $head[2];
my ($sect_prev, $sect_next, $chap_prev, $chap_next);




if ($chap == 0)
{
$sect_prev = $heading_refs[$sect-1]->[0]
unless ($sect == 0);
$sect_next = $heading_refs[$sect+1]->[0];
} else
{
$sect_prev = $heading_refs[$sect]->[0]
unless ($sect == 0);
$sect_next = $heading_refs[$sect+1]->[0];
}




if ($chap == 0 || $chap == 1)
{
$chap_prev = $heading_refs[$sect-1]->[@{$heading_refs[$sect-1]}-1]
unless ($sect == 0);
} else
{
$chap_prev = $heading_refs[$sect]->[$chap-1];
}
if ($chap == @{$heading_refs[$sect]}-1)
{
$chap_next = $heading_refs[$sect+1]->[1];
} else
{
$chap_next = $heading_refs[$sect]->[$chap+1];
}
my $nav = GBSHTML_insert_nav_3( $h, 'Top', 'Bottom', $sect_prev, $sect_next, $chap_prev, $chap_next);
$line =~ s/<\/h$h>/$nav<\/h$h>/;
push @out_lines, $line;
} else
{
push @out_lines, $line;
}
}
return @out_lines;
}




sub GBSHTML_insert_nav_2($$;$$)
{
my ($top,
$bottom,
$sect_prev,		# may be undef
$sect_next,		# may be undef
) = @_;


my @order_refs =

( [ 'medium', 'Prev. Section',   $sect_prev, 'Next Section', $sect_next ],
[ 'big',    'Top',	         $top,	     'Bottom',       $bottom ],
);

return insert_nav( \@order_refs);
}




sub GBSHTML_insert_nav_3($$$;$$$$)
{
my ($level,			# 1 = Section, 2 = Chapter
$top,
$bottom,
$sect_prev,		# may be undef
$sect_next,		# may be undef
$chap_prev,		# may be undef
$chap_next,		# may be undef
) = @_;


my @order_refs =

( [ 'small',  'Prev. Chapter',   $chap_prev, 'Next Chapter', $chap_next ],
[ 'medium', 'Prev. Section',   $sect_prev, 'Next Section', $sect_next ],
[ 'big',    'Top',	         $top,	     'Bottom',       $bottom ],
);
$order_refs[1]->[1] = 'Begin Section'	    # $p_title
if ($level == 2);

return insert_nav( \@order_refs);
}




sub insert_nav($)
{
my ($order_refs_ref) = @_;
my @lines;		# retuned as scalar string

push @lines, HTML_table_s( class => "nav");
push @lines, HTML_tr();
foreach my $ref (@{$order_refs_ref})
{
my ($image, $p_title, $prev, $n_title, $next) = @{$ref};

if ($prev)
{
push @lines, HTML_td( {title => $p_title}, HTML_llink( $prev, GBSHTML_img( "up_$image")));
} else
{
push @lines, HTML_td( GBSHTML_img( "blank_$image"));
}
if ($next)
{
push @lines, HTML_td( {title => $n_title}, HTML_llink( $next, GBSHTML_img( "down_$image")));
} else
{
push @lines, HTML_td( GBSHTML_img( "blank_$image"));
}
}
push @lines, HTML_table_e();

return "@lines";
}




sub init()
{
HTML_map( GBSDB_maps_get());

my $lock = FILE_lock_ex_wait( "$SILO_HTML_DIR/.html.lck");




foreach my $file (keys %FILES)
{
my $file_spec = "$SILO_HTML_DIR/$file";
ENV_copy_file( "$GBS_SCRIPTS_DOC_PATH/$file", $file_spec, 1);
ENV_chmod( 'u+w', $file_spec);
$FILES{$file} = $file_spec;
}




foreach my $script (keys %SCRIPTS)
{
my $file_spec = "$SILO_HTML_DIR/$script";
ENV_copy_file( "$GBS_SCRIPTS_DOC_PATH/scripts/$script", $file_spec, 1);
ENV_chmod( 'u+w', $file_spec);
$SCRIPTS{$script} = $file_spec;
}




foreach my $ref (values( %IMGS))
{
my $url = $ref->[0];		    # $url
my $file_spec = "$SILO_HTML_DIR/$url";
ENV_copy_file( "$GBS_SCRIPTS_DOC_PATH/images/$url", $file_spec, 1);
ENV_chmod( 'u+w', $file_spec);
}

FILE_unlock( $lock);

$IMAGES_DIR = $SILO_HTML_DIR;

$INITIALISED = 1;
}

return 1;
