#===============================================================
#
#   gbsjob.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbsjob;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSJOB_init
GBSJOB_sub
GBSJOB_run
GBSJOB_wait
);

use strict;
use integer;

use glo::env;
use glo::time;
use glo::spit;
use glo::slurp;
use glo::format;
use glo::shell;
use glo::genopt;
use glo::proc;
use gbscmd;
use gbsres;
use bgnotifier;





sub GBSJOB_init($$);
sub GBSJOB_sub($$$);
sub GBSJOB_run();
sub GBSJOB_wait();

sub submit();
sub subwin();
sub determine_states();
sub display_jobs($);
sub create_job_file();
sub get_sum_state($);




my $GBS_LOG_PATH = ENV_getenv_perl_path( 'GBS_LOG_PATH');

my @JOB_REFS;



my (
$SUBMODE,			    # Fixed: 'SUBWIN' or 'SUBMIT'
$JOBNAME,			    # Fixed
$TARGET,			    # Vary
$ACTION,			    # Vary: 'gen' or 'make' or audit_name
$DATE_TIME,			    # Fixed
$OS_LOGFILE,		    # Vary: '-' or logfile
$MUST_DISPLAY_NOTIFY_MSG,	    # Fixed
$ROOT,			    # Fixed
$EXEC,			    # Vary
$DISPLAY,			    # Vary
$COMMAND_ARGS,		    # Vary
$COMMENT,			    # Fixed
);

my $DELAY;
my $MUST_MAIL;

my $BG_RC;




sub GBSJOB_init($$)
{
my ($command_name,
$submit_ref,        # undef == --fg+ (foreground)

) = @_;

if (defined $submit_ref)
{
$SUBMODE = 'SUBMIT';
($DELAY,
$MUST_MAIL,
$MUST_DISPLAY_NOTIFY_MSG,
$COMMENT) = @{$submit_ref};
$OS_LOGFILE = undef;
$DISPLAY = ENV_getenv( 'DISPLAY');
} else
{
$SUBMODE =  'SUBWIN';
$DELAY = undef;
$MUST_MAIL = undef;
$MUST_DISPLAY_NOTIFY_MSG = 0;
$COMMENT = '';
$OS_LOGFILE = '-';
$DISPLAY = '-';
}
$JOBNAME = $command_name;
$DATE_TIME = GBSRES_numtime_packed();
$ROOT = ENV_getenv( 'GBS_ROOT_PATH');
my @command_args = grep( ! /^--c=/, GENOPT_get_all());
$COMMAND_ARGS = "@command_args";
}




sub GBSJOB_sub($$$)
{
my ($target,
$action,    # gen, make or <audit>
$exec) = @_;

push @JOB_REFS, [ $target, $action, $exec, undef, undef ];
}




sub GBSJOB_run()
{
if (@JOB_REFS)
{
if ($SUBMODE eq 'SUBMIT')
{
my @target_action_refs;
foreach my $ref (@JOB_REFS)
{
my ($target, $action, $exec,) = @{$ref};
$TARGET = $target;
$ACTION = $action;	    # JOBTYPE: 'gen', 'make' or $audit_name
$EXEC = "$JOBNAME $exec";
$OS_LOGFILE = ENV_os_path( GBSRES_logfile( $JOBNAME, $target, $action));
$ref->[3] = $OS_LOGFILE;
$ref->[4] = 'SUBMIT';
submit();
push @target_action_refs, [ $target, $action ];
}
BGNOTIFIER_submit( $JOBNAME, $DATE_TIME, @target_action_refs);

display_jobs( "Submitted $JOBNAME Jobs");

if (ENV_is_lunix())
{
ENV_say( 1, "Jobs queue:");
PROC_at_show();
}
} else  # SUBWIN
{
foreach my $ref (@JOB_REFS)
{
my ($target, $action, $exec,) = @{$ref};
$TARGET = $target;
$ACTION = $action;
$EXEC = "$JOBNAME $exec";
subwin();
}
}
}
}




sub submit()
{
my $job_file = create_job_file();
my $command_ref = ENV_command_ref( 'gbsjob.pl', [ $job_file ]);
my $command = ENV_command_line( $command_ref);
$command = SHELL_redirect_all( $command, $OS_LOGFILE);
my $command_name = "$JOBNAME ($TARGET / $ACTION)";






my $submit_cmd = GBSCMD_submit( $command, $command_name, $OS_LOGFILE);




PROC_at( $submit_cmd, $command_name, $DELAY, $MUST_MAIL);
}




sub subwin()
{
my $job_file = create_job_file();
my $command_ref = ENV_command_ref( 'gbsjob.pl', [ $job_file ]);
my $command = ENV_command_line( $command_ref);
my $command_name = "$JOBNAME ($TARGET / $ACTION)";

ENV_say( 1, "Subwin $JOBNAME");




my $subwin_cmd = GBSCMD_subwin( $command, $command_name);

($subwin_cmd, my $subwin_args_ref) = ENV_command_split( $subwin_cmd);
my $subwin_cmd_ref = ENV_command_ref( $subwin_cmd, $subwin_args_ref);




PROC_spawn_xterm( $subwin_cmd_ref, $command_name, undef, undef, undef, undef, undef);
}




sub GBSJOB_wait()
{

$BG_RC = 0;

if (@JOB_REFS)
{
ENV_say( 1, "Waiting...");
sleep( 2);
my $active_jobs;
do
{
$active_jobs = determine_states();
sleep( 10)
unless( $active_jobs == 0);
} while ($active_jobs > 0);
}

return $BG_RC;
}




sub determine_states()
{
my $active_jobs = 0;

my $changed = 0;
foreach my $ref (@JOB_REFS)
{
my ($target, $action, $exec, $logfile, $state) = @{$ref};
if ($state ne 'DONE' && $state ne 'FAILED')
{
if (-f "$logfile.sum")
{
my ($sum_state, $sum_rc) = get_sum_state( "$logfile.sum");
if ($sum_rc == 0)
{
$ref->[4] = 'DONE';
} else
{
$ref->[4] = 'FAILED';
$BG_RC = $sum_rc
if ($sum_rc > $BG_RC);
}
$changed = 1;
} else
{
$active_jobs++;
if ($state eq 'SUBMIT')
{
if (-f $logfile)
{
$ref->[4] = 'RUNNING';
$changed = 1;
}
}
}
}
}
display_jobs( 'Changed ' . TIME_time2num())
if ($changed);

return $active_jobs;
}




sub display_jobs($)
{
my ($head_text) = @_;

ENV_say( 1, "$head_text:");

my @line_refs;
foreach my $ref (@JOB_REFS)
{
my ($target, $action, $exec, $logfile, $state,) = @{$ref};
push @line_refs, [ $state, $target, $action, $logfile ];
}
ENV_say( 0, FORMAT_table( undef, 4, ' ', undef, @line_refs));
}




sub create_job_file()
{
my $file_name;

my @lines = map { "$_\n" }
(
$SUBMODE,
$JOBNAME,
$TARGET,
$ACTION,
$DATE_TIME,
$OS_LOGFILE,
$MUST_DISPLAY_NOTIFY_MSG,
$ROOT,
$EXEC,
$DISPLAY,
$COMMAND_ARGS,
$COMMENT,
);


$file_name = SPIT_tmp_file( 'gbsjob', [ @lines ]);

return $file_name;
}




sub get_sum_state($)
{
my ($sum_file) = @_;


my ($jobname,
$target,
$uname,
$root,
$logfile,
$time_start,
$time_end,
$time_diff,
$timer_diff,
$state,
$rc,
$action,
$date_time) = SLURP_file( $sum_file);

return ($state, $rc);
}

return 1;


