#===============================================================
#
#   gbsmake.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
# Basic 'make' flags:
#
# -d	Displays the reasons why make  chooses	to  rebuild  a
#	target;  make  displays  any and all dependencies that
#	are newer. In addition, make displays options read  in
#	from the MAKEFLAGS environment variable.
# -i	Ignores error codes returned by commands
# -k	keep-going: When a nonzero error status is returned by a
#	rule,  or when  make  cannot  find  a rule, abandons work
#	on the current target, but continues  with  other  dependency
#	branches that do not depend on it.
# -n	No execution mode.  Prints commands, but does not exe-
#	cute them. Even lines beginning with an @ are printed.
# -r	Does not read in the default makefile
#	/usr/share/lib/make/make.rules.
# -s	Silent mode.  Does not print command lines before exe-
#	cuting them.
# -S	Undoes the effect of the -k option.  Stops  processing
#	when a non-zero exit status is returned by a command.
# -t	Touches the target files (bringing them up to date)
#	rather than performing their rules.
# -B	Unconditional.
#
# -p	Print database
#===============================================================

print "*GBSMAKE*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::spit;
use glo::genopt;
use gbsenv;
use gbscmd;
use exec;
use validate;
use gbsoptenv;
use files;
use cfo;
use mm;




sub easter_egg(@);




my $GBS_ROOT_PATH;
my $GBS_SUBSYS;
my $GBS_COMPONENT;
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my $GBS_COMP_PATH;






$| = 1;
my $RC = 0;
GBSENV_init();




my @genopts = (
[ '<*>',    'comp_file_opts',   'sao', "", "bld files to make with 'make' options" ],
[ 'r',	'reverse_type',	    'bso',  0, "Specify source-types instead of build-types" ],
[ 'i',      'ignore_errors',    'bso',  0, "Continue generation after error(s)" ],
[ 'smm',    'start_makemake',   'bso',  0, "Pre-generate the make-file" ],
[ 'export', 'run_gbsexport',    'bso',  0, "Run 'gbsexport' on completion" ],
[ 'mm',     'run_makemake',	    'bso',  0, "Run 'makemake' on completion" ],
[ 'jobs',	'jobs',	       'isor1..9',  2, 'Max nr parallel jobs' ],
);
my @genenvs = qw( DEBUGGER MODE OPT MAP FLAGS_* APP_* );
GENOPT_set_optdefs( 'gbsmake', \@genopts,
"'make' one or more files",
undef);
GENOPT_set_envdefs( \&GBSOPTENV_set, \&GBSOPTENV_print, 'GBS_', \@genenvs);
GENOPT_parse();
my @COMP_FILE_OPTS = GENOPT_get( 'comp_file_opts');
my $REVERSE_TYPE = GENOPT_get( 'reverse_type');
my $IGNORE_ERRORS = GENOPT_get( 'ignore_errors');
my $START_MAKEMAKE = GENOPT_get( 'start_makemake');
my $RUN_GBSEXPORT = GENOPT_get( 'run_gbsexport');
my $RUN_MAKEMAKE = GENOPT_get( 'run_makemake');
my $JOBS = GENOPT_get( 'jobs');

easter_egg( @COMP_FILE_OPTS);




ENV_setenv( GBS_IGNORE_ERRORS => $IGNORE_ERRORS);




($GBS_ROOT_PATH, $GBS_SUBSYS) = VALIDATE_subsys_full_gbs();
$GBS_COMPONENT = ENV_getenv( 'GBS_COMPONENT');
VALIDATE_target();

$GBS_COMP_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/comp";




my $GBS_GEN_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/gen";
my $makefile = FILES_makefile_spec( $GBS_GEN_PATH, $GBS_TARGET);

if ($START_MAKEMAKE || !-e $makefile)
{
ENV_say( 1, "Pre MakeMake...");
$RC = MM_makemake();
if ($RC != 0)
{
my $text = 'Pre MakeMake (--smm) failed!';
if ($IGNORE_ERRORS)
{
ENV_error( $text);
} else
{
ENV_error_exit( $text);
}
}
}

ENV_fatal( "No Makefile generated for this Target ($GBS_TARGET)")
if ( !-e $makefile);




my $NR_CFO = CFO_parse_make( \@COMP_FILE_OPTS, $REVERSE_TYPE);
my @MAKE_OPTIONS = CFO_pre_options();




my @file_specs;
if ($NR_CFO == 0)
{
ENV_print( 1, "No make-target(s) specified - ");
if ($GBS_COMPONENT eq '')
{
@file_specs = ('ALL');
ENV_say( 0, "'ALL' assumed");
} else
{
@file_specs = ($GBS_COMPONENT);
ENV_say( 0, "Current Component ($GBS_COMPONENT) assumed");
}
} else
{



for (my $i = 0; $i < $NR_CFO; $i++)
{
my (undef, $this_comp, $files_ref, undef) = CFO_get_item( $i);
my @files = @{$files_ref};
if (@files)
{
my $bld_target_root = "$GBS_COMP_PATH/$this_comp/bld/$GBS_TARGET";
map { push @file_specs, "$bld_target_root/$_" } @files;
} else
{
push @file_specs, $this_comp;
}
}
}




{



my $make = GBSCMD_make();

my @flags_make;
push @flags_make, ENV_getenv( 'GBS_FLAGS_MAKE');
push @flags_make, @MAKE_OPTIONS;
push @flags_make, '-i' if ($IGNORE_ERRORS);
push @flags_make, '-n',	    # No execution mode. Print commands, do not execute
'-r',	    # do not use standard make-rules
'-f',	    # next is make-file
ENV_os_path( $makefile);
my @b_files = map ENV_os_path( $_), @file_specs;

my $command_ref = ENV_command_ref( $make, [ @flags_make, @b_files ]);




ENV_say( 1, "Executing 'make'...");
my @stdout_lines;
my $rc = ENV_run3( $command_ref, undef, [ 0, 1 ,2 ], \@stdout_lines, undef);
ENV_error_exit( "'make' failed (rc=$rc)")
if ($rc != 0);




my @lines;
foreach my $line (@stdout_lines)
{
next if ($line eq '');

if ($line =~ /^gbsgen (.+)/)
{
push @lines, $1;    # gbsgen lines from 'make'
} else
{
ENV_say( 0, "$line");    # other lines from 'make'
}
}

if (@lines)
{
my $gbsgen_file = SPIT_tmp_file( 'gbsmake_gbsgen.tmp');
SPIT_file_nl( $gbsgen_file, \@lines);




my @gbsgen_opts;
push @gbsgen_opts, '--i' if ($IGNORE_ERRORS);
push @gbsgen_opts, '--export' if ($RUN_GBSEXPORT);
push @gbsgen_opts, "--jobs=$JOBS";
$RC = EXEC_run( 'gbsgen.pl', [ "--inc=$gbsgen_file", @gbsgen_opts ]);
unlink $gbsgen_file;

if ($RUN_MAKEMAKE && ($RC == 0 || $IGNORE_ERRORS))
{
ENV_say( 1, "Post MakeMake...");
my $this_rc = MM_makemake();
$RC = $this_rc if ($this_rc > $RC);
if ($this_rc != 0)
{
ENV_error( "Post MakeMake (--mm) failed!");
}
}
} else
{
ENV_say( 1, "Nothing to do");
$RC = EXEC_run( 'gbsexport.pl', undef)
if ($RUN_GBSEXPORT);
}
}




ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}




sub easter_egg(@)
{
if (scalar @_ == 2 && $_[0] eq 'my' && $_[1] eq 'day')
{
ENV_say( 0, "*******************************",
"** DO YOU FEEL LUCKY TODAY,  **",
"** PUNK?                     **",
"*******************************");
ENV_beep();
ENV_exit();
}
}


