#==============================================================
#
#   gbsnew.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
print "*GBSNEW*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::spit;
use glo::ask;
use glo::slurp;
use glo::genopt;
use glo::build;
use glo::format;
use glo::scm;
use glo::file;
use gbsenv;
use gbsglo;
use gbsscm;
use gbsdb;
use roots;
use fcreate;
use fspec;
use files;
use system;
use validate;
use plugin;
use profile;




sub main_new_root($);
sub create_or_update_root($);
sub ask_new_root_data($);
sub print_root_cols();
sub format_roots();
sub validate_user();
sub main_new_other_root($);
sub main_new_subsys($);
sub ask_new_subsys_data($);
sub main_new_component($);
sub check_component($$$);
sub main_new_target($);
sub main_remove_target($);
sub select_via_menu();




my $CWD = ENV_cwd();

my $GBS_ROOT_PATH = ENV_getenv_perl_path( 'GBS_ROOT_PATH');
my $GBS_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $GBS_COMPONENT = ENV_getenv( 'GBS_COMPONENT');
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my @GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');

my @FUNCTION_REFS = (
[ 'exit',		undef			],
[ 'new_root',	\&main_new_root		],
[ 'new_other_root',	\&main_new_other_root   ],
[ 'new_subsys',	\&main_new_subsys	],
[ 'new_component',	\&main_new_component	],
[ 'new_target',	\&main_new_target	],
[ 'rem_target',	\&main_remove_target	],
);
my %FUNCTION_REFS = map { $_->[0] => $_ } @FUNCTION_REFS;










$| = 1;
my $RC = 0;
my $RESULT = '';    # The (first) item created: $new_root, $new_subsys, etc or ''
GBSENV_init();




my @genopts = (
[ '<1>',	'function', 'ssosnew_root,new_other_root,new_subsys,new_component,new_target,rem_target',  '', 'Function to be executed' ],
[ '<2>',	'func_arg', 'sso',                               '', 'Function dependent argument' ],
);
my @genconflicts = (
);
GENOPT_set_optdefs( 'gbsnew', \@genopts,
'Root/System, SubSystem, Component and Target Create/Remove/etc',
undef);
GENOPT_set_conflicts( \@genconflicts);
GENOPT_parse();
my $FUNCTION = GENOPT_get( 'function');
my $FUNC_ARG = GENOPT_get( 'func_arg');




VALIDATE_gbs();




{
if ($FUNCTION eq '')
{
select_via_menu();
} else
{
my $func = $FUNCTION_REFS{$FUNCTION}->[1];
$RESULT = &$func( $FUNC_ARG);
}
}




{




my @lines;
push @lines, "$RESULT\n";
foreach my $env_name (ENV_get_changed_envs())
{
my $value = ENV_getenv( $env_name);
push @lines, "$env_name=$value\n";
}

my $file_spec = ENV_getenv( 'GBS_NEW_FILE');
if ($file_spec)
{
SPIT_file( $file_spec, \@lines);
}
}

ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 0);
}




sub main_new_root($)
{
my ($spec_new_root) = @_;		# perl_format
my $new_root_path = '';

ENV_say( 1, "New Root ($spec_new_root)");

validate_user();




if (ROOTS_get_nr() > 0)
{
ENV_say( 1, "Currently in Roots-List:");
print_root_cols()
} else
{
ENV_say( 1, "No Roots in Roots-List");
}
ENV_say( 0, '');

my $current_root_parent_path = '';
my $last_root_parent_path = '';
my @menu_refs = (
[ '', '' ]		    # <none> entry
);


if ($spec_new_root eq '')
{
$current_root_parent_path = ($GBS_ROOT_PATH eq '') ? '' : ENV_parent_path( $GBS_ROOT_PATH);

if ($current_root_parent_path ne '')
{
ENV_say( 1, "Current Root Parent is '$current_root_parent_path'");
push @menu_refs, [ "Create a Root in the Current Parent Path ($current_root_parent_path)", 'ADD_CURPAR' ];
} else
{
my $last_root_path = GBSDB_last_root_get();
if ($last_root_path ne '')
{
$last_root_parent_path = ENV_parent_path( $last_root_path);
ENV_say( 1, "Last Used Root Parent is '$last_root_parent_path'");
push @menu_refs, [ "Create a Root in the Last Used Parent Path ($last_root_parent_path)", 'ADD_LASTPAR' ];
}
}
if ($CWD !~ /^$GBS_ROOT_PATH/ || $GBS_ROOT_PATH eq '')
{
ENV_say( 1, "Currenty (CWD) in '$CWD'");
push @menu_refs, [ 'Use the Current Path (CWD) as Root Path', 'USE_ROOT' ];
push @menu_refs, [ 'Use the Current Path (CWD) as Parent Path', 'USE_PAR' ];
}
push @menu_refs, [ 'Explicitly specify a Root Path', 'SPEC_ROOT' ];
} else
{



$spec_new_root = ENV_abs_path( undef, $spec_new_root);
my $exists = (-d $spec_new_root) ? 'Exist' : 'NOT Exists';
ENV_say( 1, "Specified '$spec_new_root' ($exists)");
push @menu_refs, [ 'Use the specified Path as Root Path', 'USE_SPEC_ROOT' ];
push @menu_refs, [ 'Use the specified Path as Parent Path', 'USE_SPEC_PAR' ];
}

my ($select_value, @more_values) = ASK_value_from_menu( 'Select Root Create action:', 1 , @menu_refs);


if ($select_value ne '')
{
my $parent_path = '';
my $root_dir = '';
my $root_path = '';	# $parent_path/$root_dir
if ($select_value eq 'ADD_CURPAR')
{
$parent_path = $current_root_parent_path;
} elsif ($select_value eq 'ADD_LASTPAR')
{
$parent_path = $last_root_parent_path;
} elsif ($select_value eq 'USE_ROOT')
{
$root_path = $CWD;
} elsif ($select_value eq 'USE_PAR')
{
$parent_path = $CWD;
} elsif ($select_value eq 'SPEC_ROOT')
{

} elsif ($select_value eq 'USE_SPEC_ROOT')
{
$root_path = $spec_new_root;
} elsif ($select_value eq 'USE_SPEC_PAR')
{
$parent_path = $spec_new_root;
} else
{
ENV_fatal( "Impossible $select_value");
}
if ($root_path ne '')
{
$parent_path = ENV_parent_path( $root_path);
$root_dir = ENV_parent_dir( $root_path, -1);
}





if ($parent_path eq '')
{



$root_path = ASK_path( "Enter New Root Path", $CWD, 0, -2);
} elsif ($root_dir eq '')
{



ENV_say( 1, "Root Parent is '$parent_path'");
ENV_say( 0, "And contains:");
my @dirs = SLURP_dir_dirs( $parent_path, 0);
ENV_say( 0, FORMAT_cols( 0, 2, '  ', 2, \@dirs));
$root_dir = ASK_dir( 'Enter New Root Dir', '', $parent_path, 0, -2);
$root_path = "$parent_path/$root_dir";
} else
{

ENV_say( 0, "OK");
}
ENV_say( 0, '');
ENV_say( 1, "Selected path:",
"$root_path");
if (! -d $root_path)
{
my $ans = ASK_YN( "Root does not exist. Create?", 'Y');
if ($ans eq 'Y')
{
FILE_mkpath( $root_path);
} else
{
$root_path = '';
}
}

if ($root_path ne '')
{



my @root_files = SLURP_dir_all( $root_path, 0);
if (@root_files)
{
ENV_say( 1, "This Root already contains data:",
"@root_files");
if (ASK_YNQ( 'Re-create this Root/System?', 'N', 1))
{
$new_root_path = $root_path;
}
} else
{
$new_root_path = $root_path;
}




if ($new_root_path ne '')
{

create_or_update_root( $new_root_path);
}
}
}

ENV_warn( "No new Root/System created or updated")
if ($new_root_path eq '');

return $new_root_path;
}




sub create_or_update_root($)
{
my ($new_root_path,
) = @_;







my ( $system_name,
@scms_items ) = ask_new_root_data( $new_root_path);




SCM_preset( $new_root_path, [ @scms_items ]);
GBSSCM_setenv( [ @scms_items ]);
GBSSCM_connect();




ENV_say( 1, "Validating Root: $new_root_path");
my $again;
do
{
$again = 0;
my $state = SCM_get_states( $new_root_path);




if ($state == -1)
{
ENV_warn( "Root directory not under SCM control");
$again = 1;
} elsif ($state == 0)
{
SCM_checkout( $new_root_path);
}
if ($again)
{
ENV_say( 1, "You now have the chance to fix things. (like Checkout WA)",
"Then enter Y to Retry. Enter N or Q to Quit");
if (ASK_YNQ( "Retry?", 'Y', 1) eq 'Y')
{
GBSSCM_preset( 0);	    # reset scm
} else
{
ENV_exit();
}
}
} while ($again);




ENV_say( 1, "Creating System ...");
my @new_dirs = FILES_mandatory_dirs( $new_root_path, 'SYSTEM');
SCM_add_dirs( \@new_dirs, \&FILES_get_ignore_spec, 0, 1);	    # $pre_validated, $verbose

my @new_files;




push @new_files, FCREATE_switch( 'switch.gbs', $new_root_path, $system_name);




push @new_files, FCREATE_sys( 'sys.gbs', $new_root_path, $system_name);
SCM_add_txt_files( \@new_files, 0, 1);		# $prevalidated, $verbose




ENV_say( 1, "Creating system file");
my $this_gbs_version = BUILD_get_version();
my $gbs_build = ENV_getenv( 'GBS_BUILD');   # VVVYYMMDD

SYSTEM_create( $new_root_path,
build_start	    => $gbs_build,
system_name	    => $system_name,
scms	    => $scms_items[0],	    # $new_scms,
scms_repository => $scms_items[1],	    # $new_scms_repository,
scms_data	    => $scms_items[2],	    # $new_scms_data,
root_version    => $this_gbs_version,
targets	    => [],
);
ENV_say( 1, "Root/System '$new_root_path' created");
$GBS_ROOT_PATH = $new_root_path;


ENV_say( 1, "Setting GBS Administrator privilege...");
my @gbs_admin = ENV_getenv( 'GBS_ADMINISTRATOR');
push @gbs_admin, $system_name;
GBSENV_setenv( GBS_ADMINISTRATOR => \@gbs_admin);
PROFILE_open( undef);
PROFILE_add( GBS_ADMINISTRATOR => $system_name);
PROFILE_close(1);

}




sub ask_new_root_data($)
{
my ($new_root) = @_;
my ($system_name, $scms, $scms_repository, $scms_data);




($system_name) = $new_root =~ m!.*[/\\](.+)!;	 #last directory
$system_name = ucfirst $system_name;
$scms = ENV_getenv( "GBS_SCMS");
$scms_data = undef;
$scms_repository = undef;

if ($scms ne '')
{
my @scms_items = ($scms, $scms_repository, $scms_data);
SCM_preset( $new_root, [ @scms_items ]);
}

do
{



$system_name = ASK_text( "Enter New System-Name", $system_name, -1);




($scms, $scms_repository, $scms_data) = SCM_select( $scms, $new_root, $scms_repository, $scms_data);




ENV_say( 1, "Summary for new Root:");
ENV_say( 0, "  Root           : $new_root",
"  System_name    : $system_name",
"  SCMS           : $scms");
ENV_say( 0, "  SCMS-data      : $scms_data") if (defined $scms_data);
ENV_say( 0, "  SCMS-repository: $scms_repository") if (defined $scms_repository);
} while (ASK_YNQ( 'OK?', 'Y', 1) eq 'N');

$scms_data = '' if (!defined $scms_data);
$scms_repository = '' if (!defined $scms_repository);
return ( $system_name, $scms, $scms_repository, $scms_data );
}




sub validate_user()
{
if (0)
{



ENV_say( 1, "==========================================================",
"== GBS is consultancy-ware",
"== The creation of a new root is restricted by the",
"== licence agreement to consultants appointed by",
"== Randy Marques Consultancy",
"==========================================================");




my $digest = 'MaGP3Dn/Ix.uY';
my $ok = ASK_password( 'Enter GBS Consultancy password', $digest);
ENV_error_exit( 'Invalid Password')
if (!$ok);
}
}




sub print_root_cols()
{
ENV_say( 0, FORMAT_cols( 0, 2, '  ', 2, [ format_roots() ]));
}




sub format_roots()
{
my @roots;

foreach my $root (ROOTS_get_all( 1))
{
my $fc = '?';
if (-e $root)
{
if (FILES_check_system( $root))
{
$fc = ' ';
}
if (FILES_check_system_other_platform( $root))
{
$fc = ($fc eq ' ') ? '+' : '*';
}
} else
{
$fc = 'x';
}
push @roots, "$fc$root";
}

return @roots;
}



sub main_new_other_root($)
{
my ($spec_new_root) = @_;		# perl_format
my $new_root_path = $spec_new_root;

ENV_say( 1, "Adjust existing Root for this Platform ($spec_new_root)");







ENV_say( 1, "Platform...");
my $system_name = SYSTEM_get( $new_root_path, 'system_name');
my $new_scms = SYSTEM_get( $new_root_path, 'scms');
my $new_scms_repository = SYSTEM_get( $new_root_path, 'scms_repository');
my $new_scms_data = SYSTEM_get( $new_root_path, 'scms_data');
ENV_say( 1, "Current SCMS      : $new_scms",
"Current Repository: $new_scms_repository");




my $must_write_repos = 0;
if ($new_scms_repository eq '')
{
my @repos;
foreach my $osname (SYSTEM_get_osnames( $new_root_path, 'scms_repository'))
{
my $repos = SYSTEM_os_get( $new_root_path, 'scms_repository', $osname);
push @repos, $repos
if ($repos);
}
if (@repos)
{
foreach my $repos (@repos)
{
ENV_say( 1, "SCM REPOSITORY: $repos");
if (ASK_YNQ( 'OK?', 'Y', 1) eq 'Y')
{
$new_scms_repository = $repos;
last;
}
}
$new_scms_repository = ASK_text( 'Enter Repository', '', 0)
if ($new_scms_repository eq '');
} else
{
$new_scms_repository = '';
}
$must_write_repos = 1;
}




my @scms_items = ($new_scms, $new_scms_repository, $new_scms_data);
SCM_preset( $new_root_path, [ @scms_items ]);
GBSSCM_setenv( [ @scms_items ]);
GBSSCM_connect();




SYSTEM_write( $new_root_path, scms_repository => $new_scms_repository)
if ($must_write_repos);




my @new_files;
push @new_files, FCREATE_switch( 'switch.gbs', $new_root_path, $system_name);
SCM_add_txt_files( \@new_files, 0, 1);	    		# $prevalidated, $verbose
ENV_say( 1, "Root '$new_root_path' adjusted for this Platform");


return $new_root_path;
}




sub main_new_subsys($)
{
my ($new_subsys) = @_;

ENV_say( 1, "New SubSys: $new_subsys");

VALIDATE_root();
VALIDATE_administrator();

GBSSCM_preset( 0);
GBSSCM_connect();




my $gbs_dev_root = "$GBS_ROOT_PATH/dev";
my @subsystems = GBSGLO_subsystems();
ENV_say( 1, "Current SubSystems:");
if (@subsystems)
{
ENV_say( 0, FORMAT_indent( 0, 2, [ 0, "@subsystems" ]));
} else
{
ENV_say( 0, '  None');
}
do
{
$new_subsys = ASK_dir( "Enter New SubSystem-name", "", $gbs_dev_root, 0, 0);
if (grep $new_subsys eq $_, @subsystems)
{
ENV_beep();
ENV_say( 2, "- SubSystem '$new_subsys' already exists");
$new_subsys = undef if (ASK_YN( "Continue with '$new_subsys'?", 'Y') eq 'N');
}
} while (!defined $new_subsys);

if ($new_subsys ne "")
{




my ($full_gbs, $import_dir, $export_dir, $res_sub_dir,
$targets_ref, $audits_ref) = ask_new_subsys_data( $new_subsys);

my $gbs_subsys_root = "$gbs_dev_root/$new_subsys";
my @new_dirs;
push @new_dirs, $gbs_subsys_root;
if ($full_gbs)
{
push @new_dirs, FILES_mandatory_dirs( $gbs_subsys_root, 'SUBSYS_GBS');
push @new_dirs, map { "$gbs_subsys_root/gen/$_" } @{$targets_ref};
push @new_dirs, map { "$gbs_subsys_root/audit/$_" } @{$audits_ref};
} else
{
push @new_dirs, FILES_mandatory_dirs( $gbs_subsys_root, 'SUBSYS_NON_GBS');
}
push @new_dirs, "$gbs_subsys_root/import"
if ($import_dir);
push @new_dirs, "$gbs_subsys_root/export"
if ($export_dir);
push @new_dirs, "$GBS_ROOT_PATH/res/$new_subsys"
if ($res_sub_dir);




ENV_say( 1, "Creating SubSystem '$new_subsys'...");
SCM_add_dirs( \@new_dirs, \&FILES_get_ignore_spec, 0, 1);	    # $pre_validated, $verbose

my @new_files;
if ($full_gbs)
{
push @new_files, FCREATE_full_gbs( $GBS_ROOT_PATH, $new_subsys);
} else
{
unlink FSPEC_full_gbs( $GBS_ROOT_PATH, $new_subsys);
push @new_files, FCREATE_export( 'export.gbs', $GBS_ROOT_PATH, $new_subsys, undef);
foreach my $script (FSPEC_subsys_script_names())
{
push @new_files, FCREATE_subsys_script( $script, $GBS_ROOT_PATH, $new_subsys);
}
}
SCM_add_txt_files( \@new_files, 0, 1);			# $prevalidated, $verbose

ENV_say( 1, "SubSystem '$new_subsys' created");
$GBS_SUBSYS = $new_subsys;

} else
{
ENV_warn( "No new SubSystem created");
}

return $new_subsys;
}




sub ask_new_subsys_data($)
{
my ($new_subsys) = @_;
my ($full_gbs, $import_dir, $export_dir, $res_sub_dir, @targets, @audits);

my @subsystems = GBSGLO_subsystems();




$full_gbs = 'Y';
$import_dir = 'N';
$export_dir = 'Y';
$res_sub_dir = (@subsystems) ? 'Y' : 'N';
@targets = ENV_getenv( 'GBS_TARGETS');
@audits = ENV_getenv( 'GBS_AUDITS');

do
{



$full_gbs = ASK_YN( "Full GBS?", $full_gbs);




$import_dir = ASK_YN( "Import directory required?", $import_dir)
if ($full_gbs eq 'Y');




$export_dir = ASK_YN( "Export directory required?", $export_dir);




$res_sub_dir = ASK_YN( "Res SubSystem directory required?", $res_sub_dir);

if ($full_gbs eq 'Y')
{


my @all_targets = ENV_getenv( 'GBS_ALL_TARGETS');
@targets = ASK_values_from_list( "Targets for SubSystem '$new_subsys'", "@targets", \@all_targets, 0);
@audits = ASK_values_from_list( "Audits for SubSystem '$new_subsys'",   "@audits", \@audits, 0)
if (@audits);
}




ENV_say( 1, "Summary for new SubSys '$new_subsys':");
ENV_say( 0, "  Full GBS    : $full_gbs");
ENV_say( 0, "  Import-dir  : $import_dir") if ($full_gbs eq 'Y');
ENV_say( 0, "  Export-dir  : $export_dir");
ENV_say( 0, "  Res-Sub-dir : $res_sub_dir");
ENV_say( 0, "  Target(s)   : @targets") if ($full_gbs eq 'Y');
ENV_say( 0, "  Audit(s)    : @audits") if ($full_gbs eq 'Y');
} while (ASK_YNQ( 'OK?', 'Y', 1) eq 'N');

$full_gbs = ($full_gbs eq 'Y');
$import_dir = ($import_dir eq 'Y' && $full_gbs);
$export_dir = ($export_dir eq 'Y');
$res_sub_dir = ($res_sub_dir eq 'Y');
return ($full_gbs, $import_dir, $export_dir, $res_sub_dir, [ @targets ], [ @audits ]);
}




sub main_new_component($)
{
my ($new_component) = @_;

ENV_say( 1, "New Component: $new_component");

VALIDATE_root();
VALIDATE_subsys_full_gbs();
VALIDATE_administrator();

GBSSCM_preset( 0);
GBSSCM_connect();

my $gbs_subsys_root = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS";
my $gbs_comp_root = "$gbs_subsys_root/comp";

my @current_comps = GBSGLO_components( $GBS_SUBSYS);
ENV_say( 1, "Current Components:");
if (@current_comps)
{
ENV_say( 0, FORMAT_indent( 0, 2, [ 0, "@current_comps" ]));
} else
{
ENV_say( 0, '  None');
}
my @all_targets = GBSGLO_subsystem_targets( $GBS_SUBSYS);
my @all_audits = GBSGLO_subsystem_audits( $GBS_SUBSYS);
my $default_targets = join( ',', @all_targets);
my $default_audits = join( ',', @all_audits);

my @comp_refs;

my $comp;
do
{



$comp = ASK_dir( "Enter New Component-name", $new_component, $gbs_comp_root, 0, 0,
[ \&check_component, [ @current_comps ] ]);
if ($comp ne "")
{



my @targets;
my @scope_comps;
my @audits;




@targets = ASK_values_from_list( "  Select Targets for Component '$comp'",
$default_targets, \@all_targets, 0);
$default_targets = join( ',', @targets);




if (@current_comps)
{
@scope_comps = ASK_values_from_list( "  Select Components in scope for Component '$comp'",
"", [ @current_comps ], 0);

}




if (@all_audits)
{
@audits = ASK_values_from_list( "  Select Audits for Component '$comp'",
$default_audits, \@all_audits, 0);
$default_audits = join( ',', @audits);
}

push @current_comps, $comp;
push @comp_refs, [ $comp, [ @targets ], [ @scope_comps ], [ @audits ] ];
}
} while ($comp ne '' && ASK_YNQ( "Add another Component?", 'N', 1) eq 'Y');

if (@comp_refs)
{
foreach my $ref (@comp_refs)
{
my ($comp_name, $targets_ref, $scopes_ref, $audits_ref) = @{$ref};
ENV_say( 1, "Creating Component: $comp_name");




my $component_root = "$gbs_comp_root/$comp_name";
my @new_dirs;
push @new_dirs, $component_root;
push @new_dirs, FILES_mandatory_dirs( $component_root, 'COMPONENT');
push @new_dirs, map { "$component_root/bld/$_" } @{$targets_ref};
push @new_dirs, map { "$component_root/aud/$_" } @{$audits_ref};
foreach my $target (@{$targets_ref})
{
push @new_dirs, map { "$component_root/aud/$_/$target" } @{$audits_ref};
}
SCM_add_dirs( \@new_dirs, \&FILES_get_ignore_spec, 0, 1);	    # $pre_validated, $verbose




my $scope_file = FCREATE_scope( 'scope.gbs', $GBS_ROOT_PATH, $GBS_SUBSYS,
$comp_name, $scopes_ref);
SCM_add_txt_files( $scope_file, 0, 1);		# $prevalidated, $verbose
}
ENV_say( 1, "". @comp_refs . " Component(s) created");




$new_component = $comp_refs[0]->[0];
$GBS_COMPONENT = $new_component;

} else
{
ENV_warn( "No new Components created");
$new_component = "";
}
return $new_component;
}




sub check_component($$$)
{
my ($values_ref,
$default_ref,
$existing_comps_ref,
) = @_;
my $error_txt = '';

my $comp = $values_ref->[0];
if (grep( $comp eq $_, @{$existing_comps_ref}))
{
$error_txt = "Component '$comp' already exists";
} elsif (grep( $comp eq $_, qw( ALL RES EXPORT)))
{
$error_txt = "'ALL', 'RES' and 'EXPORT' are reserved names";
} else
{
;
}
return $error_txt;
}




sub main_new_target($)
{
my ($default_target) = @_;	    # '' == ASK
my $new_target;

VALIDATE_root();
VALIDATE_administrator();

GBSSCM_preset( 0);

$new_target = PLUGIN_new_target( $default_target);

if ($new_target)
{
$GBS_TARGET = ENV_getenv( 'GBS_TARGET');
@GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');
}

return $new_target;
}




sub main_remove_target($)
{
my ($default_target) = @_;
my $target_to_remove;

VALIDATE_root();
VALIDATE_administrator();

GBSSCM_preset( 0);

$target_to_remove = PLUGIN_remove_target( $default_target);

if ($target_to_remove)
{
$GBS_TARGET = ENV_getenv( 'GBS_TARGET');
@GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');
}

return $target_to_remove;
}





sub select_via_menu()
{
my $again = 1;
do
{
my @menu_refs = (
[ 'Exit',					    undef ],
[ 'Create a new Root/System',		    \&main_new_root ],
[ 'Adjust existing Root for this Platform',	    \&main_new_other_root ],
);
push @menu_refs,
[ 'Create a new SubSys',			    \&main_new_subsys ]
if ($GBS_ROOT_PATH ne '');
push @menu_refs,
[ 'Create a new Component',			    \&main_new_component ]
if ($GBS_SUBSYS ne '' && GBSGLO_subsystem_is_full_gbs( $GBS_SUBSYS));
push @menu_refs,
[ 'Create a new Target',			    \&main_new_target ]
if ($GBS_ROOT_PATH ne '');
push @menu_refs,
[ 'Remove a Target',			    \&main_remove_target ]
if ($GBS_ROOT_PATH ne '');

my $function = ASK_value_from_menu( 'Select Function', 0, @menu_refs);
if (defined $function)
{
$RESULT = &$function( '');
} else
{
$again = 0;
}
} while ($again);
}



