#==============================================================
#
#   gbssfile.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbssfile;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSSFILE_get_all_structs
GBSSFILE_parse
GBSSFILE_get_section_items
GBSSFILE_get_subsection_items
GBSSFILE_get_subsection_subitems
GBSSFILE_set_init_envs
);

use strict;
use integer;

use glo::env;
use glo::path;
use glo::check;
use gbsenv;
use gbsfileglo;




sub GBSSFILE_get_all_structs($$);
sub GBSSFILE_parse($$$$$);
sub GBSSFILE_get_section_items($$@);
sub GBSSFILE_get_subsection_items($$$@);
sub GBSSFILE_get_subsection_subitems($$$$@);
sub GBSSFILE_set_init_envs($);

sub init_struct($);
sub init_sub_section($@);
sub process_item($$$);
sub type_string($$);
sub type_integer($$);
sub type_bool($$);
sub type_arrow($$);
sub type_command($$);
sub type_format($$);
sub constr_range($$);
sub constr_select($$);
sub type_format($$);
sub type_via_format($$);
sub constr_type($$);
sub validate_nr_args($$$);
sub get_failed($@);








my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();
my $IS_WIN32 = ENV_is_win32();

my $POS_PAR_TOKEN = ($IS_WIN32) ? '%' : '\$';

my %STRUCT_TYPES = (

target  => [ 'sysgen',   'build' ],
audit   => [ 'sysaudit', 'audit' ],
scadef  => [ undef,	     'audit' ],
tool    => [ 'systools', 'tools' ],
steps   => [ undef,	     'tools' ],
);

my %VALUE_TYPE_FUNCS = (
s => \&type_string,
i => \&type_integer,
b => \&type_bool,
a => \&type_arrow,
c => \&type_command,
f => \&type_format,
v => \&type_via_format,
);
my %VALUE_CONSTR_FUNCS = (
' ' => \sub {},			# do nothing
r   => \&constr_range,		# constr_data: minmax ref (integer values)
s   => \&constr_select,		# constr_data: ref to values-hash or -array (string values)
t	=> \&constr_type,		# constr_data: undef
);




my $STRUCT_REF;
my $STRUCT_FILE;	# target, audit, tool, steps
my $STRUCT_FILE_UC;	# uppercase
my $STRUCT_NAME;
my $PLUGIN_PATH;	# "$GBS_SCRIPTS_PATH/plugins/$plugin_location

my $CUR_SECTION;
my $CUR_SECTION_REF;
my $CUR_SUB_SECTION;
my $CUR_SUB_SECTION_REF;
my $CUR_ITEM;
my $SECTION_PREFIX;





sub GBSSFILE_get_all_structs($$)
{
my ($gbs_root,
$struct_type) = @_;	# target, audit, tool
my @all_structs;		# array or scalar

my $struct_location = $STRUCT_TYPES{$struct_type}->[0];
foreach my $struct_spec (glob "$gbs_root/$struct_location/*/$struct_type.gbs")
{
my $struct = (split( '/', $struct_spec))[-2]; # get the parent_dir name
push @all_structs, $struct;
}


return (wantarray) ? @all_structs : join( ',', @all_structs);
}



















sub GBSSFILE_parse($$$$$)
{
my ($struct_ref,		# OUT: Ref to HASH. here the results go. May be undef
$file_spec,		# steps.gbs, audit.gbs, target.gbs, tool.gbs (without .type)
$struct_name,		# e.g.: target_name, $audit_name, $tool_name
$inc_path_ref,		# IN: ONLY: audit.gbs, target.gbs, tool.gbs - NOT: steps.gbs
$sections_ref,		# IN: definition of syntax
) = @_;

my $full_file_spec = GBSFILEGLO_select_usr_gbs( $file_spec);    # .gbs or .usr

ENV_fatal( "File '$file_spec.gbs|.usr' does not exist")
if (!defined $full_file_spec);

$STRUCT_REF = $struct_ref;
init_struct( $sections_ref);

$STRUCT_NAME = $struct_name;			# e.g.: target_name, $audit_name, $tool_name
$STRUCT_FILE = ENV_split_spec_n( $full_file_spec);	# target, audit, tool, steps
$STRUCT_FILE_UC = uc $STRUCT_FILE;
$STRUCT_REF->{'.'}->{STRUCT_FILE} = $STRUCT_FILE;
$STRUCT_REF->{'.'}->{STRUCT_NAME} = $STRUCT_NAME;




{
my $plugin_location = $STRUCT_TYPES{$STRUCT_FILE}->[1];	# $plugin_location
$PLUGIN_PATH = ($plugin_location) ? "$GBS_SCRIPTS_PATH/plugins/$plugin_location" : undef;
GBSFILEGLO_open_file( $full_file_spec, $inc_path_ref, $PLUGIN_PATH, undef, 0);
}




$CUR_SECTION = '';
$CUR_SUB_SECTION = '';
$CUR_ITEM = '';
$SECTION_PREFIX = '';
my %sections_count = map { $_ => 0 } keys %{$sections_ref};

my $line = GBSFILEGLO_get_line();
while (defined $line)
{




GBSFILEGLO_error_exit( "Section-line expected")
if ($line =~ /^\s/);
($CUR_SECTION, my $section_value) = split( ' ', $line, 2);  # split on any whitespaces
$SECTION_PREFIX = $CUR_SECTION;
$section_value = ''
if (!defined $section_value);
my $section_def_ref = $sections_ref->{$CUR_SECTION};
GBSFILEGLO_error_exit( "Invalid section '$CUR_SECTION'")
if (!defined $section_def_ref);

$STRUCT_REF->{$CUR_SECTION} = {}
if (!defined $STRUCT_REF->{$CUR_SECTION});	# Create empty hash
$CUR_SECTION_REF = $STRUCT_REF->{$CUR_SECTION};
$CUR_SUB_SECTION = '';
$CUR_SUB_SECTION_REF = undef;
$CUR_ITEM = '';

my ($sec_type, $min_occur, $max_occur, $nr_args, $defs_ref, $section_entry_func, $section_exit_func) = @{$section_def_ref};
$sections_count{$CUR_SECTION}++;
my $is_subsection = ($sec_type eq 'S') ? 1 : 0;

if ($max_occur)
{
GBSFILEGLO_error_exit( "Maximum of $max_occur occurences of '$CUR_SECTION' allowed ($sections_count{$CUR_SECTION})")
if ($sections_count{$CUR_SECTION} > $max_occur);
}
my @section_values;
if (defined $nr_args)
{
@section_values = split( ' ', $section_value);  # split on any whitespaces
validate_nr_args( $CUR_SECTION, scalar @section_values, $nr_args);
} else
{
$section_values[0] = $section_value;
}


if ($is_subsection)
{
my %subsections = map { $_ => 1 } keys (%{$CUR_SECTION_REF});
foreach my $subsection (@section_values)
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: SubSection name cannot be '.'")
if ($subsection eq '.');
GBSFILEGLO_error_exit( "$SECTION_PREFIX: SubSection name ($subsection) already defined")
if (exists $subsections{$subsection});
$subsections{$subsection} = 1;
}
}




@section_values = &$section_entry_func( $CUR_SECTION, [ @section_values ])
if (defined $section_entry_func);




init_sub_section( $defs_ref, @section_values)
if ($is_subsection);




{
my %items_count = map { $_ => 0 } keys %{$defs_ref};


$line = GBSFILEGLO_get_line();
my $last_order = 0;
while (defined $line && $line =~ /^\s/)
{




($CUR_ITEM, my $value) = $line =~ /^\s+(\S+)\s*=\s*(.*)$/;
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Item Format must be '<space><keyword> = <value>'")
if (!defined $CUR_ITEM);
my $def_ref = $defs_ref->{$CUR_ITEM};
if (defined $def_ref)
{
my $order = $def_ref->[3];
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Wrong order for '$CUR_ITEM'. Must be placed earlier")
if (defined $order && $order < $last_order);

process_item( $is_subsection, $def_ref, $value);

$last_order = $order;
$items_count{$CUR_ITEM}++;
} else
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Unknown Item: '$CUR_ITEM'");
}
$line = GBSFILEGLO_get_line();
}




foreach my $item (keys %{$defs_ref})
{
next if (!$defs_ref->{$item}->[2]);    # $is_mandatory
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Item '$item' (mandatory) is missing")
if ($items_count{$item} == 0);
}
}




&$section_exit_func( $CUR_SECTION, [ @section_values ])
if (defined $section_exit_func);
}




foreach my $section (keys %{$sections_ref})
{
my $min_occur = $sections_ref->{$section}->[1];
if ($min_occur)
{
GBSFILEGLO_warning( "At least $min_occur occurences of '$section' needed")
if ($sections_count{$section} < $min_occur);
}
}




my ($plugin_name, $file_path_ref, $dependencies_ref) = GBSFILEGLO_close_file();
$STRUCT_REF->{'.'}->{PLUGIN_NAME} = $plugin_name;
$STRUCT_REF->{'.'}->{SEARCH_PATH} = $file_path_ref;
$STRUCT_REF->{'.'}->{DEPENDENCIES} = $dependencies_ref;



}




sub init_struct($)
{
my ($sections_ref,
) = @_;

%{$STRUCT_REF} = ();

while (my ($section, $section_ref) = each( %{$sections_ref}))
{
my ($sec_type, undef, undef, undef, $defs_ref, ) = @{$section_ref};
ENV_fatal( "Undefined defs_ref in Section $section")
if (!defined $defs_ref);
$STRUCT_REF->{$section} = {};

if ($sec_type ne 'S')
{
while (my ($item, $item_ref) = each( %{$defs_ref}))
{
$STRUCT_REF->{$section}->{$item} = ENV_clone( $item_ref->[1])	# default
}
} else
{
$STRUCT_REF->{$section}->{'.'}->{'ORDER'} = [];
}
}

}




sub init_sub_section($@)
{
my ($defs_ref,
@section_values,
)= @_;

$CUR_SUB_SECTION = shift @section_values;
$SECTION_PREFIX = "$CUR_SECTION ($CUR_SUB_SECTION)";


$CUR_SECTION_REF->{$CUR_SUB_SECTION} = {}
if (!defined $CUR_SECTION_REF->{$CUR_SUB_SECTION});
$CUR_SUB_SECTION_REF = $CUR_SECTION_REF->{$CUR_SUB_SECTION};

foreach my $item (keys %{$defs_ref})
{
my $default = $defs_ref->{$item}->[1];	# $default
$STRUCT_REF->{$CUR_SECTION}->{$CUR_SUB_SECTION}->{$item} = ENV_clone( $default);
}
$STRUCT_REF->{$CUR_SECTION}->{'.'}->{ORDER} = []
if (!defined $STRUCT_REF->{$CUR_SECTION}->{'.'}->{ORDER});
push @{$STRUCT_REF->{$CUR_SECTION}->{'.'}->{ORDER}}, $CUR_SUB_SECTION;

foreach my $subsection (@section_values)
{
$CUR_SECTION_REF->{$subsection} = $CUR_SUB_SECTION_REF;
push @{$STRUCT_REF->{$CUR_SECTION}->{'.'}->{ORDER}}, $subsection;
}

}




sub process_item($$$)
{
my ($is_subsection,
$def_ref,
$value,
) = @_;

my ($item_type, undef, undef, undef, $nr_args, $item_function, $constr_arg) = @{$def_ref};
my ($lvalue_type, $value_type, $value_constr) = split( //, $item_type);
my @values;




my $value_type_func = $VALUE_TYPE_FUNCS{$value_type};
ENV_fatal( "Value_type '$value_type' not implemented")
if (!defined $value_type_func);
@values = &$value_type_func( $value, $nr_args);




if ($value_constr)
{
my $value_constr_func = $VALUE_CONSTR_FUNCS{$value_constr};
ENV_fatal( "Value_contr '$value_constr' not implemented")
if (!defined $value_constr_func);
&$value_constr_func( [ @values ], $constr_arg);
}




@values = &$item_function( $CUR_SECTION, $CUR_SUB_SECTION, $CUR_ITEM, [ @values ], $constr_arg)
if (defined $item_function);




my $item_ref = ($is_subsection) ? $CUR_SUB_SECTION_REF : $CUR_SECTION_REF;
if ($lvalue_type eq 's')
{

$item_ref->{$CUR_ITEM} = $values[0];
} elsif ($lvalue_type eq 'a')
{

push @{$item_ref->{$CUR_ITEM}}, @values;
} elsif ($lvalue_type eq 'h')
{

foreach my $ref (@values)
{
my ($sub_key, @sub_values) = @{$ref};
$item_ref->{$CUR_ITEM}{$sub_key} = [ @sub_values];
}
} else
{
ENV_fatal( "Impossible lvalue_type ($lvalue_type)");
}
}




sub type_string($$)
{
my ($value_string,
$nr_args,
) = @_;
my @values;

if (defined $nr_args)
{
@values = ENV_split_quoted_space( $value_string);
validate_nr_args( $CUR_ITEM, scalar @values, $nr_args);
} else
{
$values[0] = $value_string;
}

return @values;
}




sub type_integer($$)
{
my ($value_string,
$nr_args,
) = @_;
my @values;

@values = type_string( $value_string, $nr_args);	# convert to array and check nr args
foreach my $value (@values)
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Value '$value' must be an integer")
if ($value !~ /^[-+]?\d+$/);
$value *= 1;
}

return @values;
}




sub type_bool($$)
{
my ($value_string,
$nr_args,
) = @_;
my @values;

@values = type_string( $value_string, $nr_args);	# convert to array and check nr args
foreach my $value (@values)
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Value '$value' must be 1 or 0")
if ($value ne '0' && $value ne '1');
}

return @values;
}




sub type_arrow($$)
{
my ($value_string,
$nr_args,
) = @_;
my @value_refs; # ([ $sub_key, @sub_values ], ...)


my ($sub_keys, $sub_value) = $value_string =~ /^(.*?)\s*=>\s*(.*)$/;
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Syntax is ' $CUR_ITEM = <name>... => <value_string>'")
if (!defined $sub_keys);

my @sub_values = type_string( $sub_value, $nr_args);    # convert to array and check nr args

foreach my $sub_key (split( ' ', $sub_keys))	    # split on any whitespaces
{
push @value_refs, [ $sub_key, @sub_values ];
}

return @value_refs;
}








sub type_command($$)
{
my ($command_sequence,  # file may be prefixed with '@', '>' or '.'
$nr_args,
) = @_;

my ($command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition);


my ($spec_command_or_file,	# file may be prefixed with '@', '>' or '.'
$ok_values_org,
$exec_condition_org
) = reverse split( /\s*;\s*/, $command_sequence, 3);







{

my @args;
my $fc = substr( $spec_command_or_file, 0, 1);
if ($fc eq '@' || $fc eq '>' || $fc eq '.')
{






$command_type = 1;  # script command
(my $rel_file_spec, @args) = ENV_split_quoted_space( substr( $spec_command_or_file, 1));
$rel_file_spec .= ENV_commandfile_ext()
if ($rel_file_spec !~ /\.sh$|\.bat$|\.cmd$|\.pl$/);

my $script_file;
if ($fc eq '@')
{
my @target_path = GBSFILEGLO_get_path();
$script_file = PATH_search_file( $rel_file_spec, \@target_path);
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Command-file '$rel_file_spec' not found in", "(@target_path)")
if ($script_file eq '');
} elsif ( $fc eq '>')
{
$script_file = "$PLUGIN_PATH/$rel_file_spec";
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Command-file '$script_file' does not exist",
"Specified: '$spec_command_or_file'")
if (!-e $script_file);
} else	# ($fc eq '.')
{
$script_file = "$GBS_SCRIPTS_PATH/$rel_file_spec";
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Command-file '$script_file' does not exist",
"Specified: '$spec_command_or_file'")
if (!-e $script_file);
}
$command_line = ENV_join_quoted_space( $script_file, @args);
} elsif ($spec_command_or_file =~ /^ECHO\s/ ||
$spec_command_or_file =~ /^SET\s/)
{



$command_type = 2;	# special command
$command_line = $spec_command_or_file;
} else
{



$command_type = 0;	# direct command
$command_line = $spec_command_or_file;
}

validate_nr_args( $CUR_ITEM, 1 + scalar @args, $nr_args);

if ($command_type == 2) 	# special command (e.g.: echo)
{
$nr_pos_arg_tokens = 0;
} else
{
$nr_pos_arg_tokens = () = $spec_command_or_file =~ /$POS_PAR_TOKEN(\*|\d)/g;
}
}




{
if (defined $ok_values_org)
{
my @values = split( /\s*,\s*/, $ok_values_org);
my @new_values;
foreach my $value (@values)
{
my ($min_value, undef, $max_value) = $value =~ /^(\d+)(\s*-\s*(\d+))?$/;
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Invadid ok-value syntax ($value)")
if (!defined $min_value);
$max_value = $min_value
if (!defined $max_value);
GBSFILEGLO_error_exit( "$SECTION_PREFIX: min_value ($min_value) > ($max_value)")
if ($min_value > $max_value);
push @new_values, $min_value..$max_value;
}
$ok_values = "@new_values";
} else
{
$ok_values = '0';
}
}




{
if (defined $exec_condition_org)
{
$exec_condition = $exec_condition_org;
} else
{
$exec_condition = '-';
}
}


return ( $command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition );
}




sub type_format($$)
{
my ($value_string,
$nr_args,
) = @_;





my ($env_type,		    # FILE or VARIABLE (default)
$format);

if ($value_string =~ /^\w+/)
{
if ($value_string =~ /^(\w+)\s+(.+)/)
{
$env_type = $1;
$format = $2;
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Type '$env_type' must be 'VARIABLE' or 'FILE'")
if ($env_type ne 'VARIABLE' && $env_type ne 'FILE');
} else
{
GBSFILEGLO_error_exit( "Syntax = [ type ] printf_format");
}
} else
{
$env_type = 'VARIABLE';
$format = $value_string;
}
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Format must contain '%s'")
if ($format !~ /%s/);


return [ $env_type, $format ];
}




sub type_via_format($$)
{
my ($value_string,
$nr_args,
) = @_;





my ($format,	    # %s, -via %s, @%s, ...
$type);		    # empty or .type

if ($value_string =~ /^(.*%s)(.*)$/)
{
$format = $1;
$type = $2;
GBSFILEGLO_error_exit( "$SECTION_PREFIX: If 'type' is specified ($type) if must conatin a ',' (DOT)")
if ($type ne '' && $type !~ /\./);
} else
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Via Format must contain '%s'")
}


return [ $format, $type ];
}




sub constr_range($$)
{
my ($values_ref,
$constr_data,	# [ value, [ min, max ], ... ]
) = @_;

foreach my $value (@{$values_ref})
{
my $error_text = CHECK_num_value( undef, $value, @{$constr_data});
GBSFILEGLO_error_exit( "$SECTION_PREFIX: $error_text")
if (defined $error_text);
}
}




sub constr_select($$)
{
my ($values_ref,
$constr_data_ref,	# possible_values_ref (hash or array
) = @_;

if (ref $constr_data_ref eq 'HASH')
{
foreach my $value (@{$values_ref})
{
if (!exists $constr_data_ref->{$value})
{
my @possible_values = keys %$constr_data_ref;
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Value '$value' must be one of (@possible_values)");
}
}
} else
{
foreach my $value (@{$values_ref})
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Value '$value' must be one of (@{$constr_data_ref})")
if (!grep( $_ eq $value, @{$constr_data_ref}));
}
}
}




sub constr_type($$)
{
my ($values_ref,
$constr_data,	# undef
) = @_;

foreach my $value (@{$values_ref})
{
GBSFILEGLO_error_exit( "$SECTION_PREFIX: Type ($value) must contain '.'")
if ($value !~ /\./);
}
}




sub validate_nr_args($$$)
{
my ($itemword,
$nr_elements,
$nr_args_or_ref,
) = @_;

if (defined $nr_args_or_ref)
{
my $error_text = CHECK_num_value( $itemword, $nr_elements, $nr_args_or_ref);
GBSFILEGLO_error_exit( "$SECTION_PREFIX: $error_text")
if (defined $error_text);
}
}




sub GBSSFILE_get_section_items($$@)
{
my ($struct_ref,
$section,
@items,
) = @_;
my @values = ();


foreach my $item (@items)
{
my $value = $struct_ref->{$section}->{$item};
get_failed( $struct_ref, $section, $item)
if (!defined $value && !exists $struct_ref->{$section}->{$item});
push @values, $value;
}
if (@items == 1)
{
return (wantarray) ? @{$values[0]} : $values[0];
} else
{
return @values;
}
}




sub GBSSFILE_get_subsection_items($$$@)
{
my ($struct_ref,
$section,
$subsection,
@items,
) = @_;
my @values = ();


foreach my $item (@items)
{
my $value = $struct_ref->{SRC}->{$subsection}->{$item};
get_failed( $struct_ref, $section, $subsection, $item)
if (!defined $value && !exists $struct_ref->{$section}->{$subsection}->{$item});
push @values, $value;
}
if (@items == 1)
{
return (wantarray) ? @{$values[0]} : $values[0];
} else
{
return @values;
}
}




sub GBSSFILE_get_subsection_subitems($$$$@)
{
my ($struct_ref,
$section,
$subsection,
$item,
@subitems,
) = @_;
my @values = ();


foreach my $subitem (@subitems)
{
my $value = $struct_ref->{$section}->{$subsection}->{$item}->{$subitem};
get_failed( $struct_ref, $section, $subsection, $item, $subitem)
if (!defined $value && !exists $struct_ref->{$section}->{$subsection}->{$item}->{$subitem});
push @values, $value;
}
if (@subitems == 1)
{
return (wantarray) ? @{$values[0]} : $values[0];
} else
{
return @values;
}
}




sub get_failed($@)
{
my ($struct_ref,
@keys,
) = @_;

my $struct_file_uc = uc $struct_ref->{'.'}->{STRUCT_FILE};
my $struct_name = $struct_ref->{'.'}->{STRUCT_NAME};
map { $_ = "->{$_}" } @keys;

ENV_fatal( "No such ${struct_file_uc}{$struct_name}" . join( '', @keys));
}





sub GBSSFILE_set_init_envs($)
{
my ($struct_ref) = @_;

{



my @keys = ($IS_WIN32) ? qw( SET_W SET) : qw( SET_X SET);
foreach my $key (@keys)
{
my $set_refs_ref = $struct_ref->{INIT}->{$key};
if (defined $set_refs_ref)
{
foreach my $ref (@{$set_refs_ref})
{
my ($name, $value) = @{$ref};

$value = ENV_expand_envs( $value);
ENV_error_exit( "No match for EnvVar in $key $name = $value")
if ($value =~ /%|\$/);
$value = ENV_perl_path_nospace( $value);
$value = ENV_os_path( $value);
GBSENV_setenv( $name => $value );

}
}
}
}

{



my $setpath_refs_ref = $struct_ref->{INIT}->{SETPATH};
if (defined $setpath_refs_ref)
{
foreach my $ref (@{$setpath_refs_ref})
{
my ($name, @values) = @{$ref};

foreach my $value (@values)
{

if ($value eq $name || $value eq "%$name%" || $value eq "\$$name")
{

$value = ENV_getenv( $name);
$value = ENV_os_path( $value);
} else
{
$value = ENV_expand_envs( $value);
ENV_error_exit( "No match for EnvVar '$value'", "in SETPATH $name = @values")
if ($value =~ /%|\$/);
$value = ENV_perl_path_nospace( $value);
$value = ENV_os_path( $value);
}

}
my $new_value = ENV_join_path( \@values);
GBSENV_setenv( $name => $new_value );

}
}
}





}

return 1;


