#===============================================================
#
#   gbssilo.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSSILO*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::genopt;
use glo::html;
use glo::spit;
use glo::slurp;
use glo::time;
use gbsenv;
use gbshtml;
use run;




sub create_index();
sub do_gbssum();
sub do_topdir($);
sub generate_sum_files($);
sub gen_sum_file($$$);




my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_SILO_PATH = "$GBS_ROOT_PATH/silo";
my $INDEX_FILE_SPEC = "$GBS_ROOT_PATH/silo/index.html";

my $GBS_SYSTEM_NAME = ENV_getenv( 'GBS_SYSTEM_NAME');
my $GBS_LOG_PATH = ENV_getenv( 'GBS_LOG_PATH');






my $RC = 0;
GBSENV_init();




my @genopts = (
[ 'ci',     'create_index',	     'bso',   1, "Create the HTML index" ],
[ 'vi',     'view_index',	     'bso',   1, "View index in HTML browser" ],
);
my @genconflicts = (
[ [ ci => 0], '=', [ vi => 0 ] ],
);
GENOPT_set_optdefs( 'gbssilo', \@genopts,
'Generate silo index and show in HTML browser',
undef);
GENOPT_set_conflicts( \@genconflicts);
GENOPT_parse();
my $CREATE_INDEX = GENOPT_get( 'create_index');
my $VIEW_INDEX = GENOPT_get( 'view_index');




create_index()
if ($CREATE_INDEX);




RUN_browser( $INDEX_FILE_SPEC)
if ($VIEW_INDEX);




ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}




sub create_index()
{
my @lines;




my $title = "$GBS_SYSTEM_NAME - Silo";

push @lines, GBSHTML_doc_start( $INDEX_FILE_SPEC, $title, 0);
push @lines, GBSHTML_doc_top( [ $title,  HTML_link( $GBS_LOG_PATH, HTML_bold( 'GBS Log Directory'))], ());
my $section = 1;




push @lines, GBSHTML_section_head( 'GBS Log Overviews', $section++);
push @lines, do_gbssum();




my @all_dirs = SLURP_dir_dirs( $GBS_SILO_PATH, 0);
if (@all_dirs)
{
my $last_dir = pop @all_dirs;
foreach my $dir (@all_dirs)
{
push @lines, GBSHTML_section_head( ucfirst( $dir), $section++);
push @lines, do_topdir( $dir);
}
push @lines, GBSHTML_section_head( ucfirst( $last_dir), -$section);
push @lines, do_topdir( $last_dir);
}




push @lines, GBSHTML_doc_bottom( $title, $section);
push @lines, GBSHTML_doc_end();

SPIT_file_nl( $INDEX_FILE_SPEC, \@lines);
ENV_say( 1, "Created Silo index");
}




sub do_gbssum()
{
my @lines;

ENV_say( 1, "gbssum...");

my $sum_file_spec = "$GBS_SILO_PATH/.gbs/gbssum/gbssyssum.log";

push @lines, '<UL>';

if (-e $sum_file_spec)
{
my ($gen_sum_file_spec, $audit_sum_file_spec) = generate_sum_files( $sum_file_spec);

push @lines, '<LI>', HTML_link( $gen_sum_file_spec, 'Generate/Make (.log)')
if ($gen_sum_file_spec);
push @lines, '<LI>', HTML_link( $audit_sum_file_spec, 'Audits (.log & .html)')
if ($audit_sum_file_spec);
}

push @lines, '</UL>';

return @lines;
}




sub do_topdir($)
{
my ($dir) = @_;
my @lines;

ENV_say( 1, "TopDir $dir...");

push @lines, '<UL>';
my $this_dir = "$GBS_SILO_PATH/$dir";
if (-f "$this_dir/index.html")
{
push @lines, '<LI>', HTML_link( "$dir/index.html", "$dir/index.html");
} else
{
my @subdirs = SLURP_dir_dirs( $this_dir, 0);
if (@subdirs)
{
foreach my $subdir (@subdirs)
{

if (-f "$this_dir/$subdir/index.html")
{
push @lines, '<LI>', HTML_link( "$this_dir/$subdir/index.html", "$dir/$subdir/index.html");
} else
{
push @lines, '<LI>', HTML_link( "$this_dir/$subdir", "$dir/$subdir");
}
}
} else
{
push @lines, '<LI>', HTML_link( "$dir", "$dir");
}
}
push @lines, '</UL>';

return @lines;
}




sub generate_sum_files($)
{
my ($sum_file_spec) = @_;
my ($gen_sum_file_spec, $audit_sum_file_spec);

my @gen_lines;
my @audit_lines;
foreach my $sum_line (SLURP_file( $sum_file_spec))
{

if (substr( $sum_line, 0, 11) eq 'gbssysaudit')
{
push @audit_lines, $sum_line;
} else	# gen and make
{
push @gen_lines, $sum_line;
}
}
if (@gen_lines)
{
$gen_sum_file_spec = "$GBS_SILO_PATH/.gbs/gbssum/gen.html";
gen_sum_file( $gen_sum_file_spec, 0, \@gen_lines)
}
if (@audit_lines)
{
$audit_sum_file_spec = "$GBS_SILO_PATH/.gbs/gbssum/audit.html";
gen_sum_file( $audit_sum_file_spec, 1, \@audit_lines)
}

return ($gen_sum_file_spec, $audit_sum_file_spec);
}




sub gen_sum_file($$$)
{
my ($file_spec,
$is_audit,			# 'Audit' or 'Generate/Make'
$lines_ref,
) = @_;

my $name = ($is_audit) ? 'Audit' : 'Generate/Make';
my @lines;




my $title = "$GBS_SYSTEM_NAME - Silo Overview: $name";

push @lines, GBSHTML_doc_start( $file_spec, $title, 1);
push @lines, GBSHTML_doc_top( [$title,  HTML_link( $GBS_LOG_PATH, HTML_bold( 'GBS Log Directory'))], ());




my @row_refs;




if ($is_audit)
{
push @lines, HTML_txt( 'Please be aware that only the HTML-summary files were saved. Not the files referred to.'),
HTML_br(),
HTML_bold_txt( 'The files referred to will always be the last audited file versions.'),
HTML_br();
}




foreach my $line (reverse( @{$lines_ref}))
{
next if ($line eq '');
my (
$jobname,
$target,
$uname,
$root,
$logfile,		    # os format
$time_start,
$time_end,
$time_diff,
$timer_diff,
$state,			    # NORMAL, FAILED, KILLED
$rc,
$action,
$date_time,
$command_args,
$comments,
undef ) = split( /\t/, $line);


$logfile = ENV_perl_paths( $logfile);

$comments = ''
if ($comments eq '<<EOF>>');

my $logfile_html = (-e $logfile) ? HTML_link( $logfile, 'VIEW') : 'Missing';
my $state_html = HTML_div( {class => $state}, $state);
$date_time = TIME_unpack( $date_time);
if ($is_audit)
{
my ($audit) = $logfile =~ /gbssysaudit_${target}_(.*)_\d/;
$audit = '*UNK*'
if (!defined $audit);
my $htmlfile = "$GBS_SILO_PATH/audits/" . ENV_split_spec_n( $logfile) . "_sum.html";
my $htmlfile_html = (-e $htmlfile) ? HTML_link( $htmlfile, 'VIEW') : 'Missing';
push @row_refs, [ $jobname, $audit, $target, $date_time, $state_html, $rc, $logfile_html, $htmlfile_html, $command_args, $comments ];
} else
{
push @row_refs, [ $jobname, $target, $date_time, $state_html, $rc, $logfile_html, $command_args, $comments ];
}
}

push @lines, HTML_br();
if ($is_audit)
{
push @lines, HTML_table( [ qw( - - - - -C -R -C -C - - ) ],
[ qw( Job Audit Target DateTime State RC LogFile HTML<BR>Sum Comand-Args Comments) ],
\@row_refs, border => 1, class => 'sortable border1');
} else
{
push @lines, HTML_table( [ qw( - - - -C -R -C - - ) ],
[ qw( Job Target DateTime State RC LogFile Comand-Args Comments) ],
\@row_refs, border => 1, class => 'sortable border1');
}




push @lines, GBSHTML_doc_bottom( $title, undef);
push @lines, GBSHTML_doc_end();

SPIT_file_nl( $file_spec, \@lines);
ENV_say( 1, "Created Silo $name");
}


