#===============================================================
#
#   gbsstats.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSSTATS*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::genopt;
use glo::format;
use glo::slurp;
use gbsenv;
use gbsglo;
use validate;
use target;




sub init_stats();
sub do_stats();
sub collect_gbs_subsys($);
sub collect_non_gbs_subsys($);
sub print_all();
sub count_files($);
sub count_lines($);
sub print_line($$);
sub print_all();




my @SUBDIRS = qw( src loc inc dat sav);
my @TARGET_SRC_FILE_TYPES;

my @ASCII_FILE_TYPES = ( qw( .c .cpp .j .h .hpp .inl .mk .sh .bat .pl .xml .txt .text .cmd .html ), '.c#');
my %ASCII_FILE_TYPES = map { $_ => 1 } @ASCII_FILE_TYPES;



my $SECTION;

my $SYSTEM_NAME;




my ($NR_SYS_DIRS, $NR_SYS_FILES);




my $NR_ALL_AUDITS;
my ($NR_SYSAUDIT_DIRS, $NR_SYSAUDIT_FILES);




my $NR_ALL_TARGETS;
my $NR_OS_TARGETS;
my ($NR_SYSGEN_DIRS, $NR_SYSGEN_FILES);




my $NR_ALL_TOOLS;
my ($NR_SYSTOOLS_DIRS, $NR_SYSTOOLS_FILES);




my ($NR_RES_DIRS, $NR_RES_FILES);
my $NR_RES_SUBS;




my ($NR_EXT_DIRS, $NR_EXT_FILES);
my $NR_ALL_SUBS;




my ($NR_DEV_DIRS, $NR_DEV_FILES);
my $NR_GBS_SUBS;
my $NR_NON_GBS_SUBS;




my $NR_GBS_COMPS = 0;

my %SUBDIR_FILE_TYPE_TOTALS;


my %GBS_SUBDIR_TOTALS = map { $_ => [ undef, 0, 0 ] } @SUBDIRS;






my ($NR_NON_GBS_DIRS, $NR_NON_GBS_FILES);
my %NON_GBS_FILE_TYPE_TOTALS;





my %FILE_TYPE_TOTALS;







$| = 1;
my $RC = 0;
GBSENV_init();




my @genopts = (
[ 'lines',  'count_lines',	     'bso',   0, "Count lines in src, inc and loc" ],

);
GENOPT_set_optdefs( 'gbsstats', \@genopts,
'Show GBS System Statistics',
undef);
GENOPT_parse();
my $COUNT_LINES = GENOPT_get( 'count_lines');




my $GBS_ROOT_PATH = VALIDATE_root();




init_stats();
do_stats();




ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}




sub init_stats()
{


my %src_types;
foreach my $target (ENV_getenv( 'GBS_TARGETS'))
{
map { $src_types{$_} = 1 } TARGET_get_src_types( $target);
}
@TARGET_SRC_FILE_TYPES = sort( keys( %src_types));
map { $FILE_TYPE_TOTALS{$_} = [ undef, 0, 0 ] } @TARGET_SRC_FILE_TYPES;

map { $ASCII_FILE_TYPES{$_} = 1 } @TARGET_SRC_FILE_TYPES;
@ASCII_FILE_TYPES = keys( %ASCII_FILE_TYPES);
}




sub do_stats()
{
ENV_say( 1, "Collecting data...");
$SYSTEM_NAME = ENV_getenv( 'GBS_SYSTEM_NAME');




($NR_SYS_DIRS, $NR_SYS_FILES) = count_files( "$GBS_ROOT_PATH/sys");




my @all_audits = ENV_getenv( 'GBS_AUDITS');
$NR_ALL_AUDITS = @all_audits;

($NR_SYSAUDIT_DIRS, $NR_SYSAUDIT_FILES) = count_files( "$GBS_ROOT_PATH/sysaudit");




my @all_targets = ENV_getenv( 'GBS_ALL_TARGETS');
$NR_ALL_TARGETS = @all_targets;

my @targets = ENV_getenv( 'GBS_TARGETS');
$NR_OS_TARGETS = @targets;

($NR_SYSGEN_DIRS, $NR_SYSGEN_FILES) = count_files( "$GBS_ROOT_PATH/sysgen");




my @all_tools = ENV_getenv( 'GBS_TOOLS');
$NR_ALL_TOOLS = @all_tools;

($NR_SYSTOOLS_DIRS, $NR_SYSTOOLS_FILES) = count_files( "$GBS_ROOT_PATH/systools");




($NR_RES_DIRS, $NR_RES_FILES) = count_files( "$GBS_ROOT_PATH/res");
my @res_subs = GBSGLO_select_direntries( SLURP_dir_dirs( "$GBS_ROOT_PATH/res", 0));
$NR_RES_SUBS = @res_subs;




($NR_EXT_DIRS, $NR_EXT_FILES) = count_files( "$GBS_ROOT_PATH/ext");
my @all_subs = GBSGLO_subsystems();
$NR_ALL_SUBS = @all_subs;




ENV_say( 1, "General...");
($NR_DEV_DIRS, $NR_DEV_FILES) = count_files( "$GBS_ROOT_PATH/dev");




{
my @gbs_subs = GBSGLO_subsystems_full_gbs();
$NR_GBS_SUBS = @gbs_subs;

if ($NR_GBS_SUBS > 0)
{
ENV_say( 1, "Full GBS Subsystems...");
map { collect_gbs_subsys( $_) } @gbs_subs
}
}




{
my @non_gbs_subs = GBSGLO_subsystems_non_gbs();
$NR_NON_GBS_SUBS = @non_gbs_subs;

if ($NR_NON_GBS_SUBS > 0)
{
ENV_say( 1, "Non GBS SubSystems...");
map { collect_non_gbs_subsys( $_) } @non_gbs_subs
}
}




print_all();

}




sub collect_gbs_subsys($)
{
my ($subsys) = @_;

my @comps = GBSGLO_components( $subsys);
$NR_GBS_COMPS += @comps;
foreach my $comp (@comps)
{
foreach my $subdir (@SUBDIRS)		# src loc inc dat sav
{
ENV_chdir( "$GBS_ROOT_PATH/dev/$subsys/comp/$comp/$subdir");
my @files = GBSGLO_glob( '*');
foreach my $file (@files)
{
if (-d $file)
{
my $file_type = "dir($file)";
my ($nr_dirs, $nr_files) = count_files( $file);
$nr_dirs++;
$SUBDIR_FILE_TYPE_TOTALS{$subdir}->{$file_type}->[0] += $nr_dirs;
$SUBDIR_FILE_TYPE_TOTALS{$subdir}->{$file_type}->[1] += $nr_files;
$GBS_SUBDIR_TOTALS{$subdir}->[0] += $nr_dirs;
$GBS_SUBDIR_TOTALS{$subdir}->[1] += $nr_files;
$FILE_TYPE_TOTALS{$file_type}->[0] += $nr_dirs;
$FILE_TYPE_TOTALS{$file_type}->[1] += $nr_files;
} else
{
my $file_type = ENV_split_spec_t( $file);
$file_type = 'no type'
if ($file_type eq '');
$SUBDIR_FILE_TYPE_TOTALS{$subdir}->{$file_type}->[1]++;
$GBS_SUBDIR_TOTALS{$subdir}->[1]++;
$FILE_TYPE_TOTALS{$file_type}->[1]++;
if ($COUNT_LINES && exists( $ASCII_FILE_TYPES{$file_type}))
{
my $nr_lines = count_lines( $file);
$SUBDIR_FILE_TYPE_TOTALS{$subdir}->{$file_type}->[2] += $nr_lines;
$GBS_SUBDIR_TOTALS{$subdir}->[2] += $nr_lines;
$FILE_TYPE_TOTALS{$file_type}->[2] += $nr_lines;
}
}
}
}
}
}




sub collect_non_gbs_subsys($)
{
my ($subsys) = @_;

my $subsys_root = "$GBS_ROOT_PATH/dev/$subsys";
foreach my $ref (SLURP_dir_tree( $subsys_root, 0)) # [ $is_dir, $rel_file_or_dir ]
{
if ($ref->[0])
{
$NR_NON_GBS_DIRS++;
} else
{
$NR_NON_GBS_FILES++;
my $file_spec = "$subsys_root/$ref->[1]";
my ($file_name, $file_type) = ENV_split_spec_nt( $file_spec);
my $file = "$file_name$file_type";
my $type;
if (lc $file eq 'makefile')
{
$type = 'Makefile';
} else
{
if ($file_type eq '')
{
$type = 'No Type';
} else
{
if ($ASCII_FILE_TYPES{$file_type})
{
$type = $file_type;
}
}
}
if (defined $type)
{
$NON_GBS_FILE_TYPE_TOTALS{ $type}->[1]++;	# $nr_files
$FILE_TYPE_TOTALS{ $type}->[1]++;		# $nr_files
if ($COUNT_LINES && exists( $ASCII_FILE_TYPES{$type}))
{
my $nr_lines = count_lines( $file_spec);
$NON_GBS_FILE_TYPE_TOTALS{ $type}->[2] += $nr_lines;
$FILE_TYPE_TOTALS{ $type}->[2] += $nr_lines;
}
}
}
}
}




sub print_all()
{
ENV_say( 1, "");
ENV_say( 1, "$SYSTEM_NAME ($GBS_ROOT_PATH)");

{
my @row_refs;

$SECTION = 'SYS';
push @row_refs, print_line( 'All dirs / files', [ $NR_SYS_DIRS, $NR_SYS_FILES ]);
push @row_refs, [ '--', '', '' ];

$SECTION = 'SYSAUDIT';
push @row_refs, print_line( 'All dirs / files', [ $NR_SYSAUDIT_DIRS, $NR_SYSAUDIT_FILES ]);
push @row_refs, print_line( 'All Audits', [ undef, $NR_ALL_AUDITS ]);
push @row_refs, [ '--', '', '' ];

$SECTION = 'SYSGEN';
push @row_refs, print_line( 'All dirs / files', [ $NR_SYSGEN_DIRS, $NR_SYSGEN_FILES ]);
push @row_refs, print_line( 'All Targets', [ undef, $NR_ALL_TARGETS ]);
push @row_refs, [ '--', '', '' ];

$SECTION = 'SYSTOOLS';
push @row_refs, print_line( 'All dirs / files', [ $NR_SYSTOOLS_DIRS, $NR_SYSTOOLS_FILES ]);
push @row_refs, print_line( 'All Tools', [ undef, $NR_ALL_TOOLS ]);
push @row_refs, [ '--', '', '' ];

$SECTION = 'RES';
push @row_refs, print_line( 'dirs / files', [ $NR_RES_DIRS, $NR_RES_FILES ]);
push @row_refs, print_line( 'SubSystems', [ undef, $NR_RES_SUBS ]);
push @row_refs, [ '--', '', '' ];

$SECTION = 'EXT';
push @row_refs, print_line( 'dirs / files', [ $NR_EXT_DIRS ,$NR_EXT_FILES ]);
push @row_refs, [ '--', '', '' ];

$SECTION = 'DEV';
push @row_refs, print_line( 'dirs / files', [ $NR_DEV_DIRS, $NR_DEV_FILES ]);
push @row_refs, print_line( 'SubSystems', [ undef, $NR_ALL_SUBS ]);
push @row_refs, print_line( 'Full GBS SubSystems', [ undef, $NR_GBS_SUBS ]);
push @row_refs, print_line( 'Non  GBS SubSystems', [ undef, $NR_NON_GBS_SUBS ]);
push @row_refs, [ '--', '', '' ];




if ($NR_GBS_SUBS > 0)
{
$SECTION = 'Full GBS';
push @row_refs, print_line( 'GBS Components', [ undef, $NR_GBS_COMPS ]);
push @row_refs, map { print_line( "Files $_", $GBS_SUBDIR_TOTALS{$_}) } @SUBDIRS;
if (keys %FILE_TYPE_TOTALS)
{
push @row_refs, [ '', '--', '' ];
foreach my $subdir (@SUBDIRS)
{
foreach my $file_type ( sort( keys %{$SUBDIR_FILE_TYPE_TOTALS{$subdir}}))
{
my $count_ref = $SUBDIR_FILE_TYPE_TOTALS{$subdir}->{$file_type};
push @row_refs, print_line( "Files $subdir $file_type", $count_ref);
}
}
}
push @row_refs, [ '--', '', '' ];
}




if ($NR_NON_GBS_SUBS > 0)
{
$SECTION = 'Non GBS';

push @row_refs, print_line( 'dirs / files', [ $NR_NON_GBS_DIRS, $NR_NON_GBS_FILES ]);
push @row_refs, map { print_line( "Files $_", $NON_GBS_FILE_TYPE_TOTALS{$_}) } sort( keys %NON_GBS_FILE_TYPE_TOTALS);
push @row_refs, [ '--', '', '' ];
}




$SECTION = 'Totals';
push @row_refs, map { print_line( "Files $_", $FILE_TYPE_TOTALS{$_}) } sort( keys %FILE_TYPE_TOTALS);




ENV_say( 0, FORMAT_table( 0, 0, ' ', undef, @row_refs));
}

if ($COUNT_LINES)
{
my $total_files = $GBS_SUBDIR_TOTALS{src}->[1] +
$GBS_SUBDIR_TOTALS{loc}->[1] +
$GBS_SUBDIR_TOTALS{inc}->[1];
my $total_lines = $GBS_SUBDIR_TOTALS{src}->[2] +
$GBS_SUBDIR_TOTALS{loc}->[2] +
$GBS_SUBDIR_TOTALS{inc}->[2];
foreach my $type (@ASCII_FILE_TYPES)
{
if ($NON_GBS_FILE_TYPE_TOTALS{$type})
{
$total_files += $NON_GBS_FILE_TYPE_TOTALS{$type}->[1];
$total_lines += $NON_GBS_FILE_TYPE_TOTALS{$type}->[2];
}
}

ENV_say( 0, '--');
ENV_say( 1, "Total Files / Lines for GBS Licence (src+inc+loc+NonGBS): $total_files / $total_lines");
}

}




sub count_files($)
{
my ($dir) = @_;
my ($nr_dirs, $nr_files) = (0,0);

foreach my $ref (SLURP_dir_tree( $dir, 0)) # [ $is_dir, $file_or_dir ]
{
if ($ref->[0])
{
$nr_dirs++;
} else
{
$nr_files++;
}
}

return ($nr_dirs, $nr_files);
}




sub count_lines($)
{
my ($file) = @_;
my $nr_lines = 0;

open( IN, $file) || ENV_fatal( "Unable to open file '$file'", "- $!");
while (<IN>)
{
$nr_lines++;
}
close IN;

return $nr_lines;
}




sub print_line($$)
{
my ($item,
$count_ref) = @_;
my $row_ref;

my ($nr_dirs, $nr_files, $nr_lines) = @{$count_ref};
my $lines = ($COUNT_LINES && defined $nr_lines) ? " ($nr_lines)" : '';
if ($nr_dirs)
{
$row_ref = [ '', "$SECTION: Nr. $item", ": $nr_dirs / $nr_files$lines" ];
} else
{
$nr_files = sprintf( "%3d", $nr_files);
$row_ref = [ '', "$SECTION: Nr. $item", ": $nr_files$lines" ];
}

return $row_ref;
}


