#==============================================================
#
#   gbsswr.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
print "*GBSSWR*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::list;
use glo::shell;
use glo::ask;
use glo::genopt;
use glo::slurp;
use glo::format;
use gbsenv;
use gbscmd;
use gbsglo;
use run;
use roots;
use files;
use system;
use validate;
use swt;
use swc;
use sws;
use swr;
use swglo;
use gbsdb;




sub ask_root();
sub do_add_root($);
sub ask_parent_root($);
sub check_gbs_root($$$);
sub do_remove_root($);
sub do_show_roots();
sub select_root($$);
sub print_root_cols();
sub format_roots();
sub add_to_roots_list($);
sub check_and_write_roots_list();




my $CWD = ENV_cwd();
my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();

my $OLD_ROOT = GBSENV_perl_gbs_root();
my $OLD_SYSTEM_NAME = ENV_getenv( 'GBS_SYSTEM_NAME');
my $OLD_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $OLD_COMPONENT = ENV_getenv( 'GBS_COMPONENT');
my $OLD_TARGET = ENV_getenv( 'GBS_TARGET');

my $NEW_AUDIT = ENV_getenv( 'GBS_AUDIT');
my $NEW_ROOT = GBSENV_perl_gbs_root();
my $NEW_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $NEW_COMPONENT = ENV_getenv( 'GBS_COMPONENT');
my $NEW_TARGET = ENV_getenv( 'GBS_TARGET');

my $CANCELLED = 0;
my $MUST_RUN_SWT_NEW = 0;






$| = 1;
my $RC = 0;
GBSENV_init();




my @genopts = (
[ '<1>',	'root', 	   'sso',  '', "System / Root: '' == ASK, '-' == none, '.' == current, '!' == CWD, '!<abs_file/dir_spec>', '!!' == CWD = GBS_ROOT_PATH" ],
[ '<2>',	'subsystem',	   'sso',  '', "SubSystem: '-' == none, '.' == current" ],
[ '<3>',	'component',	   'sso',  '', "Component: '-' == none, '.' == current" ],
[ 'target',	'target',	   'sso',  '.', "Target: '-' == none, '.' == current" ],
[ 'new',	'create_new_root', 'bso',   0, "Create a new System / Root" ],
[ 'add',	'add_root',	   'bso',   0, "Add a Root to the Roots List" ],
[ 'remove', 'remove_root',	   'bso',   0, "Remove a Root from the Roots List" ],
[ 'show',   'show_roots',	   'bso',   0, "Show the Roots List" ],
[ 'exec',	'execute_command', 'sso',  '', "(GBS-)Command(s) to execute. 'none' = OK" ],
[ 'nocd',	'no_cd',	   'bso',   0, "Do not 'cd' to the selected directories "],
);
my @genconflicts = (
[ [ new => 1],	'=', 'add', '=', 'remove', '=', 'show',   '=', 'exec', '=', '<2>', '=', '<3>', '=', 'target' ],
[ [ add => 1],	'=', 'new', '=', 'remove', '=', 'show',   '=', 'exec', '=', '<2>', '=', '<3>', '=', 'target' ],
[ [ remove => 1],	'=', 'new', '=', 'add',    '=', 'show',   '=', 'exec', '=', '<2>', '=', '<3>', '=', 'target' ],
[ [ show => 1],	'=', 'new', '=', 'add',    '=', 'remove', '=', 'exec', '=', '<1>', '=', '<2>', '=', '<3>', '=', 'target' ],
[ [ '<1>' => '-' ], '=', '<2>' ],
[ [ '<1>' => '-' ], '=', '<3>' ],
[ [ '<1>' => '-' ], '=', 'target' ],
[ [ '<2>' => '-' ], '=', '<3>' ],
);
GENOPT_set_optdefs( [ 'gbsswr', 'swr' ], \@genopts,
[ 'Set a current Root/System or',
'Create a new one, Add / Remove one from the Roots List or Show the Roots List'],
"You can only '--add' a Root if it was originally created by 'swr --new'");
GENOPT_set_conflicts( \@genconflicts);
GENOPT_parse();
my $ROOT = GENOPT_get( 'root');     	    # Unix or Windows format
my $SUBSYS = GENOPT_get( 'subsystem');
my $COMPONENT = GENOPT_get( 'component');
my $TARGET = GENOPT_get( 'target');
my $NEW = GENOPT_get( 'create_new_root');
my $ADD = GENOPT_get( 'add_root');
my $REMOVE = GENOPT_get( 'remove_root');
my $SHOW_ROOTS = GENOPT_get( 'show_roots');
my $EXEC = GENOPT_get( 'execute_command');
my $NO_CD = GENOPT_get( 'no_cd');

$ROOT = ENV_perl_paths( $ROOT)		    # Perl format
if ($ROOT ne '');





VALIDATE_gbs();




if ($NEW)
{



my $new_root = RUN_gbsnew( 'new_root', $ROOT);
if ($new_root ne '')
{
$ROOT = $new_root;
$SUBSYS = '-';
$COMPONENT = '-';
$TARGET = '-';




$MUST_RUN_SWT_NEW = 1;
add_to_roots_list( $ROOT);
} else
{
$ROOT = '.';
$CANCELLED = 1;
}
} elsif ($ADD)
{



my $added_root = do_add_root( $ROOT);
$ROOT = ($added_root eq '') ? '.' : $added_root;
} elsif ($REMOVE)
{



my $removed_root = do_remove_root( ENV_perl_path_nospace( $ROOT));
my $current_root = GBSENV_perl_gbs_root();


if ($removed_root eq $current_root)
{
ENV_say( 1, "Current Root removed");
$ROOT = '-';	# No current
} else
{
$ROOT = '.';	# Keep current
$CANCELLED = 1;
}

} elsif ($SHOW_ROOTS)
{



do_show_roots();
}

if (!$CANCELLED)
{



if ($ROOT eq '')
{
$ROOT = ask_root();
}

($NEW_TARGET, $NEW_COMPONENT, $NEW_SUBSYS, $NEW_ROOT, $NEW_AUDIT) = SWGLO_validate( $TARGET, $COMPONENT, $SUBSYS, $ROOT);









{
SWC_set( $OLD_ROOT, $OLD_SUBSYS, '', 0);    # Clear Current Component
SWS_set( $OLD_ROOT, '', 0);		    # Clear Current SubSys
SWT_set( $OLD_ROOT, '', 0);		    # Clear Current Target
SWR_set( '', 0);			    # Clear Current Root

my $must_write = $NEW_ROOT ne '' || $ROOT eq '-';
SWR_set( $NEW_ROOT, $must_write);
ENV_setenv( GBS_AUDIT => $NEW_AUDIT);
}

if ($NEW_SUBSYS ne ENV_getenv( 'GBS_SUBSYS'))
{
my $must_write = $NEW_SUBSYS ne '' || $SUBSYS eq '-';
SWS_set( $NEW_ROOT, $NEW_SUBSYS, $must_write);
}

if ($NEW_COMPONENT ne ENV_getenv( 'GBS_COMPONENT'))
{
my $must_write = $NEW_COMPONENT ne '' || $COMPONENT eq '-';
SWC_set( $NEW_ROOT, $NEW_SUBSYS, $NEW_COMPONENT, $must_write);
}

if ($NEW_TARGET ne ENV_getenv( 'GBS_TARGET'))
{
SWT_set( $NEW_ROOT, $NEW_TARGET, 1);
}
if ($NEW_ROOT eq '' && $ROOT ne '-')
{
ENV_warn( "No current Root");
} else
{
GBSDB_last_root_write( $NEW_ROOT)
if (GBSENV_mode_is_interactive());
}
ENV_warn( "No current SubSystem") if ($NEW_SUBSYS eq '' && $SUBSYS ne '-' && $ROOT ne '-');
ENV_warn( "No current Component") if ($NEW_COMPONENT eq '' &&
$COMPONENT ne '-' &&
GBSGLO_subsystem_is_full_gbs( $NEW_SUBSYS, $NEW_ROOT));
ENV_warn( "No current Target") if ($NEW_TARGET eq '' && $TARGET ne '-' &&
$NEW_ROOT ne '');








{
my $new_cwd = ($NO_CD) ? $CWD : SWGLO_new_cwd( $NEW_ROOT, $NEW_SUBSYS, $NEW_COMPONENT);

my @lines;

{
push @lines, SWC_exit( $OLD_ROOT, $OLD_SUBSYS, $OLD_COMPONENT) if ($OLD_COMPONENT ne '');
push @lines, SWS_exit( $OLD_ROOT, $OLD_SUBSYS) if ($OLD_SUBSYS ne '');
push @lines, SWR_exit( $OLD_ROOT, $OLD_SYSTEM_NAME) if ($OLD_ROOT ne '');

push @lines, GBSENV_changed_setenv_commands( 0);




if ($NEW_ROOT ne '')
{
push @lines, SHELL_echo( '')
if ($OLD_ROOT ne '');
push @lines, SHELL_conditional_eq( GBS_RC => 0, SWR_entry( $NEW_ROOT));
push @lines, SHELL_conditional_eq( GBS_RC => 0, SWS_entry( $NEW_ROOT, $NEW_SUBSYS))
if ($NEW_SUBSYS ne '');
push @lines, SHELL_conditional_eq( GBS_RC => 0, SWC_entry( $NEW_ROOT, $NEW_SUBSYS, $NEW_COMPONENT))
if ($NEW_COMPONENT ne '');
push @lines, SHELL_conditional_ne( GBS_RC => 0, SHELL_echo( "Execution of switch.gbs failed. GBS_SITE not ok?"));
}
push @lines, SHELL_conditional_eq( GBS_RC => 0, SHELL_cd( $new_cwd));
push @lines, SHELL_conditional_ne( GBS_RC => 0, SHELL_setenv( GBS_ROOT_PATH => '""'),
SHELL_setenv( GBS_SUBSYS => '""'),
SHELL_setenv( GBS_COMPONENT => '""'),
SHELL_setenv( GBS_TARGET => '""'),
SWR_entry( ''),
SHELL_cd( $NEW_ROOT));




push @lines, SHELL_conditional_eq( GBS_RC => 0, GBSCMD_command( 'gbsshow', '--brief'));
push @lines, SHELL_conditional_ne( GBS_RC => 0, SHELL_echo_failed());

if ($MUST_RUN_SWT_NEW)
{



push @lines, SHELL_conditional_eq( GBS_RC => 0, GBSCMD_command( 'gbsswt', '--new'));
} elsif ($EXEC && $EXEC ne 'none')
{



my ($command, @rest) = split( ' ', $EXEC);
$command = 'gbsshow'
if ($command eq 'gbs');
my $script_command = ENV_commandfile_spec( "$GBS_SCRIPTS_PATH/$command");   # TBS --brief ????
my $command_ref;
if (-f $script_command)
{
$command_ref = ENV_command_ref( $script_command, \@rest);
} elsif (-f "$GBS_SCRIPTS_PATH/$command.pl")
{
$command_ref = ENV_command_ref( "$command.pl", \@rest);
} else
{
ENV_error_exit( "Invalid/Empty '--exec' command: $EXEC");
}
my $exec_command = ENV_command_line( $command_ref);

push @lines, SHELL_conditional_eq( GBS_RC => 0, $exec_command);
}
}

GBSENV_write_result_script( \@lines, 0);
}
}

ENV_exit( $RC);




sub END
{
if ($CANCELLED)
{
ENV_print_end_msg( 1);
} else
{
ENV_print_end_msg( 0);
}
}




sub ask_root()
{
my $root;	    # Perl format




my @roots = ROOTS_get_all( 1);	    # Perl paths
if (@roots == 0)
{
ENV_error_exit( "No Root(s) defined for this user",
"use 'swr --add' to use an existing Root/System",
"or  'swr --new' to define a new Root/System");

} elsif (@roots == 1)
{
$root = ENV_perl_path_nospace( $roots[0]);
ENV_say( 1, "Single Root '$root' taken by default");
} else
{



my $default_root = ENV_getenv_perl_path( 'GBS_ROOT_PATH');
$default_root = GBSDB_last_root_get()
if ($default_root eq '');
$root = select_root( "Select Root/System Directory", \$default_root);
}

return $root;
}




sub do_add_root($)
{
my ($root) = @_;	# From CommandLine - Perl format

ENV_say( 1, "Add Root/System ($root)");

if ($root eq '')
{



ENV_say( 1, "Please note that you can only ADD a Root that was originally created with 'swr --new'");
if (FILES_check_system( $CWD))
{



ENV_say( 1, "Currently in '$CWD',",
"Which seems to be the Top of a GBS Root");
$root = $CWD
if (ASK_YNQ( "Do you want to add it?", 'N', 1) eq 'Y');   # exit on quit
}

if ($root eq '')
{



$root = ask_parent_root( $CWD);
}

my $root_parent_path = ENV_parent_path( $OLD_ROOT);
if ($root eq '' && $root_parent_path ne '' && $root_parent_path ne $CWD)
{



$root = ask_parent_root( $root_parent_path);
}

if ($root eq '')
{



if (ROOTS_get_nr() > 0)
{
ENV_say( 1, "Currently in Roots List:");
print_root_cols();
}
$root = ASK_path( "Enter existing GBS Root Directory",
'', 1, 0, [ \&check_gbs_root, undef ]);
$root = ENV_perl_paths( $root);
}
} else
{



$root = ENV_perl_path_nospace( $root);
ENV_say( 1, "Adding $root");
ENV_error_exit( "No such directory: $root")
if (!-d $root);
ENV_error_exit( "Not a GBS Root: $root")
if (!FILES_check_system( $root));
}

if ($root eq '')
{
ENV_warn( "No Root/System added!");
$CANCELLED = 1;
} else
{
if (FILES_check_system_other_platform( $root))
{
ENV_say( 1, "Adding Other Platform...");
$root = RUN_gbsnew( 'new_other_root', $root);
if ($root ne '')
{



$TARGET = '-';




my @this_platform_targets = SYSTEM_get( $root, 'targets');

$MUST_RUN_SWT_NEW = 1
if (@this_platform_targets == 0);
add_to_roots_list( $root);
} else
{
$CANCELLED = 1;
}
} else
{
add_to_roots_list( $root);
}
}

return $root;
}




sub ask_parent_root($)
{
my ($parent_path) = @_;
my $root_path = '';




my @root_refs;

my $nr_this_platform = 0;
my $nr_other_platform = 0;
my $nr_in_roots = 0;
foreach my $dir (SLURP_dir_dirs( $parent_path, 0))
{
my $is_this_platform = 0;
my $is_other_platform = 0;
my $this_path = "$parent_path/$dir";

if (FILES_check_system( $this_path))
{
$is_this_platform = 1;
$nr_this_platform++;
}
if (FILES_check_system_other_platform( $this_path))
{
$is_other_platform = 1;
$nr_other_platform++;
}

if ($is_this_platform || $is_other_platform)
{
my $is_in_roots = ROOTS_exists( $this_path);
$nr_in_roots++
if ($is_in_roots);

push @root_refs, [ $dir, $is_this_platform, $is_other_platform, $is_in_roots ]
}
}

if (@root_refs)
{



if ($parent_path eq $CWD)
{
ENV_say( 1, "Currently in '$parent_path',",
"Which seems to be a Parent of GBS Roots");
} else
{
ENV_say( 1, "Currently in GBS Root '$parent_path'");
}
my $nr_not_in_roots = @root_refs - $nr_in_roots;
if ($nr_not_in_roots > 0)
{
ENV_say( 1, "It contains $nr_not_in_roots non-added GBS Root directories:");
my @dirs;
foreach my $ref (@root_refs)
{
my ($dir, $is_this_platform, $is_other_platform, $is_in_roots) = @{$ref};
if (!$is_in_roots)
{
if ($is_other_platform)
{
$dir = "*$dir";			    # mark '*' for other platform
} else
{
$dir = " $dir";			    # mark ' ' for this platform
}
push @dirs, $dir;
}
}
ENV_say( 0, FORMAT_cols( 0, 4, '  ', 3, \@dirs));
if (ASK_YNQ( 'Add one of these?', 'N', 1) eq 'Y')
{
my $selected_dir;
if (@dirs == 1)
{
$selected_dir = $dirs[0];
ENV_say( 1, "Single GBS Root taken");
} else
{
$selected_dir = ASK_value_from_menu( "Select Root/System Directory", 0, ( '', @dirs));
}
if ($selected_dir ne '')
{
$root_path = "$parent_path/" . substr( $selected_dir, 1);	# remove leading ' ' or '*'
} else
{
ENV_say( 1, "No GBS Root selected");
}

}
} else
{
ENV_say( 1, "It contains NO non-added GBS Root directories");
}
}

return $root_path;
}




sub check_gbs_root($$$)
{
my ($values_ref,
$default_ref,
$args_ref,
) = @_;
my $error_txt = '';

my $root = $values_ref->[0];

if ($root ne '')
{
if (-d $root)
{
$error_txt = "Not a GBS Root: $root"
if (!FILES_check_system( $root));
} else
{
$error_txt = "No such directory '$root'";
}
}

return $error_txt;
}




sub do_remove_root($)
{
my ($root) = @_;	# Perl format

ENV_say( 1, "Remove Root/System $root");

if ($root eq '')
{
$root = select_root( "Select Root to remove from Roots-List:", 0);
}

if ($root eq '-')
{
$CANCELLED = 1;
} else
{
$root = ENV_perl_path_nospace( $root);
if (ROOTS_delete( $root))
{
ENV_say( 1, "Removed Root '$root'",
"Note that the Root Directories/Files are NOT deleted")
if (-e $root);
} else
{
ENV_error_exit( "Root '$root' not in Roots-List");
}
}

check_and_write_roots_list();

return $root;
}




sub do_show_roots()
{
ENV_say( 1, "Show Roots-List");

print_root_cols();

check_and_write_roots_list();
$CANCELLED = 1;
}




sub select_root($$)
{
my ($prompt,		# for ASK_value_from_menu
$default,		# for ASK_value_from_menu
) = @_;
my $selected_root;

my @roots = format_roots();

my $this_default = $default;
if (ref $default)
{
$this_default = $$default;
if (LIST_contains_str( "+$this_default", \@roots))
{
$this_default = \"+$this_default";
} else
{
$this_default = \" $this_default";
}
}

unshift @roots, '';		# Add <none>

$selected_root = ASK_value_from_menu( $prompt, $this_default, @roots);
if ($selected_root ne '')
{
$selected_root = substr( $selected_root, 1);
} else
{
$selected_root = '-';
}

return $selected_root;
}




sub print_root_cols()
{
ENV_say( 0, FORMAT_cols( 0, 2, '  ', 2, [ format_roots() ]));
}




sub format_roots()
{
my @roots;

foreach my $root (ROOTS_get_all( 1))
{
my $fc = '?';
if (-e $root)
{
if (FILES_check_system( $root))
{
$fc = ' ';
}
if (FILES_check_system_other_platform( $root))
{
$fc = ($fc eq ' ') ? '+' : '*';
}
} else
{
$fc = 'x';
}
push @roots, "$fc$root";
}

return @roots;
}




sub add_to_roots_list($)
{
my ($new_root) = @_;   # Perl format

if (ROOTS_add( $new_root))
{
ENV_say( 1, "Root '$new_root' added to Roots List...");
check_and_write_roots_list();
} else
{
ENV_warn( "Root '$new_root' already in Roots List");
}
}




sub check_and_write_roots_list()
{
foreach my $root (ROOTS_get_all( 1))
{
if (!-e $root)
{
ENV_say( 1, "Root '$root'",
"in Roots-list does not exist (anymore)");
ROOTS_delete( $root)
if (ASK_YN( 'Remove?', 'Y') eq 'Y');
}
}
ROOTS_write();
}


