#===============================================================
#
#   ask.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package glo::ask;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
ASK_text
ASK_not_text
ASK_word
ASK_not_word
ASK_wild_words
ASK_int
ASK_password
ASK_pause
ASK_values_from_list
ASK_YN
ASK_YNQ
ASK_YNEQ
ASK_YNAE
ASK_YNAEQ
ASK_more
ASK_menu
ASK_index_from_menu
ASK_value_from_menu
ASK_dir
ASK_path
ASK_other_path
ASK_file
ASK_file_path
);

use strict;
use integer;

use glo::env;
use glo::check;
use glo::list;
use glo::format;




sub ASK_text($$$;$;$);
sub ASK_not_text($$$$);
sub ASK_word($$$);
sub ASK_not_word($$$@);
sub ASK_wild_words($$$$);
sub ASK_int($$$$);
sub ASK_password($$);
sub ASK_pause($);
sub ASK_values_from_list($$$$);
sub ASK_YN($$);
sub ASK_YNQ($$$);
sub ASK_YNEQ($$$);
sub ASK_YNAE($$);
sub ASK_YNAEQ($$$);
sub ASK_more($$$);
sub ASK_menu($$$);
sub ASK_index_from_menu($$@);
sub ASK_value_from_menu($$@);
sub ASK_path($$$$;$);
sub ASK_other_path($$$;$);
sub ASK_dir($$$$$;$);
sub ASK_file($$$$$);
sub ASK_file_path($$$$);

sub gen_menu($$);
sub gen_ask_fc_list($$$);
sub gen_ask_s($$$$$);
sub gen_ask_a($$$$$$);
sub read_line($$$);
sub check_length($$$);
sub set_menu_entry($$$);
sub set_wildcard_values($$$);
sub set_path($$$);
sub set_other_path($$$);
sub check_chars($$$);
sub check_not_chars($$$);
sub check_not_allowed_values($$$);
sub check_integer($$$);
sub check_num_values($$$);
sub check_nr_elements($$$);
sub check_allowed_values($$$);
sub check_duplicate_values($$$);
sub check_abs_path($$$);
sub check_dir_exists($$$);
sub check_file_exists($$$);
sub show_menu($$$);
sub show_files($);
sub opt_checks($);
sub min_max($);
sub local_prompt($$);
sub print_error($);
sub quit();




my $IS_WIN32 = (ENV_is_win32());
my $IS_LUNIX = !$IS_WIN32;

my @WORD_CHARS = ( '\w', 'A-Z a-z 0-9 _' );
my @WILD_WORD_CHARS = ( '\w*?', 'A-Z a-z 0-9 _ * ?' );

my $NOT_PATH_CHARS = ':&()<>^;|,*? ';
my $NOT_DIR_CHARS =  ':&()<>^;|,*? .';
my $NOT_FILE_CHARS = ':&()<>^;|,*? /\\';

my $MENU_LEVEL = 0;

















sub ASK_text($$$;$;$)
{
my ($prompt,
$default,
$length,	    # >0 = max, <0 = min or [ $min, $max ]
$opt_checks_ref,    # optional: [ [ $check_func, $app_data_ref ]... ]
$help_ref,	    # optional: $help or [ @help ] or \&help_func
) = @_;
my $line;

my $checks_ref = [
opt_checks( $opt_checks_ref),
];

$line = gen_ask_s( $prompt, $default, $length, $checks_ref, $help_ref);

return $line;
}









sub ASK_not_text($$$$)
{
my ($prompt,
$default,
$invalid_values_ref,
$length,	# >0 = max, <0 = min or [ $min, $max ]
) = @_;
my $value;

ENV_say( -1, "$prompt, NOT one of (@{$invalid_values_ref})");
$prompt = 'Enter value';

my $checks_ref = [
[ \&check_not_allowed_values, $invalid_values_ref ],
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref, undef);

return $value;
}








sub ASK_word($$$)
{
my ($prompt,
$default,
$length,	# >0 = max, <0 = min or [ $min, $max ]
) = @_;
my $value;

my $checks_ref = [
[ \&check_chars, \@WORD_CHARS ],
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref, undef);

return $value;
}









sub ASK_not_word($$$@)
{
my ($prompt,
$default,
$invalid_values_ref,
$length,	# >0 = max, <0 = min or [ $min, $max
) = @_;
my $value;

ENV_say( -1, "$prompt, NOT one of:",
"@{$invalid_values_ref}");

my $checks_ref = [
[ \&check_chars, \@WORD_CHARS ],
[ \&check_not_allowed_values, $invalid_values_ref ],
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref, undef);

return $value;
}









sub ASK_wild_words($$$$)
{
my ($prompt,
$default,
$length,	# >0 = max, <0 = min or [ $min, $max ]
$elements,	# >0 = max, <0 = min or [ $min, $max ]
) = @_;


my $checks_ref = [
[ \&check_chars,	    \@WILD_WORD_CHARS ],
[ \&check_duplicate_values, undef ],
];

if (wantarray)
{
return gen_ask_a( $prompt, $default, $length, $elements, $checks_ref, undef);
} else
{
return gen_ask_s( $prompt, $default, $length, $checks_ref, undef);
}
}









sub ASK_int($$$$)
{
my ($prompt,
$default,
$values_ref,	# [ value, [ min, max ], ... ]
$length,	# >0 = max, <0 = min or [ $min, $max ]
) = @_;
my $value;

my $checks_ref = [
[ \&check_integer, undef ],
[ \&check_num_values, $values_ref ],
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref, undef);

return $value;
}







sub ASK_password($$)
{
my ($prompt,
$digest,
) = @_;
my $ok_or_value;




$prompt .= ' '
if (index( ' )]', substr($prompt, -1, 1)) == -1);
$prompt .= '[]: ';

my $count = 0;
do
{
ENV_reap_beeps();
my $password;
if ($IS_WIN32)
{
my $gbs_scripts = ENV_getenv( 'GBS_SCRIPTS_PATH');
my $command =  "$gbs_scripts\\win32utils\\password.exe";

my ($prefix, undef) = ENV_get_prefix( -1);
my $command_ref = ENV_command_ref( $command, [ "$prefix: $prompt" ]);
my $password_text;
ENV_run3( $command_ref, undef, 0, \$password_text, undef);

if ($password_text =~ /^Password=<(.*)>$/)
{
$password = $1;
} else
{
ENV_fatal( "Error calling '$command': ", "- $password_text");
}
} else
{
ENV_print( -1, $prompt);
system( "stty -echo");
chomp( $password = <STDIN>);
system( "stty echo");
ENV_say( 0, '');
}


if ($password eq '!')
{
ENV_beep();
ENV_warn_exit( "Terminating...")
if (ASK_YN( 'Quit?', 'Y') eq 'Y');
} elsif ($password eq '?')
{
print_error( 'No help available');
} else
{
if (defined $digest)
{
$ok_or_value = (crypt( $password, $digest) eq $digest);
if ($ok_or_value)
{
$count = 4;	    # done
} else
{
print_error( 'Password invalid - try again');
$count++;
}
} else
{
$ok_or_value = $password;
$count = 4;	    # done
}
}
} while ($count < 3);

return $ok_or_value;
}




sub ASK_pause($)
{
my ($prompt) = @_;          # may be undef

ENV_reap_beeps();

if (defined $prompt)
{
ENV_say( -1, $prompt);
}
ENV_say( -2, ''); $| = 1;   # and flush


my $command = ($IS_WIN32) ? 'pause' : 'read -r -p "Press <ENTER> to continue . . ." key;';
ENV_system( $command, 0, undef);
}









sub ASK_values_from_list($$$$)
{
my ($prompt,
$default,
$valid_values_ref,
$elements,	# >0 = max, <0 = min or [ $min, $max ]
) = @_;
my @values;

ENV_say( -1, $prompt);
my @possible_values = @{$valid_values_ref};
my $values_text = "(@{$valid_values_ref} *)";
map { ENV_say( 0, $_) } FORMAT_hanging( 0, 4, 1, $values_text);

$prompt = local_prompt( 'Enter list', $prompt);

my $checks_ref = [
[ \&set_wildcard_values, $valid_values_ref ],
[ \&check_allowed_values, $valid_values_ref ],
];

@values = gen_ask_a( $prompt, $default, 0, $elements, $checks_ref, [ 'Wildcards are allowed' ]);

return @values;
}







sub ASK_YN($$)
{
my ($prompt,
$default,
) = @_;
my $ans;

$ans = gen_ask_fc_list( $prompt, $default, [ qw( Yes No) ]);

return $ans;
}







sub ASK_YNQ($$$)
{
my ($prompt,
$default,
$exit_on_quit,
) = @_;
my $ans;

$ans = gen_ask_fc_list( $prompt, $default, [ qw( Yes No Quit) ]);

quit()
if ($ans eq 'Q' && $exit_on_quit);

return $ans;
}







sub ASK_YNEQ($$$)
{
my ($prompt,
$default,
$exit_on_quit,
) = @_;
my $ans;

$ans = gen_ask_fc_list( $prompt, $default,
[ 'Yes',
'No',
'End and keep previous',
'Quit'
]);

quit()
if ($ans eq 'Q' && $exit_on_quit);

return $ans;
}







sub ASK_YNAE($$)
{
my ($prompt,
$default,
) = @_;
my $ans;

$ans = gen_ask_fc_list( $prompt, $default,
[ 'Yes',
'No',
'All or the rest',
'End and keep previous',
]);

return $ans;
}







sub ASK_YNAEQ($$$)
{
my ($prompt,
$default,
$exit_on_quit,
) = @_;
my $ans;

$ans = gen_ask_fc_list( $prompt, $default,
[ 'Yes',
'No',
'All or the rest',
'End and keep previous',
'Quit',
]);

quit()
if ($ans eq 'Q' && $exit_on_quit);

return $ans;
}








sub ASK_more($$$)
{
my ($slack,
$header,
$lines_ref,
) = @_;

(undef, my $nr_rows) = ENV_get_term_size();
$nr_rows -= $slack + 1;
my $nr_lines = @{$lines_ref};


if ($nr_lines < $nr_rows)
{
ENV_say( 0, $header)
if ($header);
ENV_say( 0, $lines_ref);
} else
{
my $last_page = $nr_lines / $nr_rows;
my $page = 0;

do
{
my $from = $page * $nr_rows;
my $to = $from + $nr_rows - 1;
$to = $nr_lines - 1 if ($to > $nr_lines - 1);

ENV_say( 0, $header)
if ($header);
ENV_say( 0, @{$lines_ref}[$from..$to]);

my $default;
my @answers;
if ($page == 0)
{
@answers = qw( Next Last End);
$default = 'N';
} elsif ($page == $last_page)
{
@answers = qw( Previous First End);
$default = 'E';
} else
{
@answers = qw( Next Previous First Last End);
$default = 'N';
}
my $prompt = ($page == $last_page) ? 'Last' : 'More';
my $ans = gen_ask_fc_list( $prompt, $default, \@answers);
if ($ans eq 'N')
{
$page++;
} elsif ($ans eq 'P')
{
$page--;
} elsif ($ans eq 'F')
{
$page = 0;
} elsif ($ans eq 'L')
{
$page = $last_page;
} elsif ($ans eq 'E')
{
$page = -1;
}
$page = -1 if ($page > $last_page);
} while ($page >= 0);
}
}










sub ASK_menu($$$)
{
my ($prompt,
$menu_refs_ref,	    	# ($text, $function, $func_data_ref)
$entries_ref,		# may be undef
) = @_;

$MENU_LEVEL++;

if (defined $entries_ref && @{$entries_ref})
{
my @entries = @{$entries_ref};
my $entry = shift @entries;

if ($entry < @{$menu_refs_ref})
{
my $entry_ref = $menu_refs_ref->[$entry];
my (undef, $func, $func_data_ref) = @{$entry_ref};
&$func( $entry, $func_data_ref, [ @entries ]);
} else
{
my $entry_level = $MENU_LEVEL + 1;
my $entry_nr = $entry + 1;
ENV_warn( "No such menu sub-entry-$entry_level: $entry_nr");
gen_menu( $prompt, $menu_refs_ref);
}
} else
{
gen_menu( $prompt, $menu_refs_ref);
}

$MENU_LEVEL--;
}




sub gen_menu($$)
{
my ($prompt,
$menu_refs_ref,	    	#  ($text, $function, $func_argument_ref, @nestings)
) = @_;

my @menu_refs = @{$menu_refs_ref};

if ($MENU_LEVEL == 1)
{
unshift @menu_refs, [ "Exit (program)" ];
} else
{
unshift @menu_refs, [ "Exit (to previous menu)" ];
}
my $max = @menu_refs - 1;
my @texts = map { $_->[0] } @menu_refs;
my $choice_prompt = local_prompt( "Choice (0-$max)", $prompt);

my $checks_ref = [
[ \&set_menu_entry, \@texts ],
[ \&check_num_values, [ [ 0, $max ] ]],
];

my $value = 0;
do
{
ENV_say( -1, $prompt);
show_menu( 0, 0, \@texts);

$value = gen_ask_s( $choice_prompt, 0, -1, $checks_ref, undef);

if ($value > 0)
{
my ($text, $func, @args) = @{$menu_refs[$value]};
&$func( $value - 1, @args);
ENV_say( 0, '');
}
} while ($value != 0);
}










sub ASK_index_from_menu($$@)
{
my ($prompt,
$default_index,	    # ref == default_value, -1 == no default, 0 = first
@refs,		    # [ $display_text, $value, @more_values ]
) = @_;
my $index;

map { $_ = [ $_, $_ ] if (!ref $_) } @refs;
my @texts = map { $_->[0] } @refs;
my $start_index = 1;
if ($texts[0] eq '')
{
$texts[0] = '<none>';
$start_index = 0;
} elsif ($texts[0] =~ /^<.*>$/)
{
$start_index = 0;
}

if (ref $default_index)
{
my $default_value = $$default_index;
if (defined $default_value)
{
$default_index = LIST_firstidx_ref_str( $default_value, \@refs, 1);
} else
{
$default_index = -1;
}
}
$default_index = 0
if ($start_index == 0 && $default_index == -1);

ENV_say( -1, $prompt);
show_menu( $start_index, $default_index, \@texts);

unshift @texts, ''
if ($start_index == 1);  # array starts at '1'
my $max = @refs - 1 + $start_index;
my $default_txt = ($default_index == -1) ? '' : $default_index + $start_index;
$prompt = local_prompt( "Choice ($start_index-$max)", $prompt);

my $checks_ref = [
[ \&set_menu_entry, \@texts ],
[ \&check_num_values, [ [ $start_index, $max ] ] ],
];

my $index_value = gen_ask_s( $prompt, $default_txt, -1, $checks_ref, undef);

$index = $index_value - $start_index;

quit()
if ($refs[$index]->[0] eq '<quit>');	    # $name

return $index;
}










sub ASK_value_from_menu($$@)
{
my ($prompt,
$default_index,	    # ref == default_value, -1 == no default, 0 = first
@refs,		    # [ $display_text, $value, @more_values ]
) = @_;
my @values;

map { $_ = [ $_, $_ ] if (!ref $_) } @refs;
my @texts = map { $_->[0] } @refs;
my $start_index = 1;
if ($texts[0] eq '')
{
$texts[0] = '<none>';
$start_index = 0;
} elsif ($texts[0] =~ /^<.*>$/)
{
$start_index = 0;
}

if (ref $default_index)
{
my $default_value = $$default_index;
if (defined $default_value)
{
$default_index = LIST_firstidx_ref_str( $default_value, \@refs, 1);
} else
{
$default_index = -1;
}
}
$default_index = 0
if ($start_index == 0 && $default_index == -1);

ENV_say( -1, $prompt);
show_menu( $start_index, $default_index, \@texts);

unshift @texts, ''
if ($start_index == 1);  # array starts at '1'
my $max = @refs - 1 + $start_index;
my $default_txt = ($default_index == -1) ? '' : $default_index + $start_index;
$prompt = local_prompt( "Choice ($start_index-$max)", $prompt);

my $checks_ref = [
[ \&set_menu_entry, \@texts ],
[ \&check_num_values, [ [ $start_index, $max ] ] ],
];

my $index_value = gen_ask_s( $prompt, $default_txt, -1, $checks_ref, undef);

my $index = $index_value - $start_index;

quit()
if ($refs[$index]->[0] eq '<quit>');	    # $name

@values = @{$refs[$index]};
shift @values;  # remove $display_text

return (wantarray) ? @values : $values[0];
}






sub ASK_path($$$$;$)
{
my ($prompt,
$default,
$must_exist,	    # 1=must exist 0= may exist -1=may not exist
$length,	    # >0 = max, <0 = min or [ $min, $max ]
$opt_checks_ref,    # optional: [ [ $check_func, $app_data ]... ]
) = @_;
my $value;

my $checks_ref = [
[ \&set_path, 1 ],
[ \&check_abs_path, undef ],
[ \&check_dir_exists, [ $must_exist, '' ] ],
opt_checks( $opt_checks_ref),
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref, undef);

return $value;
}







sub ASK_other_path($$$;$)
{
my ($prompt,
$default,
$length,	    # >0 = max, <0 = min or [ $min, $max ]
$opt_checks_ref,    # optional: [ [ $check_func, $app_data ]... ]
) = @_;
my $value;

my $checks_ref = [
[ \&set_other_path, 1 ],
opt_checks( $opt_checks_ref),
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref, undef);

return $value;
}











sub ASK_dir($$$$$;$)
{
my ($prompt,
$default,
$root_dir,
$must_exist,	    # 1=must exist 0= may exist -1=may not exist
$length,	    # >0 = max, <0 = min or [ $min, $max ]
$opt_checks_ref,    # optional: [ [ $check_func, $app_data ]... ]
) = @_;
my $value;

$prompt .= ($root_dir eq '') ? " (<dirname>)" : "(<dirname>/?)";

my $checks_ref = [
[ \&check_not_chars, \$NOT_DIR_CHARS ],
[ \&check_dir_exists, [ $must_exist, $root_dir ] ],
opt_checks( $opt_checks_ref),
];

$value = gen_ask_s( $prompt, $default, $length, $checks_ref,
[ \&show_files, $root_dir ]);

return $value;
}










sub ASK_file($$$$$)
{
my ($prompt,
$default,
$root_dir,
$must_exist,	    	# 1=must exist 0= may exist -1=may not exist
$length,	    	# >0 = max, <0 = min or [ $min, $max ]
) = @_;
my $value;

$prompt = local_prompt( "$prompt (<file-name>/?)", $prompt);

my $checks_ref = [
[ \&check_not_chars, \$NOT_FILE_CHARS ],
[ \&check_file_exists, [ $must_exist, $root_dir ] ],
];
$value = gen_ask_s( $prompt, $default, $length, $checks_ref,
[ \&show_files, $root_dir ]);

return $value;
}









sub ASK_file_path($$$$)
{
my ($prompt,
$default,
$must_exist,	    	# 1=must exist 0= may exist -1=may not exist
$length,	    	# >0 = max, <0 = min or [ $min, $max ]
) = @_;
my $value;

my $checks_ref = [
[ \&set_path, 0 ],	    # TBS: 0 can become 1?
[ \&check_abs_path, undef ],
[ \&check_file_exists, [ $must_exist, '' ] ],
];
$value = gen_ask_s( $prompt, $default, $length, $checks_ref,
undef);

return $value;
}




sub gen_ask_fc_list($$$)
{
my ($prompt,
$default,			# uppercase
$possible_values_ref,		# [ Yes, No, Etc ]
) = @_;
my $value;

my @prompt_list;
my @allowed_values;
my @q_list;
foreach my $possible_value (@{$possible_values_ref})
{
my $fc = lc substr( $possible_value, 0, 1);
push @allowed_values, ($fc, uc $fc);
push @prompt_list, $fc;
push @q_list, "$fc=$possible_value";
}
my $prompt_list = join( '/', @prompt_list);
my $q_list = join( ', ', @q_list);
$prompt = "$prompt ($prompt_list)";

my $checks_ref = [
[ \&check_allowed_values, \@allowed_values ],
];

$default = lc $default;
$value = gen_ask_s( $prompt, $default, -1, $checks_ref, $q_list);

return uc $value;
}





sub gen_ask_s($$$$$)
{
my ($prompt,
$default,
$length,	# >0 = max, <0 = min or [ $min, $max ]
$checks_ref,	# [ $check_func, $app_data ]...
$help,
) = @_;
my @values;		# $values[0]

my $error_txt;
do
{
$error_txt = '';
my ($line, $full_prompt, $typed_line) = read_line( $prompt, $default,
$help);
@values = ($line);
my @check_refs = ([ \&check_length, $length ],
@{$checks_ref},
);
foreach my $ref (@check_refs)
{
my ($check_func, $app_data_ref) = @{$ref};
$error_txt = $check_func->( \@values, \$default, $app_data_ref);	# May change @values!!
last if ($error_txt);
}

if ($typed_line ne '' && $typed_line ne ' ' && $values[0] ne $typed_line)
{
ENV_say( -1, "$full_prompt$values[0]");
}
print_error( $error_txt)
if ($error_txt);
} while ($error_txt);

return $values[0];
}





sub gen_ask_a($$$$$$)
{
my ($prompt,
$default,
$length,	# >0 = max, <0 = min or [ $min, $max ]
$elements,	# >0 = max, <0 = min or [ $min, $max ]
$checks_ref,	# [ $check_func, $app_data ]...
$help,
) = @_;
my @values;

my $error_txt;
do
{
$error_txt = '';
my ($line, $full_prompt, $typed_line) = read_line( $prompt, $default,
$help);
$line =~ s/\s*,\s*/ /g;
@values = split( ' ', $line);

my @check_refs = ( [ \&check_nr_elements, $elements ],
[ \&check_length, $length ],
@{$checks_ref},
);
foreach my $ref (@check_refs)
{
my ($check_func, $app_data_ref) = @{$ref};
$error_txt = $check_func->( \@values, \$default, $app_data_ref);
last if ($error_txt);
}

if ($typed_line ne '' && "@values" ne $typed_line)
{
ENV_say( -1, "$full_prompt$values[0]");
}
print_error( $error_txt)
if ($error_txt);
} while ($error_txt);

return @values;
}




sub read_line($$$)
{
my ($prompt,
$default,
$help_ref,  	#$help, @help, or &helpfunc
) = @_;
my ($line, $full_prompt, $typed_line);




$prompt .= ' ' if (index( ' )]', substr($prompt, -1, 1)) == -1);
$full_prompt = "$prompt\[$default\]: ";

my $again;
do
{
ENV_reap_beeps();
$again = 0;
ENV_print( -1, $full_prompt);
$typed_line = <STDIN>;
$typed_line = "\a\n"
if (!defined $typed_line);	#SIGINT
chomp $typed_line;
$line = ($typed_line eq '') ? $default : $typed_line;
$line =~ s/\s+$//;		# remove trailing whitespace
$line =~ s/^\s+//;		# remove leading whitespace
if ($line eq '?')
{
my @help_lines;
if ($help_ref)
{
$help_ref = [ $help_ref ] if (!ref $help_ref);
if (ref $help_ref->[0] eq 'CODE')
{
my ($function, @args) = @{$help_ref};
@help_lines = &$function( @args);
} else
{
@help_lines = @{$help_ref};
}
}
if (@help_lines)
{
map { ENV_say( 0, "  $_") } @help_lines;
} else
{
print_error( 'No help available');
}
$again = 1;
} elsif ($line eq '!')
{
ENV_beep();
ENV_warn_exit( "Terminating...")
if (ASK_YN( 'Quit?', 'Y') eq 'Y');
$again = 1;
}
} while ($again);
return ( $line, $full_prompt, $typed_line);
}




sub check_length($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,  	# $min_max_length or [ $min_length, $max_length ]
) = @_;
my $error_txt = '';

my @values = @{$values_ref};

$values[0] = ''
if (@values == 0);

foreach my $value (@values)
{
my $value_length = length( $value);
my ($min_length, $max_length) = min_max( $app_data_ref);

if ($max_length > 0)
{
if ($value_length > $max_length)
{
$error_txt = "Maximum length = $max_length ($value)";
last;
}
}
if ($min_length > 0)
{
if ($value_length <  $min_length)
{
if ($value_length == 0)
{
$error_txt = 'Empty reply not allowed here';
} else
{
$error_txt = "Minimum length = $min_length ($value)";
}
last;
}
}
}

return $error_txt;
}




sub set_menu_entry($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,
) = @_;
my $error_txt = '';

foreach my $value (@{$values_ref})
{
if ($value =~ /^\d+$/)   	# numeric
{
$value = $value * 1;
} else
{
my $m_line = quotemeta $value;
my $match_count = grep( /^$m_line/, @{$app_data_ref});

if ($match_count == 0)
{
$error_txt = "No match ($value)";
last;
} elsif ($match_count == 1)
{
$value = 0;
while ((@{$app_data_ref}[$value] !~ /^$m_line/))
{
$value++;
}
} else
{
$error_txt = "Ambiguous reply ($value)";
last;
}
}
}

return $error_txt;
}




sub set_wildcard_values($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,
) = @_;
my $error_txt = '';

my @non_matches;
my @new_values = ENV_wildcards( $values_ref, $app_data_ref, \@non_matches);
if (@non_matches)
{
$error_txt = "Unmatched wild-card(s): '@non_matches'";
} else
{
@{$values_ref} = @new_values;
}

return $error_txt;
}




sub set_path($$$)
{
my ($values_ref,
$default_ref,
$use_default,
) = @_;
my $error_txt = '';

foreach my $path (@{$values_ref})
{
if ($path ne '')
{
$path = ENV_perl_path_nospace( ENV_expand_envs( $path));
if ($use_default && substr( $path, 0, 1) eq '.')
{
my $default = $$default_ref;
$path = ENV_perl_path_nospace( ENV_abs_path( $default, $path));
}
$path = ENV_canonicalize_path( $path);

if ($IS_WIN32)
{
if (length $path >= 2)	    # X:
{
my $test_text = substr( $path, 2);	#remove 'X:' for further tests
$error_txt = check_not_chars( [$test_text ], $default_ref, \$NOT_PATH_CHARS);
} else
{
$error_txt = "At least 2 chars expected";
}
} else
{
$error_txt = check_not_chars( [$path ], $default_ref, \$NOT_PATH_CHARS);
}
last if ($error_txt ne '');
}
}

return $error_txt;
}




sub set_other_path($$$)
{
my ($values_ref,
$default_ref,
$use_default,
) = @_;
my $error_txt = '';

foreach my $path (@{$values_ref})
{
if ($path ne '')
{
if ($use_default && substr( $path, 0, 1) eq '.')
{
my $default_path = $$default_ref;
$path = "$default_path/$path";
}
$path = ENV_perl_paths( $path);

if ($IS_WIN32)
{



$error_txt = check_not_chars( [$path ], $default_ref, \$NOT_PATH_CHARS);
if ($error_txt eq '')
{
$error_txt = "Abs path on Lunix must start with '/' or '~/'"
if ($path !~ m!^(/|~/)!);
}
} else
{



if (length $path >= 2)	    # X:
{
my $test_text = substr( $path, 2);	#remove 'X:' for further tests
$error_txt = check_not_chars( [$test_text ], $default_ref, \$NOT_PATH_CHARS);
} else
{
$error_txt = "At least 2 chars expected";
}
$error_txt = check_not_chars( [$path ], $default_ref, \$NOT_PATH_CHARS);
if ($error_txt eq '')
{
$error_txt = "Abs path on Windows must start with '//' or '<letter>:/'"
if ($path !~ m!^(//.+|[A-Za-z]:/)!);
}
}
last if ($error_txt ne '');
}
}

return $error_txt;
}




sub check_chars($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,
) = @_;
my $error_txt = '';

my ($valid_chars, $valid_chars_msg) = @{$app_data_ref};

foreach my $value (@{$values_ref})
{
if ($value ne '')
{

if ($value =~ /[^$valid_chars]/)
{
$error_txt = "Word may only contain '$valid_chars_msg' ($value)";
last;
}
}
}
return $error_txt;
}




sub check_not_chars($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,
) = @_;
my $error_txt = '';

my $invalid_chars = $$app_data_ref;
my $not_mask = quotemeta $invalid_chars;

foreach my $value (@{$values_ref})
{
if ($value ne '')
{
if ($value =~ /[$not_mask]/)
{
$error_txt = "Input may not contain '$invalid_chars' ($value)";
last;
}
}
}
return $error_txt;
}




sub check_integer($$$)
{
my ($values_ref,
$default_ref,
undef,		# not used
) = @_;
my $error_txt = '';

foreach my $value (@{$values_ref})
{
if ($value ne '')
{
if ($value !~ /^-?\d+$/)
{
$error_txt = "Not an integer ($value)";
last;
}
}
}

return $error_txt;
}




sub check_num_values($$$)
{
my ($values_ref,
$default_ref,
$possible_values_ref, 	    # 	# [ value, [ min, max ], ... ]
) = @_;
my $error_txt = '';

if (defined $values_ref)
{
foreach my $value (@{$values_ref})
{
if ($value ne '')
{
$error_txt = CHECK_num_value( undef, $value, @{$possible_values_ref});
if (defined $error_txt)
{
last;
} else
{
$error_txt = '';
}
}
}
}

return $error_txt;
}




sub check_nr_elements($$$)
{
my ($values_ref,
$default_ref,
$min_max_ref,  	# $min_max_elements or [ $min_elements, $max_elements ]
) = @_;
my $error_txt = '';

if (defined $min_max_ref)
{
my ($min_elements, $max_elements) = min_max( $min_max_ref);
my $nr_elements = @{$values_ref};

if ($max_elements > 0)
{
if ($nr_elements > $max_elements)
{
$error_txt = "Maximum nr. elements = $max_elements ($nr_elements)";
}
}
if ($min_elements > 0)
{
if ($nr_elements < $min_elements)
{
if ($nr_elements == 0)
{
$error_txt = "Empty reply not allowed here";
} else
{
$error_txt = "Minimum nr. elements = $min_elements ($nr_elements)";
}
}
}
}

return $error_txt;
}




sub check_allowed_values($$$)
{
my ($values_ref,
$default_ref,
$valid_values_ref,
) = @_;
my $error_txt = '';

if (defined $valid_values_ref)
{
foreach my $value (@{$values_ref})
{
if ($value ne '')
{
if (LIST_firstidx_str( $value, $valid_values_ref) == -1)
{
$error_txt = "Must be one of (@{$valid_values_ref}) ($value)";
last;
}
}
}
}

return $error_txt;
}




sub check_not_allowed_values($$$)
{
my ($values_ref,
$default_ref,
$invalid_values_ref,
) = @_;
my $error_txt = '';

if (defined $invalid_values_ref)
{
foreach my $value (@{$values_ref})
{
if ($value ne '')
{
if (grep( $_ eq $value, @{$invalid_values_ref}))
{
$error_txt = "'$value' not allowed";
last;
}
}
}
}

return $error_txt;
}




sub check_duplicate_values($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,	    # undef
) = @_;
my $error_txt = '';

foreach my $value (@{$values_ref})
{
if ($value ne '')
{
if (grep( $value eq $_, @{$values_ref}) > 1)
{
$error_txt = "'$value' may appear only once";
last;
}
}
}

return $error_txt;
}





sub check_abs_path($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,	    # undef
) = @_;
my $error_txt = '';

foreach my $value (@{$values_ref})
{
if ($value ne '')
{
if (!ENV_is_abs_path_perl( $value))
{
my $start = ($IS_WIN32) ? "'<letter>:' or '\\\\'" : "'/' or '~/'";
$error_txt = "'$value' must be absolute path (start with $start)";
last;
}
}
}

return $error_txt;
}





sub check_dir_exists($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,	    # [ $must_exist, $path ]
) = @_;
my $error_txt = '';

my ($must_exist, $path) = @{$app_data_ref};

foreach my $dir (@{$values_ref})
{
if ($dir ne '')
{
my $dir_spec = ($path) ? "$path/$dir" : $dir;
if (-e $dir_spec)
{
if (-d _)
{
if ($must_exist == -1)	    # may not exist
{
$error_txt = "Directory '$dir_spec' already exists";
last;
}
} else
{
if ($must_exist == 1 || $must_exist == 0) # must or may exist
{
$error_txt = "'$dir_spec' is not a directory";
last;
}
}
} else
{
if ($must_exist == 1)	    # must exists
{
$error_txt = "Directory '$dir_spec' does not exist";
last;
}
}
}
}

return $error_txt;
}





sub check_file_exists($$$)
{
my ($values_ref,
$default_ref,
$app_data_ref,	    # [ $must_exist, $path ]
) = @_;
my $error_txt = '';

my ($must_exist, $path) = @{$app_data_ref};

foreach my $file (@{$values_ref})
{
if ($file ne '')
{
$file = "$path/$file"
if ($path);
if (-e $file)
{
if (-f _)
{
if ($must_exist == -1)	    # may not exist
{
$error_txt = "File '$file' already exists";
last;
}
} else
{
if ($must_exist == 1 || $must_exist == 0) # must or may exist
{
$error_txt = "'$file' is not a file";
last;
}
}
} else
{
if ($must_exist == 1)	    # must exists
{
$error_txt = "File '$file' does not exist";
last;
}
}
}
}

return $error_txt;
}




sub show_menu($$$)
{
my ($first_index, 	    #(0 or 1)
$default_index,
$texts_ref,
) = @_;

my $max = @{$texts_ref};
my $width = length( $max - 1 + $first_index);

my @list_lines;
for (my $i = 0; $i < $max; $i++)
{
my $arrow = ($i == $default_index) ? '->' : '  ';
my $text = @{$texts_ref}[$i];
$text =~ s/[ \t]+$//;					    #remove trailing spaces or TABs
my ($leading_nls, $plain_text, $trailing_nls) = $text =~ /^(\n*)(.*)(\n*)$/;

map { push @list_lines, '' } (1..length($leading_nls));
push @list_lines, sprintf( "$arrow%${width}d. %s", $first_index + $i, $plain_text);
map { push @list_lines, '' } (1..length($trailing_nls));
}

@list_lines = FORMAT_cols( 0, 0, ' | ', 2, [@list_lines])
if (@list_lines > 3);
ENV_say( 0, @list_lines);
}




sub show_files($)
{
my ($path,
) = @_;
my @lines;

foreach my $file (ENV_readdir( $path))
{
next if (substr( $file, 0, 1) eq '.');
my $type;
if (-d "$path/$file")
{
$type = 'd';
} elsif (-f _)
{
$type = 'f';
} else
{
$type = '~';
}
push @lines, "-$type $file";
}
return @lines;
}




sub opt_checks($)
{
my ($opt_checks_ref,
) = @_;
my @refs = ();

if ($opt_checks_ref)
{


if (ref $opt_checks_ref->[0] eq 'ARRAY')
{
@refs = @{$opt_checks_ref};
} else
{
@refs = ($opt_checks_ref);
}
}
return @refs;
}




sub min_max($)
{
my ($min_max_ref,
) = @_;
my ($min, $max) = (0, 0);

if (defined $min_max_ref)
{
if (ref $min_max_ref)
{
($min, $max) = @{$min_max_ref};
} else
{
if ($min_max_ref > 0)
{
$max = $min_max_ref;
} else
{
$min = abs $min_max_ref;
}
}
}
return ($min, $max);
}




sub local_prompt($$)
{
my ($prompt,
$original_prompt,
) = @_;
my $local_prompt;

my ($leading_spaces) = $original_prompt =~ /^(\s*)/;
$local_prompt = "$leading_spaces$prompt";

return $local_prompt;
}




sub print_error($)
{
my ($line,
) = @_;

ENV_beep();
ENV_say( -2, "- $line");
}




sub quit()
{
ENV_say( -1, 'Quitting...');
ENV_exit();
}

return 1;
