#===============================================================
#
#   build.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package glo::build;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
BUILD_get_version_dir
BUILD_get_version
BUILD_get_version_date
BUILD_get_show_version
BUILD_get_lic
BUILD_version_2_build
BUILD_build_2_version
BUILD_get_all_versions
BUILD_get_all_spec_build_files
BUILD_get_build
BUILD_file_read
BUILD_file_write
);

use strict;
use integer;

use glo::env;
use glo::time;
use glo::slurp;
use glo::spit;




sub BUILD_get_version_dir();
sub BUILD_get_version();
sub BUILD_get_version_date();
sub BUILD_get_show_version();
sub BUILD_get_lic();
sub BUILD_version_2_build($$);
sub BUILD_build_2_version($);
sub BUILD_get_all_versions($);
sub BUILD_get_all_spec_build_files($);
sub BUILD_get_build($);
sub BUILD_file_read($);
sub BUILD_file_write($$$$$$$);

sub get_build_data();




my $APPLICATION;	# Do not initialize!	e.g.: GBS
my $APP_ROOT_PATH;	# Do not initialize!
my $VERSION_DIR;	# Do not initialize!
my $VERSION;		# Do not initialize!
my $VERSION_DATE;	# Do not initialize!
my $LICN;		# Do not initialize!
my $LICS;		# Do not initialize!
my $LICE;		# Do not initialize!
my $LICD;		# Do not initialize!
my @DEV_EXCLUDE_DIRS;	# Do not initialize!




sub BUILD_get_version_dir()
{
get_build_data()
if (!defined $VERSION_DATE);

return $VERSION_DIR;
}




sub BUILD_get_version()
{
get_build_data()
if (!defined $VERSION_DATE);

return $VERSION;
}




sub BUILD_get_version_date()
{
get_build_data()
if (!defined $VERSION_DATE);

return $VERSION_DATE;
}




sub BUILD_get_show_version()
{
my $show_version;

get_build_data()
if (!defined $VERSION_DATE);

my $parent_dir = ENV_parent_dir( $APP_ROOT_PATH, -1);
if ($parent_dir ne $VERSION || $VERSION_DIR ne $VERSION)
{
$show_version = "$VERSION/$VERSION_DIR/$parent_dir";
} else
{
$show_version = $VERSION;
}
$show_version .= " [ $VERSION_DATE ]";

return $show_version;
}




sub BUILD_get_lic()
{
get_build_data()
if (!defined $VERSION_DATE);

return wantarray ?($LICS, $LICN, $LICE, $LICD) : $LICN;
}




sub get_build_data()
{
$APP_ROOT_PATH = ENV_get_application_root_path();



($APPLICATION, $VERSION_DIR, $VERSION, $VERSION_DATE, my @rest) = BUILD_file_read( $APP_ROOT_PATH);




my $app_name = ENV_get_application_name();
ENV_fatal( "Application_name mismatch: BUILD=$APPLICATION, ENV=$app_name")
if ($APPLICATION ne $app_name);




if (substr( $VERSION_DATE, 0, 1) eq '#')
{
my @now = localtime;
$VERSION_DATE = sprintf( "%4d-%02d-%02d", $now[5] + 1900, $now[4] + 1, $now[3]);
ENV_say( 1, "$APPLICATION Development - Version_Date set to $VERSION_DATE");
@DEV_EXCLUDE_DIRS = @rest;
($LICN, $LICS) = ("3k3l3y3y3l3m3l3l3m3m3byo6k6e6l7xomld6k7o7k7y6y73oml36f6e737y6c7l6k6e637x", "2009-01-01");
} else
{
($LICN, $LICS) = @rest;
$LICN = "3m3byl686x73omlc6x636y6e736yom686k73om6o6yoml86k636b6y6lok" if (length($LICN) < 10);
}
$LICN=~tr/klmoxy/140295/;$LICN=~s/([a-fA-F0-9][a-fA-F0-9])/chr(hex($1))/eg;
($LICE, $LICN) = split( ';', $LICN);
$LICD = ($LICE - time) / (60 * 60 * 24);
$LICE = TIME_time2numdate( $LICE);

}





sub BUILD_version_2_build($$)
{
my ($app_version,			# 2.02
$app_version_date,		# 2008-09-12
) = @_;
my $app_build;			# 202080912

$app_build = $app_version . substr( $app_version_date, 2);
$app_build =~ s/\.//;
$app_build =~ s/-//g;

return $app_build;
}






sub BUILD_build_2_version($)
{
my ($app_build) = @_;
my ($app_version, $app_version_date);

if ($app_build)
{
$app_version = substr( $app_build, 0, -8) . '.' . substr( $app_build, -8, 2);

$app_version_date = substr( $app_build, -6, 6);
my (@parts) = $app_version_date =~ /(\d\d)/g;
$app_version_date = '20' . join( '-', @parts);
return (wantarray) ? ($app_version, $app_version_date) : "$app_version [$app_version_date]";
} else
{
return (wantarray) ? ('', '') : '';
}
}





sub BUILD_get_all_versions($)
{
my ($app_scripts_root) = @_;  # ENV_getenv_perl_path( 'GBS_SCRIPTS_PATH');
my @versions;

my @dirs = SLURP_dir_dirs( $app_scripts_root, 0);
@versions = grep( -f "$app_scripts_root/$_/build.dat", @dirs);

return @versions;
}




sub BUILD_get_all_spec_build_files($)
{
my ($app_scripts_root) = @_;  # ENV_getenv_perl_path( 'GBS_SCRIPTS_PATH');

return glob( "$app_scripts_root/*.??_20*/build.dat");
}




sub BUILD_get_build($)
{
my ($application_root_path,  # get from BUILD_get_all_spec_build_files
) = @_;

my (undef, undef, $version, $version_date) = BUILD_file_read( $application_root_path);

return BUILD_version_2_build( $version, $version_date);
}






sub BUILD_file_read($)
{
my ($application_root_path,
) = @_;


my $in_build_filespec = "$application_root_path/build.dat";

ENV_error_exit( "'build.dat' does not exist in '$application_root_path'",
'CWD?')
if (!-e $in_build_filespec);

if (wantarray)
{
return (SLURP_file( $in_build_filespec));
} else
{
return $in_build_filespec;
}
}




sub BUILD_file_write($$$$$$$)
{
my ($build_path,
$application,	    # e.g.: GBS
$version_dir,	    # e.g.: beta
$version,	    # e.g.: 3.00
$version_date,	    # YYYY-MM-DD
$licn,		    # encrypted
$lics,		    # YYYY-MM-DD
) = @_;
my $out_build_filespec = "$build_path/build.dat";

chmod( 0755, $out_build_filespec);    # u:rwx go:rx
SPIT_file_nl( $out_build_filespec, ( $application, $version_dir, $version, $version_date, ($licn, $lics)) );
chmod( 0555, $out_build_filespec);    # ugo:rwx

return $out_build_filespec;
}

return 1;
