#===============================================================
#
#   scmcvs.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#   File states:
#    -2 => 'NO-FILE',
#    -1 => 'NOT-CCM',
#     0 => 'CHECKED-IN',
#     1 => 'CHECKED-OUT',
#===============================================================

package glo::scmcvs;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCMCVS_init
SCMCVS_connect
SCMCVS_get_states
SCMCVS_checkout
SCMCVS_uncheckout
SCMCVS_checkin
SCMCVS_remove
SCMCVS_move
SCMCVS_add_dirs
SCMCVS_add_files
SCMCVS_setup
);

use strict;
use integer;

use File::Path;
use glo::env;
use glo::ask;

ENV_say( 1, "Loading CVS Plugin...");




sub SCMCVS_init($$);
sub SCMCVS_connect($$$);
sub SCMCVS_get_states($);
sub SCMCVS_checkout($);
sub SCMCVS_uncheckout($);
sub SCMCVS_checkin($);
sub SCMCVS_remove($);
sub SCMCVS_move($);
sub SCMCVS_add_dirs($);
sub SCMCVS_add_files($);
sub SCMCVS_setup($$$);

sub check_repos($$$);
sub exec_version();
sub exec_status($);
sub exec_commit($);
sub exec_update($);
sub exec_remove($);
sub exec_move($$);
sub exec_add($);
sub exec_database_version($);




my $APP_NAME;
my $REPOSITORY_OS;
my $APP_ROOT_PATH;
my @SKIPTYPES = qw( ^CVS$ );	# Keep in synch with %PREINFO_DATA in scm.pm

my $EXEC_OS = 'cvs';

my %FSTATES;





sub SCMCVS_init($$)
{
($APP_NAME,
$EXEC_OS) = @_;
}





sub SCMCVS_connect($$$)
{
($REPOSITORY_OS,
$APP_ROOT_PATH,
undef) = @_;	    # DATA
my $connect_texts_ref;

my $old_env = ENV_getenv( 'CVSROOT');
if ($old_env ne $REPOSITORY_OS)
{
ENV_setenv( CVSROOT => $REPOSITORY_OS);
ENV_say( 1, "'CVSROOT' set to $REPOSITORY_OS");
}

$connect_texts_ref = exec_version();
%FSTATES = ();	    # clear previous file-states

return ( undef, undef, $connect_texts_ref );
}








sub SCMCVS_get_states($)
{
my ($spec_refs_ref) = @_;

my @specs;
foreach my $spec_ref (@{$spec_refs_ref})
{
my ($spec, $state) = @{$spec_ref};
if (!defined $state)	    # $state may be -2 (no such file/dir)
{
$state = $FSTATES{$spec};
if (defined $state)
{
$spec_ref->[1] = $state;
} else
{
push @specs, $spec;
}
}
}

if (@specs)
{
exec_status( \@specs);
foreach my $spec_ref (@{$spec_refs_ref})
{
my ($spec, $state) = @{$spec_ref};
$spec_ref->[1] = $FSTATES{$spec}
if (!defined $state)
}
}
}




sub SCMCVS_checkout($)
{
my ($specs_ref) = @_;

ENV_chmod( 'u+w', @{$specs_ref});

map { $FSTATES{$_} = 1 } @{$specs_ref}; # checkout
}




sub SCMCVS_checkin($)
{
my ($specs_ref) = @_;

exec_commit( $specs_ref);
ENV_chmod( 'ugo-w', @{$specs_ref});
map { $FSTATES{$_} = 0 } @{$specs_ref}; # checkin
}




sub SCMCVS_uncheckout($)
{
my ($specs_ref) = @_;

exec_update( $specs_ref);
ENV_chmod( 'ugo-w', @{$specs_ref});
map { $FSTATES{$_} = 0 } @{$specs_ref}; # checkin
}




sub SCMCVS_remove($)
{
my ($specs_ref) = @_;

exec_remove( $specs_ref);
map { rmtree( $_, 0, 1) } @{$specs_ref};
map { $FSTATES{$_} = -1 } @{$specs_ref}; # notscm
}




sub SCMCVS_move($)
{
my ($pair_refs_ref) = @_;

foreach my $ref (@{$pair_refs_ref})
{
my ($old_spec, $new_spec) = @{$ref};
ENV_chmod( 'u+w', $old_spec);
exec_move( $old_spec, $new_spec);
$FSTATES{$old_spec} = -1;	# notscm
$FSTATES{$new_spec} = 1;    	# checkout
}
}




sub SCMCVS_add_dirs($)
{
my ($dir_refs_ref) = @_;    #	[ $dir, $ignores_ref ]

my @dir_specs;
foreach my $ref (@{$dir_refs_ref})
{
my ($dir, undef) = @{$ref};
ENV_mkdir( $dir, 0777)
if (!-d $dir);
push @dir_specs, $dir;
}

exec_add( \@dir_specs);

map { $FSTATES{$_} = 1 } @dir_specs; # checkout
}




sub SCMCVS_add_files($)
{
my ($files_ref) = @_;

map { ENV_touch_file( $_) if (!-f $_) } @{$files_ref};
exec_add( $files_ref);
map { $FSTATES{$_} = 1 } @{$files_ref}; # checkout
}




sub SCMCVS_setup($$$)
{
my (undef,				# $app_root
$scms_repository,
undef,				# $scms_data
) = @_;


my $default_repository = ENV_getenv( 'CVSROOT');
$default_repository = $scms_repository if (!$default_repository);
$default_repository = '' if (!defined $default_repository);
$scms_repository = ASK_text( 'Enter CVS Repository (CVSROOT)',
$default_repository, -1,
[ \&check_repos, undef ]);

return ( $scms_repository, '' );
}




sub check_repos($$$)
{
my ($values_ref,
$default_ref,
$args_ref,
) = @_;
my $error_txt = '';

my $scms_repository = $values_ref->[0];

my $rc = exec_database_version( $scms_repository);
$error_txt = "Not a valid CVS Repository URL"
if ($rc == 1);

return $error_txt;
}









sub exec_version()
{
my @connect_texts;

my $stdout;
my $command = "$EXEC_OS version";
ENV_run3( $command, undef, [0], \$stdout, undef);
if ($stdout !~ /^Concurrent Versions System/)
{
ENV_print( 0, $stdout);
ENV_fatal( "Failed to start CVS session");
}
@connect_texts = split( '\n', $stdout);

return [ @connect_texts ];
}

































sub exec_status($)
{
my ($specs_ref) = @_;

my $file_count = 0;
my $stdout;
my $stderr;
foreach my $spec (@{$specs_ref})
{
my $command = "$EXEC_OS status -l $spec";
ENV_run3( $command, undef, [0, 1], \$stdout, \$stderr);
my $state;
if (-d $spec)
{
if ($stdout =~ /no version/)
{
$state = -1;	# Not SCM
} else
{
$state = (-w $spec) ? 1 : 0;	# Checked Out : Checked In
}
} else
{
if ($stdout =~ /Status: Unknown/)
{
$state = -1;	# Not SCM
} else
{
$state = (-w $spec) ? 1 : 0;	# Checked Out : Checked In
}
}
$FSTATES{$spec} = $state;
$file_count++;
}

if ($file_count != @{$specs_ref})
{
ENV_fatal( "Unexpected return text with 'cvs status ...'",
"STDOUT: $stdout",
"STDERR: $stderr");
}


}








sub exec_commit($)
{
my ($specs_ref) = @_;

my $command = "$EXEC_OS commit -l -m \"Checkin by $APP_NAME\"";
ENV_run3( $command, $specs_ref, [0], undef, '');
}






sub exec_update($)
{
my ($specs_ref) = @_;

my $command = "$EXEC_OS update -C";
ENV_run3( $command, $specs_ref, [0], undef, '');
}






sub exec_remove($)
{
my ($specs_ref) = @_;


my $command = "$EXEC_OS remove -lf";
ENV_run3( $command, $specs_ref, [0], undef, '');
}






sub exec_add($)
{
my ($specs_ref) = @_;

my $command = "$EXEC_OS add -m \"Added by $APP_NAME\"";
ENV_run3( $command, $specs_ref, [0], undef, '');
}






sub exec_move($$)
{
my ($old_spec,
$new_spec,
) = @_;

ENV_fatal( "MOVE not implemented!");    # TBS
}







sub exec_database_version($)
{
my ($repository) = @_;
my $rc;

my $command = "$EXEC_OS -d \"$repository\" version";
$rc = ENV_run3( $command, undef, [0, 1], undef, '');

return $rc;
}

return 1;

