#===============================================================
#
#   shell.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package glo::shell;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SHELL_conditional_eq
SHELL_conditional_ne
SHELL_cd
SHELL_delete_files
SHELL_delete_dirs
SHELL_echo
SHELL_echo_failed
SHELL_exit_rc
SHELL_concat
SHELL_redirect_stderr
SHELL_redirect_stdout
SHELL_append_stdout
SHELL_redirect_all
SHELL_setenv
SHELL_unsetenv
SHELL_setfunc
SHELL_unsetfunc
SHELL_setrc
SHELL_source
SHELL_batch_header
SHELL_at
SHELL_at_show
SHELL_subwin
SHELL_xterm_start
SHELL_xterm_follow_up
SHELL_xtitle
SHELL_xcolor
SHELL_xgeometry
);

use strict;
use integer;

use glo::env;




sub SHELL_conditional_eq($$@);
sub SHELL_conditional_ne($$@);
sub SHELL_cd($);
sub SHELL_delete_files(@);
sub SHELL_delete_dirs(@);
sub SHELL_echo(@);
sub SHELL_echo_failed();
sub SHELL_exit_rc();
sub SHELL_concat(@);
sub SHELL_redirect_stderr($$);
sub SHELL_redirect_stdout($$);
sub SHELL_append_stdout($$);
sub SHELL_redirect_all($$);
sub SHELL_setenv($$);
sub SHELL_unsetenv($);
sub SHELL_setfunc($$;$);
sub SHELL_unsetfunc($;$);
sub SHELL_setrc();
sub SHELL_batch_header();
sub SHELL_source($;@);
sub SHELL_at($$$$);
sub SHELL_at_show();
sub SHELL_subwin($$);
sub SHELL_xterm_start($$);
sub SHELL_xterm_follow_up();
sub SHELL_xtitle($);
sub SHELL_xcolor($$);
sub SHELL_xgeometry($$);




my $SPREFIX = ENV_get_main_name() . '*';
my $SPREFIX_ = ' ' x (length( $SPREFIX) + 1);

my $IS_WIN32 = (ENV_is_win32());
my $IS_LUNIX = !$IS_WIN32;

my $DOS_SPECIAL_CHARS = '()<>|&^';
my ($MAX_COL, $MAX_ROW) = ENV_get_term_size_default();

my %COLORS = (

black	    => [ '0', 30 ],
red		    => [ '4', 31 ],
green	    => [ '2', 32 ],
yellow	    => [ '6', 33 ],
blue	    => [ '1', 34 ],
magenta	    => [ '5', 35 ],
cyan	    => [ '3', 36 ],
white	    => [ '7', 37 ],
bright_black    => [ '8', 90 ],
bright_red	    => [ 'C', 91 ],
bright_green    => [ 'A', 92 ],
bright_yellow   => [ 'E', 93 ],
bright_blue	    => [ '9', 94 ],
bright_magenta  => [ 'D', 95 ],
bright_cyan	    => [ 'B', 96 ],
bright_white    => [ 'F', 97 ]
);





sub SHELL_conditional_eq($$@)
{
my ($env_var,
$value,
@commands) = @_;

if ($IS_WIN32)
{
@commands = map { "if [%$env_var%]==[$value] $_"  } @commands;
} else
{
@commands = map { "[[ \$$env_var == $value ]] && $_" } @commands;
}
return @commands;
}





sub SHELL_conditional_ne($$@)
{
my ($env_var,
$value,
@commands) = @_;

if ($IS_WIN32)
{
@commands = map { "if NOT [%$env_var%]==[$value] $_"  } @commands;
} else
{
@commands = map { "[[ \$$env_var != $value ]] && $_" } @commands;
}
return @commands;
}





sub SHELL_cd($)
{
my ($dir) = @_;

if ($IS_WIN32)
{
$dir =~ tr!/!\\!;
return ( "cd /D $dir\\ > NUL");
} else
{
return ( "\\cd $dir");
}
}




sub SHELL_delete_files(@)
{
if ($IS_WIN32)
{
my @files = grep { $_ =~ tr!/!\\! } @_;
return ( "del /F /Q @files" );
} else
{
return ( "\\rm -f @_" );
}
}




sub SHELL_delete_dirs(@)
{
if ($IS_WIN32)
{
my @dirs = grep { $_ =~ tr!/!\\! } @_;
return ( "rmdir /S /Q @dirs" );
} else
{
return ( "\\rm -rf @_" );
}
}





sub SHELL_echo(@)
{
my @in_lines = @_;
my @shell_lines;

my $prefix = $SPREFIX;

if ($IS_WIN32)
{
foreach my $line (@in_lines)
{
$line =~ s/\s*$//;			# remove trailing whitespace
$line = ENV_escape_os_text( $line);	# escape special characters







push @shell_lines, "echo:$prefix: $line";
$prefix = $SPREFIX_;
}
} else
{
foreach my $line (@in_lines)
{
push @shell_lines, "\\echo \"$prefix: $line\"";
$prefix = $SPREFIX_;
}
}

return @shell_lines;
}





sub SHELL_echo_failed()
{
my $this_prefix = (ENV_beeps_are_enabled()) ? "\a$SPREFIX" : $SPREFIX;

if ($IS_WIN32)
{
return ( "echo $this_prefix: Failed (%GBS_RC%)" );
} else
{
return ( "\\echo \"$this_prefix: Failed " . "(\$GBS_RC)\"" );	# need to split because perlxref thinks it is a call (TBS?)
}
}





sub SHELL_exit_rc()
{
my @lines;

if ($IS_WIN32)
{
push @lines, 'exit /B %GBS_RC%';

} else
{
push @lines, "\\exit \$GBS_RC";
}
return @lines;
}




sub SHELL_concat(@)
{
my @commands = grep( defined $_, @_);
my $command_line;

if ($IS_WIN32)
{
$command_line = join( '&', @commands);
} else
{
$command_line = join( ';', @commands);
}

return $command_line;
}




sub SHELL_redirect_stderr($$)
{
my ($command,
$file_to
) = @_;

return "$command 2> $file_to";
}




sub SHELL_redirect_stdout($$)
{
my ($command,
$file_to
) = @_;

return "$command > $file_to";
}




sub SHELL_append_stdout($$)
{
my ($command,
$file_to
) = @_;

return "$command >> $file_to";
}




sub SHELL_redirect_all($$)
{
my ($command,
$file_to
) = @_;

return "$command > $file_to 2>&1";
}




sub SHELL_setenv($$)
{
my ($name,
$value
) = @_;



if ($IS_WIN32)
{
$value =~ s/%(\d)%/%%$1%%/g;

return ( "set $name=$value" );
} else
{
$value =~ s/\$(\d)/\\\$$1/g;

if (substr( $value, 0, 1) ne '"' && $value =~ /\s/)
{
return ( "\\export $name=\"$value\"" );
} else
{
return ( "\\export $name=$value" );
}
}
}




sub SHELL_unsetenv($)
{
my ($name) = @_;


if ($IS_WIN32)
{
return ( "set $name=" );
} else
{
return ( "\\unset $name" );
}
}




sub SHELL_setfunc($$;$)
{
my ($name,
$function,
$is_alias,	    # Opional, LUnix only
) = @_;



if ($IS_WIN32)
{
return ( "doskey $name=$function \$*" );
} else
{
if ($is_alias)
{
return ( "\\alias $name='$function'" );
} else
{
return ( "function $name { $function \$* ; }" );
}
}
}





sub SHELL_unsetfunc($;$)
{
my ($name,
$is_alias,	    # Opional, LUnix only
) = @_;


if ($IS_WIN32)
{
return ( "doskey $name=" );
} else
{
if ($is_alias)
{
return ( "\\unalias $name" );
} else
{
return ( "\\unset -f $name" );
}
}
}




sub SHELL_setrc()
{
if ($IS_WIN32)
{
return ('set GBS_RC=%ERRORLEVEL%');
} else
{
return ('GBS_RC=$?');
}
}





sub SHELL_batch_header()
{

if ($IS_WIN32)
{
return ( '@ECHO OFF',
'if not defined GBS_RC set GBS_RC=0' );
} else
{
return ( '#!/bin/bash' );
}
}




sub SHELL_source($;@)
{
my ($file,
@args
) = @_;

if ($IS_WIN32)
{
$file =~ tr!/!\\!;
return ( "call $file @args" );
} else
{
return ( ". $file @args" );
}
}




sub SHELL_at($$$$)
{
my ($command,
$name,			# jobname
$delay,			# now | [=+]hh:mm
$mail			# bool
) = @_;

if ($IS_WIN32)
{



if ($delay eq 'now' || $delay eq '')
{
$delay = '+00:00';
}
if (substr($delay,0,1) eq '=')
{
my ($sec, $min, $hour) = localtime();
my ($hours, $minutes) = split (':', substr($delay,1));
my $d_hours = $hours - $hour;
my $d_minutes = $minutes - $min;
if ($d_minutes < 0)
{
$d_hours--;
$d_minutes = 60 + $d_minutes;
}
$d_hours = 24 + $d_hours if ($d_hours < 0);
$delay = "+$d_hours:$d_minutes";
}
my ($hours, $minutes) = split (':', substr($delay,1));
my $time = $hours * 60 + $minutes;




if ($mail)
{
ENV_say( 1, "submit: mail option not implemented for Windows");
}

$command = ENV_command_line( ENV_command_ref( 'glo/winat.pl',
[ 'Q', $name, $time, $mail, $command ]));

} else # Assume LUNIX
{



my $at_options;

if ($delay eq 'now' || $delay eq '')
{
$at_options = 'now';
} elsif (substr($delay,0,1) eq '=')
{
$at_options = substr($delay,1);
} elsif (substr($delay,0,1) eq '+')
{
my ($hours, $minutes) = split (':', substr($delay,1));
$minutes += $hours * 60;
$at_options = "now + $minutes minutes";
} else
{
ENV_error_exit( "submit: Invalid delay: $delay");
}




if ($mail)
{
$at_options = "-m $at_options";
}

$command = "\\cd; \\at $at_options <<EOF\n$command\nEOF\n";
}

return $command;
}




sub SHELL_at_show()
{
if ($IS_WIN32)
{
return ( "REM -- No 'at' function on WIN32");
} else # Assume LUNIX
{
return ( "\\atq");
}
}




sub SHELL_subwin($$)
{
my ($title,
$command
) = @_;

if ($IS_WIN32)
{
$command = "start \"$title\" $command";
} else # Assume LUNIX
{
$command = "\\xterm -title \"$title\" -geometry $MAX_COL -bg white -fg darkblue -C -sb -ls -e $command \&";





}


return $command;
}




sub SHELL_xterm_start($$)
{
my ($title,
$command_line) = @_;
my $xterm_command_line;

if ($IS_WIN32)
{
$xterm_command_line = "start cmd.exe /T:F1 /C $command_line";
} else # Assume LUNIX
{
}
return $xterm_command_line;
}




sub SHELL_xterm_follow_up()
{
my $command_line;

if ($IS_WIN32)
{
$command_line = join( '&', (SHELL_xgeometry( undef, undef),
SHELL_xcolor( undef, undef),	# bg=bright white fg=blue
));
} else # Assume LUNIX
{
$command_line = 'echo -n';				# not for unix';
}

return $command_line;
}




sub SHELL_xtitle($)
{
my ($title) = @_;
my $line;

if ($IS_WIN32)
{
$line = "TITLE $title";
} else
{




$line = "echo -ne \"\e]0;$title\a\"";
}


return $line;
}




sub SHELL_xcolor($$)
{
my ($fg,
$bg) = @_;
my $command_line;

$fg = 'blue' if (!defined $fg);
$bg = 'bright_white' if (!defined $bg);

my @code_refs;
foreach my $color ($fg, $bg)
{
my $code_ref = $COLORS{lc $color};
ENV_fatal( "Unknown color: '$color'")
if (!defined $code_ref);
push @code_refs, $code_ref;
}
if ($IS_WIN32)
{
my $fgc = $code_refs[0]->[0];
my $bgc = $code_refs[1]->[0];
$command_line = "COLOR $bgc$fgc";
} else
{
my $fgc = $code_refs[0]->[1];
my $bgc = $code_refs[1]->[1] + 10;
$command_line = "echo -n \"\e[${fgc};${bgc}m\"";
}

return $command_line;
}




sub SHELL_xgeometry($$)
{
my ($width,
$height) = @_;
my $command_line;

$width = $MAX_COL if (!defined $width);
$height = $MAX_ROW if (!defined $height);

if ($IS_WIN32)
{
$command_line = "mode con cols=$width lines=$height";
} else
{
$command_line = "echo -n";	    # not possible - do nothing";
}


return $command_line;
}

return 1;

