#=================================================
#
#   plugin.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package plugin;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
PLUGIN_new_target
PLUGIN_remove_target
PLUGIN_dialog
);
use strict;
use integer;

use glo::env;
use glo::list;
use glo::ask;
use glo::slurp;
use glo::file;
use glo::format;
use glo::scm;
use gbsenv;
use gbsglo;
use fspec;
use fcreate;
use files;
use target;
use audit;
use tool;
use gbsscm;
use system;
use plg;




sub PLUGIN_new_target($);
sub PLUGIN_remove_target($);
sub PLUGIN_dialog($$);

sub main_remove_item($$$);
sub main_update_dialog($$$);
sub main_new_item($$$);
sub remove_item($$);
sub update_dialog($$);
sub update_plugin($$$);
sub new_build($$);
sub build_create_dirs($$$$);
sub build_update_dirs($$$);
sub new_audit($$);
sub audit_create_dirs($$$$$);
sub audit_update_dirs($$$);
sub new_tool($$);
sub tools_create_dirs($$);
sub tools_update_dirs($$$);
sub select_new_item($);
sub select_item($$);
sub ask_new_data($$);
sub select_plugin_name($);
sub select_plugin_item($);
sub do_install($$$);
sub get_plugins();
sub do_plugin_file($$$);
sub copy_1_file($$$$);
sub select_selector_items($$$);
sub add_dirs($);
sub system_add($);
sub system_store($);
sub setup_type($);
sub setup_plugin($);
sub setup_item($);




my $IS_WIN32 = ENV_is_win32();
my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();
my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();

my $DEFAULT_GEN_NAME = ($IS_WIN32) ? 'mvs' : 'linux';











my %PLUGIN_SETUP = (

audit   => [ qw( audit Audit  audit.gbs  sysaudit audit audits)	],
build   => [ qw( build Target target.gbs sysgen   gen   targets)    ],	# We want platform Targets only
tools   => [ qw( tools Tool   tool.gbs   systools tools tools)      ],
);

my %PLUGIN_TYPE_GET_DERIVED_ITEMS_FUNCS = (
audit   => \&AUDIT_get_derived_items,
build   => \&TARGET_get_derived_items,
tools   => \&TOOL_get_derived_items,
);

my %EXCLUDE_ITEM_NAMES = (
audit   => [ qw (ALL) ],
build   => [ qw (ALL makemake_stubs) ],
tools   => [ qw (ALL) ],
);




my $PLUGIN_TYPE;	    # 'audit',     'build',      'tools'
my $PLUGIN_TYPE_NAME;	    # 'Audit',     'Target',     'Tool'
my $PLUGIN_FILE;	    # 'audit.gbs', 'target.gbs', 'tool.gbs'
my $SYS_DIR;		    # 'sysaudit',  'sysgen',     'systools'
my $SUBSYS_DIR;		    # 'audit',     'gen',        'tools'
my $PLUGIN_TYPE_ID;	    # 'audits',    'targets',    'tools'	# also used for $ITEMS_ENV_NAME

my $PLUGIN_TYPE_NAME_S;		# 'Audits',     'Targets',     'Tools'

my $ITEM_ENV_NAME;		# 'GBS_AUDIT', 'GBS_TARGET', 'GBS_TOOL'
my $ITEMS_ENV_NAME;		# 'GBS_AUDITS', 'GBS_TARGETS', 'GBS_TOOLS'
my $ALL_ITEMS_ENV_NAME;		# 'GBS_ALL_AUDITS', 'GBS_ALL_TARGETS', 'GBS_ALL_TOOLS'

my $PLUGIN_NAME;		# e.g.: mingw win32
my $GBS_SCRIPTS_PLUGINS_PATH;	# $GBS_SCRIPTS_PATH/plugins/$PLUGIN_TYPE
my $GBS_SCRIPTS_PLUGIN_PATH;	# $GBS_SCRIPTS_PATH/plugins/$PLUGIN_TYPE/$PLUGIN_NAME

my $ITEM_NAME;			# e.g.: GBS_TARGET
my $GBS_ITEMS_PATH;		# $GBS_ROOT_PATH/$SYS_DIR
my $GBS_ITEM_PATH;		# $GBS_ROOT_PATH/$SYS_DIR/$ITEM_NAME

my $GBS_ITEM_VALUE;
my @GBS_ITEM_VALUES;
my @GBS_ALL_ITEM_VALUES;




my %PLUGINS;

my $PLUGINS_MAX_NAME;




sub PLUGIN_new_target($)
{
my ($default_target) = @_;

return new_build( 'build', $default_target);
}




sub PLUGIN_remove_target($)
{
my ($default_target) = @_;

return remove_item( 'build', $default_target);
}




sub PLUGIN_dialog($$)
{
my ($menu_entry,
$data_ref,
$entries_ref) = @_;

ENV_say( 1, "GBS Plugins");

GBSSCM_preset( 0);

my @menu_items = (


[ 'Build: Update',		\&main_update_dialog, 'build' ],

[ "\nAudit: Define/Install",	\&main_new_item,      'audit' ],
[ 'Audit: Update',		\&main_update_dialog, 'audit' ],
[ 'Audit: Remove',		\&main_remove_item,   'audit' ],
[ 'Tools: Define/Install',	\&main_new_item,      'tools' ],
[ 'Tools: Update',		\&main_update_dialog, 'tools' ],
[ 'Tools: Remove',		\&main_remove_item,   'tools' ],
);

ASK_menu( 'Select function to perform:', \@menu_items, $entries_ref);
}




sub main_remove_item($$$)
{
my ($menu_entry,
$data_ref,		    # $plugin_type
$entries_ref,
) = @_;

my $plugin_type = $data_ref;
GBSSCM_connect();

remove_item( $plugin_type, '');	# no default_item
}




sub main_update_dialog($$$)
{
my ($menu_entry,
$data_ref,		    # $plugin_type
$entries_ref,
) = @_;

my $plugin_type = $data_ref;
GBSSCM_connect();

update_dialog( $plugin_type, '');	# no default_item
}




sub main_new_item($$$)
{
my ($menu_entry,
$data_ref,		    # $plugin_type
$entries_ref,
) = @_;

my $plugin_type = $data_ref;
GBSSCM_connect();

if ($plugin_type eq 'build')
{
new_build( $plugin_type, '');
} elsif ($plugin_type eq 'audit')
{
new_audit( $plugin_type, '');
} elsif ($plugin_type eq 'tools')
{
new_tool( $plugin_type, '');
} else
{
ENV_fatal( "Impossible $plugin_type");
}
}




sub remove_item($$)
{
my ($plugin_type,		# audit build tools
$default_item,
) = @_;
my $removed_item = '';	# '' if nothing removed

setup_type( $plugin_type);
ENV_say( 0, '');
ENV_say( 1, "Remove $PLUGIN_TYPE_NAME $default_item");

ENV_say( 1, "Current $PLUGIN_TYPE_NAME_S = '@GBS_ITEM_VALUES'");

my $item_to_remove = select_item( remove => $default_item);
if ($item_to_remove ne '')
{



setup_item( $item_to_remove);

my @new_gbs_item_values = grep( $item_to_remove ne $_, @GBS_ITEM_VALUES);
system_store( \@new_gbs_item_values);

ENV_say( 1, "$PLUGIN_TYPE_NAME $item_to_remove removed",
"Note that the $PLUGIN_TYPE_NAME Directories/Files are NOT deleted");
ENV_say( 1, "Remaining $PLUGIN_TYPE_NAME_S: '@GBS_ITEM_VALUES'");

$removed_item = $item_to_remove;
setup_item( undef);
}

return $removed_item;
}




sub update_dialog($$)
{
my ($plugin_type,		# audit build tools
$default_item,
) = @_;

setup_type( $plugin_type);
ENV_say( 0, '');
ENV_say( 1, "Update $PLUGIN_TYPE_NAME $default_item");




my ($item_name, $plugin_name) = select_item( update => '');
if ($item_name ne '')
{



setup_item( $item_name);

my $update_dirs_func;
if ($PLUGIN_TYPE eq 'audit')
{
$update_dirs_func = \&audit_update_dirs;
} elsif ($PLUGIN_TYPE eq 'build')
{
$update_dirs_func = \&build_update_dirs;
} else #($PLUGIN_TYPE eq 'tools')
{
$update_dirs_func = \&tools_update_dirs;
}

my @menu_items = (
[ 'Add/Repair Directories',
$update_dirs_func, $item_name ],
[ "Update $PLUGIN_TYPE_NAME $item_name [$plugin_name]",
\&update_plugin, [ $item_name, $plugin_name ] ],
);
ASK_menu( "Select Update for $item_name", \@menu_items, undef);

}
}




sub update_plugin($$$)
{
my ($menu_entry,
$data_ref,		    # [ $item_name, $plugin_name ]
$entries_ref) = @_;

my ($item_name, $plugin_name) = @{$data_ref};

ENV_say( 1, "Update $PLUGIN_TYPE_NAME $item_name [$plugin_name]");

$plugin_name = select_plugin_name( $item_name)
if (!$plugin_name);

do_plugin_file( $plugin_name, $item_name, 'U')
if ($plugin_name);
}




sub new_build($$)
{
my ($plugin_type,		# audit build tools
$default_item,
) = @_;
my $new_target;

setup_type( $plugin_type);
ENV_say( 0, '');
ENV_say( 1, "New $PLUGIN_TYPE_NAME $default_item");

my $default_target = $default_item;	    #TBR

$new_target = select_new_item( $default_target);




if ($new_target ne '')
{




setup_item( $new_target);
my ($action,	    # REF ORG COPY or OWN
$plugin_name_or_item,
undef,	    # $targets_ref
$subsys_ref,
$comp_selector,
$components_sel_ref) = ask_new_data( $new_target, '');




ENV_say( 1, "Creating Target '$new_target'...");




build_create_dirs( $new_target, $subsys_ref, $comp_selector, $components_sel_ref);




do_install( $action, $plugin_name_or_item, $new_target);




system_add( $new_target);
GBSENV_setenv( GBS_TARGET => $new_target);

ENV_say( 1, "Target '$new_target' created");
} else
{
ENV_warn( "No new Target selected!");
}

return $new_target;
}




sub build_create_dirs($$$$)
{
my ($target,
$subsys_ref,
$comps_selector,
$components_sel_ref,
) = @_;

my @new_dirs;
push @new_dirs, "$GBS_ROOT_PATH/sysgen/$target";
push @new_dirs, "$GBS_ROOT_PATH/sysgen/makemake_stubs/$target";
push @new_dirs, "$GBS_ROOT_PATH/tmp/$target";

my @subsystems = @{$subsys_ref};
foreach my $subsys (@subsystems)
{
push @new_dirs, "$GBS_ROOT_PATH/dev/$subsys/gen/$target";

my @components = select_selector_items( $comps_selector, $components_sel_ref, [ GBSGLO_components( $subsys) ]);
foreach my $component (@components)
{
push @new_dirs, "$GBS_ROOT_PATH/dev/$subsys/comp/$component/bld/$target";
}
}
add_dirs( \@new_dirs);
}




sub build_update_dirs($$$)
{
my ($menu_entry,
$data_ref,		    # $target
$entries_ref) = @_;

my $target = $data_ref;

ENV_say( 1, "Add/Repair Dirs for Target $target");

my (undef,			# $action
undef,			# $plugin_name_or_item
undef,			# $targets_ref
$subsys_ref,
$comp_selector,
$components_sel_ref) = ask_new_data( $target, undef);

ENV_say( 1, "Updating Target '$target'...");
build_create_dirs( $target, $subsys_ref, $comp_selector, $components_sel_ref);
}




sub new_audit($$)
{
my ($plugin_type,		# audit build tools
$default_item,
) = @_;
my $new_audit;

setup_type( $plugin_type);
ENV_say( 0, '');
ENV_say( 1, "New $PLUGIN_TYPE_NAME $default_item");

$new_audit = select_new_item( '');




if ($new_audit ne '')
{
setup_item( $new_audit);
my ($action,	    # REF ORG COPY or OWN
$plugin_name_or_item,
$targets_ref,
$subsys_ref,
$comp_selector,
$components_sel_ref) = ask_new_data( $new_audit, '');




ENV_say( 1, "Creating Audit '$new_audit'...");




audit_create_dirs( $new_audit, $targets_ref, $subsys_ref, $comp_selector, $components_sel_ref);




do_install( $action, $plugin_name_or_item, $new_audit);




system_add( $new_audit);
GBSENV_setenv( GBS_AUDIT => $new_audit);

ENV_say( 1, "Audit '$new_audit' created");
} else
{
ENV_warn( "No new Audit selected!");
}

return $new_audit;
}




sub audit_create_dirs($$$$$)
{
my ($audit,
$targets_ref,
$subsys_ref,
$comps_selector,
$components_sel_ref,
) = @_;

my @new_dirs;
push @new_dirs, "$GBS_ROOT_PATH/sysaudit/$audit";
push @new_dirs, "$GBS_ROOT_PATH/tmp/$audit";

my @targets = @{$targets_ref};
foreach my $target (@targets)
{
my @subsystems = @{$subsys_ref};
foreach my $subsys (@subsystems)
{
if (-d "$GBS_ROOT_PATH/dev/$subsys/gen/$target")
{
push @new_dirs, "$GBS_ROOT_PATH/dev/$subsys/audit/$audit";

my @components = select_selector_items( $comps_selector, $components_sel_ref, [ GBSGLO_components( $subsys) ]);
foreach my $component (@components)
{
push @new_dirs, "$GBS_ROOT_PATH/dev/$subsys/comp/$component/aud/$audit";
push @new_dirs, "$GBS_ROOT_PATH/dev/$subsys/comp/$component/aud/$audit/$target";
}
} else
{
ENV_say( 0, "SubSys '$subsys' skipped (does not generate for Target '$target')");
}
}
}
add_dirs( \@new_dirs);
}




sub audit_update_dirs($$$)
{
my ($menu_entry,
$data_ref,		    # $audit
$entries_ref) = @_;

my $audit = $data_ref;

ENV_say( 1, "Add/Repair Dirs for Audit $audit");

my (undef,			    # $action
undef,			    # $plugin_name_or_item
$targets_ref,
$subsys_ref,
$comp_selector,
$components_sel_ref) = ask_new_data( $audit, undef);

ENV_say( 1, "Updating Audit '$audit'...");
audit_create_dirs( $audit, $targets_ref, $subsys_ref, $comp_selector, $components_sel_ref);
}




sub new_tool($$)
{
my ($plugin_type,		# audit build tools
$default_item,
) = @_;
my $new_tool;

setup_type( $plugin_type);
ENV_say( 0, '');
ENV_say( 1, "New $PLUGIN_TYPE_NAME $default_item");

$new_tool = select_new_item( '');




if ($new_tool ne '')
{
setup_item( $new_tool);
my ($action,	    # REF ORG COPY or OWN
$plugin_name_or_item,
undef,	    # $targets_ref
$subsys_ref,
undef,	    # $comps_selector
undef) = ask_new_data( $new_tool, '');




ENV_say( 1, "Creating Tool '$new_tool'...");




tools_create_dirs( $new_tool, $subsys_ref);




do_install( $action, $plugin_name_or_item, $new_tool);




system_add( $new_tool);
GBSENV_setenv( GBS_TOOL => $new_tool);

ENV_say( 1, "Tool '$new_tool' created");
} else
{
ENV_warn( "No new Tool selected!");
}

return $new_tool;
}




sub tools_create_dirs($$)
{
my ($tool,
$subsys_ref,
) = @_;

my @new_dirs;
push @new_dirs, "$GBS_ROOT_PATH/systools/$tool";
push @new_dirs, "$GBS_ROOT_PATH/tmp/$tool";

my @subsystems = @{$subsys_ref};
foreach my $subsys (@subsystems)
{
push @new_dirs, "$GBS_ROOT_PATH/dev/$subsys/tools/$tool";
}
add_dirs( \@new_dirs);
}




sub tools_update_dirs($$$)
{
my ($menu_entry,
$data_ref,		    # $tool
$entries_ref) = @_;

my $tool = $data_ref;

ENV_say( 1, "Add/Repair Dirs for Tool $tool");

my (undef,		# $action
undef,		# $plugin_name_or_item
undef,		# $targets_ref
$subsys_ref,
undef,		# $comps_selector
undef) = ask_new_data( $tool, undef);

ENV_say( 1, "Updating Tool '$tool'...");
tools_create_dirs( $tool, $subsys_ref);
}




sub select_item($$)
{
my ($function,	    # remove update copy
$wanted_item,	    # '' == force ask. Not eq '' == do not ask, take it if it exists
) = @_;
my ($item, $plugin_name) = ( '', undef);

$wanted_item = ''
if (!defined $wanted_item);
my $get_plugin_name_func_ref = $PLUGIN_TYPE_GET_DERIVED_ITEMS_FUNCS{$PLUGIN_TYPE};
if ($wanted_item eq '')
{
if (@GBS_ITEM_VALUES)
{
my @menu_items = ( [ '', '', undef ]);
foreach my $item_name (@GBS_ITEM_VALUES)
{
my $plugin_name = $get_plugin_name_func_ref->( $item_name, 'PLUGIN_NAME');
push @menu_items, [ "$item_name [$plugin_name]", $item_name, $plugin_name ];
};
my $default_item = ($function eq 'remove') ? '' : $GBS_ITEM_VALUE;

($item, $plugin_name) = ASK_value_from_menu( "Select $PLUGIN_TYPE_NAME to $function",
\$default_item, @menu_items);
} else
{
ENV_say( 1, "No $PLUGIN_TYPE_NAME_S defined!");
}
} else
{
if (grep $_ eq $wanted_item, @GBS_ITEM_VALUES)
{
$item = $wanted_item;
$plugin_name = $get_plugin_name_func_ref->( $item, 'PLUGIN_NAME')
if (wantarray);
} else
{
ENV_error_exit( "$PLUGIN_TYPE_NAME '$wanted_item' does not exists");
}
}

if ($item ne '')
{
ENV_say( 1, "Selected: $item [$plugin_name]");
} else
{
ENV_say( 1, "Selected: None");
}

return wantarray ? ($item, $plugin_name) : $item;
}




sub select_new_item($)
{
my ($default_item) = @_;
my $new_item;	    # '' if not selected

ENV_say( 1, "Create New $PLUGIN_TYPE_NAME");

my @gbs_all_targets = SYSTEM_get_all( $GBS_ROOT_PATH, 'targets');
my @gbs_all_audits  = SYSTEM_get_all( $GBS_ROOT_PATH, 'audits');
my @gbs_all_tools   = SYSTEM_get_all( $GBS_ROOT_PATH, 'tools');

{



ENV_say( 1, "All Items:");
my @row_refs = ( [ 'All Targets', "@gbs_all_targets" ],
[ 'All_Audits',  "@gbs_all_audits" ],
[ 'ALL Tools',   "@gbs_all_tools" ],
);
ENV_say( 0, FORMAT_table( 0, 2, ': ', undef, @row_refs));
}

{



my $get_plugin_name_func_ref = $PLUGIN_TYPE_GET_DERIVED_ITEMS_FUNCS{$PLUGIN_TYPE};
if (@GBS_ITEM_VALUES)
{
ENV_say( 1, "Current $PLUGIN_TYPE_NAME_S:");
my @items;
foreach my $item_name (@GBS_ITEM_VALUES)
{
my $plugin_name = $get_plugin_name_func_ref->( $item_name, 'PLUGIN_NAME');
push @items, "$item_name [$plugin_name]";
};
ENV_say( 0, FORMAT_cols( 0, 2, '  |  ', 2, \@items));
} else
{
ENV_say( 1, "No $PLUGIN_TYPE_NAME_S defined");
}
}




my @exclude_names = ( @gbs_all_targets, @gbs_all_audits, @gbs_all_tools, @{$EXCLUDE_ITEM_NAMES{$PLUGIN_TYPE}} );
$default_item = ''
if (grep( $default_item eq $_, @exclude_names));
do
{
$new_item = ASK_not_word( "Enter New $PLUGIN_TYPE_NAME-name", $default_item, \@exclude_names, 20);
if ($new_item ne '')
{
my $item_path = "$GBS_ROOT_PATH/$SYS_DIR/$new_item";
if (-d $item_path)
{
my @files = SLURP_dir_all( $item_path, 0);
ENV_say( 1, "  $PLUGIN_TYPE_NAME directory '$new_item' already exists",
"  It Contains:",
"    @files");
$new_item = undef
if (ASK_YNQ( "  Use it ($new_item) anyway?", 'N', 1) eq 'N');
}
}
} while (!defined $new_item);

return $new_item;
}




sub ask_new_data($$)
{
my ($item_name,		# name of target, audit or tool
$default_plugin,	# undef=OK. ask plugin if defined
) = @_;
my ($action, $plugin_name_or_item, $targets_ref, $subsys_ref, $comps_selector, $components_sel_ref);

my $ask_plugin = (defined $default_plugin);
my $ask_targets = ($PLUGIN_TYPE eq 'audit');
my $ask_subsystems = 1;
my $ask_components = ($PLUGIN_TYPE eq 'build' || $PLUGIN_TYPE eq 'audit');




my @gbs_targets = ENV_getenv( 'GBS_TARGETS');
my $targets_selector;
my @targets = @gbs_targets;
my @all_subsystems = GBSGLO_subsystems_full_gbs();
my $subsystems_selector;
my @subsystems = @all_subsystems;
my @all_components;	# cannot determine now
my $components_selector;
my @components;

$plugin_name_or_item = $default_plugin;	# default
my $all_ok;
do
{
$all_ok = 1;




if ($ask_plugin)
{
setup_plugin( undef);
my @menu_item_refs;

push @menu_item_refs, [ '<quit>', undef ];
push @menu_item_refs, [ "Create a Reference to a $PLUGIN_TYPE_NAME Plugin", 'REF',  \&select_plugin_name ];
push @menu_item_refs, [ "Create a Copy of a $PLUGIN_TYPE_NAME Plugin",      'ORG',  \&select_plugin_name ];
push @menu_item_refs, [ "Create a Copy of a $PLUGIN_TYPE_NAME",             'COPY', \&select_plugin_item ];
push @menu_item_refs, [ "Define your own from scratch",	                'OWN',  sub { return ''; } ];
map { $_->[0] =~ s/ a A/ an A/ } @menu_item_refs;	    # $display_text

($action, my $func_ref) =
ASK_value_from_menu( "Select action for Creating new $PLUGIN_TYPE_NAME", 1, @menu_item_refs);
$plugin_name_or_item = $func_ref->( $plugin_name_or_item);
if (defined $plugin_name_or_item)
{
setup_plugin( $plugin_name_or_item)
if ($action eq 'REF' || $action eq 'ORG');
} else
{
$all_ok = 0;
}
} else
{
$action = '';
}

if ($all_ok)
{
ENV_say( 0, '');
if ( $ask_targets && @gbs_targets > 1 ||
$ask_subsystems && @all_subsystems > 1 ||
$ask_components && @all_subsystems == 1 && @all_components > 1 ||
$ask_components && @all_subsystems > 1)
{
ENV_say( 1, "Now select which items must use this $PLUGIN_TYPE_NAME");
}




if ($ask_targets)
{
if (@gbs_targets == 0)
{
$targets_selector = 'None';
ENV_say( 1, "  No Targets");
} elsif (@gbs_targets == 1)
{
$targets_selector = 'ALL';
ENV_say( 1, "  Target @targets");
} else
{
ENV_say( 1, "  Possible Targets:",
"  @gbs_targets");
$targets_selector = ASK_value_from_menu( "  Select Targets-Selector",
0, qw( All Only Exclude));
my @selectors;
if ($targets_selector eq 'Only' || $targets_selector eq 'Exclude')
{
push @selectors, ASK_values_from_list( "$targets_selector Targets:", '', \@gbs_targets, -1);
}
@targets = select_selector_items( $targets_selector, \@selectors, \@gbs_targets);
ENV_say( 1, "  Targets: @targets");
}
}




if ($ask_subsystems)
{
if (@all_subsystems == 0)
{
$subsystems_selector = 'None';
ENV_say( 1, "  No SubSystems");
} elsif (@all_subsystems == 1)
{
$subsystems_selector = 'All';
ENV_say( 1, "  SubSystem: @all_subsystems");
} else
{
ENV_say( 1, "  Possible SubSystems:",
"  @all_subsystems");
$subsystems_selector = ASK_value_from_menu( "  Select SubSystems-Selector",
0, qw( All Only Exclude None));
my @selectors;
if ($subsystems_selector eq 'Only' || $subsystems_selector eq 'Exclude')
{
push @selectors, ASK_values_from_list( "  $subsystems_selector SubSystems:", '', \@all_subsystems, -1);
}
@subsystems = select_selector_items( $subsystems_selector, \@selectors, \@all_subsystems);
ENV_say( 1, "  SubSystems: @subsystems");
}
} else
{
if (@all_subsystems == 0)
{
$subsystems_selector = 'None';
} else
{
$subsystems_selector = 'All';
}
}




if ($ask_components)
{
if (@subsystems == 0)
{
ENV_say( 1, "  No Components");
$components_selector = 'None';
} elsif (@subsystems == 1)
{
@all_components = GBSGLO_components( $subsystems[0]);
@components = @all_components;
if (@all_components == 0)
{
$components_selector = 'None';
ENV_say( 1, "  No Components");
} elsif (@all_components == 1)
{
$components_selector = 'All';
ENV_say( 1, "  Component: @all_components");
} else
{
ENV_say( 1, "  Possible Components:",
"  @all_components");
$components_selector = ASK_value_from_menu( "  Select Components-Selector",
0, qw( All Only Exclude None));
my @selectors;
if ($components_selector eq 'Only' || $components_selector eq 'Exclude')
{
push @selectors, ASK_values_from_list( "  $components_selector Components:", '', \@all_components, -1);
}
@components = select_selector_items( $components_selector, \@selectors, \@all_components);
ENV_say( 1, "  Components: @components");
}

} else
{
ENV_say( 1, "  Multiple SubSystems are involved: use wildcards for Components Selection");
$components_selector = ASK_value_from_menu( "  Select Components-Selector",
0, qw( All Only Exclude None));
if ($components_selector eq 'Only' || $components_selector eq 'Exclude')
{



@components = ASK_wild_words( "  $components_selector (wild) Components:", '', -1, -1);
}
}
}




my $item_1;
if ($action eq 'COPY')
{




$item_1 = substr( "$PLUGIN_TYPE_NAME   ", 0, 7);
} else  # REF ORG or OWN
{
$item_1 = 'Plugin ';
}
ENV_say( 0, '');
ENV_say( 1, "Summary for $PLUGIN_TYPE_NAME '$item_name':");
ENV_say( 0, "  $item_1    : $plugin_name_or_item")		if ($ask_plugin);
ENV_say( 0, "  Targets    : $targets_selector: @targets")	if ($ask_targets);
ENV_say( 0, "  SubSystems : $subsystems_selector: @subsystems")	if ($ask_subsystems);
ENV_say( 0, "  Components : $components_selector: @components")	if ($ask_components);
} else
{
ENV_say( 0, '');
ENV_say( 1, 'Retry');
$all_ok = 0;
}
} while (!$all_ok || ASK_YNQ( 'OK?', 'Y', 1) eq 'N');

$targets_ref = \@targets
if ($ask_targets);
$subsys_ref = \@subsystems
if ($ask_subsystems);
if ($ask_components)
{
$comps_selector = $components_selector;
$components_sel_ref = [ @components ];
}
return ($action, $plugin_name_or_item, $targets_ref, $subsys_ref, $comps_selector, $components_sel_ref);
}





sub select_plugin_name($)
{
my ($default_plugin,	# May be undef. Not used
) = @_;
my $plugin_name;

get_plugins();




my @menu_refs;
push @menu_refs, ([ '' ]);		# Menu entry 0 : <none> Nothing

foreach my $ref (@{$PLUGINS{$PLUGIN_TYPE}})
{
my ($name, $description) = @{$ref};
my $display_text = sprintf( "%-*s: %s", $PLUGINS_MAX_NAME, $name, $description);
push @menu_refs, [ $display_text, $name ];
}




my $index = ASK_index_from_menu( "  Select $PLUGIN_TYPE plugin",
\$DEFAULT_GEN_NAME, ( @menu_refs));
$plugin_name = $menu_refs[$index]->[1];	# $name
ENV_say( 1, "  Selected: $plugin_name")
if (defined $plugin_name);

return $plugin_name;
}





sub select_plugin_item($)
{
my ($default_item,	# current target or audit. May be undef
) = @_;
my $item_name;

$item_name = select_item( copy => '');  # force ask

$item_name = undef
if ($item_name eq '');

return $item_name;
}




sub do_install($$$)
{
my ($action,	# REF ORG COPY or OWN
$plugin_name_or_item,
$item_name,
) = @_;

my $plugin_filespec = FSPEC_plugin_file( $PLUGIN_FILE, $GBS_ROOT_PATH, $item_name);


if ($action eq 'REF')
{
if (-e $plugin_filespec)
{
SCM_assert_co_file( $plugin_filespec);
ENV_unlink( $plugin_filespec);
}
FCREATE_plugin_ref_file( $PLUGIN_FILE, $GBS_ROOT_PATH, $item_name, $plugin_name_or_item);
do_plugin_file( $plugin_name_or_item, $item_name, 'I');
} elsif ($action eq 'ORG')
{
do_plugin_file( $plugin_name_or_item, $item_name, 'U');
} elsif ($action eq 'COPY')
{
my $from_plugin_item = $plugin_name_or_item;
my $from_dir = "$GBS_ITEMS_PATH/$from_plugin_item";
my @files = SLURP_dir_files( $from_dir, 0);
if (@files)
{
ENV_say( 1, "$PLUGIN_TYPE_NAME $from_plugin_item contains the following files",
"@files");
my @to_file_specs;
foreach my $file (@files)
{
my $from_spec = "$from_dir/$file";
my $to_spec = "$GBS_ITEM_PATH/$file";
ENV_say( 0, " copy $from_spec -> $GBS_ITEM_PATH/");
SCM_assert_co_file( $to_spec);			# TBS add ITEM-translate
FILE_copy_translate( $from_spec, $to_spec);
push @to_file_specs, $to_spec;
}
ENV_say( 1, scalar @to_file_specs . " copied");
SCM_add_txt_files( \@to_file_specs, 0, 1)
} else
{
ENV_say( 1, "No files in $PLUGIN_TYPE_NAME $from_plugin_item");
}
} elsif ($action eq 'OWN')
{

} else
{
ENV_fatal( "Impossible $action");
}




if (! -e $plugin_filespec)
{



FCREATE_plugin_file( $PLUGIN_FILE, $GBS_ROOT_PATH, $item_name);
}
SCM_add_txt_files( $plugin_filespec, 0, 1);		# $prevalidated, $verbose_level
}




sub get_plugins()
{
if (! $PLUGINS{$PLUGIN_TYPE})
{
my @plugin_refs = PLG_get_plugins( $PLUGIN_TYPE);





$PLUGINS_MAX_NAME = 0;
foreach my $ref (@plugin_refs)
{
my $plugin_name = $ref->[0];
my $name_length = length( $plugin_name);
$PLUGINS_MAX_NAME = $name_length
if ($name_length > $PLUGINS_MAX_NAME);
}
$PLUGINS{$PLUGIN_TYPE} = [ @plugin_refs ];
}
}




sub do_plugin_file($$$)
{
my ($plugin_name,
$item_name,
$install_or_update, # 'I' or 'U'
) = @_;

my $to_dir = "$GBS_ROOT_PATH/$SYS_DIR/$item_name";
my $plugin_dir = "$GBS_SCRIPTS_PATH/plugins/$PLUGIN_TYPE";
my $from_dir = "$plugin_dir/$plugin_name";

my @data_refs = PLG_get_plugin_data( $plugin_name, $PLUGIN_TYPE, $install_or_update);




if (@data_refs)
{



ENV_say( 1, "Plugin Data overview:");
foreach my $ref (@data_refs)
{
my ( $func, $command, $ask_default, @args ) = @{$ref};
if ($command eq 'C')
{



ENV_say( 0, "  Copy @args");
} elsif ($command eq 'R')
{



ENV_say( 0, "  Run @args");
} elsif ($command eq 'D')
{



ENV_say( 0, "  Create Dir(s) @args");
} else # ($command eq 'E')
{



ENV_say( 0, "  Echo @args");
}
}




ENV_say( 1, "Executing...");
foreach my $ref (@data_refs)
{
my ( $func, $command, $ask_default, @args ) = @{$ref};
if ($command eq 'C')
{



my ($from_file, $to_file) = @args;	    # names may contain dirs!
my @file_refs;

if (ENV_is_wildcard( $from_file))
{
my @from_files = glob( "$from_dir/$from_file");
foreach my $from_file (@from_files)
{
$from_file =~ s!^$from_dir/!!;
my $file = ENV_split_spec_f( $from_file);
push @file_refs, [ $file, $from_file, $file ];
}
} else
{
my $file = ENV_split_spec_f( $from_file);
$to_file = $file
if (! $to_file);
push @file_refs, [ $file, $from_file, $to_file ];
}
my @from_file_names = map { $_->[1] } @file_refs;
ENV_say( 1, "Files to copy:");
ENV_say( 0, "    @from_file_names");

my $ans = ($ask_default eq '+') ? 'A' : $ask_default;
foreach my $ref (@file_refs)
{
my ($file, $from_file, $to_file) = @{$ref};
my $from_file_spec = "$from_dir/$from_file";
my $to_file_spec   = "$to_dir/$to_file";

$ans = ASK_YNAE( "Copy file: $from_file?", $ask_default)
if ($ans ne 'A');
if ($ans eq 'Y' || $ans eq 'A')
{
copy_1_file( $file, $from_file_spec, $to_file_spec, $plugin_name);
} elsif ($ans eq 'E')
{
last;
}
}
} elsif ($command eq 'R')
{



my ($script, $script_spec) = @args;

my $ans = 'Y';
if ($ask_default ne '+')
{
$ans = ASK_YN( 'Run Installer?', $ask_default);
}
if ($ans eq 'Y')
{
ENV_say( 1, "Executing $script...");
my $command_ref = ENV_command_ref( $script_spec, [ $func, $from_dir, $to_dir ]);
my $rc = ENV_system( $command_ref, undef, undef);
ENV_error( "Install failed")
if ($rc != 0);
}
} elsif ($command eq 'D')
{



my $dir = $args[0];
my $dir_spec = "$to_dir/$dir";
if (!-d $dir_spec)
{
ENV_say( 1, "Creating Directory '$dir'");
SCM_add_dirs( [ $dir_spec ], \&FILES_get_ignore_spec, 0, 1);	    # $pre_validated, $verbose
}
} else # ($command eq 'E')
{



ENV_say( 1, "$args[0]");
}
}
} else
{

}
}




sub copy_1_file($$$$)
{
my ($file_name,
$in_file_spec,
$to_file_spec,
$plugin_name,
) = @_;

my $ans = 'Y';
if (-e $to_file_spec)
{
$ans = ASK_YN( "File already exists. Replace?", "N");
}
if ($ans eq 'Y')
{
SCM_assert_co_file( $to_file_spec);

my $file_name_uc = uc $file_name;
$file_name_uc =~ s/\..*$//;		    # remove .type
my $type_name_uc = uc $PLUGIN_TYPE_NAME;    # TARGET AUDIT TOOL
FILE_copy_template( $in_file_spec, $to_file_spec,
( SCRIPT => $file_name_uc, $type_name_uc => $plugin_name ));

ENV_chmod ('ug+x', $to_file_spec)
if (ENV_is_commandfile( $to_file_spec));
SCM_add_txt_files( $to_file_spec, 1, 1);	# $prevalidated, $verbose_level
ENV_say( 1, "File copied");
}
}




sub select_selector_items($$$)
{
my ($selector,			# 'All', 'Only', "Exclude', 'None'
$selections_ref,		# Do not modify! ??
$all_items_ref,
) = @_;
my @selection;

my @selection_list = @{$selections_ref};    # copy. We do not want to modify
if ($selector eq 'All')
{
@selection =  @{$all_items_ref};
} elsif ($selector eq 'Only')
{
@selection = ENV_wildcards( \@selection_list, $all_items_ref, undef);
} elsif ($selector eq 'Exclude')
{
@selection = ENV_not_wildcards( \@selection_list, $all_items_ref);
} elsif ($selector eq 'None')
{
@selection = ();
} else
{
ENV_fatal( "Impossible selector: $selector");
}


return @selection;
}




sub add_dirs($)
{
my ($dirs_ref) = @_;

ENV_say( 0, $dirs_ref);

my @new_dirs = grep( !-d $_, @{$dirs_ref});		# remove already existing
@new_dirs = LIST_unique( \@new_dirs);		# remove duplicates
SCM_add_dirs( \@new_dirs, \&FILES_get_ignore_spec, 0, 1);	    # $pre_validated, $verbose_level
}




sub system_add($)
{
my ($new_item,
) = @_;




my @current_items = @GBS_ITEM_VALUES;
push @current_items, $new_item;
my @new_current_items = sort( LIST_unique( \@current_items));

system_store( \@new_current_items);
}




sub system_store($)
{
my ($new_gbs_item_values_ref,
) = @_;




@GBS_ITEM_VALUES = @{$new_gbs_item_values_ref};

SYSTEM_write( $GBS_ROOT_PATH, $PLUGIN_TYPE_ID => [ @GBS_ITEM_VALUES ]);
@GBS_ALL_ITEM_VALUES = SYSTEM_get_all( $GBS_ROOT_PATH, $PLUGIN_TYPE_ID);




GBSENV_setenv( $ITEMS_ENV_NAME => [ @GBS_ITEM_VALUES ]);
GBSENV_setenv( $ALL_ITEMS_ENV_NAME => [ @GBS_ALL_ITEM_VALUES ] );
}




sub setup_type($)
{
my ($plugin_type) = @_;

if (defined $plugin_type)
{
($PLUGIN_TYPE, $PLUGIN_TYPE_NAME, $PLUGIN_FILE, $SYS_DIR, $SUBSYS_DIR, $PLUGIN_TYPE_ID) = @{$PLUGIN_SETUP{$plugin_type}};

$PLUGIN_TYPE_NAME_S = "${PLUGIN_TYPE_NAME}s";

$GBS_SCRIPTS_PLUGINS_PATH   = "$GBS_SCRIPTS_PATH/plugins/$PLUGIN_TYPE";
$GBS_ITEMS_PATH		    = "$GBS_ROOT_PATH/$SYS_DIR";

my $plugin_type_id_uc = uc $PLUGIN_TYPE_ID;
$ITEMS_ENV_NAME	    = "GBS_$plugin_type_id_uc";		# 'GBS_AUDITS', 'GBS_TARGETS', 'GBS_TOOLS'
$ALL_ITEMS_ENV_NAME = "GBS_ALL_$plugin_type_id_uc";	# 'GBS_ALL_AUDITS', 'GBS_ALL_TARGETS', 'GBS_ALL_TOOLS'
$ITEM_ENV_NAME      = substr( $ITEMS_ENV_NAME, 0, -1);	# 'GBS_AUDIT', 'GBS_TARGET', 'GBS_TOOL'

$GBS_ITEM_VALUE = ENV_getenv( $ITEM_ENV_NAME);
@GBS_ITEM_VALUES = ENV_getenv( $ITEMS_ENV_NAME);
@GBS_ALL_ITEM_VALUES = ENV_getenv( $ALL_ITEMS_ENV_NAME);
} else
{
($PLUGIN_TYPE, $PLUGIN_TYPE_NAME, $PLUGIN_FILE, $SYS_DIR, $SUBSYS_DIR, $PLUGIN_TYPE_ID) = ();

$PLUGIN_TYPE_NAME_S = undef;

$GBS_SCRIPTS_PLUGINS_PATH = undef;
$GBS_ITEMS_PATH = undef;

my $plugin_type_id_uc = undef;
$ITEMS_ENV_NAME	= undef;
$ALL_ITEMS_ENV_NAME = undef;
$ITEM_ENV_NAME = undef;

$GBS_ITEM_VALUE = undef;
@GBS_ITEM_VALUES = undef;
@GBS_ALL_ITEM_VALUES = undef;
}

setup_plugin( undef);
setup_item( undef);
}




sub setup_plugin($)
{
my ($plugin_name) = @_;

if (defined $plugin_name)
{
$PLUGIN_NAME = $plugin_name;
$GBS_SCRIPTS_PLUGIN_PATH = "$GBS_SCRIPTS_PATH/plugins/$PLUGIN_TYPE/$PLUGIN_NAME";
} else
{
$PLUGIN_NAME = undef;
$GBS_SCRIPTS_PLUGIN_PATH = undef;
}
}




sub setup_item($)
{
my ($item_name) = @_;

if (defined $item_name)
{
$ITEM_NAME = $item_name;
$GBS_ITEM_PATH = "$GBS_ROOT_PATH/$SYS_DIR/$ITEM_NAME";
} else
{
$ITEM_NAME = undef;
$GBS_ITEM_PATH = undef;
}
}

return 1;
