#===============================================================
#
#   prqa.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#   Generic package for qac and qac++
#===============================================================
#
#   Call:
#	arg_1	= file_name	(without type)
#	arg_2	= in_file_spec	(relative file spec - file_name + type
#	arg_3	= out_file_base (relative file spec - with in file type)
#	arg_4	= out_file_dir	(relative file spec - directory only)
#	arg_5	= out_file_type
#	arg_6-N = command_flags (from command line)
#   Environment:
#	General:
#	    GBS_INCS		Include paths ( -I<dir> ... )
#	    GBS_SYSINCS 	System include paths ( -SI<dir> ... )
#	    GBS_AUDIT_FLAGS	Passed from session
#	    GBS_FLAGS		From flags_<type> files
#	    GBS_SYSFLAGS	From bld_flags_<type> file
#	    GBS_AUDIT_OPTIONS	($MUST_SHOW_STDOUT)
#	Application:
#	    GBS_TARGET
#	    GBS_AUDIT
#	    GBSEXT_tool_REL
#	    GBSEXT_tool_PATH
#	    GBS_COMPONENT
#	    toolHELPFILES   (optional)
#============================================================================

































package prqa;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
PRQA_init
PRQA_pre
PRQA_run
PRQA_post
PRQA_load
);

use strict;
use integer;

use glo::env;
use glo::slurp;
use gbsenv;
use scadef;
use scaglo;
use scafile;
use lib "$ENV{GBS_SCRIPTS_PATH}/plugins/audit/qac/bin";
use prqaglo;




sub PRQA_init($$$$);
sub PRQA_pre($$$$);
sub PRQA_run();
sub PRQA_post();
sub PRQA_load();

sub handle_unkn_msg($$$);
sub do_metrics();
sub handle_st_data($$);
sub create_msg_file($);




my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_TARGET = $ENV{GBS_TARGET};
my $GBS_AUDIT = $ENV{GBS_AUDIT};




my $PLUGIN_NAME;		# qac or qacpp
my $GBSEXT_PLUGIN_PATH;
my $TARGET_PLUGIN_NAME;

my $FOR_QAC;			# bool
my $FOR_QACPP;			# bool
my $PLUGIN_NAME_UC;		# QAC or QACPP
my $GBS_SYSAUDIT_AUDIT_PATH;
my $PLUGIN_HELPFILES_PATH;
my $GBSEXT_PLUGIN_BIN_PATH;	# depends on GBSEXT_PLUGIN_USE_STUB




my @GBS_DEFAULTS;	    # @QAC_SETTINGS / @QACPP_SETTINGS, @ANA_SETTINGS
my @COMPILER_SETTINGS;	    #
my @GBS_SETTINGS;	    # @ERRDISP_SETTINGS




my %MET_DEFS;





my $SRC_PATH;		    # CWD
my $COMPONENT;
my $SOURCE_FILE;
my $GOUT_FILE_SPEC;

my $SOURCE_FILE_SPEC;
my $OS_SOURCE_FILE_SPEC;
my $MET_FILE_SPEC;	    # qac/qacpp output file

my $USR_MSG_FILESPEC;	    # if undef: file does not exist




my %MET_DATA;

my @MET_DATA_ORDER;




my $USR_MSG_FILE_TYPE = '.gbs';

my @QAC_SETTINGS = (
qw( -comment a ),
);

my @QACPP_SETTINGS = (
);

my @ANA_SETTINGS = (		# QAC & QACPP (-D && -Q)
-D => 'PRQA',
);

my @PAL_SETTINGS = (
);

my @ERRDSP_SETTINGS = (
qw( -maxcount 50 ),		# Limit the number of times that any particular warning message is displayed
qw( -references+ ),		# The optional verbose/reference text specified in the message file is displayed
qw( -messagesonly+ ),	# -m+ == diaglist, -m- == annsrc
-format => '"\n*GBS*%F*%l*%C*%u*%g*%n*%t\n--%v\n--%B"',


);




sub PRQA_init($$$$)
{
(	$PLUGIN_NAME,
$GBSEXT_PLUGIN_PATH,
$TARGET_PLUGIN_NAME,
my $extra_gbs_defaults_ref,	# e.g. -quiet options
) = @_;


$FOR_QAC = ($PLUGIN_NAME eq 'qac') ? 1 : 0;
$FOR_QACPP = ($PLUGIN_NAME eq 'qacpp') ? 1 : 0;
$PLUGIN_NAME_UC = uc $PLUGIN_NAME;
$GBS_SYSAUDIT_AUDIT_PATH = "$GBS_ROOT_PATH/sysaudit/$GBS_AUDIT";

PRQAGLO_init( $PLUGIN_NAME);

map { $MET_DEFS{$_} = 1 } @{SCADEF_get_section_items( $GBS_AUDIT, 'METRICS', 'ENABLE')};
map { $MET_DEFS{$_} = 0 } @{SCADEF_get_section_items( $GBS_AUDIT, 'METRICS', 'IGNORE')};

SCAGLO_sanity_check( [ PRQAGLO_get_msg_filespecs() ], \&create_msg_file,
undef, undef);	    # help_file




push @ANA_SETTINGS, @{$extra_gbs_defaults_ref};

$USR_MSG_FILESPEC = "$GBS_SYSAUDIT_AUDIT_PATH/$PLUGIN_NAME.usr$USR_MSG_FILE_TYPE";	# .gbs
@GBS_DEFAULTS = (($FOR_QAC) ? @QAC_SETTINGS : @QACPP_SETTINGS, @ANA_SETTINGS);





{
my @files = ("comp_$GBS_TARGET.p_c");
push @files, "comp_$TARGET_PLUGIN_NAME.p_c"
if ($TARGET_PLUGIN_NAME ne $GBS_TARGET);
my $compiler_filespec = SCAGLO_find_settings_filespec( @files);


@COMPILER_SETTINGS = ( '-via', ENV_shortest_path( undef, $compiler_filespec))
if (defined $compiler_filespec);
}





if (-e $USR_MSG_FILESPEC)
{
push @ERRDSP_SETTINGS, "-usrfile $USR_MSG_FILE_TYPE";	    # .gbs
push @ERRDSP_SETTINGS, "-usrpath $GBS_SYSAUDIT_AUDIT_PATH";
} else
{
$USR_MSG_FILESPEC = undef;
}
@GBS_SETTINGS = (@ERRDSP_SETTINGS);





$PLUGIN_HELPFILES_PATH = ENV_getenv( "${PLUGIN_NAME_UC}HELPFILES");
$PLUGIN_HELPFILES_PATH = "$GBSEXT_PLUGIN_PATH/help"
if ($PLUGIN_HELPFILES_PATH eq '' && -d "$GBSEXT_PLUGIN_PATH/help");

$ENV{"${PLUGIN_NAME_UC}BIN"} = ENV_os_path( "$GBSEXT_PLUGIN_PATH/bin");
$ENV{"${PLUGIN_NAME_UC}TEMPFILES"} = $ENV{GBS_TMP_PATH};
$ENV{"${PLUGIN_NAME_UC}HELPFILES"} = $PLUGIN_HELPFILES_PATH;

{
my $gbsext_use_stub = ENV_getenv( "GBSEXT_${PLUGIN_NAME}_USE_STUB");
$gbsext_use_stub = ($gbsext_use_stub =~ /^(1|Y|y)/) ? 1 : 0;
if ($gbsext_use_stub)
{
ENV_say( 1, "** Running in Development Mode from STUBS...");
$GBSEXT_PLUGIN_BIN_PATH = "$ENV{GBS_SCRIPTS_PATH}/STUBS/qac/bin";
} else
{
$GBSEXT_PLUGIN_BIN_PATH = "$GBSEXT_PLUGIN_PATH/bin";
}
ENV_setenv( "GBSEXT_${PLUGIN_NAME_UC}_BIN_PATH" => ENV_os_path( $GBSEXT_PLUGIN_BIN_PATH));
}

return ( '* ', \&handle_unkn_msg);
}





sub PRQA_pre($$$$)
{
(	$SRC_PATH,
$COMPONENT,
$SOURCE_FILE,
$GOUT_FILE_SPEC,
) = @_;
my @printable_out_files;

$SOURCE_FILE_SPEC = "$SRC_PATH/$SOURCE_FILE";
$OS_SOURCE_FILE_SPEC = ENV_os_path( $SOURCE_FILE_SPEC);
my $out_sub_path = "../aud/$GBS_AUDIT";
my $out_file_path = "$out_sub_path/$GBS_TARGET";
my $out_file_base = "$out_file_path/$SOURCE_FILE";




$ENV{"${PLUGIN_NAME_UC}OUTPATH"} = ENV_os_path( $out_file_path);

$MET_FILE_SPEC = "$out_file_base.met";

SCAFILE_garg_write( '* ', \@GBS_DEFAULTS, \@COMPILER_SETTINGS, \@GBS_SETTINGS);

if ($FOR_QAC)
{
ENV_setenv( "GBS_QAC_FLAGS" => "@QAC_SETTINGS @ANA_SETTINGS @COMPILER_SETTINGS");
} else
{
ENV_setenv( "GBS_QACPP_FLAGS" => "@QACPP_SETTINGS @ANA_SETTINGS @COMPILER_SETTINGS");
}

my $style_filespec = ENV_shortest_path( undef, "$GBS_ROOT_PATH/sysaudit/$GBS_AUDIT/style_ALL.nrf");
ENV_setenv( GBS_PAL_FLAGS => "@PAL_SETTINGS -nrf $style_filespec");

ENV_setenv( GBS_ERRDSP_FLAGS => "@ERRDSP_SETTINGS");

@printable_out_files = ( $GOUT_FILE_SPEC );

return @printable_out_files;
}






sub PRQA_run()
{
my $rc = 0;

ENV_fatal( "The PRQA_run function is obsolete");

return $rc;
}





sub PRQA_post()
{

my @warning_refs;


my @total_refs;


my @metric_refs;





{
















my @lines = SCAFILE_gout_read();

my $nr_lines = @lines;
my $i = 0;
while ($i < $nr_lines)
{



my $line1 = $lines[$i++];		# *GBS*...
next if ($line1 eq '');
my $line2 = $lines[$i++];		# --<Extra message>
my $line3 = $lines[$i++];		# --<file>

if ($line1 !~ /^\*GBS\*/ ||
$line2 !~ /^--/ ||
$line3 !~ /^--/ )
{
ENV_say( 0, "=== $GOUT_FILE_SPEC");
map { ENV_say( 0, sprintf( "%7d $lines[$_]", $_+1)) } $i-3..$i-1;
ENV_say( 0, "===");
ENV_error_exit( "Invalid .gout file format. ");
}
my (undef, $gbs_text, $os_file_spec, $line_nr, $col_nr, $not_used_msg_depth, $msg_level, $msg_nr, $msg_text) = split( '\*', $line1, 9);
my $src_line_nr;
my $level_text;
my $help_file_url = '';	# filled by GBSBGSCA
my @extra_src_lines;

if ($os_file_spec eq $OS_SOURCE_FILE_SPEC)	    # In source
{
$os_file_spec = '-';
$src_line_nr = $line_nr;
} else					    # In include
{
$src_line_nr = -1;
}

$level_text = ($msg_level == 0) ? 'INFO' :
($msg_level == 9) ? 'ERROR' :
($msg_level  > 9) ? 'SEVERE ERROR' :
'WARNING';

my $extra_msg_line = substr( $line2, 2);	# Skip '--'
if ($extra_msg_line ne '')
{
@extra_src_lines = ($extra_msg_line);
}

push @warning_refs, [ $src_line_nr, $col_nr, $os_file_spec, $line_nr, $msg_level, $level_text, $msg_nr, $msg_text, $help_file_url, @extra_src_lines ];


}
}




my ($total_refs_ref, $metric_refs_ref) = do_metrics();

return (\@warning_refs, $total_refs_ref, $metric_refs_ref);
}




sub do_metrics()
{

my @total_refs;


my @metric_refs;


my $cur_file_spec = '';
my $cur_st_size;
my $cur_st_name;
my $cur_st_type;    # 1 = Function, 2 = Class, 3 = File
my @cur_st_refs;    # [ $st_name, $st_value ]

my $s_nr_lines = -1;
my $t_nr_lines = -1;
my $s_nr_funcs = -1;
my $t_nr_funcs = -1;
my $s_nr_class = -1;
my $t_nr_class = -1;
my $s_nr_const = -1;
my $t_nr_const = -1;

foreach my $line (SLURP_file( $MET_FILE_SPEC), '<S>')
{
next if ($line eq '');
next if ($line !~ /^(<DEFINE>|<S>)/);


if (substr( $line, 0, 8) eq '<DEFINE>')
{



my ( undef, undef, $file_spec, undef, undef, $def, $space, undef, undef, $type ) =
split( /\s+/, $line);
$cur_file_spec = $file_spec;
next if ($space eq 'OM');                   # skip macros
next if ($def eq 'DC' && $space eq 'OF');   # skip function-prototypes
my $nr_consts = $type =~ tr/=/=/;


$t_nr_const += $nr_consts;
$s_nr_const += $nr_consts
if ($cur_file_spec eq $SOURCE_FILE_SPEC);

} elsif ($line eq '<S>')
{
if (@cur_st_refs)
{
if ($cur_st_type == 1)
{

foreach my $ref (@cur_st_refs)
{
handle_st_data( $ref->[0], $ref->[1])
if ($cur_st_size > 0);
}
} elsif ($cur_st_type == 2)
{

foreach my $ref (@cur_st_refs)
{
handle_st_data( $ref->[0], $ref->[1]);
}
} else
{

foreach my $ref (@cur_st_refs)
{
handle_st_data( $ref->[0], $ref->[1]);
}
}
$cur_st_size = 0;
$cur_st_name = '';
$cur_st_type = 0;
@cur_st_refs = ();      # [ $st_name, $st_value ]
}
} elsif (substr( $line, 0, 5) eq '<S>ST')
{



my ($full_key, @values) = split( ' ', $line);
my $metric_name = substr( $full_key, 5);
if ($metric_name eq 'FIL')          # <S>STFIL
{
$cur_file_spec = $values[0];

} elsif ($metric_name eq 'NAM')     # <S>STNAM  (function or class)
{
my $name = $values[0];

if ($name eq $SOURCE_FILE_SPEC)
{

$cur_st_type = 3;   # File
} else
{
$cur_st_name = $name;
if (substr( $name , 0, 2) eq '::' && substr( $name, -1, 1) ne ')' )
{

$cur_st_type = 2;   # Class
$t_nr_class++;
$s_nr_class++
if ($cur_file_spec eq $SOURCE_FILE_SPEC);
} else
{

$cur_st_type = 2;   # Func
$t_nr_funcs++;
$s_nr_funcs++
if ($cur_file_spec eq $SOURCE_FILE_SPEC);
}
}
} elsif ($metric_name eq 'XLN')     # <S>STXLN  (Number of Executable Statements)
{
$cur_st_size = $values[0];
push @cur_st_refs, [ $metric_name, $values[0] ];
} elsif (exists $MET_DEFS{$metric_name})
{
push @cur_st_refs, [ $metric_name, $values[0] ]
if ($MET_DEFS{$metric_name} == 1);
} else
{
ENV_warn( "Unknown met-key '$full_key'");
}
} else
{

}
}

$MET_DATA{'TLN'} = [ 0, 0, 0 ] if (!exists $MET_DATA{'TLN'});
$MET_DATA{'TPP'} = [ 0, 0, 0 ] if (!exists $MET_DATA{'TPP'});

$s_nr_lines = $MET_DATA{'TLN'}->[0]; # Total source lines
$t_nr_lines = $MET_DATA{'TPP'}->[0]; # Totals pre-processed lines




@total_refs = (
[ 0, $s_nr_lines, $t_nr_lines ],	# Lines
[ 1, $s_nr_funcs, $t_nr_funcs ],	# Funcs
[ 2, $s_nr_class, $t_nr_class ],	# Class
[ 3, $s_nr_const, $t_nr_const ],	# Const
);




foreach my $metric_name (@MET_DATA_ORDER)
{
my ($min, $max, $total) = @{$MET_DATA{$metric_name}};
push @metric_refs, [ $metric_name => $min, $max, $total ];
}

return ( \@total_refs, \@metric_refs );
}





sub handle_st_data($$)
{
my ($metric_name,
$value,
) = @_;

no integer;

if (exists $MET_DATA{$metric_name})
{
my ($min, $max, $total) = @{$MET_DATA{$metric_name}};
$min = $value if ($value < $min);
$max = $value if ($value > $max);
$total += $value;
$MET_DATA{$metric_name} = [ $min, $max, $total ];
} else
{
$MET_DATA{$metric_name} = [ $value, $value, $value ];
push @MET_DATA_ORDER, $metric_name;
}

use integer;
}





sub create_msg_file($)
{
my ($plugin_msg_files_ref) = @_;	 # 2, not used here

ENV_say( 1, "Creating msg file...");

my @msg_refs = PRQAGLO_read_qac_msg();

SCAGLO_write_msg_file( \@msg_refs);
}





sub handle_unkn_msg($$$)
{
my ($msg_nr,		# defined
$msg_level,		# defined
$level_text,		# defined
) = @_;
my $help_file_url = '';


$help_file_url = "$msg_nr.html";		# Create a 'msg does not exist.html' file?

return ($msg_level, $level_text, $help_file_url);
}





sub PRQA_load()
{
ENV_say( 1, "Loaded");
}

return 1;



