#===============================================================
#
#   prqaopt.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#   Generic package for qac and qac++
#===============================================================

package prqaopt;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
PRQAOPT_init
PRQAOPT_get_pos_args
PRQAOPT_get_opts
);

use strict;
use integer;

use glo::env;
use glo::slurp;
use glo::check;




sub PRQAOPT_init($$$$$);
sub PRQAOPT_get_pos_args();
sub PRQAOPT_get_opts(@);

sub parse_args(@);
sub do_echo($$);
sub do_forget($$);
sub do_help($$);
sub do_via($$);
sub print_error($@);








my @PROGRAM_NAMES = qw( qac qacpp pal errdsp viewer errsum errwrt);
my %PROGRAM_INDICES;
{
my $i = 0;
%PROGRAM_INDICES = map { $_ => $i++ } @PROGRAM_NAMES;
}

my $FORMAT = '%?u==0%(%?C%(%^%)%)%?u==0%(%?h%(Err%:Msg%)%:-->%)(%g:%N) %R(%u, )%t';
my $IT_REF = [ 'size_t=unsigned int', 'ptrdiff_t=long', 'wchar_t=unsigned char'];
my $PO_REF = [ 'df::analyse_header_functions-', 'df::inter=0', 'df::map_old_messages-', 'df::query_timeout=10000' ];
my $STYLE  = 'exdented';
my @OPT_DEFS = (




[ [ qw( a		align			    ) ], 2,      [], [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# [type]=[bytes]
[ [ qw( accd	allowconstconvdeduction	    ) ], 1,       0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( aicpm	allowimplicitconvptrmbr	    ) ], 1,       0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( ar		arithrsh		    ) ], 1,       0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( atanpc	allowtmplargnullptrconv	    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( atncrb	allowtempnonconstrefbind    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( audi	alwaysusedirectinitializer  ) ], 1,       0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( bits	bitsigned		    ) ], 1,       0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( co		comment			    ) ], 2,     'n', [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# {a, n, i}
[ [ qw( cmaf	crossmoduleanalysisfile	    ) ], 2,      '', [ 0, 0, 1, 0, 0, 0, 0 ], undef ],	# filename
[ [ qw( c++03auto				    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( c++03decltype			    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( d					    ) ], 2,      [], [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# ident[=[value]]
[ [ qw( dpft	delayparsefntemplates	    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( duc99dilt	dontusec99decintlittypes    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( e		echo			    ) ], 2,   undef, [ 1, 1, 0, 1, 0, 0, 0 ], \&do_echo ],   # text
[ [ qw( ed		enabledataflow		    ) ], 1,	  0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( emhm	expandmultihomedmessages    ) ], 2,      [], [ 0, 0, 0, 1, 1, 0, 0 ], undef ],	# message_number_list
[ [ qw( en		encoding		    ) ], 2,   'ASC', [ 1, 1, 1, 1, 1, 1, 1 ], undef ],	# {ASC , EUC, NEWJ, OLDJ, NECJ, SJ, UTF}
[ [ qw( ex		extensions		    ) ], 2,      '', [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# {ANSIPC, PC, ASM, C++, DOLLAR, LONGLONG, ALL}
[ [ qw( fi		forceinclude		    ) ], 2,      '', [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# filename
[ [ qw( file				    ) ], 2,      '', [ 0, 0, 0, 1, 0, 1, 0 ], undef ],	# filename
[ [ qw( forget	forgetall		    ) ], 2,   undef, [ 1, 1, 1, 1, 1, 1, 1 ], \&do_forget ], # option
[ [ qw( format				    ) ], 2, $FORMAT, [ 0, 0, 0, 1, 1, 0, 0 ], undef ],	# string
[ [ qw( fpc		foldplainchar		    ) ], 1,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( h		help			    ) ], 0,   undef, [ 1, 1, 1, 1, 1, 1, 1 ], \&do_help ],
[ [ qw( hdr		hdrsuppress		    ) ], 1,       0, [ 0, 0, 0, 1, 1, 0, 0 ], undef ],
[ [ qw( hm		hiddenmessage		    ) ], 1,       0, [ 0, 0, 0, 0, 0, 0, 1 ], undef ],
[ [ qw( hmw		hidemacrowarnings	    ) ], 1,       0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( html				    ) ], 1,       0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( hw		hiddenwarnings		    ) ], 1,       0, [ 0, 0, 0, 1, 1, 1, 0 ], undef ],
[ [ qw( i					    ) ], 2,      [], [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# path
[ [ qw( ifn		introducefriendnames	    ) ], 1,       0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( il		indentlevel		    ) ], 2,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# value
[ [ qw( it		intrinsictype		    ) ], 2, $IT_REF, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# {size_t | ptrdiff_t | wchar_t}=datatype
[ [ qw( k+r					    ) ], 2,      [], [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# [sun, a, b, c, e, g, i, l, n, p, s, t, u, v, x]
[ [ qw( lf		layoutfile		    ) ], 2,	 '', [ 0, 1, 0, 0, 0, 0, 0 ], undef ],	# file
[ [ qw( lih		layoutignoreheaders	    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( liw		layoutindentwidth	    ) ], 2,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],	# value
[ [ qw( luhi	layoutusehangingindent	    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( l		line			    ) ], 1,       0, [ 0, 0, 0, 1, 1, 0, 0 ], undef ],
[ [ qw( list				    ) ], 2,      [], [ 0, 0, 1, 1, 1, 1, 0 ], undef ],	# filelist.lst
[ [ qw( m		messagesonly		    ) ], 1,	  0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( max		maxcount		    ) ], 2,       0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],	# value
[ [ qw( maxerr	maxerrors		    ) ], 2,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# value
[ [ qw( mll		maxlinelength		    ) ], 2,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# value
[ [ qw( met		metrics			    ) ], 1,       1, [ 1, 1, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( mrtnd  maxrecursivetemplatenestingdepth ) ], 2,     200, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( no		nomsg		n	    ) ], 2,	 [], [ 1, 1, 0, 1, 1, 1, 0 ], undef ],	# message_number_list
[ [ qw( nosort				    ) ], 1,       0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( na		namelength		    ) ], 2,	 31, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# value
[ [ qw( nrf		namerulefile		    ) ], 2,	 '', [ 0, 0, 1, 0, 0, 0, 0 ], undef ],	# rulefile
[ [ qw( o		only			    ) ], 2,	 [], [ 1, 1, 0, 1, 1, 1, 0 ], undef ],	# message_number_list
[ [ qw( one		onelineonly		    ) ], 1,       0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( op		outpath			    ) ], 2,	 '', [ 1, 1, 1, 1, 1, 1, 1 ], undef ],	# path (%QACOUTPATH%)
[ [ qw( po		prodoption		    ) ], 2, $PO_REF, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# df::analyse_header_functions{+|-}
[ [ qw( ppf		ppfilename		    ) ], 1,       0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( ppl		pplist			    ) ], 1,	  0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( ppm		ppmetrics		    ) ], 1,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( ppo		pponly			    ) ], 1,       0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( pp32bt	preproc32bittype	    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( ptp		propagatetemplateparameters ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( q		quiet			    ) ], 2,	 [], [ 1, 1, 1, 0, 0, 0, 0 ], undef ],	# {filename | path}
[ [ qw( ref		references		    ) ], 1,       0, [ 0, 0, 0, 1, 1, 0, 0 ], undef ],
[ [ qw( rem		remark			    ) ], 2,   undef, [ 1, 1, 1, 1, 1, 1, 1 ], sub { return 0 } ],
[ [ qw( s		size			    ) ], 2,      [], [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# [type]=[bits]
[ [ qw( sa		showalias		    ) ], 1,	  1, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( sd		systemdefine		    ) ], 2,	 [], [ 0, 1, 0, 0, 0, 0, 0 ], undef ],	# ident[=[value]]
[ [ qw( sdep	searchdependentbaseclasses  ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( set		settings		    ) ], 1,	  0, [ 1, 1, 0, 1, 0, 1, 0 ], undef ],
[ [ qw( si		systeminclude		    ) ], 2,	 [], [ 0, 1, 0, 0, 0, 0, 0 ], undef ],	# path
[ [ qw( sig		stdisglobal		    ) ], 1,	  0, [ 0, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( sl		slashwhite		    ) ], 1,	  0, [ 1, 1, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( spragma	skippragma		    ) ], 2,	 [], [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# [pragma_start],[pragma_end]
[ [ qw( sr		trictrank		    ) ], 2,	  0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( ss          strictsignedness	    ) ], 1,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# (obsolete)
[ [ qw( st		standard		    ) ], 1,	  0, [ 0, 0, 0, 0, 1, 0, 0 ], undef ],
[ [ qw( sty		style			    ) ], 2,  $STYLE, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# {k+r, exdented, indented}
[ [ qw( su		suppresslvl		    ) ], 2,	 [], [ 0, 0, 0, 1, 1, 1, 0 ], undef ],	# message_level
[ [ qw( summ	summary			    ) ], 1,	  0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( tab		tabstop		t	    ) ], 2,	  8, [ 1, 1, 0, 1, 1, 0, 0 ], undef ],	# columns
[ [ qw( text	textonly		    ) ], 1,	  0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( thresh	threshold		    ) ], 2,      [], [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# metric{<|<=|=|>|>=}value[:msg_no]
[ [ qw( total				    ) ], 1,       0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( tsumm	tablesummary		    ) ], 1,       0, [ 0, 0, 0, 1, 0, 0, 0 ], undef ],
[ [ qw( u		unsignedchar		    ) ], 1,       0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( up		usrpath			    ) ], 2,	 '', [ 0, 0, 0, 1, 1, 1, 0 ], undef ],	# path (%QACBIN%)
[ [ qw( usr		usrfile			    ) ], 2,	 '', [ 0, 0, 0, 1, 1, 1, 0 ], undef ],	# extension (type)
[ [ qw( ver		version			    ) ], 1,	  0, [ 1, 1, 1, 1, 1, 1, 1 ], undef ],
[ [ qw( via					    ) ], 2,      [], [ 1, 1, 1, 1, 1, 1, 1 ], \&do_via ],
[ [ qw( wc		warncall		    ) ], 2,	 [], [ 1, 1, 0, 0, 0, 0, 0 ], undef ],	# [function[=msg_no]]
[ [ qw( wk		warnkeyword		    ) ], 2,	 [], [ 0, 1, 0, 0, 0, 0, 0 ], undef ],	# [keyword[=msg_no]]
[ [ qw( wp		warnpragma		    ) ], 2,	 [], [ 0, 1, 0, 0, 0, 0, 0 ], undef ],	# [pragma[*]=[0 | msg_no]
[ [ qw( xc		xcase			    ) ], 1,	  0, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],
[ [ qw( xn		xnamelength		    ) ], 2,	  6, [ 1, 0, 0, 0, 0, 0, 0 ], undef ],	# value
);
my %OPT_KEYS;

foreach my $ref (@OPT_DEFS)
{
my @mnems = @{$ref->[0]};	#   $mnems_ref
map { $OPT_KEYS{$_} = $ref } @mnems;
}




my $PROGRAM_NAME;	#   qac, qacpp, pal, errdsp, viewer, errsum, errwrt
my $USAGE_TEXT;
my $HELP_TEXT;		#   May be undef

my $QAC;		#   qac or qacpp
my $QAC_UC;		#   QAC or QACPP
my $PROGRAM_INDEX;




my $QACOUTPATH;
my $QACBIN;
my $QACTEMPFILES;
my $QACHELPFILES;




my %OPTS;



my @POS_ARGS;

my %OPTS_COUNT;


my @VIA_FILES;	    # files stack. First is Command-line. Last is Current.




sub PRQAOPT_init($$$$$)
{
(   $PROGRAM_NAME,		# qac, qacpp, pal, errdsp, viewer, errsum, errwrt
my $nr_req_pos_args,	# value or [min, max]
my $mandatory_options_ref,
$USAGE_TEXT,
$HELP_TEXT,		# undef = OK
) = @_;


my $nr_args = @ARGV;
ENV_error_exit( "At least 1 parameter required")
if ($nr_args  == 0);




if ($nr_args == 1)
{
if ($ARGV[0] eq '--h')
{
ENV_say( 0, "Usage: $USAGE_TEXT");
exit (0);
} elsif ($ARGV[0] eq '--help')
{
ENV_say( 0, "Usage: $USAGE_TEXT");
if (defined $HELP_TEXT)
{
ENV_say( 0, $HELP_TEXT);
} else
{
ENV_say( 0, '*No additional help*');
}
exit( 0);
}
}




foreach my $ref (@OPT_DEFS)
{
my $mnem = $ref->[0]->[0];
my $default = $ref->[2];
$OPTS{$mnem} = $default;
}




$QAC = lc $ARGV[0];
ENV_error_exit( "First arg ($ARGV[0]) must be 'qac' or 'qacpp or qac++'")
if ($QAC ne 'qac' && $QAC ne 'qacpp' && $QAC ne 'qac++');
$QAC = 'qacpp'
if ($QAC eq 'qac++');
$QAC_UC = uc $QAC;
$PROGRAM_NAME = $QAC
if (!defined $PROGRAM_NAME);
$PROGRAM_INDEX = $PROGRAM_INDICES{$PROGRAM_NAME};




$QACOUTPATH	    = ENV_getenv( "${QAC_UC}OUTPATH");
$QACBIN	    = ENV_getenv( "${QAC_UC}BIN");



$OPT_KEYS{op}->[2] = $QACOUTPATH;	# $default outpath;
$OPT_KEYS{up}->[2] = $QACBIN;	# $default userpath;






push @VIA_FILES, 'Command-Line';
parse_args( @ARGV);
pop @VIA_FILES;




{
my $err_text = CHECK_num_value( 'Nr ARGs', scalar @POS_ARGS, $nr_req_pos_args);
ENV_error_exit( $err_text, @POS_ARGS)
if (defined $err_text);
}





if (defined $mandatory_options_ref)
{
foreach my $opt (@{$mandatory_options_ref})
{
ENV_error_exit( "Mandatory option '$opt' not specified")
if (!defined $OPTS_COUNT{$opt});
}
}

return $QAC;   # qac or qacpp
}




sub PRQAOPT_get_pos_args()
{
return @POS_ARGS;
}




sub PRQAOPT_get_opts(@)
{
my @opts = @_;
my @values;

foreach my $opt (@opts)
{
my $value = $OPTS{$opt};
if (defined $value)
{
push @values, (ref $value) ? @{$value} : $value;
} else
{
ENV_error_exit( "No such option '$opt");
}
}

return (wantarray) ? @values : $values[0];
}




sub parse_args(@)
{
my @args = @_;


my $nr_args = @args;
my $i = 0;
while ($i < $nr_args)
{
my $arg = $args[$i];
if (substr( $arg, 0, 1) eq '-')
{
my $opt = substr( $arg, 1);
my ($key, $value) = $opt =~ /^(\w+|\w+[+]+\w+)(.*)$/;	# (aaa | aa++bb) (+|-)
if (defined $key)
{
$key = lc $key;
$value = ''
if (!defined $value);

my $found_ref = $OPT_KEYS{$key};
if (defined $found_ref)
{
my ($mnems_ref, $type, $default, $valid_for_ref, $action_func) = @{$found_ref};



my $mnem = $mnems_ref->[0];
$OPTS_COUNT{$mnem}++;
if (!$valid_for_ref->[$PROGRAM_INDEX])
{
my @actives;
for (my $i = 0; $i < 7; $i++)
{
if ($valid_for_ref->[$i])
{
push @actives, $PROGRAM_NAMES[$i];
}
}
print_error( 'I', "Option '$arg' not active for '$PROGRAM_NAME' (@actives)")
}
my $opt_is_ok;
if ($type == 0)		    # plain
{
$opt_is_ok = 1;
} elsif ($type == 1)	    # bool
{
if ($value eq '+')
{
$value = 1;
$opt_is_ok = 1;
} elsif ($value eq '-')
{
$value = 0;
$opt_is_ok = 1;
} else
{
print_error( 'E', "Option ($arg) value must be '+' or '-'");
}
} else	# ($type == 2)	    # value in next arg
{
if ($value eq '')
{
$value = $args[++$i];
if (defined $value)
{
$opt_is_ok = 1;
} else
{
print_error( 'E', "Arg value missing at end of Option-list ($arg)");
}
} elsif ($value =~ /^\s+(.*)/)
{
$value =~ $1;
$opt_is_ok = 1;
} else
{
print_error( 'E', "Invalid option syntax ($arg)");
}
}
if ($opt_is_ok)
{

if (ref $default)
{
push @{$OPTS{$mnem}}, $value;
} elsif (defined $default)
{
$OPTS{$mnem} = $value;
}
if (defined $action_func)
{
my $must_continue = $action_func->( $mnem, $value);
$i = $nr_args
if (!$must_continue);
}
}
} else
{
print_error( 'E', "Unknown option '$arg'");
}
} else
{
print_error( 'E', "Invalid option '$arg'");
}
} else
{
push @POS_ARGS, $arg;

}
$i++;
}
}




sub do_echo($$)
{
my ($mnem,	    # echo, e
$value,
) = @_;
my $must_continue = 1;

ENV_say( 0, $value);

return $must_continue;
}




sub do_forget($$)
{
my ($mnem,	    # forget, forgetall
$option,
) = @_;
my $must_continue = 1;

my $opt_def_ref = $OPT_KEYS{$option};
if (defined $opt_def_ref)
{
$mnem = $opt_def_ref->[0]->[0];	    # $mnems_ref
ENV_say( 1, "Forgetall: $option");
if (ref $OPTS{$mnem})
{
$OPTS{$mnem} = [];
} else
{
$OPTS{$mnem} = '';
}
} else
{
print_error( 'E', "Invalid option ($option) in arg (-$mnem $option)");
}

return $must_continue;
}



sub do_help($$)
{
my ($mnem,	    # help, h
$value,
) = @_;
my $must_continue = 1;

ENV_say( 0, "Usage: $USAGE_TEXT");
if (defined $HELP_TEXT)
{
ENV_say( 0, $HELP_TEXT);
} else
{
ENV_say( 0, '*No additional help*');
}

return $must_continue;
}




sub do_via($$)
{
my ($mnem,	    # via
$file_spec,
) = @_;
my $must_continue = 1;

my $full_file_spec = ENV_expand_envs($file_spec);

if (-e $full_file_spec)
{

push @VIA_FILES, $full_file_spec;
foreach my $line (SLURP_file( $full_file_spec))
{
next if ($line eq '' || substr( $line, 0, 1) eq '*');

parse_args( ENV_split_quoted_space( $line));
}
pop @VIA_FILES;

} else
{
print_error( 'EE', "-via file '$full_file_spec' not found");
}

return $must_continue;
}




sub print_error($@)
{
my ($type,	    # 'I', 'W', 'E', 'EE' or 'F'
@msg_lines
) = @_;

my $cur_file = "$VIA_FILES[-1]:";
if ($type eq 'W')
{
ENV_warn( $cur_file, @msg_lines);
} elsif ($type eq 'E')
{
ENV_error( $cur_file, @msg_lines);
} elsif ($type eq 'EE')
{
ENV_error_exit( $cur_file, @msg_lines);
} elsif ($type eq 'F')
{
ENV_fatal( $cur_file, @msg_lines);
} elsif ($type eq 'I')
{
ENV_info( $cur_file, @msg_lines);
} else
{
ENV_fatal( "Invalid type ($type)");
}
}

return 1;


