#===============================================================
#
#   sca.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package sca;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCA_init
SCA_get_filespec
SCA_msg_is_enabled
SCA_get_format_name
SCA_get_quiet_paths
SCA_get_threshold_refs
);

use strict;
use integer;

use glo::env;
use gbsenv;
use gbsfile;
use gbsfilecom;
use scamet;
use scadef;




sub SCA_init($$$$$);
sub SCA_get_filespec();
sub SCA_msg_is_enabled($);
sub SCA_get_format_name();
sub SCA_get_quiet_paths();
sub SCA_get_threshold_refs();

sub read_sca($$$$$);
sub do_errornumber($$$);
sub do_format($$$);
sub do_quiet($$$);
sub do_metrics_threshold($$$);




my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_SYSAUDIT_PATH = "$GBS_ROOT_PATH/sysaudit";

my $CUR_SCA_AUDIT = '';
my $CUR_SCA_TARGET = '';
my $CUR_SCA_REF;

my %SCAS = ();







my %SPECIAL_MSG_NRS;	    # used only during read_sca





my @FORMAT_VALUES = qw( NONE DEFAULT PCLINT VISUAL_C GNU_C ECLIPSE);
my %FORMAT_VALUES = map { $_ => 1 } @FORMAT_VALUES;







my %SCA_DEF = (

E	=> [ -1,  \&do_errornumber  ],
F	=> [  1,  \&do_format],
Q	=> [  1,  \&do_quiet ],
T   => [  1,  \&do_metrics_threshold ],
);
my @TYPES = sort keys %SCA_DEF;





sub SCA_init($$$$$)
{
my ($audit,
$target,
$all_msg_nrs_ref,	    # All possible msg_nrs
$fatal_msg_nrs_ref,
$specw_msg_nrs_ref,
) = @_;




read_sca( $audit, $target, $all_msg_nrs_ref, $fatal_msg_nrs_ref, $specw_msg_nrs_ref);
}




sub SCA_get_filespec()
{
return $CUR_SCA_REF->[0];	# file_spec
}




sub SCA_msg_is_enabled($)
{
my ($msg_nr,
) = @_;
my $enabled;

$enabled = $CUR_SCA_REF->[1]->{$msg_nr};

if (!defined $enabled)
{
my @keys = sort keys %{$CUR_SCA_REF->[1]};	# $msg_nrs
ENV_say( 0, "@keys");
ENV_error( "Unknow Msg $msg_nr in $CUR_SCA_AUDIT: $CUR_SCA_TARGET");
$enabled = 0;
}

return $enabled;
}




sub SCA_get_format_name()
{
return $CUR_SCA_REF->[2];	    # $wanted_format
}




sub SCA_get_quiet_paths()
{
return @{$CUR_SCA_REF->[3]};    # @quiet_paths
}




sub SCA_get_threshold_refs()
{
return @{$CUR_SCA_REF->[4]};    # @metric_threshold_refs

}




sub read_sca($$$$$)
{
my ($audit,
$target,
$all_msg_nrs_ref,	    # All possible msg_nrs
$fatal_msg_nrs_ref,
$specw_msg_nrs_ref,
) = @_;

$CUR_SCA_REF = $SCAS{$audit}->{$target};
$CUR_SCA_AUDIT = $audit;
$CUR_SCA_TARGET = $target;
if (exists( $SCAS{$audit}->{$target}))
{

} else
{




$CUR_SCA_REF->[0] = undef;	# $file_spec
$CUR_SCA_REF->[1] = {};		# $msg_nrs
$CUR_SCA_REF->[2] = 'DEFAULT';	# $wanted_format
$CUR_SCA_REF->[3] = [];		# @quiet_list
$CUR_SCA_REF->[4] = [];		# @metric_threshold_refs




map { $CUR_SCA_REF->[1]->{$_ * 1} = 1 } @{$all_msg_nrs_ref};




%SPECIAL_MSG_NRS = map { $_ * 1 => 1 } (@{$fatal_msg_nrs_ref}, @{$specw_msg_nrs_ref});




my $file_spec = "$GBS_SYSAUDIT_PATH/$audit/sca_$target";
my @inc_path = ( "$GBS_SYSAUDIT_PATH/$audit",
);

my $cur_file_spec = GBSFILE_open( $file_spec, 0, \@inc_path, undef);	# $must_exist, $allow_target_line_selection
if ($cur_file_spec)
{
$CUR_SCA_REF->[0] = $cur_file_spec;

my $line = GBSFILE_get_line();
while ($line)
{




my ($valid_line, $comment) = split( /[^\\]#/, $line, 2);




my ($op, @values) = split( / /, $valid_line);
GBSFILE_error_exit( "Line syntax = +|-<type> <value(s)>")
if (length( $op) != 2 || @values == 0);
my $plus_minus = substr( $op, 0, 1);
my $type = substr( $op, 1, 1);
my $enable = ($plus_minus eq '+') ? 1 : ($plus_minus eq '-') ? 0 : undef;
GBSFILE_error_exit( "Line syntax = +|-<type> <value(s)>")
if (!defined $enable);




my $sca_ref = $SCA_DEF{$type};
GBSFILE_error_exit( "Invalid type ($type). Must be one of (@TYPES)")
if (!defined $sca_ref);
my ($max_args, $do_func_ref) = @{$sca_ref};
GBSFILE_error_exit( "Only $max_args value(s) allowed")
if ($max_args >= 0 && @values > $max_args);
$do_func_ref->( $enable, $type, \@values);

$line = GBSFILE_get_line();
}
ENV_warn( "No Msgs enabled for $CUR_SCA_AUDIT: $CUR_SCA_TARGET")
if (keys %{$CUR_SCA_REF->[1]} == 0);
} else
{
ENV_say( 1, "All messages enabled for $CUR_SCA_AUDIT: $CUR_SCA_TARGET");
}


%SPECIAL_MSG_NRS = ();	    # cleanup
}
}




sub do_errornumber($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

foreach my $msg_nr (@{$values_ref})
{
if ($msg_nr eq 'ALL')
{
if ($enable)
{
map { $CUR_SCA_REF->[1]->{$_} = 1 } keys %{$CUR_SCA_REF->[1]};
} else
{
map { $CUR_SCA_REF->[1]->{$_} = 0 } keys %{$CUR_SCA_REF->[1]};
map { $CUR_SCA_REF->[1]->{$_} = 1 } keys %SPECIAL_MSG_NRS;
}
} else
{
$msg_nr *= 1;
if (exists $CUR_SCA_REF->[1]->{$msg_nr})
{
if ($enable)
{
$CUR_SCA_REF->[1]->{$msg_nr} = 1;
} else
{
$CUR_SCA_REF->[1]->{$msg_nr} = 0;
if (exists $SPECIAL_MSG_NRS{$msg_nr})
{
GBSFILE_info( "Implicit disable of fatal or special message ($msg_nr)");
}
}
} else
{
GBSFILE_error( "Unknown Msg_nr $msg_nr in $CUR_SCA_AUDIT: $CUR_SCA_TARGET");
}
}
}
}




sub do_format($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

my $value = $values_ref->[0];
GBSFILE_error_exit( "Invalid FORMAT value '$value' (@FORMAT_VALUES)")
if (!exists $FORMAT_VALUES{$value});

$CUR_SCA_REF->[2] = $value;			# $wanted_format
}




sub do_quiet($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

my $path = GBSFILECOM_validate_path( $values_ref->[0]);

push @{$CUR_SCA_REF->[3]}, $path;		# $quiet_paths
}




sub do_metrics_threshold($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

my $threshold = $values_ref->[0];
my ($metric_name, $min_value, $max_value, $msg_nr) = uc( $threshold) =~ /^([A-Z0-9]+)=(\d+|)(,\d+|)(:\d+|)$/;

if (defined $metric_name)
{
$max_value = substr( $max_value, 1)	    # remove leading ','
if ($max_value ne '');
$msg_nr = substr( $msg_nr, 1)	    	    # remove leading ':'
if ($msg_nr ne '');
GBSFILE_error_exit( "At least one Threshold Limit (min and/or max) must be specified")
if ($min_value eq '' && $max_value eq '');

if (SCADEF_metric_is_enabled( $CUR_SCA_AUDIT, $metric_name))
{

push @{$CUR_SCA_REF->[4]}, [ $metric_name, $min_value, $max_value, $msg_nr ];	# @metric_threshold_refs
} else
{
if (SCAMET_name_is_valid( $metric_name))
{
GBSFILE_error_exit( "Metric mnemonic '$metric_name' not enabled")
} else
{
GBSFILE_error_exit( "Unknown metric: '$metric_name'");
}
}
} else
{
GBSFILE_error_exit( "Threshold value syntax is: <metric_name>=[<min_value>][,<max_value>][:<msg_nr>]");
}
}

return 1;


