#===============================================================
#
#   scaglo.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package scaglo;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCAGLO_init
SCAGLO_sanity_check
SCAGLO_write_msg_file
SCAGLO_read_msg_file
SCAGLO_write_help_file
SCAGLO_find_settings_filespec
);

use strict;
use integer;

use glo::env;
use glo::spit;
use glo::slurp;
use glo::file;
use glo::html;
use gbshtml;
use gbsenv;




sub SCAGLO_init($$$);
sub SCAGLO_sanity_check($$$$);
sub SCAGLO_write_msg_file($);
sub SCAGLO_read_msg_file();
sub SCAGLO_write_help_file($);
sub SCAGLO_find_settings_filespec(@);




my $GBS_SCRIPTS_PATH = GBSENV_perl_gbs_scripts();
my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_AUDIT = ENV_getenv( 'GBS_AUDIT');
my $GBS_SILO_CACHE_PATH = "$GBS_ROOT_PATH/silo/.gbs/cache";




my $PLUGIN_NAME;
my $PLUGIN_REL;
my $PLUGIN_NICE_NAME;

my $GBS_SILO_PLUGIN_PATH;    # $GBS_SILO_CACHE_PATH/$PLUGIN_NAME/$PLUGIN_REL

my $MSG_FILESPEC;	    # $GBS_SILO_PLUGIN_PATH/msg.txt
my $HELP_FILESPEC;	    # $GBS_SILO_PLUGIN_PATH/msg_help.html




sub SCAGLO_init($$$)
{
($PLUGIN_NAME,
$PLUGIN_REL,
$PLUGIN_NICE_NAME,
) = @_;

$GBS_SILO_PLUGIN_PATH = "$GBS_SILO_CACHE_PATH/$PLUGIN_NAME/$PLUGIN_REL";
if (! -e $GBS_SILO_PLUGIN_PATH)
{
ENV_say( 1, "Creating Silo Plugin dir...");
mkdir "$GBS_SILO_CACHE_PATH/$PLUGIN_NAME";
mkdir "$GBS_SILO_CACHE_PATH/$PLUGIN_NAME/$PLUGIN_REL";
}

my $gbs_silo_audit_path  = "$GBS_SILO_CACHE_PATH/$GBS_AUDIT";   # still needed now gcon has been replaced?
if (! -e $gbs_silo_audit_path)
{
ENV_say( 1, "Creating Silo Audit dir...");
mkdir $gbs_silo_audit_path;
}

$MSG_FILESPEC  = "$GBS_SILO_PLUGIN_PATH/msg.txt";
$HELP_FILESPEC = "$GBS_SILO_PLUGIN_PATH/msg_help.html";
}




sub SCAGLO_sanity_check($$$$)
{
my ($plugin_msg_files_ref,		# mandatory - may also be scalar
$plugin_msg_file_read_func,	# callback
$plugin_help_files_ref,		# optional - may also be scalar (may be undef)
$plugin_help_file_read_func,	# callback (may be undef)
) = @_;




$plugin_msg_files_ref  = [ $plugin_msg_files_ref ] if (!ref( $plugin_msg_files_ref));
foreach my $plugin_msg_filespec (@{$plugin_msg_files_ref})
{
if (ENV_file_is_newer( $plugin_msg_filespec, $MSG_FILESPEC))
{
$plugin_msg_file_read_func->( $plugin_msg_files_ref);
last;
}
}




if (defined $plugin_help_files_ref)
{
$plugin_help_files_ref = [ $plugin_help_files_ref ] if (!ref( $plugin_help_files_ref));
foreach my $plugin_help_filespec (@{$plugin_help_files_ref})
{
if (ENV_file_is_newer( $plugin_help_filespec, $HELP_FILESPEC))
{
$plugin_help_file_read_func->( $plugin_help_files_ref);
last;
}
}
}
}






sub SCAGLO_write_msg_file($)
{
my ($msg_refs_ref,
) = @_;


ENV_say( 1, "Writing Silo Msg file...");

my @msg_refs = sort( { $a->[0] <=> $b->[0] } @{$msg_refs_ref});
my @lines = map { join( "\t", @{$_}) } @msg_refs;

my $fh = FILE_lock_ex_wait( "$MSG_FILESPEC.lck");
SPIT_file_nl( $MSG_FILESPEC, \@lines);
FILE_unlock( $fh);

return $MSG_FILESPEC;
}





sub SCAGLO_read_msg_file()
{
my @msg_refs;

my $fh = FILE_lock_sh_wait( "$MSG_FILESPEC.lck");
@msg_refs = map { [ split( "\t", $_) ] } SLURP_file( $MSG_FILESPEC);
FILE_unlock( $fh);

return @msg_refs;
}






sub SCAGLO_write_help_file($)
{
my ($help_refs_ref,	    # [ 'M', $msg_nr, $level_text, $heading, @lines ]

) = @_;


my $title = "$PLUGIN_NICE_NAME ($PLUGIN_REL) Warnings Helpfile";
my @out_lines;
push @out_lines, GBSHTML_doc_start( $HELP_FILESPEC, "$title - Created by GBS", 0);
push @out_lines, GBSHTML_doc_top( $title, ());
push @out_lines, HTML_br();
foreach my $ref (@{$help_refs_ref})
{
my $ref_type = $ref->[0];
if ($ref_type eq 'M')
{



my (undef, $msg_nr, $level_text, $heading, @lines) = @{$ref};
push @out_lines, HTML_h( 2, { id => $msg_nr }, "$msg_nr ($level_text) -- $heading");
push @out_lines, HTML_pre_s();
my $line = join( "\n", @lines);
push @out_lines, HTML_pre_txt( $line);
push @out_lines, HTML_pre_e();
} else # ($ref_type eq 'T')
{



my (undef, $heading, @lines) = @{$ref};
push @out_lines, HTML_h( 2, $heading);
push @out_lines, HTML_pre_s();
my $line = join( "\n", @lines);
push @out_lines, HTML_pre_txt( $line);
push @out_lines, HTML_pre_e();
}
}
push @out_lines, GBSHTML_doc_bottom( $title, undef);
push @out_lines, GBSHTML_doc_end();


my $fh = FILE_lock_ex_wait( "$HELP_FILESPEC.lck");
SPIT_file_nl( $HELP_FILESPEC, \@out_lines);
FILE_unlock( $fh);

return $HELP_FILESPEC;
}




sub SCAGLO_find_settings_filespec(@)
{
my @files = @_;
my $filespec;

my @dirs = (
"$GBS_ROOT_PATH/sysaudit/$PLUGIN_NAME",
"$GBS_SCRIPTS_PATH/plugins/audit/$PLUGIN_NAME",
);

foreach my $dir (@dirs)
{
foreach my $file (@files)
{
my $this_filespec = "$dir/$file";
if (-f $this_filespec)
{
$filespec = $this_filespec;
last;
}
}
last if (defined $filespec);
}
if (defined $filespec)
{
ENV_say( 1, "Using '$filespec'");
} else
{
ENV_error_exit( "File(s) '@files' not found in (@dirs)");
}

return $filespec;
}

return 1;

