#===============================================================
#
#   scasum.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================



































package scasum;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCASUM_init
SCASUM_run
);

use strict;
use integer;

use glo::env;
use glo::html;
use glo::spit;
use glo::slurp;
use gbsenv;
use gbshtml;
use gfl;
use scadef;
use scamet;
use scaglo;
use scafile;
use sca;
use audit;
use audithist;




sub SCASUM_init($);
sub SCASUM_run();

sub init_msg();
sub init_metrics();
sub update_metric_thresholds();

sub lvl_grand_total();
sub lvl_subsys_head($);
sub lvl_subsys_break();
sub lvl_component_head($);
sub lvl_component_break();
sub lvl_handle_file($);
sub print_totals($$);
sub create_ano_html_file();
sub ano_line_html($$);
sub ano_print_nav($$);
sub get_file_line($$);
sub create_log_html_file($$);
sub generate_levels_summary();
sub print_file_line();
sub generate_message_summary();
sub generate_metrics_summary($);
sub generate_suspicious_files_list();
sub generate_worse_files_list();
sub generate_special_warnings_summary();
sub write_lines_or_link($$$);
sub get_diff($$$);
sub color_text($$$);
sub color_value($$$);
sub color_value_f($$$);
sub delta_value($$);
sub clear_totals($);
sub add_totals($);
sub calc_levels_sum($);
sub calc_density($$$$);
sub ano_href($$$);
sub print_legend();
sub dot_print($$);
sub read_gmet_file();
sub read_hist_gmet_file();
sub store_warning($$$$$);
sub store_total($$$$);
sub extract_date_from_filename($);




my $GBS_DOC_PATH = ENV_getenv( 'GBS_SCRIPTS_PATH') . '/doc';
my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my $GBS_AUDIT = ENV_getenv( 'GBS_AUDIT');





my $RARR = '&'. 'rarr;';	    # right-arrow

my %MET_DEFS;






my @MET_DEFS_ORDER;

my %DIFFS = (

clean       => [ 0, 'OK: Clean',        'Clean'  ],
gone        => [ 1, 'OK: Gone',         'Gone'   ],
better      => [ 2, 'OK: Improved',     'Better' ],
same        => [ 3, 'Same',             'Same'   ],
worse       => [ 4, 'NOK: Worse',       'Worse'  ],
better9     => [ 5, 'NOK: Syntax',      'Fatal: Better' ],
same9       => [ 6, 'NOK: Syntax',      'Fatal: Same' ],
worse9      => [ 7, 'NOK: Syntax',      'Fatal: Worse' ],
);
my @DIFFS;
map { $DIFFS[$DIFFS{$_}->[0]] = $_ } keys %DIFFS;




my $PLUGIN_NICE_NAME;
my $SUM_FILE_SPEC;
my $SUM_FILE_DIR;
my $SUM_FILE_NAME;	    # gbs<job>_<target>_<audit>_YYMMDD-hhmmss_sum
my $SUM_FILE_TYPE;	    # .html
my $SUM_FILE_DATETIME;
my $RELEASE_TEXT;

my %SEVERITY_INDEX;

my @LEVEL_TEXTS;
my %LEVEL_INDEX;

my @FATAL_LEVELS;
my %FATAL_LEVELS;

my @FATAL_LEVEL_INDEXES;
my %FATAL_LEVEL_INDEXES;

my $NR_LEVELS;
my $LAST_LEVEL_I;		# $NR_LEVELS - 1;

my @EXTRA_LEVEL_TEXTS = (
"Nr.\nLines",
"Nr.\nFuncs",
"Nr.\nClass",
"Nr.\nconst",
);
my $NR_LINES_I = 0;
my $NR_FUNCS_I = 1;
my $NR_CLASS_I = 2;
my $NR_CONST_I = 3;
my $LAST_TOT_I = 3;
my $NR_EXTRA_LEVELS = @EXTRA_LEVEL_TEXTS;

my $HELP_FILES_PATH,    # undef if no msg help files
my %MSG_REFS;


my %SPECIAL_MSG_NRS;	    # Fatal and SPECWS





my %SPECWS;






my $TOTAL_SUBSYS_COUNT = 0;
my $TOTAL_COMP_COUNT = 0;
my $TOTAL_FILE_COUNT = 0;
my @SYSTEM_MSG_SUM_REFS;

my @SYSTEM_TOT_SUM_REFS;





my $CUR_SUBSYS = '';
my $SUBSYS_COUNT = 0;
my @SUBSYS_MSG_SUM_REFS;

my @SUBSYS_TOT_SUM_REFS;





my $CUR_COMPONENT = '';
my $CUR_SRC_DIR;
my $CUR_AUD_TARGET_DIR;
my $COMP_COUNT = 0;
my @COMP_MSG_SUM_REFS;

my @COMP_TOT_SUM_REFS;





my $CUR_FILE = '';
my $CUR_FILE_SPEC = '';
my $CUR_OS_FILE_SPEC = '';
my $FILE_COUNT = 0;
my $IS_NEW_FILE = 0;
my $FILE_LOG_FILE_SPEC;
my $FILE_ANO_FILE_SPEC;
my %FILE_MSG_COUNT;

my @FILE_MSG_SUM_REFS;

my @FILE_TOT_SUM_REFS;

my %FILE_MET_DATA;


my @FILE_MET_DATA_ORDER;

my %FILE_SPECIAL_MSGS;


my @ROW_REFS;           # HTML rows
my @SUM_REFS = (

[ \@FILE_MSG_SUM_REFS,   \@FILE_TOT_SUM_REFS ],	    # 0	= File
[ \@COMP_MSG_SUM_REFS,   \@COMP_TOT_SUM_REFS ],	    # 1 = Component
[ \@SUBSYS_MSG_SUM_REFS, \@SUBSYS_TOT_SUM_REFS ],	    # 2 = SubSys
[ \@SYSTEM_MSG_SUM_REFS, \@SYSTEM_TOT_SUM_REFS ],	    # 3 = System
);




my %TOTAL_MSG_COUNT;



my %TOTAL_MET_DATA;

my @TOTAL_MET_DATA_ORDER;

my @TOTAL_SUSP_FILE_REFS;


my %TOTAL_WORSE_FILES;


my %TOTAL_SPECIAL_MSGS;





my %CODE_FILES_REF;





sub SCASUM_init($)
{
(	$SUM_FILE_SPEC,
) = @_;

my ($plugin_name, $plugin_rel) = AUDIT_get_plugin_data( $GBS_AUDIT);
ENV_say( 1, "Audit $GBS_AUDIT [$plugin_name ($plugin_rel)] for Target $GBS_TARGET",
$SUM_FILE_SPEC);

%SPECWS = map { @{$_} } @{SCADEF_get_section_items( $GBS_AUDIT, 'MESSAGES', 'SPECIAL')};
$PLUGIN_NICE_NAME = SCADEF_get_section_items( $GBS_AUDIT, 'SETUP', 'NICE_NAME');

$RELEASE_TEXT = "$PLUGIN_NICE_NAME ($plugin_rel)";

SCAGLO_init( $plugin_name, $plugin_rel, $PLUGIN_NICE_NAME);

($SUM_FILE_DIR, $SUM_FILE_NAME, $SUM_FILE_TYPE) = ENV_split_spec_dnt( $SUM_FILE_SPEC);
$SUM_FILE_DATETIME = sprintf( "20%s-%s-%s %s:%s:%s", extract_date_from_filename( $SUM_FILE_NAME));




init_msg();





init_metrics();
}




sub init_msg()
{



my @msg_level_refs = @{SCADEF_get_section_items( $GBS_AUDIT, 'LEVELS', 'LEVEL')};

foreach my $ref (@msg_level_refs)
{
my ($level, $severity, $text, $severity_index) = @{$ref};

if ($text =~ s/_/\n/g)
{
$text .= " ($level)";
} else
{
$text .= "\n($level)";
}
$SEVERITY_INDEX{$level} = $severity_index;
my $level_index = scalar @LEVEL_TEXTS;
$LEVEL_INDEX{$level} = $level_index;
push @LEVEL_TEXTS, $text;
my $is_fatal_level = ($severity eq 'F') ? 1 : 0;
push @FATAL_LEVELS, $level
if ($severity eq 'F');
push @FATAL_LEVEL_INDEXES, $level_index
if ($severity eq 'F');
$FATAL_LEVELS{$level} = $is_fatal_level;
$FATAL_LEVEL_INDEXES{$level_index} = $is_fatal_level;
}

$NR_LEVELS = @LEVEL_TEXTS;
$LAST_LEVEL_I = $#LEVEL_TEXTS;




%MSG_REFS = map { $_->[0] => $_ } SCAGLO_read_msg_file();

{
my @all_msg_nrs;
my @fatal_msg_nrs;
foreach my $ref (values %MSG_REFS)
{
my ($msg_nr, $level) = @{$ref};
push @all_msg_nrs, $msg_nr;
push @fatal_msg_nrs, $msg_nr
if ($SEVERITY_INDEX{$level} == 2);
}
SCA_init( $GBS_AUDIT, $GBS_TARGET, \@all_msg_nrs, \@fatal_msg_nrs, [ keys %SPECWS ]);
%SPECIAL_MSG_NRS = map { $_ => 1 } (@fatal_msg_nrs,  keys %SPECWS);
}




$HELP_FILES_PATH = SCADEF_get_section_items( $GBS_AUDIT, 'SETUP', 'HELPFILES_PATH');
$HELP_FILES_PATH = ENV_perl_path_nospace( ENV_expand_envs( $HELP_FILES_PATH));





map { clear_totals( $_) } 0 .. 3;	#   $sum_level 0 - 3


}




sub init_metrics()
{
my @metric_names =  @{SCADEF_get_section_items( $GBS_AUDIT, 'METRICS', 'ENABLE')};
if (!@metric_names)
{
push @metric_names, 'TLN';	# 'Total preprocessed source lines'
push @metric_names, 'TPP';	# 'Total unpreprocessed source lines'
}






my %met_defs_sorted;

my $met_defs_sorted = 0;
foreach my $metric_name (@metric_names)
{

$MET_DEFS{$metric_name} = [ SCAMET_get_metric_data( $metric_name) ];


if (!$met_defs_sorted{$metric_name})	    # prevent duplicates (should have been done earlier)
{
push @MET_DEFS_ORDER, $metric_name;
$met_defs_sorted = 0;
$met_defs_sorted{$metric_name} = 1;
}
}




if (!$met_defs_sorted)
{

my @order_refs = ( [], [], [], []);		    # function, class, file, project
foreach my $key (@MET_DEFS_ORDER)
{
my $group = $MET_DEFS{$key}->[1];
push @{$order_refs[$group]}, $key;
}
@MET_DEFS_ORDER = map { @{$_} } @order_refs;

$met_defs_sorted = 1;
}




update_metric_thresholds();
}




sub update_metric_thresholds()
{
my @metric_threshold_refs = SCA_get_threshold_refs();


foreach my $ref (@metric_threshold_refs)
{
my ($metric_name, $min_value, $max_value, $msg_nr ) = @{$ref};	    # msg_nr not used yet

my $defs_ref = $MET_DEFS{$metric_name};


if ($min_value ne '')
{
if (defined $defs_ref->[2])			# $min_ref
{
my $old_value = $defs_ref->[2]->[0];	# $min_ref->$limit
$defs_ref->[2]->[0] = $min_value;	# $min_ref->$limit
ENV_say( 1, "METRIC $metric_name min limit changed from $old_value to $min_value");
} else
{
ENV_warn( "No min limit for metric $metric_name - Ignored");
}
}

if ($max_value ne '')
{
if (defined $defs_ref->[3])			# $max_ref
{
my $old_value = $defs_ref->[3]->[0];	# $max_ref->$limit
$defs_ref->[3]->[0] = $max_value;	# $max_ref->$limit
ENV_say( 1, "METRIC $metric_name max limit changed from $old_value to $max_value");
} else
{
ENV_warn( "No max limit for metric $metric_name - Ignored");
}
}

}
}




sub SCASUM_run()
{



my $title = "$PLUGIN_NICE_NAME Summary $SUM_FILE_NAME";

my @sum_lines;
push @sum_lines, GBSHTML_doc_start( $SUM_FILE_SPEC, $title, 0);

$title = "$PLUGIN_NICE_NAME Summary Report ($GBS_AUDIT)";
push @sum_lines,
GBSHTML_doc_top( [ $title, HTML_link( '.', HTML_bold( 'Log Directory')) ],
( [ System => $ENV{GBS_SYSTEM_NAME},	'' ],
[ Root   => $GBS_ROOT_PATH,		'' ],
[ Target => $ENV{GBS_TARGET},    	'' ],
[ Summary=> $SUM_FILE_NAME,		'' ],
[ Date   => $SUM_FILE_DATETIME,  	'' ],
));
my $section = 1;
push @sum_lines, $RELEASE_TEXT;




push @sum_lines, GBSHTML_section_head( "$PLUGIN_NICE_NAME Message Summary", $section++);
push @sum_lines, print_legend();
push @sum_lines, generate_levels_summary();




push @sum_lines, GBSHTML_section_head( "$PLUGIN_NICE_NAME Warning Message Summary",  $section++);
push @sum_lines, generate_message_summary();




push @sum_lines, GBSHTML_section_head( "$PLUGIN_NICE_NAME Metrics Summary",  $section++);
push @sum_lines, generate_metrics_summary( 'SUMMARY');




push @sum_lines, GBSHTML_section_head( "$PLUGIN_NICE_NAME Suspicious Files Summary",  $section++);
push @sum_lines, generate_suspicious_files_list();




push @sum_lines, GBSHTML_section_head( "$PLUGIN_NICE_NAME Decreased Quality Files Summary",  $section++);
push @sum_lines, generate_worse_files_list();




push @sum_lines, GBSHTML_section_head( "$PLUGIN_NICE_NAME Special Warnings Summary", -$section);  # last section!
push @sum_lines, generate_special_warnings_summary();




push @sum_lines, GBSHTML_doc_bottom( $title, $section);
push @sum_lines, GBSHTML_doc_end();


SPIT_file_nl( $SUM_FILE_SPEC, \@sum_lines);
}




sub generate_levels_summary()
{
my @lines;




my ($audit, undef, @subsys_refs) = GFL_read( $GBS_AUDIT);
foreach my $subsys_ref (@subsys_refs)
{



my ($subsys, undef, @comp_refs) = @{$subsys_ref};

lvl_subsys_head( $subsys);
foreach my $comp_ref (@comp_refs)
{



my ($component, undef, @files) = @{$comp_ref};

lvl_component_head( $component);
foreach my $in_file (@files)
{




lvl_handle_file( $in_file);
print_file_line();
$CUR_FILE = '';
$CUR_FILE_SPEC = '';
$CUR_OS_FILE_SPEC = '';
}
lvl_component_break();
$CUR_COMPONENT = '';
}
lvl_subsys_break();
$CUR_SUBSYS = '';
}




lvl_grand_total();
my @formats = qw( -TL -TRN);			    # Name, Legend
map { push @formats, qw( -TR ) } 1 .. $NR_LEVELS;	    # msg-levels
push @formats, qw( -TR);				    # total Msgs
map { push @formats, qw( -TR ) } 1 .. $NR_EXTRA_LEVELS; # Total MSG,


push @lines, HTML_table( [ @formats ],
undef,
[ @ROW_REFS ],
class => 'border1-padding1');
@ROW_REFS = ();	# cleanup

return @lines;
}




sub print_file_line()
{
my $out_base = "$CUR_AUD_TARGET_DIR/$CUR_FILE";
my $garg_file_spec = "$out_base.garg";		    # In

$FILE_ANO_FILE_SPEC = "$out_base.html";
$FILE_LOG_FILE_SPEC = "$out_base.log.html";

SCAFILE_init( $CUR_SUBSYS, $CUR_COMPONENT, $CUR_FILE);  # returns $gout_file_spec




$IS_NEW_FILE = read_hist_gmet_file();
if (read_gmet_file())
{



create_ano_html_file();
create_log_html_file( $garg_file_spec, 1);




print_totals( 0, $CUR_FILE);

add_totals( 0);	# File




{
my $diff_count = 0;
foreach my $ref (@FILE_MSG_SUM_REFS)
{
my (undef, undef, $t_cur_msg_sum, $t_hist_msg_sum) = @{$ref};
my $diff = $t_cur_msg_sum - $t_hist_msg_sum;
$diff_count += $diff
if ($diff > 0);
}
my $fatal_count = 0;
map { $fatal_count += $FILE_MSG_SUM_REFS[$_]->[2] } @FATAL_LEVEL_INDEXES;  # $t_cur_msg_sum

if ($diff_count > 0)
{
my $key = "$CUR_SUBSYS:$CUR_COMPONENT:$CUR_FILE";
$TOTAL_WORSE_FILES{$key} = [ $IS_NEW_FILE, $diff_count, $fatal_count, 0 ];

}
}






while (my ($msg_nr, $ref) = each %FILE_MSG_COUNT)
{


my $t_cur_msg_count = $ref->[2];
if ($t_cur_msg_count > 0)				# hist may also contain errors, so select only current
{
$TOTAL_MSG_COUNT{$msg_nr} = [ 0, 0 ]		# $msg_count, $msg_file_count
if (!defined $TOTAL_MSG_COUNT{$msg_nr});
my $msg_ref = $TOTAL_MSG_COUNT{$msg_nr};
$msg_ref->[0] += $t_cur_msg_count;
$msg_ref->[1]++;
push @{$msg_ref}, [ $CUR_SUBSYS, $CUR_COMPONENT, $CUR_FILE, $t_cur_msg_count ];
}
}




%FILE_MSG_COUNT = ();
%FILE_MET_DATA = ();
@FILE_MET_DATA_ORDER = ();
%FILE_SPECIAL_MSGS = ();
} else
{
$FILE_ANO_FILE_SPEC = $CUR_FILE_SPEC
if (!-e $FILE_ANO_FILE_SPEC);

create_log_html_file( $garg_file_spec, 0);

my $line1 = HTML_txt( '   ') .
HTML_link( {target => '_blank'}, $FILE_ANO_FILE_SPEC, HTML_txt( $CUR_FILE));
my $line2 = HTML_txt( '   ') .
HTML_link( {target => '_blank'}, $FILE_LOG_FILE_SPEC, HTML_txt( "$CUR_FILE.log"));

push @ROW_REFS, [ HTML_div( {align => 'left'}, $line1 . '<br>' . $line2),
HTML_div( {align => 'left', style => "color:red"}, 'No Analysis Data') ];

my @lines;
push @lines, "$CUR_COMPONENT:$CUR_FILE: No Analysis Data file found ($CUR_FILE.gmet)";
push @lines, "$CUR_COMPONENT:$CUR_FILE: No Annotated file found ($CUR_FILE.html)"
if (!-e $FILE_ANO_FILE_SPEC);
push @lines, "$CUR_COMPONENT:$CUR_FILE: No Arguments Log file found ($CUR_FILE.garg)"
if (!-e $garg_file_spec);
ENV_warn( @lines);
}
}




sub lvl_grand_total()
{
lvl_component_break()
if ($CUR_COMPONENT);
lvl_subsys_break()
if ($CUR_SUBSYS);

if ($SUBSYS_COUNT > 1)
{
lvl_subsys_head( '');
push @ROW_REFS, [ '' ];
print_totals( 3, $SUBSYS_COUNT);
}
}




sub lvl_subsys_head($)
{
my ($subsys) = @_;

my @row;

$CUR_SUBSYS = $subsys;

if ($CUR_SUBSYS eq '')
{
push @row, '';
} else
{
my $hist_date = AUDITHIST_select_subsys( $subsys);
push @row, HTML_div( {align => 'left'}, HTML_bold_txt( "SubSystem:\n$CUR_SUBSYS"),
HTML_txt( "\nHist-date: $hist_date"));
}
push @row, map { HTML_bold_txt( $_) } ( "Legend ",
@LEVEL_TEXTS,
"Total\nMsgs",
@EXTRA_LEVEL_TEXTS,
);
push @ROW_REFS, [ @row ];

clear_totals( 2);	    # SubSys
$COMP_COUNT = 0;
}




sub lvl_subsys_break()
{
$SUBSYS_COUNT++;
$TOTAL_SUBSYS_COUNT++;
if ($COMP_COUNT > 1 || $SUBSYS_COUNT > 1)
{
print_totals( 2, $COMP_COUNT);
}
add_totals( 2);	# SubSys
}





sub lvl_component_head($)
{
my ($component) = @_;

$CUR_COMPONENT = $component;
$CUR_SRC_DIR = "$GBS_ROOT_PATH/dev/$CUR_SUBSYS/comp/$CUR_COMPONENT/src";
$CUR_AUD_TARGET_DIR = "$GBS_ROOT_PATH/dev/$CUR_SUBSYS/comp/$CUR_COMPONENT/aud/$GBS_AUDIT/$GBS_TARGET";

my @row = ( [ {align => 'left'}, HTML_bold_txt( "  $CUR_COMPONENT") ]);
push @ROW_REFS, [ @row ];

clear_totals( 1);	    # Component
$FILE_COUNT = 0;
}





sub lvl_component_break()
{
$COMP_COUNT++;
$TOTAL_COMP_COUNT++;
if ($FILE_COUNT > 1 || $COMP_COUNT > 1 || $SUBSYS_COUNT > 1)
{
print_totals( 1, $FILE_COUNT);
}
add_totals( 1);	# Component
}




sub lvl_handle_file($)
{
my ($new_file,
) = @_;

$CUR_FILE = $new_file;
$CUR_FILE_SPEC = "$CUR_SRC_DIR/$CUR_FILE";
$CUR_OS_FILE_SPEC = ENV_os_path( $CUR_FILE_SPEC);
$FILE_COUNT++;
$TOTAL_FILE_COUNT++;
clear_totals( 0);	    # file
}




sub print_totals($$)
{
my ($sum_level,             # 0 = line, 1 = comp, 2 = subsys, 3 = system
$info,                  # name, nr_files, nr_comp or nr_subsys
) = @_;
my @line;

my ($msg_sum_refs_ref, $tot_sum_refs_ref) = @{$SUM_REFS[$sum_level]};

my ($s_cur_nr_lines,
$s_hist_nr_lines,
$t_cur_nr_lines,
$t_hist_nr_lines)= @{$tot_sum_refs_ref->[$NR_LINES_I]};

my $line1;
my $line2;
my $line3;




if ($sum_level == 0)
{
my $name = $info;
$line1 = HTML_txt( '   ') .
HTML_link( {target => '_blank'}, $FILE_ANO_FILE_SPEC,
HTML_txt( $name));
$line2 = HTML_txt( '   ') .
HTML_link( {target => '_blank'}, $FILE_LOG_FILE_SPEC,
HTML_txt( "$name.log"));
$line3 = ($IS_NEW_FILE) ? HTML_txt( '   (new)') : '';
} elsif ($sum_level == 1)
{
$line1 = HTML_bold_txt( "  $CUR_COMPONENT");
$line2 = HTML_bold_txt( "  Component Totals:");
$line3 = HTML_txt( "  $info File(s)");
} elsif ($sum_level == 2)
{
$line1 = HTML_bold_txt( " $CUR_SUBSYS");
$line2 = HTML_bold_txt( " Subsys Totals:");
$line3 = HTML_txt( "  $info Component(s)");
} else #($sum_level == 3)
{
$line1 = '';
$line2 = HTML_bold( "System Totals:");
$line3 = HTML_txt( "  $info SubSystem(s)");
}
push @line, HTML_div( $line1) . HTML_div( $line2) . HTML_div( $line3);




push @line, HTML_div( HTML_bold( "Source$RARR")) .
HTML_div( HTML_bold( "Total$RARR")) .
HTML_div( HTML_bold( "Density$RARR"));




my $s_is_fatal_level = 0;
my $t_is_fatal_level = 0;
for (my $level_index = 0; $level_index < $NR_LEVELS; $level_index++)
{
my ($s_cur_msg_sum, $s_hist_msg_sum,
$t_cur_msg_sum, $t_hist_msg_sum) = @{$msg_sum_refs_ref->[$level_index]};
my $is_fatal_level = $FATAL_LEVEL_INDEXES{$level_index};
$line1 = color_value( $s_cur_msg_sum, $s_hist_msg_sum, $is_fatal_level);    # Source
$line2 = color_value( $t_cur_msg_sum, $t_hist_msg_sum, $is_fatal_level);    # Total
my $density      = calc_density( $t_cur_msg_sum,  $t_cur_nr_lines,  $s_cur_msg_sum,  $s_cur_nr_lines);
my $hist_density = calc_density( $t_hist_msg_sum, $t_hist_nr_lines, $s_hist_msg_sum, $s_hist_nr_lines);
$line3 = color_value_f( $density, $hist_density, $is_fatal_level);	    # Density
push @line, "$line1$line2$line3";
$s_is_fatal_level = 1
if ($is_fatal_level && $s_cur_msg_sum > 0);
$t_is_fatal_level = 1
if ($is_fatal_level && $t_cur_msg_sum > 0);
}




my ($s_cur_msg_count, $s_hist_msg_count,
$t_cur_msg_count, $t_hist_msg_count) = calc_levels_sum( $msg_sum_refs_ref );
my $t_cur_density =  calc_density( $t_cur_msg_count,  $t_cur_nr_lines,  $s_cur_msg_count,  $s_cur_nr_lines);
my $t_hist_density = calc_density( $t_hist_msg_count, $t_hist_nr_lines, $s_hist_msg_count, $s_hist_nr_lines);

$line1 = color_value( $s_cur_msg_count, $s_hist_msg_count, $s_is_fatal_level);	# Source Total
$line2 = color_value( $t_cur_msg_count, $t_hist_msg_count, $t_is_fatal_level);	# Total Total
$line3 = color_value_f( $t_cur_density, $t_hist_density,   $t_is_fatal_level);	# Density (Average?)
push @line, "$line1$line2$line3";




foreach my $ref ( @{$tot_sum_refs_ref})
{
my ($s_cur_tot_sum, $s_hist_tot_sum, $t_cur_tot_sum, $t_hist_tot_sum) = @{$ref};
$line1 = delta_value( $s_cur_tot_sum, $s_hist_tot_sum);			    # Source
$line2 = delta_value( $t_cur_tot_sum, $t_hist_tot_sum);			    # Total
push @line, "$line1$line2";
}

push @ROW_REFS, [ @line ];
}




sub create_ano_html_file()
{



my @counts = map { 0 } (0..$NR_LEVELS);




my @warning_refs = SCAFILE_gerr_read();
my $nr_warnings = @warning_refs;




my @src_line_refs = map { [ $_ , undef ] } ( '', SLURP_file( $CUR_FILE_SPEC), '<EOF>');  # $line_nr indexes from 1

my $nr_lines = @src_line_refs;
for (my $w = 0;  $w < $nr_warnings; $w++)
{
my $ref = $warning_refs[$w];


my $msg_nr = $ref->[6];
if (SCA_msg_is_enabled( $msg_nr) || exists $SPECIAL_MSG_NRS{$msg_nr})
{
my $src_line_nr = $ref->[0];
$src_line_nr = $#src_line_refs
if ($src_line_nr >= $nr_lines);
push @{$src_line_refs[$src_line_nr]->[1]}, $w;

my $msg_level = $ref->[4];
$counts[$LEVEL_INDEX{$msg_level}]++;
}
}

my @ano_lines;




my $heading = "$CUR_FILE - $CUR_COMPONENT - $CUR_SUBSYS - $GBS_ROOT_PATH";
push @ano_lines, GBSHTML_doc_start( $FILE_ANO_FILE_SPEC, $heading, 0);




my $title = "$CUR_FILE [ $GBS_AUDIT - $GBS_TARGET ]";
push @ano_lines, GBSHTML_doc_top( [ $title, HTML_link( $CUR_FILE_SPEC, HTML_bold('Actual Source File'))],
( [ $heading, ''] ));
my $section = 1;




push @ano_lines, "Nr. messages: $nr_warnings";




push @ano_lines, GBSHTML_section_head( "Annotated source: $CUR_FILE_SPEC", $section++);




push @ano_lines, HTML_anchor( 'FIRST', '');
push @ano_lines, ano_print_nav( \@warning_refs, -1);
push @ano_lines, HTML_pre_s();
for (my $line_nr = 0; $line_nr < $nr_lines; $line_nr++)
{
my ($line, $warn_indexes_ref) = @{$src_line_refs[$line_nr]};
if (defined $warn_indexes_ref)
{



push @ano_lines, HTML_pre_e();
push @ano_lines, HTML_anchor( "ERR_LINE_$line_nr", '');
push @ano_lines, HTML_pre_s();
push @ano_lines, ano_line_html( $line_nr, $line);
foreach my $w (@{$warn_indexes_ref})
{

my ($src_line_nr, $col_nr, $file_spec, $line_nr, $msg_level, $level_text,
$msg_nr, $msg_text, $help_file_url, @extra_src_lines) = @{$warning_refs[$w]};




if ($file_spec ne '-')
{
push @ano_lines, HTML_span( { style => "color:green" }, $file_spec);
push @ano_lines, ano_line_html( $line_nr, get_file_line( $file_spec, $line_nr));
}




my $color = ($FATAL_LEVELS{$msg_level}) ? 'red' : 'blue';
if ($col_nr >= 0)
{
my $lead = ' ' x $col_nr;
push @ano_lines, HTML_span( { style => "color:$color" }, "    $lead....^");
}




push @ano_lines, map { sprintf( "%5s:  %s", '', HTML_pre_txt( $_)); } @extra_src_lines;





push @ano_lines, HTML_pre_e();



my $html_text = HTML_txt( $msg_text);
if ($help_file_url)
{
$html_text = HTML_link( { target => '_blank', style => "color:$color"} ,
"$HELP_FILES_PATH/$help_file_url", $html_text);
}
$html_text =  "** $html_text" . HTML_br() . "\n" .
"** Level: $level_text ($msg_level) -- Msg nr: $msg_nr";

push @ano_lines, HTML_div( { style => "color:$color" }, HTML_bold( $html_text));




push @ano_lines, ano_print_nav( \@warning_refs, $w);




push @ano_lines, HTML_pre_s();
}
} else
{



push @ano_lines, ano_line_html( $line_nr, $line);
}
}
push @ano_lines, HTML_pre_e();
push @ano_lines, HTML_anchor( 'END', '');
push @ano_lines, ano_print_nav(  \@warning_refs, $nr_warnings);




push @ano_lines, GBSHTML_section_head( "Metrics Summary: $CUR_FILE_SPEC ($nr_lines Lines)", -$section);
push @ano_lines, HTML_br();

no integer;
my @row_refs;
my $t_count = 0;
for (my $level_index = 0; $level_index < $NR_LEVELS; $level_index++)
{
my $count = $counts[$level_index];
my $level = $LEVEL_TEXTS[$level_index];
$level =~ s/\\n/ /;
my $density = ($nr_lines == 0) ? 0.0 : ($count * 1.0) / $nr_lines;
push @row_refs, [ $level, sprintf( "%1.4f", $density), sprintf( "%5d", $count) ];
$t_count += $count;
}
my $t_density = ($nr_lines == 0) ? 0.0 : ($t_count * 1.0) / $nr_lines;
push @row_refs, [ '------ ', ' ------- ', ' ------------- ' ];
push @row_refs, [ 'Totals',  sprintf( "%1.4f", $t_density), sprintf( "%5d", $t_count) ];
use integer;

my @formats = qw( -L -C -C );
my @headings = ( 'Error Level', 'Density', 'No. of Errors');
push @ano_lines, HTML_table( \@formats, \@headings,
\@row_refs, class => 'border1');




push @ano_lines, GBSHTML_doc_bottom( $title, $section);
push @ano_lines, GBSHTML_doc_end();


SPIT_file_nl( $FILE_ANO_FILE_SPEC, \@ano_lines);

}




sub ano_line_html($$)
{
my ($line_nr,
$line,
) = @_;
my @html_lines;

push @html_lines, sprintf( "%5d:  %s", $line_nr, HTML_pre_txt( ENV_detab( $line)));

return @html_lines;
}




sub ano_print_nav($$)
{
my ($warning_refs_ref,
$warn_index,
) = @_;
my $out_line;

my $last_warn_index = $#{$warning_refs_ref};

if ($warn_index >= 0)
{
$out_line = HTML_link( "#FIRST", HTML_txt( "<top>"));
} else
{
$out_line = HTML_txt( '<top>');
}

if ($warn_index >= 1)
{
my $prev_nr = $warning_refs_ref->[$warn_index-1]->[0];
$out_line .= HTML_link( "#ERR_LINE_$prev_nr", HTML_txt( "<prev>"));
} else
{
$out_line .= HTML_txt( '<prev>');
}

if ($warn_index < $last_warn_index)
{
my $next_nr = $warning_refs_ref->[$warn_index+1]->[0];
$out_line .= HTML_link( "#ERR_LINE_$next_nr", HTML_txt( "<next>"));
} else
{
$out_line .= HTML_txt( '<next>');
}

if ($warn_index <= $last_warn_index)
{
$out_line .= HTML_link( "#END", HTML_txt( "<bottom>"));
} else
{
$out_line .= HTML_txt( '<bottom>');
}

return (HTML_bold( $out_line) . HTML_br());
}




sub get_file_line($$)
{
my ($file_spec,
$line_nr,
) = @_;


my $lines_ref = $CODE_FILES_REF{$file_spec};	#   $file_spec=> $lines_ref;
if (!defined $lines_ref)
{

$lines_ref = $CODE_FILES_REF{$file_spec} = [ SLURP_file( $file_spec) ];
} else
{

}

return $lines_ref->[ $line_nr - 1 ];
}




sub create_log_html_file($$)
{
my ($garg_file_spec,
$data_available,
) = @_;

my @log_lines;




my $heading = "$CUR_FILE.log - $CUR_COMPONENT - $CUR_SUBSYS - $GBS_ROOT_PATH";
push @log_lines, GBSHTML_doc_start( $FILE_LOG_FILE_SPEC, $heading, 0);




my $title = "$CUR_FILE.log [ $GBS_AUDIT - $GBS_TARGET ]";
push @log_lines, GBSHTML_doc_top( [ $title, HTML_link(  $FILE_ANO_FILE_SPEC, HTML_bold( 'Annotated Source')) ],
( [ $heading, '', ''] ));
my $section = 1;




push @log_lines, GBSHTML_section_head( 'Warnings Summary', $section++);

if ($data_available)
{
my @msg_nrs = sort( { $a <=> $b } keys( %FILE_MSG_COUNT));
if (@msg_nrs)
{
my @row_refs;
foreach my $msg_nr (@msg_nrs)
{

my $msg_ref = $MSG_REFS{$msg_nr};
ENV_fatal( "Unknown message '$msg_nr'")
if (!defined $msg_ref);
my (undef, $level, $level_text, $msg_text, $help_file_url) = @{$msg_ref};
my $s_cur_msg_count = $FILE_MSG_COUNT{$msg_nr}->[0];
my $s_hist_msg_count = $FILE_MSG_COUNT{$msg_nr}->[1];
my $t_cur_msg_count = $FILE_MSG_COUNT{$msg_nr}->[2];
my $t_hist_msg_count = $FILE_MSG_COUNT{$msg_nr}->[3];

my $is_fatal_level = $FATAL_LEVELS{$level};
my $s_diff_txt = color_text( $s_cur_msg_count, $s_hist_msg_count, $is_fatal_level);
my $t_diff_txt = color_text( $t_cur_msg_count, $t_hist_msg_count, $is_fatal_level);

my $msg_nr_txt = $msg_nr;
if ($help_file_url)
{
$msg_nr_txt = HTML_link( {target => '_blank'},
"$HELP_FILES_PATH/$help_file_url",
$msg_nr);
}
my $warning_txt;
if ($is_fatal_level)
{
$warning_txt = 'Error';
} else
{
$warning_txt = 'Warning';
}
$level_text = '****' if (!defined $level_text);
$msg_text = '****' if (!defined $msg_text);
push @row_refs, [ $warning_txt, $msg_nr_txt, $level, $level_text,
$s_hist_msg_count, $s_cur_msg_count, $s_diff_txt,
$t_hist_msg_count, $t_cur_msg_count, $t_diff_txt,
$msg_text ];
}
my @formats = qw( - -C -C - -R -R -N -R -R -N -N);
my @headings = ( 'Severity', 'MsgNr', 'Level', 'Level-Txt',
'SRC<BR>Base', 'SRC<BR>Now', 'SRC<BR>Diff',
'TOT<BR>Base', 'TOT<BR>Now', 'TOT<BR>Diff',
'Message Text');
push @log_lines, HTML_table( \@formats, \@headings,
\@row_refs, class => 'border1');
} else
{
push @log_lines, HTML_p( "No Warnings for this file");
}
} else
{
push @log_lines, HTML_p( "No Data Available");
}




push @log_lines, GBSHTML_section_head( 'Metrics Summary:', $section++);
if ($data_available)
{
push @log_lines, generate_metrics_summary( 'FILE');
} else
{
push @log_lines, HTML_p( "No Data Available");
}




push @log_lines, GBSHTML_section_head( 'Special Messages:', $section++);
if ($data_available)
{
if (%FILE_SPECIAL_MSGS)
{
foreach my $text (sort( keys( %FILE_SPECIAL_MSGS)))
{
push @log_lines, HTML_h( 3, "$text:");
foreach my $item (keys( %{$FILE_SPECIAL_MSGS{$text}}))
{
if ($item eq '')
{
push @log_lines, HTML_italic_txt( "  <Undefined>\n")
} else
{
push @log_lines, HTML_txt( "  $item\n");
}
}
}
} else
{
push @log_lines, HTML_p( "No Special Messages for this file");
}
} else
{
push @log_lines, HTML_p( "No Data Available");
}




push @log_lines, GBSHTML_section_head( 'Special Messages:', $section++);
my $sca_flags_file_spec = SCA_get_filespec();	# undef if does not exist
if (defined $sca_flags_file_spec)
{
push @log_lines, HTML_link( {target => '_blank'}, $sca_flags_file_spec, HTML_txt( $sca_flags_file_spec));
} else
{
push @log_lines, HTML_p( "No SCA Flags file for this file");
}




{
push @log_lines, GBSHTML_section_head( 'Master Settings file:', -$section);
my @lines = SCAFILE_garg_read();
if (@lines)
{
push @log_lines, '<PRE>';
foreach my $line (@lines)
{
if ($line =~ /\s*(-via\s+)(.*)$/ ||	    # QAC/QAC++	    # TBR via SCADEF
$line =~ /\s*()([^-].*\.lnt)$/)	    # PC-Lint	    # TBR via SCADEF
{
my $via = $1;			    # '-via ' or ''
my $via_file_spec = $2;

push @log_lines, $via .
HTML_link( {target => '_blank'}, $via_file_spec, HTML_txt( $via_file_spec));
} else
{
push @log_lines, $line;
}
}
push @log_lines, '</PRE>';
} else
{
push @log_lines, HTML_txt( 'No data');
}

}




push @log_lines, GBSHTML_doc_bottom( $title, $section);
push @log_lines, GBSHTML_doc_end();


SPIT_file_nl( $FILE_LOG_FILE_SPEC, \@log_lines);
}




sub generate_message_summary()
{
my @lines;

my @row_refs;
my $total_count = 0;
foreach my $msg_nr (sort( { $a <=> $b } keys %TOTAL_MSG_COUNT))
{



my $msg_ref = $TOTAL_MSG_COUNT{$msg_nr};



my ($msg_count, $msg_file_count, @file_refs) = @{$msg_ref};


my (undef, $level, $level_text, $msg_text, $help_file_url) = @{$MSG_REFS{$msg_nr}};
my $msg_link_or_txt = $msg_nr;
if (defined $help_file_url)
{
$msg_link_or_txt = HTML_link( {target => '_blank'},
"$HELP_FILES_PATH/$help_file_url",
$msg_nr);
}
$level = HTML_div( {class => 'worse9'}, $level)
if ($FATAL_LEVELS{$level});
push @row_refs, [ $msg_link_or_txt, $level, $level_text, $msg_count, $msg_file_count, $msg_text ];




my $nr_rows = int ((scalar @file_refs + 1) / 2);
while (@file_refs)
{
my @line = ( '', undef, undef, undef, undef );
for (my $i = 0; $i < 2 && @file_refs; $i++)
{
my $file_ref = shift @file_refs;
my ($subsys, $comp, $file, $count) = @{$file_ref};
my $href = ano_href( $subsys, $comp, $file);
push @line, "$href ($count)";
}
push @line, '' if ($nr_rows > 1 && scalar @line < 6);
push @row_refs, [ @line ];
}
$total_count += $msg_count;
}

push @row_refs, [ HTML_bold( 'Total:'), undef, undef, $total_count, '' ] ;

my @formats = qw( -BR -BC -BL -BR -BR -BN -BN -BN );
my @headings = (
'MsgNr',
'Level',
'Level-Txt',
'Count',
'Nr.<BR>Files',
'Message<BR>Affected Files'
);
my @table_lines = HTML_table( \@formats,\@headings,
\@row_refs, class => 'border1');

push @lines, write_lines_or_link( 'Warning Message Summary', 'ms', \@table_lines);

return @lines;
}




sub generate_metrics_summary($)
{
my ($s_type,         # 'SUMMARY' or 'FILE'
) = @_;
my @lines;




my $nr_files;
my $nr_class;
my $nr_funcs;
my $header_line;
my $met_data_ref;
if ($s_type eq 'SUMMARY')
{
$nr_files = $TOTAL_FILE_COUNT;
$nr_class = $SYSTEM_TOT_SUM_REFS[$NR_CLASS_I]->[2];   # $t_cur_tot_sum
$nr_funcs = $SYSTEM_TOT_SUM_REFS[$NR_FUNCS_I]->[2];   # $t_cur_tot_sum
$header_line = "SubSystems: $TOTAL_SUBSYS_COUNT, Components: $TOTAL_COMP_COUNT, Files: $nr_files";
$met_data_ref = \%TOTAL_MET_DATA;
} else
{
$nr_files = 1;
$nr_class = $FILE_TOT_SUM_REFS[$NR_CLASS_I]->[2];   # $t_cur_tot_sum
$nr_funcs = $FILE_TOT_SUM_REFS[$NR_FUNCS_I]->[2];   # $t_cur_tot_sum
$header_line = '';
$met_data_ref = \%FILE_MET_DATA;
}
$header_line .= ", Functions: $nr_funcs"
if ($nr_funcs >= 0);
$header_line .= ", Classes: $nr_class"
if ($nr_class >= 0);
$header_line =~ s/^, //;		    # remove possible leading ', '
push @lines, HTML_p( HTML_bold( $header_line));

$nr_funcs = -1 if ($nr_funcs == 0);
$nr_files = -1 if ($nr_files == 0);
$nr_class = -1 if ($nr_class == 0);

my @violation_refs;

my @row_refs;
my $prev_group = 0;
no integer;
foreach my $metric_name (@MET_DEFS_ORDER)
{
my ($title, $group, $min_ref, $max_ref, $do_tot, $do_avg) = @{$MET_DEFS{$metric_name}};
if ($group != $prev_group)
{



my $metric_group_name = SCAMET_group2txt( $group);
push @row_refs, [ map { HTML_bold( $_) } ('Key  ', "$metric_group_name Metrics", 'Min', 'Max', 'Total', 'Average') ];
$prev_group = $group;
}




my $metric_name_href = HTML_link( "$GBS_DOC_PATH/metrics.html#ST$metric_name", $metric_name);

if (exists $met_data_ref->{$metric_name})
{
my ($met_min, $met_max, $met_tot) = @{$met_data_ref->{$metric_name}};


my @sev_classes = qw(same worse worse9);

my $html_min = '';
if (defined $min_ref)
{
my ($min_min, $min_sev) = @{$min_ref};

my @min_div = (title => "< $min_min");
if ($met_min < $min_min)
{
push @min_div, class => $sev_classes[$min_sev];
push @violation_refs, [ $metric_name, "$met_min < $min_min" ];
} else
{
push @min_div, class => 'clean';
}
$html_min = HTML_div( {@min_div}, dot_print( $met_min, 3));
}

my $html_max = '';
if (defined $max_ref)
{
my ($max_max, $max_sev) = @{$max_ref};
my @max_div = (title => "> $max_max");
if ($met_max > $max_max)
{
push @max_div, class => $sev_classes[$max_sev];
push @violation_refs, [ $metric_name, "$met_max > $max_max" ];
} else
{
push @max_div, class => 'clean';
}
$html_max = HTML_div( {@max_div}, dot_print( $met_max, 3));
}

my $html_tot = '';
if ($do_tot)
{
$html_tot = dot_print( $met_tot, 2);
}

my $html_avg = '';
if ($do_avg)
{
my $avg;
my $name;
my $average = '';
if ($group == 1)        # function based
{
$avg = $met_tot / $nr_funcs;
$name = 'func';
} elsif ($group == 2)   # class based
{
$avg = $met_tot / $nr_class;
$name = 'class';
} elsif ($group == 3)   # file based
{
$avg = $met_tot / $nr_files;
$name = 'file';
} else                  # unknown
{
;
}
if (defined $name)
{
$average = sprintf "%8.2f per $name", $avg;
}
$html_avg = $average;
if (defined $max_ref)
{
my ($max_max, $max_sev) = @{$max_ref};
my @max_div = (title => "> $max_max");
if ($avg > $max_max)
{
push @max_div, class => $sev_classes[$max_sev];
} else
{
push @max_div, class => 'clean';
}
$html_avg = HTML_div( {@max_div}, $average);
}
}
push @row_refs, [ $metric_name_href, $title, $html_min, $html_max, $html_tot, $html_avg ];
} else
{
my $html_no_data = HTML_div( {align=>'left'}, 'No data');
push @row_refs, [ $metric_name_href, $title, $html_no_data, undef, '', '' ];
}
}
use integer;

if (@violation_refs && $s_type eq 'FILE')
{
my $file = "$CUR_SUBSYS:$CUR_COMPONENT:$CUR_FILE";
push @TOTAL_SUSP_FILE_REFS, [ $file, @violation_refs ];
}

my @formats = qw( -L -L -R -R -R -R);
push @lines, HTML_table( \@formats, undef, \@row_refs, class => 'border1');

return @lines;
}




sub generate_suspicious_files_list()
{
my @lines;

my $nr_susp_files = @TOTAL_SUSP_FILE_REFS;
if ($nr_susp_files > 0)
{
my @row_refs;
foreach my $ref (@TOTAL_SUSP_FILE_REFS)
{
my ($file_id , @violation_refs) = @{$ref};
my ($subsys, $comp, $file) = split ':', $file_id;
my $file_href = ano_href( $subsys, $comp, "$file.log");
my $nr_rows = int ((scalar @violation_refs + 3) / 4);

my @line = ( $file_href );
while (@violation_refs)
{
for ( my $i = 0; $i < 4 && @violation_refs; $i++)
{

my ($metric_name, $violation_text) = @{shift @violation_refs};
my $title = $MET_DEFS{$metric_name}->[0];
push @line, HTML_span({title => $title},
HTML_link( "$GBS_DOC_PATH/metrics.html#ST$metric_name", $metric_name) .
HTML_txt( ": $violation_text"));
}
push @line, '' if (scalar @line < 5);
push @row_refs, [@line ];
@line = (undef);
}
}
my @formats = qw( -T -N -N -N -N );
my @headings = ( 'File', 'Metrics Violations');
my @table_lines = HTML_table( \@formats, \@headings, \@row_refs, class => 'border1');
push @lines, write_lines_or_link( 'Suspicious Files Summary', 'sf', \@table_lines);

ENV_say( 1, "$nr_susp_files Suspicious file(s) found");
} else
{
my $line = "No Suspicious Files found.";
push @lines, HTML_p( HTML_bold( $line));
ENV_say( 1, $line);
}

return @lines;
}




sub generate_worse_files_list()
{
my @lines;

if (%TOTAL_WORSE_FILES)
{
my @row_refs;
foreach my $key (sort keys %TOTAL_WORSE_FILES)
{
my @line;
my ($subsys, $comp, $file) = split ':', $key;
my ($is_new_file, $total_count, $fatal_count, $density_is_same) = @{$TOTAL_WORSE_FILES{$key}};

my $new = ($is_new_file) ? 'new' : '';
my $pfatal_count = ($fatal_count > 0) ? " $fatal_count" : '';

my $href = ano_href( $subsys, $comp, $file);
push @row_refs, [ $href, "+$total_count", $new, $pfatal_count ];

ENV_say( 1, "**> $key: +$total_count $new Decreased Quality$pfatal_count");
}
push @lines, HTML_table( [ qw( - -R -C -R ) ],
[ qw( File Warnings New Errors ) ],
\@row_refs, class => 'border1');
} else
{
my $line = "All $TOTAL_FILE_COUNT file(s) Same or Increased Quality";
push @lines, HTML_p( $line);
ENV_say( 1, $line);
}

return @lines;
}




sub generate_special_warnings_summary()
{
my @lines;

if (%TOTAL_SPECIAL_MSGS)
{
my @row_refs;
foreach my $text (sort( keys( %TOTAL_SPECIAL_MSGS)))
{
foreach my $item (sort( keys( %{$TOTAL_SPECIAL_MSGS{$text}})))
{
my $count = $TOTAL_SPECIAL_MSGS{$text}->{$item};
my $html_item = $item;
if ($item eq '')
{
$item = '<Undefined>';
$html_item = HTML_italic_txt( $item);
}
ENV_say( 1, "****> $text: $item ($count)");
push @row_refs, [ $text, $html_item, $count ];
}
}
push @lines, HTML_table( [ qw( - - -R) ],
[ qw( Warning/Error Item Occurs) ],
\@row_refs, class => 'border1');
} else
{
push @lines, HTML_p( HTML_bold( 'No Special Messages'));
}

return @lines;
}




sub write_lines_or_link($$$)
{
my ($heading,
$type,		    # without the dot! e.g.: 'ms', 'sf'
$lines_ref,
) = @_;
my @lines;

if (@{$lines_ref} > 100)

{
my $file_name = "$SUM_FILE_NAME.$type";
my $file = "$file_name$SUM_FILE_TYPE";		    # .html
my $file_spec = "$SUM_FILE_DIR/$file";


{
my @file_lines;
push @file_lines, GBSHTML_doc_start( $file_spec,
"$PLUGIN_NICE_NAME $heading $file", 0);

my $title = "$PLUGIN_NICE_NAME $heading";
push @file_lines, GBSHTML_doc_top( [ $title, HTML_link( $SUM_FILE_SPEC, HTML_bold('Summary_file'))] , ());
push @file_lines, HTML_br();
push @file_lines, @{$lines_ref};
push @file_lines, GBSHTML_doc_bottom( $title, undef);
push @file_lines, GBSHTML_doc_end();
SPIT_file_nl( $file_spec, \@file_lines);
}
push @lines, HTML_p(
HTML_txt( "$heading very large - Created in seperate file: "),
HTML_link( {target => '_blank'}, $file_spec, "$file"));
return @lines;
} else
{
return @{$lines_ref};
}
}




sub get_diff($$$)
{

my ($msg_count,
$hist_count,
$is_fatal_level,
) = @_;
my ($diff_class, $diff_type);

if ($msg_count > $hist_count)
{
$diff_type = 'worse';
} elsif ($msg_count == $hist_count)
{
if ($msg_count == 0)
{
$diff_type = 'clean';
} else
{
$diff_type = 'same';
}
} else
{
if ($msg_count == 0)
{
$diff_type = 'gone';
} else
{
$diff_type = 'better';
}
}

if ($is_fatal_level)
{
$diff_class = "${diff_type}9";
} else
{
$diff_class = $diff_type;
}

return wantarray ? ($diff_class, $diff_type) : $diff_class;
}




sub color_text($$$)
{
my ($value,
$hist_value,
$is_fatal_level,
) = @_;

my ($diff_class, $diff_type) = get_diff( $value, $hist_value, $is_fatal_level);
my $diff_text = $DIFFS{$diff_type}->[1];

return HTML_div( {class => $diff_class}, $diff_text);
}




sub color_value($$$)
{
my ($value,
$hist_value,
$is_fatal_level,
) = @_;

my $diff_class = get_diff( $value, $hist_value, $is_fatal_level);
my @div = (class => $diff_class);
push @div, (title => $hist_value) if ($value != $hist_value);

return HTML_div( {@div}, $value);
}




sub color_value_f($$$)
{
my ($value,
$hist_value,
$is_fatal_level,
) = @_;
my $html;

my $diff_class = get_diff( $value, $hist_value, $is_fatal_level);
my @div = (class => $diff_class);

no integer;
push @div, (title => sprintf( "%0.3f", $hist_value / 1000.0))
if ($value != $hist_value);
$html = HTML_div( {@div}, sprintf( "%0.3f", $value / 1000.0));
use integer;

return $html;
}




sub delta_value($$)
{
my ($cur_value,
$hist_value,
) = @_;
my $html;

if ($cur_value >= 0)
{
if ($cur_value == $hist_value)
{
$html = HTML_div( HTML_txt( "$cur_value="));
} else
{
my $diff = ($cur_value > $hist_value) ? '>' : '<';
$html = HTML_div( {title => $hist_value}, HTML_txt( "$cur_value$diff"));
}
} else
{
$html = HTML_txt( ' ');
}

return $html;
}




sub clear_totals($)
{
my ($sum_level) = @_;   # 0 = line, 1 = comp, 2 = subsys, 3 = system

my ($msg_sum_refs_ref, $tot_sum_refs_ref) = @{$SUM_REFS[$sum_level]};

foreach my $i (0 .. 3)
{
map { $msg_sum_refs_ref->[$_]->[$i] = 0 } 0 .. $LAST_LEVEL_I;
map { $tot_sum_refs_ref->[$_]->[$i] = 0 } 0 .. $LAST_TOT_I;

}
}





sub add_totals($)
{
my ($sum_level,             # 0 = line, 1 = comp, 2 = subsys, 3 = system
) = @_;

my ($from_msg_sum_refs_ref, $from_tot_sum_refs_ref) = @{$SUM_REFS[$sum_level]};
my ($to_msg_sum_refs_ref,   $to_tot_sum_refs_ref)   = @{$SUM_REFS[$sum_level + 1]};

foreach my $i (0 .. 3)
{
map { $to_msg_sum_refs_ref->[$_]->[$i] += $from_msg_sum_refs_ref->[$_]->[$i] } 0 .. $LAST_LEVEL_I;
map { $to_tot_sum_refs_ref->[$_]->[$i] += $from_tot_sum_refs_ref->[$_]->[$i] } 0 .. $LAST_TOT_I;
}
}




sub calc_levels_sum($)
{
my ($sum_refs_ref) = @_;
my @sums = (0,0,0,0);

foreach my $ref (@{$sum_refs_ref})
{
map { $sums[$_] += $ref->[$_] } ( 0 .. 3);
}

return @sums;
}




sub calc_density($$$$)
{
my ($total_nr_msg,
$total_nr_lines,
$source_nr_msg,	    # in case total_nr_lines == 0
$source_nr_lines,   # in case total_nr_lines == 0
) = @_;
my $density = 0;


if ($total_nr_msg > 0)
{
if ($total_nr_lines > 0)
{
$total_nr_msg *= 1000;
$density = $total_nr_msg / $total_nr_lines;
} elsif ($source_nr_lines > 0)
{
$source_nr_msg *= 1000;
$density = $source_nr_msg / $source_nr_lines;
}
$density = 1
if $density == 0;
$density = - $density
if ($total_nr_lines <= 0);
}
return $density;
}




sub ano_href($$$)
{
my ($subsys,
$component,
$file,
) = @_;

my $name = "${subsys}::${component}:$file";
my $file_spec = "$GBS_ROOT_PATH/dev/$subsys/comp/$component/aud/$GBS_AUDIT/$GBS_TARGET/$file.html";

return HTML_link( {target => '_blank'}, $file_spec, $name);
}




sub print_legend()
{
my @lines;

push @lines, HTML_bold_txt( "Legend:");

my @row;
my $cell;
foreach my $type (@DIFFS)
{
my $severity = $DIFFS{$type}->[0];
my $color_text = $DIFFS{$type}->[2];
$cell .= HTML_div( {class => $type}, $color_text) if ($severity <= 1);
}
push @row, $cell;
$cell = '';
foreach my $type (@DIFFS)
{
my $severity = $DIFFS{$type}->[0];
my $color_text = $DIFFS{$type}->[2];
$cell .= HTML_div( {class => $type}, $color_text)
if ($severity >= 2 && $severity <= 4);
}
push @row, $cell;
$cell = '';
foreach my $type (@DIFFS)
{
my $severity = $DIFFS{$type}->[0];
my $color_text = $DIFFS{$type}->[2];
$cell .= HTML_div( {class => $type}, $color_text)
if ($severity >= 5);
}
push @row, $cell;

$cell = HTML_txt( "Source\nTotal\nDensity");
push @row, scalar HTML_table( undef, undef, [[$cell]], class => 'border1');
push @row, HTML_bold_txt( "Nr.Lines:\n") .
HTML_txt( "  Source = Nr lines in source\n" .
"  Total = Nr lines after preprocessing");
push @row, HTML_bold_txt( "Density:\n").
HTML_txt( "  Positive: TotalNr.Msgs / Nr.TotalLines  (lines after preprocessing)\n".
"  Negative: SourceNr.Msgs / Nr.SourceLines (lines in source)");

push @lines, HTML_table( [ qw( -T -T -T -T) ],
undef,
[ [ @row ] ],
cellspacing => 10);

return @lines;
}




sub dot_print($$)
{
my ($value,
$decimals,
) = @_;

if (index( $value, '.') >= 0)
{
$value = sprintf( "%.${decimals}f", $value);
}
return $value;
}




sub read_gmet_file()
{
my $file_exists = SCAFILE_gmet_read( undef);

if ($file_exists)
{



foreach my $ref (SCAFILE_gmet_get_counts())
{
my ($msg_level, $msg_nr, $s_msg_count, $t_msg_count) = @{$ref};
if (SCA_msg_is_enabled( $msg_nr) || exists $SPECIAL_MSG_NRS{$msg_nr})	    # TBS Sure????
{
store_warning( 0, $msg_nr, $msg_level, $s_msg_count, $t_msg_count);	# cur
}
}




foreach my $ref (SCAFILE_gmet_get_totals())
{
my ($index, $total_name, $s_count, $t_count) = @{$ref};
store_total( 0, $index, $s_count, $t_count);	    # cur values
}




foreach my $ref (SCAFILE_gmet_get_specws())
{
my ($warning, $item) = @{$ref};
$FILE_SPECIAL_MSGS{$warning}->{$item}++;
$TOTAL_SPECIAL_MSGS{$warning}->{$item}++;
}




no integer;     # enable floats
foreach my $ref (SCAFILE_gmet_get_stats())
{
my ($metric_name, $new_min, $new_max, $new_total) = @{$ref};




if (exists $FILE_MET_DATA{$metric_name})
{
my ($min, $max, $total) = @{$FILE_MET_DATA{$metric_name}};
$min = $new_min if ($new_min < $min);
$max = $new_max if ($new_max > $max);
$total += $new_total;
$FILE_MET_DATA{$metric_name} = [ $min, $max, $total ];
} else
{
$FILE_MET_DATA{$metric_name} = [ $new_min, $new_max, $new_total ];
push @FILE_MET_DATA_ORDER, $metric_name;
}




if (exists $TOTAL_MET_DATA{$metric_name})
{
my ($min, $max, $total) = @{$TOTAL_MET_DATA{$metric_name}};
$min = $new_min if ($new_min < $min);
$max = $new_max if ($new_max > $max);
$total += $new_total;
$TOTAL_MET_DATA{$metric_name} = [ $min, $max, $total ];
} else
{
$TOTAL_MET_DATA{$metric_name} = [ $new_min, $new_max, $new_total ];
push @TOTAL_MET_DATA_ORDER, $metric_name;
}
}
use integer;    # disable floats
}

return $file_exists;
}




sub read_hist_gmet_file()
{
my $is_new_file;

my $gmet_file = "$CUR_FILE.gmet";
my @lines = AUDITHIST_get_file( $CUR_SUBSYS, $CUR_COMPONENT, $gmet_file);
if (@lines)
{
SCAFILE_gmet_read( \@lines);



foreach my $ref (SCAFILE_gmet_get_totals())
{
my ($index, $total_name, $s_count, $t_count) = @{$ref};
store_total( 1, $index, $s_count, $t_count);	# hist
}




foreach my $ref (SCAFILE_gmet_get_counts())
{
my ($msg_level, $msg_nr, $s_msg_count, $t_msg_count) = @{$ref};
if (SCA_msg_is_enabled( $msg_nr) || exists $SPECIAL_MSG_NRS{$msg_nr})
{
if (exists $MSG_REFS{$msg_nr})
{
my $level = $MSG_REFS{$msg_nr}->[1];
store_warning( 1, $msg_nr, $level, $s_msg_count, $t_msg_count);	# hist
} else
{
ENV_warn( "Message $msg_nr does not exist anymore: skipped");
}
}
}





$is_new_file = 0;
} else
{
ENV_say( 1, "No hist for $CUR_SUBSYS, $CUR_COMPONENT, $gmet_file");
$is_new_file = 1;
}
}




sub store_warning($$$$$)
{
my ($sum_index,        # 0 = cur, 1 = hist
$msg_nr,
$msg_level,
$s_count,
$t_count,
) = @_;



$FILE_MSG_COUNT{$msg_nr} = [ 0, 0, 0, 0 ]
if (!exists $FILE_MSG_COUNT{$msg_nr});

$FILE_MSG_COUNT{$msg_nr}->[$sum_index] += $s_count;			# $s_index_msg_count
$FILE_MSG_COUNT{$msg_nr}->[$sum_index + 2] += $t_count;		# $t_index_msg_count

my $msg_level_index = $LEVEL_INDEX{$msg_level};
$FILE_MSG_SUM_REFS[$msg_level_index]->[$sum_index] += $s_count;	# $s_index_msg_sum
$FILE_MSG_SUM_REFS[$msg_level_index]->[$sum_index + 2] += $t_count;	# $t_index_msg_sum
}




sub store_total($$$$)
{
my ($sum_index,     # 0 = cur, 2 = hist
$index,		# 0 = nr_lines, 1 = nr_funcs, 2 = nr_class, 3 = nr_const
$s_count,	# -1 == not available
$t_count,	# -1 == not available
) = @_;

$FILE_TOT_SUM_REFS[$index]->[$sum_index] = $s_count;	# $s_index_tot_sum
$FILE_TOT_SUM_REFS[$index]->[$sum_index + 2] = $t_count;	# $t_index_tot_sum
}




sub extract_date_from_filename($)
{
my ($file_spec) = @_;    # gbs<job>_<target>_<audit>_YYMMDD-hhmmss_sum
my ($Y,$M,$D,$h,$m,$s);

($Y,$M,$D,$h,$m,$s) = $file_spec =~ /.*(\d{2})(\d{2})(\d{2})-(\d{2})(\d{2})(\d{2})/;
ENV_fatal( "No YYMMDD-hhmmss in file '$file_spec'")
if (!defined $Y);

return ($Y,$M,$D,$h,$m,$s);
}

return 1;


