#===============================================================
#
#   steps.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package steps;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
STEPS_analyze_steps
STEPS_select_steps
STEPS_get_command
STEPS_get_gbs_subsystems
);

use strict;
use integer;

use glo::env;
use glo::format;
use gbsenv;
use gbsglo;
use gbssfile;
use gbsfileglo;




sub STEPS_analyze_steps($$);
sub STEPS_select_steps($$);
sub STEPS_get_command($);
sub STEPS_get_gbs_subsystems();

sub de_alias_step_names(@);
sub validate_step_names(@);
sub expand_step_names(@);
sub select_steps_by_execute(@);
sub display_steps();
sub read_steps();
sub init_steps();
sub step_entry($$);
sub do_step_command($$$$$);
sub do_step_exec($$$$$);
sub do_step_targets($$$$$);
sub do_step_audits($$$$$);
sub step_exit($$);
sub alias_entry($$);
sub do_alias_steps($$$$$);
sub do_alias_target($$$$$);




my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();

my @ALL_SUBSYSTEMS = GBSGLO_subsystems();
my %ALL_SUBSYSTEMS = map { $_ => 1 } @ALL_SUBSYSTEMS;

my @GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');		# All platforms
my %GBS_ALL_TARGETS = map { ($_, 1) } @GBS_ALL_TARGETS;

my @GBS_TARGETS = ENV_getenv( 'GBS_TARGETS');			# Current platform only
my %GBS_TARGETS = map { ($_, 1) } @GBS_TARGETS;

my @GBS_AUDITS = ENV_getenv( 'GBS_AUDITS');		# All platforms
my %GBS_AUDITS = map { ($_, 1) } @GBS_AUDITS;




my $STEPS_READ = 0;
my $CUR_STEP;
my $CUR_ALIAS;
my $CUR_SUBSYS;

my @STEP_NAMES_ORDER;
my %STEPS_DATA;









my @ALIAS_NAMES_ORDER;
my %ALIASES;





my %ALL_EXEC = (
SELECT  => 1,	# Will be executed if selected
SPECIFY => 1,	# Will only be executed if explicitly specifed
ALLWAYS => 1,	# Will allways be executed, not if job fails
FORCE   => 1,	# Will allways be executed, even if job fails
);

my %STEPS;


























my %STEP_DEFS = (
COMMAND	    => [ 'sc' , undef,     0, undef, undef,  \&do_step_command, undef      ],
EXEC	    => [ 'sss', 'SELECT',  0, undef,     1,  \&do_step_exec,    \%ALL_EXEC ],
TARGETS	    => [ 'ss' , ['*'],     0, undef,   [1],  \&do_step_targets, undef      ],
AUDITS	    => [ 'ss' , ['*'],     0, undef,   [1],  \&do_step_audits,  undef      ],
);

my %ALIAS_DEFS = (
STEPS	    => [ 'ss' , [],        0, undef,   [1],  \&do_alias_steps,  undef ],
TARGET	    => [ 'ss' , '',	   0, undef,     1,  \&do_alias_target, undef ],
);










my %SECTION_DEFS = (
STEP     => [ 'S', 1, 0,   1, \%STEP_DEFS,  \&step_entry,  \&step_exit ],
ALIAS    => [ 'S', 0, 0,   1, \%ALIAS_DEFS, \&alias_entry, undef ],
);




my %SPECIFIED_STEPS;


my $JOBTYPE;	    # gen, make or audit
my @EXPANDED_STEPS;
my %ALIASED_TARGETS;




sub STEPS_analyze_steps($$)
{
my ($jobtype,    # 'gen', 'make' or 'audit'
$steps_ref,
) = @_;
my @selectable_targets;

$JOBTYPE = $jobtype;
my @steps = @{$steps_ref};
@EXPANDED_STEPS = ();

ENV_say( 1, "@steps")
if (@steps && $steps[0] ne '?');




read_steps()
if (!$STEPS_READ);






if (!@steps)
{
@steps = @STEP_NAMES_ORDER;
} elsif ($steps[0] eq '?')
{
ENV_say( 1, "Showing steps...");
display_steps();
ENV_exit( 0);
} else
{



@steps = de_alias_step_names( @steps);
@steps = validate_step_names( @steps);
@steps = expand_step_names( @steps);
}
@EXPANDED_STEPS = select_steps_by_execute( @steps);


if (%ALIASED_TARGETS)
{
@selectable_targets = keys( %ALIASED_TARGETS);
ENV_say( 1, "Aliased Target(s): @selectable_targets");
} else
{
if (%SPECIFIED_STEPS)
{
my %selected_targets;
foreach my $step (keys %SPECIFIED_STEPS)
{

my $targets_ref = $STEPS_DATA{$step}->[6];	    # $targets_ref
map { $selected_targets{$_} = 1 } keys( %{$targets_ref});;
}
@selectable_targets = keys( %selected_targets);
ENV_say( 1, "Selectable Target(s): @selectable_targets");
} else
{
@selectable_targets = @GBS_TARGETS;
ENV_say( 1, "Selectable Target(s): All Targets");
}
}

return @selectable_targets;
}







sub STEPS_select_steps($$)
{
my ($target,
$audit_name, # undef for gen & make
) = @_;
my @selected_steps;





foreach my $step_name (@EXPANDED_STEPS)
{

my (undef, undef, undef, undef, undef, undef, $targets_ref, $audits_ref) = @{$STEPS_DATA{$step_name}};

if ($targets_ref->{$target})
{
if ($audit_name)
{
if ($audits_ref->{$audit_name})
{
push @selected_steps, $step_name;
} else
{
if (exists $SPECIFIED_STEPS{$step_name})
{

}
}
} else
{
push @selected_steps, $step_name;
}
} else
{
if (exists $SPECIFIED_STEPS{$step_name})
{

}
}
}


return @selected_steps;
}




sub de_alias_step_names(@)
{
my @in_steps = @_;
my @out_steps;




my %changed_steps;

foreach my $step_name (@in_steps)
{
if ($step_name eq '.')
{
$step_name = ENV_getenv( 'GBS_SUBSYS');
ENV_error_exit( "Steps: No current SubSystem (.)")
if ($step_name eq '');
push @out_steps, $step_name;
} elsif ($step_name eq '-')
{
push @out_steps, '-';
} elsif (exists $ALIASES{$step_name})
{
my ($steps_ref, $alias_target) = @{$ALIASES{$step_name}};
my @alias_steps = @{$steps_ref};
push @out_steps, @alias_steps
if (@alias_steps);		# ALIAS possibly changes TARGET only
if ($alias_target)
{
@alias_steps = @STEP_NAMES_ORDER
if (!@alias_steps);
foreach my $step (@alias_steps)
{
my $changed_target = $changed_steps{$step};
if (!defined $changed_target)
{
$changed_steps{$step} = $alias_target;
} else
{
ENV_error_exit( "Alias '$step_name' changed Target more than once for STEP '$step'",
"( $changed_target => $alias_target)")
if ($alias_target ne $changed_target);
}
$STEPS_DATA{$step}->[6] = { $alias_target => 1 };	# $targets_ref
}
$ALIASED_TARGETS{ $alias_target } = 1;

}
} else
{
push @out_steps, $step_name;
}
}

return @out_steps;
}




sub validate_step_names(@)
{
my @in_steps = @_;
my @out_steps;




my $previous_step_index = -1;
if (@in_steps == 1 && $in_steps[0] eq 'ALL' && !exists $STEPS_DATA{'ALL'})
{
ENV_say( 1, "Selection forced to 'ALL'");
map { $SPECIFIED_STEPS{$_} = 1 } @STEP_NAMES_ORDER;
@out_steps = @STEP_NAMES_ORDER;
} else
{
foreach my $step_name (@in_steps)
{
if ($step_name eq '-')
{

} else
{
ENV_error_exit( "Unknown step '$step_name'",
"Use 'gbssys... ?' to show all steps")
if (!exists $STEPS_DATA{$step_name});

my $this_step_index = $STEPS_DATA{$step_name}->[0];	# $step_nr
if ($this_step_index < $previous_step_index)
{
my $previous_step_name = $STEP_NAMES_ORDER[$previous_step_index];
ENV_error_exit( "'$previous_step_name' must appear after '$step_name'",
"Proper order is: @STEP_NAMES_ORDER");
} elsif ($this_step_index == $previous_step_index)
{
ENV_error_exit( "'$step_name' specified more than once");
}
$previous_step_index = $this_step_index;
$SPECIFIED_STEPS{$step_name} = 1;
}
}
@out_steps = @in_steps;
}
}




sub expand_step_names(@)
{
my @in_steps = @_;
my @out_steps;

if (!@in_steps)
{
@out_steps = @STEP_NAMES_ORDER;
} else
{



my @working_steps = @in_steps;
unshift @working_steps, $STEP_NAMES_ORDER[0] if ($in_steps[0] eq '-');
push @working_steps, $STEP_NAMES_ORDER[-1] if ($in_steps[-1] eq '-');


while (@working_steps)
{
my $step1 = shift @working_steps;

if (@working_steps && $working_steps[0] eq '-')
{
shift @working_steps;	    # skip '-'
my $step2 = shift @working_steps;
ENV_error_exit( "Subsequent hyphens (- -) not allowed")
if ($step2 eq '-');;
my $i1 = $STEPS_DATA{$step1}->[0];			# $step_nr
my $i2 = $STEPS_DATA{$step2}->[0];			# $step_nr
push @out_steps, @STEP_NAMES_ORDER[$i1..$i2];
} else
{
push @out_steps, $step1;
}
}
}

return @out_steps;
}




sub select_steps_by_execute(@)
{
my @in_steps = @_;
my @out_steps;

my %in_steps = map { $_ => 1 } @in_steps;
foreach my $step (@STEP_NAMES_ORDER)
{
my $execute = $STEPS_DATA{$step}->[1];	    # $execute
if (exists $in_steps{$step})
{
if ($execute eq 'SPECIFY' )
{
push @out_steps, $step
if (exists $SPECIFIED_STEPS{$step});
} else  # SELECT, ALLWAYS or FORCE
{
push @out_steps, $step;
}
} else
{
if ($execute eq 'ALLWAYS' || $execute eq 'FORCE')
{
ENV_say( 1, "Step added: '$step' ($execute)");
push @out_steps, $step;
} else
{

}
}
}

return @out_steps;
}




sub STEPS_get_command($)
{
my ($step_name) = @_;
my ($subsys, $command, $force_execute);

read_steps() if (!$STEPS_READ);

ENV_error_exit( "Unknown step '$step_name'")
if (!exists $STEPS_DATA{$step_name});

(undef, my $execute, undef, $subsys, undef, $command) = @{$STEPS_DATA{$step_name}};
$force_execute = ($execute eq 'FORCE');

return ($subsys, $command, $force_execute);
}





sub STEPS_get_gbs_subsystems()
{
my @all_subsys;

read_steps() if (!$STEPS_READ);

foreach my $step_name (@STEP_NAMES_ORDER)
{
push @all_subsys, $step_name
if ($STEPS_DATA{$step_name}->[4]);		# $is_full_gbs
}

return @all_subsys;
}




sub display_steps()
{
read_steps() if (!$STEPS_READ);




{
my @row_refs = ( [ 'Step', 'Exec', 'Type', 'Audits', 'Targets' ] );

foreach my $step_name (@STEP_NAMES_ORDER)
{
my (undef, $execute, $step, $subsys, $is_full_gbs, undef, $targets_ref, $audits_ref) = @{$STEPS_DATA{$step_name}};

my $type;
if ($subsys ne '')
{
$type = ($is_full_gbs) ? 'Full-GBS' : 'Non-GBS';
} else
{
$type = 'Command';
}
my $audits_comma_list = join( ',', keys %{$audits_ref});
my $targets_comma_list = join( ',', keys %{$targets_ref});
push @row_refs, [ $step_name, $execute, $type, $audits_comma_list, $targets_comma_list ];
}

ENV_say( 0, "- Steps for $GBS_ROOT_PATH:",
FORMAT_table( 0, 2, '  ', undef, @row_refs));
}




{
my @row_refs = ( [ 'Step', 'Command' ] );

foreach my $step_name (@STEP_NAMES_ORDER)
{
my (undef, undef, $step, undef, undef, $command ) = @{$STEPS_DATA{$step_name}};
if ($command)
{
push @row_refs, [ $step_name, $command ];
}
}
if (@row_refs > 1)
{
ENV_say( 0, '',
"- Commands for $GBS_ROOT_PATH:",
FORMAT_table( 0, 2, '  ', undef, @row_refs));
}
}




{
my @row_refs = ([ 'Alias', 'Target', 'Steps' ]);
foreach my $alias (@ALIAS_NAMES_ORDER)
{
my ($steps_ref, $target) = @{$ALIASES{$alias}};
push @row_refs, [ $alias, $target, "@{$steps_ref}" ];
}
if (@row_refs > 1)
{
ENV_say( 0, '',
"- Aliases for $GBS_ROOT_PATH:",
FORMAT_table( 0, 2, '	', undef, @row_refs));
} else
{
ENV_say( 0, '',
"- No Aliases for $GBS_ROOT_PATH");
}
}
}




sub read_steps()
{
init_steps();

my $step_file = "$GBS_ROOT_PATH/sys/steps";
my $full_file_spec = GBSFILEGLO_select_usr_gbs( $step_file);    # .gbs or .usr
if (defined $full_file_spec)
{




GBSSFILE_parse( \%STEPS, $step_file, '', undef, \%SECTION_DEFS);

foreach my $subsys (@ALL_SUBSYSTEMS)
{
ENV_error( "SubSystem: '$subsys' not specified in any STEP")
if (!defined $STEPS_DATA{$subsys}->[0]);
}
} else
{



if (@ALL_SUBSYSTEMS > 1)
{
ENV_say( 1, "No 'steps.gbs' defined",
"SubSystems will be taken in alphabetical order...");
}
foreach my $subsys (@ALL_SUBSYSTEMS)
{
$STEPS_DATA{$subsys}->[0] = scalar @STEP_NAMES_ORDER;	# $step_nr
push @STEP_NAMES_ORDER, $subsys;
}
}

$STEPS_READ = 1;
}




sub init_steps()
{
foreach my $subsys (@ALL_SUBSYSTEMS)
{
my $stepname = $subsys;
my $is_full_gbs = GBSGLO_subsystem_is_full_gbs( $subsys);
my $command = '';
my %cur_targets = map { $_ => 1 } GBSGLO_subsystem_targets( $subsys);
my %cur_audits = map { $_ => 1 } GBSGLO_subsystem_audits( $subsys);
$STEPS_DATA{$stepname} = [ undef, 'SELECT', $stepname, $subsys, $is_full_gbs, $command,
\%cur_targets, \%cur_audits ];
}
}




sub step_entry($$)
{
my ($section,	# STEP
$args_ref,	# $stepname, $command, @command_args
) = @_;

my ($stepname) =  @{$args_ref};





GBSFILEGLO_error_exit( "Invalid stepname '$stepname'")
if ($stepname eq '-');




if (exists $ALL_SUBSYSTEMS{$stepname})
{
$CUR_SUBSYS = $stepname;
} else
{
$CUR_SUBSYS = '';
my $is_full_gbs = 0;
my $command = '';
my %cur_targets = map { $_ => 1 } @GBS_TARGETS;
my %cur_audits;
$STEPS_DATA{$stepname} = [ undef, 'SELECT', $stepname, $CUR_SUBSYS, $is_full_gbs, $command, \%cur_targets, \%cur_audits ];
}

$STEPS_DATA{$stepname}->[0] = scalar @STEP_NAMES_ORDER;	# $step_nr
push @STEP_NAMES_ORDER, $stepname;

$CUR_STEP = $stepname;

return ( $stepname );
}




sub do_step_command($$$$$)
{
my ($section,	    # STEP
$subsection,	    # $step_name
$item,		    # COMMAND
$values_ref,	    # [$command_type, $nr_pos_arg_tokens, $command_line, $ok_values, $exec_condition];

$constr_arg,
) = @_;




my (undef, undef, $command_line, undef, undef) = @{$values_ref};
GBSFILEGLO_error_exit( "Cannot specify COMMAND for SubSystem STEP '$subsection'", "($command_line)")
if ($CUR_SUBSYS ne '');


$STEPS_DATA{$subsection}->[5] = $command_line;    # $command;

return $values_ref;
}




sub do_step_exec($$$$$)
{
my ($section,   # STEP
$subsection,# $step_name
$item,	    # EXEC
$values_ref,
$constr_arg,
) = @_;

my $value = $values_ref->[0];

$STEPS_DATA{$CUR_STEP}->[1] = $value;    # $execute

return $value;
}




sub do_step_targets($$$$$)
{
my ($section,   # STEP
$subsection,# $step_name
$item,	    # TARGETS
$values_ref,
$constr_arg,
) = @_;


my @targets = @{$values_ref};

my $target_op = '=';
if (@targets > 1 && $targets[0] eq '!')
{
$target_op = shift @targets;	# '!'
}

my $targets_ref = $STEPS_DATA{$CUR_STEP}->[6];

if (@targets == 1 && $targets[0] eq '*')
{
@targets = keys( %{$targets_ref});
}




my %targets;
if (@targets == 1 && $targets[0] eq '-')
{
$targets{'-'} = 1;
} else
{
foreach my $target (@targets)
{
GBSFILEGLO_error_exit( "No such Target: $target")
if (!exists $GBS_ALL_TARGETS{$target});
GBSFILEGLO_warning( "Step '$CUR_STEP' does not execute for Target '$target'")
if (!exists $targets_ref->{$target});
GBSFILEGLO_error_exit( "Duplicate Target: $target")
if (exists $targets{$target});
$targets{$target} = 1;
}
}




if ($target_op eq '=')
{
%{$targets_ref} = ();
map { $targets_ref->{$_} = 1 } keys( %targets);
} else  # $target_op eq '!'
{
map { delete $targets_ref->{$_} } keys( %targets);
}

return [ @{$values_ref} ];
}




sub do_step_audits($$$$$)
{
my ($section,   # STEP
$subsection,# $step_name
$item,	    # AUDITS
$values_ref,
$constr_arg,
) = @_;


my @audits = @{$values_ref};

my $audit_op = '=';
if (@audits > 0 && $audits[0] eq '!')
{
$audit_op = shift @audits;	# '!'
GBSFILEGLO_error_exit( "No Audit(s) specified after '!'")
if (@audits == 0);
}

my $audits_ref = $STEPS_DATA{$CUR_STEP}->[7];
if ($CUR_SUBSYS eq '' && keys %{$audits_ref} == 0)
{
map { $audits_ref->{$_} = 1 } @GBS_AUDITS;
}

if (@audits == 1 && $audits[0] eq '*')
{
@audits = keys( %{$audits_ref});
}




my %audits;
if (@audits == 1 && $audits[0] eq '-')
{
$audits{'-'} = 1;
} else
{
foreach my $audit (@audits)
{
GBSFILEGLO_error_exit( "No such Audit: $audit")
if (!exists $GBS_AUDITS{$audit});
GBSFILEGLO_warning( "Step '$CUR_STEP' does not audit for Audit '$audit'")
if (!exists $audits_ref->{$audit});
GBSFILEGLO_error_exit( "Duplicate Audit: $audit")
if (exists $audits{$audit});
$audits{$audit} = 1;
}
}




if ($audit_op eq '=')
{
%{$audits_ref} = ();
map { $audits_ref->{$_} = 1 } keys( %audits);
} else  # $audit_op eq '!'
{
map { delete $audits_ref->{$_} } keys( %audits);
}

return [ @{$values_ref} ];
}




sub step_exit($$)
{
my ($section,	# STEP
$args_ref,	# $stepname
) = @_;

my ($stepname) = @{$args_ref};




if (!exists $ALL_SUBSYSTEMS{$stepname} &&
$STEPS_DATA{$stepname}->[5] eq '')    # $command
{
GBSFILEGLO_error_exit( "COMMAND required after non-SubSystem STEP (STEP=$stepname)");
}
}




sub alias_entry($$)
{
my ($section,	# ALIAS
$args_ref,	# $alias
) = @_;

my $alias = $args_ref->[0];




$ALIASES{$alias} = [ [], '' ];

push @ALIAS_NAMES_ORDER, $alias;

$CUR_ALIAS = $alias;

return ($alias);
}




sub do_alias_steps($$$$$)
{
my ($section,   # ALIAS
$subsection,# $alias_name
$item,	    # STEPS
$values_ref,
$constr_arg,
) = @_;



my @steps = @{$values_ref};
my @step_names_order = @STEP_NAMES_ORDER;
foreach my $step (@steps)
{
next if ($step eq '-');
GBSFILEGLO_error_exit( "No such step '$step'")
if (!exists $STEPS_DATA{$step});

while (@step_names_order && $step ne $step_names_order[0])
{
shift @step_names_order;
}
GBSFILEGLO_error_exit( "Wrong step-sequence for step '$step' (@STEP_NAMES_ORDER)")
if (!@step_names_order);
}
$ALIASES{$CUR_ALIAS}->[0] = [@steps];   # $step_refs

return [ @steps ];
}




sub do_alias_target($$$$$)
{
my ($section,   # ALIAS
$subsection,# $alias_name
$item,	    # STEP
$value_ref,
$constr_arg,
) = @_;



my $value = $value_ref->[0];
GBSFILEGLO_error_exit( "No such Target: $value")
if (!exists $GBS_ALL_TARGETS{$value});
$ALIASES{$CUR_ALIAS}->[1] = $value;	    # $target

return $value;
}

return 1;


