#==============================================================
#
#   swr.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package swr;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SWR_validate
SWR_validate_base
SWR_find_root
SWR_set
SWR_entry
SWR_exit
);

use strict;
use integer;

use glo::env;
use glo::shell;
use glo::build;
use glo::scm;
use gbsenv;
use gbs;
use gbsrc;
use roots;
use system;
use gbsres;
use sys;
use files;




sub SWR_validate($);
sub SWR_validate_base($);
sub SWR_find_root($);
sub SWR_set($$);
sub SWR_entry($);
sub SWR_exit($$);

sub find_switch_files($);

my $GBS_ROOT_PATH = GBSENV_perl_gbs_root();
my $GBS_VERSION = ENV_getenv( 'GBS_VERSION');








sub SWR_validate($)
{
my ($root_path,	# '.', <number>, <spec>

) = @_;
my $new_root = '';	# perl_spec



if ($root_path eq '.')
{
my $os_root = GBSRC_get_base( 'root');
ENV_error_exit( "No current Root/System")
if ($os_root eq '');
$new_root = ENV_perl_paths( $os_root);
} elsif ($root_path =~ /^\d+$/)	# numeric
{
my @all_os_roots = ROOTS_get_all( 0);
my $nr_roots = @all_os_roots;
if ($root_path >= 1 && $root_path <= $nr_roots)
{
$new_root = ENV_perl_paths( $all_os_roots[$root_path - 1]);
} else
{
ENV_error_exit( "Numeric Root-selection must be 1-$nr_roots");
}
} else
{
$new_root = ENV_perl_path_nospace( $root_path);
}

if ($new_root ne '')
{
if (-d $new_root)
{
if (!FILES_check_system( $new_root))
{
ENV_error_exit( "Root '$root_path' ($new_root) is not a valid GBS Root");
}
} else
{
ENV_error_exit( "Root '$root_path' ($new_root) does not exist (anymore)");
}
}




{
my $build_now = ENV_getenv( 'GBS_BUILD');
my $build_start = SYSTEM_get( $new_root, 'build_start');
my $build_end = SYSTEM_get( $new_root, 'build_end');

if ($build_now < $build_start ||
($build_end && $build_now > $build_end))
{
my $build_now_print = BUILD_build_2_version( $build_now);
my $build_start_print = BUILD_build_2_version( $build_start);
my $build_end_print = BUILD_build_2_version( $build_end);
if ($build_now < $build_start)
{
ENV_error_exit( "Current GBS version: $build_now_print ($build_now) too old.",
"Need at least GBS version $build_start_print ($build_start).",
"Range: [$build_start-$build_end]");

} else
{
ENV_error_exit( "Current GBS version: $build_now_print ($build_now) too new.",
"Need latest GBS version $build_end_print ($build_end).",
"Range: [$build_start-$build_end]");
}
$new_root = '';
}
}

return $new_root;
}




sub SWR_validate_base($)
{
my ($root_path) = @_;

if (! ROOTS_exists( $root_path))
{
ENV_warn( "'$root_path' is not listed in Roots List",
"Use 'swr --add' to add the Root to the Roots List",
"Continueing anyway...");
}
}




sub SWR_find_root($)
{
my ($cwd) = @_;
my ($root_path, $subsys, $component) = FILES_split_spec( $cwd);

if ($root_path ne '')
{

if ($subsys ne '')
{

if ($component ne '')
{

} else
{
$component = '-';
}
} else
{
$subsys = '-';
$component = '-';
}
ENV_say( 1, "Found: $root_path $subsys $component");
} else
{
ENV_error_exit( "No GBS-Root found in '$cwd'");
}

return ($root_path, $subsys, $component, $cwd);
}




sub SWR_set($$)
{
my ($root_path,
$must_write,
) = @_;

ENV_debug( "Set $root_path");

if (GBSENV_mode_is_interactive())
{
if ($must_write)
{

GBSRC_write_base( root => $root_path);
}
}




my $root_parent_dir = ENV_parent_dir( $root_path, -1);
my @scm_skiptypes;
my @all_skiptypes;
my @env_refs = (



[ ROOT_PATH	    => 0, 'f', sub { return $root_path } ],
[ ROOT_PARENT       => 0, 'f', sub { return $root_parent_dir} ],
[ SUBSYS	    => 0, 'c', undef ],
[ COMPONENT	    => 0, 'c', undef ],
[ TARGET	    => 0, 'c', undef ],
[ AUDIT		    => 0, 'c', undef ],
[ TOOL		    => 0, 'c', undef ],


[ EXT_PATH	    => 0, 'p', undef ],
[ DEV_PATH	    => 0, 'p', undef ],
[ RES_PATH	    => 0, 'p', undef ],
[ TMP_PATH	    => 0, 'p', undef ],
[ SILO_PATH	    => 0, 'p', undef ],
[ SYS_PATH	    => 0, 'p', undef ],
[ SYSAUDIT_PATH	    => 0, 'p', undef ],
[ SYSGEN_PATH	    => 0, 'p', undef ],
[ SYSTOOLS_PATH	    => 0, 'p', undef ],


[ SYSTEM_NAME	    => 1, 's', undef ],
[ SCMS		    => 1, 's', undef ],
[ SCMS_DATA	    => 1, 's', undef ],
[ SCMS_REPOSITORY   => 1, 's', undef ],
[ ROOT_VERSION	    => 1, 's', undef ],
[ TARGETS	    => 1, 's', undef ],
[ AUDITS	    => 1, 's', undef ],
[ TOOLS		    => 1, 's', undef ],


[ ALL_TARGETS	    => 1, 'a', undef ],
[ ALL_AUDITS	    => 1, 'a', undef],
[ ALL_TOOLS	    => 1, 'a', undef ],
[ SKIPTYPES	    => 1, 'f', sub { return \@all_skiptypes } ],
[ SCM_SKIPTYPES	    => 1, 'f', sub { return \@scm_skiptypes } ],
[ IS_INTEGRATOR	    => 0, 'f', sub { GBSENV_is_integrator() } ],
[ IS_ADMINISTRATOR  => 0, 'f', sub { GBSENV_is_administrator() } ],
);

GBSENV_set_gbs_root_path( $root_path);
if ($root_path eq '')
{



map { GBSENV_setenv( "GBS_$_->[0]", '') } @env_refs;
ENV_set_tmpdir_path( undef);
} else
{



@scm_skiptypes = @{SCM_preinfo( SYSTEM_get( $root_path, 'scms'), $root_path)};
@all_skiptypes = ( GBSRES_skiptypes(), @scm_skiptypes, SYS_skiptypes( $root_path) );
foreach my $ref (@env_refs)
{
my ($name, $is_list, $type, $arg) = @{$ref};

my $env_name = "GBS_$name";
my $value;
if ($type eq 'p')
{
my $path_part = lc( substr( $name, 0, -5));	    # minus "_PATH"
$value = "$root_path/$path_part";
} elsif ($type eq 'c')
{
$value = '';
} elsif ($type eq 's')
{
my $system_part = lc $name;
$value = [ SYSTEM_get( $root_path, $system_part) ];
} elsif ($type eq 'a')
{
my $system_part = lc( substr( $name, 4));	    # minus "ALL_"
$value = [ SYSTEM_get_all( $root_path, $system_part) ];
} elsif ($type eq 'f')
{
$value = $arg->();
} else
{
ENV_fatal( "Impossible type ($type)");
}

if ($is_list)
{
my @values = @{$value};

GBSENV_setenv( $env_name => [ @values ]);
} else
{
$value = $$value
if (ref $value);

GBSENV_setenv( $env_name => $value );
}
}
ENV_set_tmpdir_path( "$root_path/tmp");

my $gbs_root_version = ENV_getenv( 'GBS_ROOT_VERSION');
if ($gbs_root_version lt $GBS_VERSION)
{
ENV_warn( "GBS ROOT-VERSION ($gbs_root_version) mismatch ($GBS_VERSION)",
"Please run gbsmaint 7 9 (Upgrade)");
}
}
}




sub SWR_entry($)
{
my ($root_path) = @_;
my @lines;

if ($root_path ne '')
{
my $gbs_system_name = $ENV{GBS_SYSTEM_NAME};
push @lines, SHELL_echo( "Starting $root_path($gbs_system_name)...");




foreach my $cust_env_name (GBSENV_get_custom_envs())
{
my $all_env_name = 'GBSALL_' . substr( $cust_env_name, 4);   # replace 'GBS_' by 'GBSALL_'
my $all_value = ENV_getenv( $all_env_name);

if ($all_value ne '')
{
push @lines, SHELL_echo( "$all_env_name => $cust_env_name => $all_value");
push @lines, SHELL_setenv( $cust_env_name => $all_value);
}
}



foreach my $env (keys %ENV)
{
if (substr( $env, 0, 10) eq 'GBSALLEXT_')
{
my $all_value = ENV_getenv( $env);
my $cust_env_name = 'GBSEXT_' . substr( $env, 10);
push @lines, SHELL_echo( "$env => $cust_env_name => $all_value");
push @lines, SHELL_setenv( $cust_env_name => $all_value);
}
}





foreach my $switch_spec (find_switch_files( $root_path))
{
if ($switch_spec =~ /switch\.usr\./)
{
my $short_switch_spec = substr( $switch_spec, length( $root_path) + 1);
push @lines, SHELL_echo( "*** Executing System $gbs_system_name $short_switch_spec entry ***");
}
push @lines, SHELL_source( "$switch_spec entry");
}
push @lines, GBS_set_title( $root_path);
} else
{
push @lines, GBS_set_title();
}

return SHELL_conditional_eq( GBS_RC => 0, @lines);
}




sub SWR_exit($$)
{
my ($root_path,
$system_name,
) = @_;
my @lines;

push @lines, GBS_set_title();
if ($root_path ne '')
{



foreach my $switch_spec (reverse find_switch_files( $root_path))
{
if ($switch_spec =~ /switch\.usr\./)
{
my $short_switch_spec = substr( $switch_spec, length( $root_path) + 1);
push @lines, SHELL_echo( "*** Executing System $system_name $short_switch_spec exit ***");
}
push @lines, SHELL_source( "$switch_spec exit");
}




push @lines, map { SHELL_unsetenv( $_) } GBSENV_get_envs_to_reset();


}

return @lines;
}




sub find_switch_files($)
{
my ($root_path) = @_;
my @files;

foreach my $switch_file ('switch.gbs', 'switch.usr')
{
my $switch_spec = ENV_commandfile_spec( "$root_path/$switch_file");
if (-f $switch_spec)
{
push @files, $switch_spec;
}
}

return @files;
}

return 1;


