#===============================================================
#
#   system.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package system;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SYSTEM_get
SYSTEM_get_all
SYSTEM_get_osnames
SYSTEM_os_get
SYSTEM_create
SYSTEM_write
SYSTEM_close
SYSTEM_fix
);

use strict;
use integer;

use glo::env;
use glo::scm;
use gbsfile;




sub SYSTEM_get($$);
sub SYSTEM_get_all($$);
sub SYSTEM_get_osnames($$);
sub SYSTEM_os_get($$$);
sub SYSTEM_create($;@);
sub SYSTEM_write($;@);
sub SYSTEM_close();
sub SYSTEM_fix($;@);

sub read_file($);
sub get_item($$);
sub put_item($$);
sub write_file();
sub general_init();
sub check_values($$$);








my $CUR_OS = $^O;
my @SUPPORTED_SCMS = SCM_get_names();
my $DEFAULT_SCMS = $SUPPORTED_SCMS[0];

my %DEFS = (




id		    => [ 0, 0, 'system',       0, [ 'system' ] ],
system_name     => [ 0, 0, 'NoName',       0, undef ],
build_start	    => [ 0, 0, '0',            0, qr/^\d{9}$/ ],		# VVVYYMMDD
build_end	    => [ 0, 0, '',             1, qr/^\d{9}$/ ],		# VVVYYMMDD
scms	    => [ 1, 0, $DEFAULT_SCMS,  0, \@SUPPORTED_SCMS ],
scms_data	    => [ 0, 0, '',             1, undef ],
scms_repository => [ 1, 0, '',             1, undef ],
root_version    => [ 0, 0, '0.00',         0, qr/^\d+\.\d\d$/ ],		# r.vv
targets	    => [ 1, 1, [],             0, undef ],
audits	    => [ 1, 1, [],             1, undef ],
tools	    => [ 1, 1, [],             1, undef ],
);
my @ITEMS = qw( id system_name root_version build_start build_end
scms scms_data scms_repository targets audits tools);




my $CURRENT_FILE = '';

my %DATA;



my %OTHER_OS;


my @MISSING_ITEMS;
my $RUNNING_FIX = 0;




sub SYSTEM_get($$)
{
my ($root,
$wanted_key,
) = @_;

read_file( $root);

return get_item( $wanted_key, $CUR_OS);
}




sub SYSTEM_get_all($$)
{
my ($root,
$wanted_key,
) = @_;
my @all_values;

read_file( $root);

ENV_fatal( "No such SYSTEM key ($wanted_key)")
if (!exists $DEFS{$wanted_key});

foreach my $osname (keys %{$DATA{$wanted_key}})
{
my @values = get_item( $wanted_key, $osname);
push @all_values, @values
if (@values && defined $values[0]);
}


return @all_values;
}




sub SYSTEM_get_osnames($$)
{
my ($root,
$wanted_key,
) = @_;
my @osnames;

read_file( $root);

ENV_fatal( "No such SYSTEM key ($wanted_key)")
if (!exists $DEFS{$wanted_key});

@osnames = keys( %{$DATA{$wanted_key}});

return @osnames;
}




sub SYSTEM_os_get($$$)
{
my ($root,
$wanted_key,
$osname,
) = @_;

read_file( $root);

return get_item( $wanted_key, $osname);
}




sub SYSTEM_create($;@)
{
my ($root,
@item_pairs,
) = @_;

general_init();
while (@item_pairs)
{
my $name = shift @item_pairs;
my $value = shift @item_pairs;
put_item( $name, $value);
}
$CURRENT_FILE = "$root/system.gbs";
write_file();
}




sub SYSTEM_write($;@)
{
my ($root,
@item_pairs,
) = @_;

read_file( $root);

while (@item_pairs)
{
my $name = shift @item_pairs;
my $value = shift @item_pairs;
put_item( $name, $value);
}
write_file();
}




sub SYSTEM_close()
{
if ($CURRENT_FILE ne '')
{
$CURRENT_FILE = '';
%DATA = ();
%OTHER_OS = ();
}
}




sub SYSTEM_fix($;@)
{
my ($root,
@item_pairs,
) = @_;
my $is_updated = 0;

$RUNNING_FIX = 1;
read_file( $root);
while (@item_pairs)
{
my $name = shift @item_pairs;
my $value = shift @item_pairs;
put_item( $name, $value);
$is_updated = 1;
}

if (@MISSING_ITEMS)
{
if (get_item( 'build_start', $CUR_OS) == 0)
{
my $gbs_build = ENV_getenv( 'GBS_BUILD');   # VVVYYMMDD
put_item( build_start => $gbs_build);
}
@MISSING_ITEMS = ();
$is_updated = 1;
}
write_file();
$RUNNING_FIX = 0;

return $is_updated;
}




sub read_file($)
{
my ($root) = @_;

my $file_spec = "$root/system.gbs";

if ($file_spec ne $CURRENT_FILE)
{
general_init();

GBSFILE_open( $file_spec, 1, undef, undef);
my $line = GBSFILE_get_line();
while ($line)
{

my ($name, $value) = split( /=/,$line, 2);
$value = '' if (!defined $value);
my ($key, $osname) = split( /-/, $name, 2);
$osname = '' if (!defined $osname);
my $def_ref = $DEFS{$key};
if (defined $def_ref)
{
my ($is_os_item, $is_list) = @{$def_ref};
my @values = ($is_list) ? split( ',', $value) : ($value);
check_values( $key, $def_ref, \@values);
if ($is_os_item)
{
if ($is_list)
{
push @{$DATA{$key}->{$osname}}, @values;
} else
{
$DATA{$key}->{$osname} = $values[0];
}
$OTHER_OS{$osname} = 1
if ($osname ne $CUR_OS);
} else
{
if ($is_list)
{
push @{$DATA{$key}}, @values;
} else
{
$DATA{$key} = $values[0];
}
}
} else
{
GBSFILE_warning( "Invalid item '$name'");
}
$line = GBSFILE_get_line();
}




foreach my $item (@ITEMS)
{
my (undef, undef, $default, $empty_ok) = @{$DEFS{$item}};
if (!$empty_ok && !defined $DATA{$item})
{
GBSFILE_warning( "Item '$item' missing");
$DATA{$item} = $default;
push @MISSING_ITEMS, $item;
}
}

ENV_warn( "Errors found. Use gbsmaint 7 5 (Fix system.gbs) to repair")
if (@MISSING_ITEMS && !$RUNNING_FIX);

$CURRENT_FILE = $file_spec;
}
}




sub get_item($$)
{
my ($key,
$osname,
) = @_;
my @values;

my ($is_os_item, $is_list, $default) = @{$DEFS{$key}};
my $value;

ENV_fatal( "No such SYSTEM key ($key)")
if (!exists $DEFS{$key});

if ($is_os_item)
{
$value = $DATA{$key}->{$osname};
if (!defined $value)
{
my @available_keys = sort( grep( $DATA{$key}->{$_}, keys( %OTHER_OS)));
ENV_warn( "$key: No data for this OS ($osname)",
"Available for OS: (@available_keys)",
"Assume: $default");
$value = $default;
$DATA{$key}->{$osname} = $value;
}
} else
{
$value = $DATA{$key};
}
if ($is_list)
{
ENV_fatal( "get_item: $key: wantarray and is_list mismatch")
if (!wantarray);
@values = @{$value};
return @values;
} else
{
return $value;
}
}





sub put_item($$)
{
my ($key,
$value,
) = @_;

my $def_ref = $DEFS{$key};
ENV_fatal( "No such SYSTEM key ($key)")
if (!defined $def_ref);

my ($is_os_item, $is_list) = @{$def_ref};
my $default = ($is_list) ? [] : '';
ENV_fatal( "Value_type key '$key' mismatch!")
if (ref $default ne ref $value);

my $old_value;
if ($is_os_item)
{
$DATA{$key}->{$CUR_OS} = $value;
} else
{
$DATA{$key} = $value;
}
}





sub write_file()
{
my @lines;
push @lines, '#';
push @lines, '#   SYSTEM.GBS';
push @lines, '#';

foreach my $key (@ITEMS)
{
my ($is_os_item, $is_list, $default) = @{$DEFS{$key}};
if ($is_os_item)
{
while (my ($os, $value) = each %{$DATA{$key}})
{
my $value = $DATA{$key}->{$os};
$value = $default
if (!defined $value);
my @values = ($is_list) ? @{$value} : ($value);
push @lines, map { "$key-$os=$_" } @values;
}
} else
{
my $value = $DATA{$key};
my @values = ($is_list) ? @{$value} : ($value);
push @lines, map { "$key=$_" } @values;
}
}

push @lines, '###EOF###';
map { $_ .= "\n" } @lines;


ENV_copy_file( $CURRENT_FILE, "$CURRENT_FILE.old", 0)
if (-e $CURRENT_FILE);
SCM_store_file( $CURRENT_FILE, \@lines);
}




sub general_init()
{
%DATA = ();
%OTHER_OS = ();
foreach my $item (@ITEMS)
{
my ($is_os_item, $is_list, undef, $empty_ok) = @{$DEFS{$item}};
my $default = ($is_list) ? [] : ($empty_ok) ? '' : undef;
if ($is_os_item)
{
$DATA{$item}->{$CUR_OS} = $default;
} else
{
$DATA{$item} = $default;
}
}
$DATA{id} = 'system';
}




sub check_values($$$)
{
my ($item,
$def_ref,
$values_ref,
) = @_;

my (undef, undef, $default, $empty_ok, $check) = @{$def_ref};


if (!$empty_ok)
{
$values_ref->[0] = undef
if (@{$values_ref} == 0);
foreach my $value (@{$values_ref})
{
if (!$value)
{
GBSFILE_warning( "Item must have a value");
$value = $default;
}
}
}

if (defined $check)
{
if (ref $check eq 'ARRAY')
{
my %ok_values = map { $_ => 1 } @{$check};
foreach my $value (@{$values_ref})
{

if ($value && !defined $ok_values{$value})
{
my $old_value = $value;
my %lc_ok_values = map { lc( $_) => $_ } @{$check};
if (defined $lc_ok_values{lc $value})
{
$value = $lc_ok_values{lc $value};
} else
{
$value = $default;
}
GBSFILE_warning( "Invalid value: $old_value",
"Possible value(s): @{$check}",
"Assumed: $value");
}
}
} elsif (ref $check eq 'Regexp')
{
foreach my $value (@{$values_ref})
{

if ($value && $value !~ $check)
{
GBSFILE_warning( "Invalid format value: $value",
"Regexp: $check");
$value = $default;
}
}
} else
{
ENV_fatal( 'Impossible else. $check=' . ref $check);
}
}

}

return 1;
